// ===================================================================
// $Id: brdfman.cpp,v 1.3 2005/11/02 12:59:01 xkrivanj Exp $
//
// brdfman.h
//    Manages the basis function representation of BRDFs.
//
// Initial coding by Jaroslav Krivanek - December 2004.

// GOLEM headers
#include "config.h"
#include "brdfman.h"
#include "brdfwrap.h"
#include "surface.h"
#include "timer.h"

// SPHAR headers
#include "shbasis.h"
#include "shbrdf.h"
#include "shlib.h"

// standard headers
#include <string>

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  CBrdfRepManager::CBrdfRepData::DestroyRep()
// --------------------------------------------------------------------------
void 
CBrdfRepManager::CBrdfRepData::DestroyRep()
{ 
  delete raw; 
  if(adapted != raw) delete adapted;
  raw = adapted = 0;
}

// --------------------------------------------------------------------------
//  CBrdfRepManager::CBrdfRepManager()
// --------------------------------------------------------------------------
CBrdfRepManager::CBrdfRepManager(const CSphBasis<float,float> *shBasis,
                                 int   channels,
                                 int   order,
                                 int   xRes,
                                 int   yRes,
                                 float maxError,
                                 bool  adaptiveOrder,
                                 bool  adaptiveDeletion,
                                 int   numDirections,
                                 bool  loadBrdfRep,
                                 bool  saveBrdfRep)
{
  // first some sanity checking
  if(shBasis==0) {
    FATAL << "shBasis is NULL (and it should not be)." << std::endl;
    FATAL_EXIT;
  }

  if(order<1 || order>MAX_FIXED_SH_ORDER) {
    FATAL << "order=" << order << " is out of bounds." << std::endl;
    FATAL_EXIT;
  }

  if(xRes<1) {
    FATAL << "xRes must be at least 1." << std::endl;
    FATAL_EXIT;
  }
    
  if(yRes<1) {
    FATAL << "yRes must be at least 1." << std::endl;
    FATAL_EXIT;
  }

  if(channels<1) {
    FATAL << "channels must be at least 1." << std::endl;
    FATAL_EXIT;
  }

  if(numDirections<1) {
    FATAL << "numDirections must be at least 1." << std::endl;
    FATAL_EXIT;
  }

  // now copy the parameters
  _shBasis          = shBasis;
  _channels         = channels;
  _order            = order;
  _xRes             = xRes;
  _yRes             = yRes;
  _maxError         = maxError;
  _adaptiveOrder    = adaptiveOrder;
  _adaptiveDeletion = adaptiveDeletion;
  _loadBrdfRep      = loadBrdfRep;
  _saveBrdfRep      = saveBrdfRep;

  // convert order to a string
  char buff[200];
  sprintf(buff,"%d",order);
  _orderStr = buff;

  // convert number of channels to a string
  sprintf(buff,"%d",channels);
  _channelsStr = buff;

  // convert resolution to a string
  sprintf(buff,"%d-%d",xRes,yRes);
  _resStr = buff;

  // convert number of sample directions to a string
  sprintf(buff,"%d",numDirections);

  // create the signature part independent of the brdf
  _surfIndependentSignature =  
    "ch" + _channelsStr + "-res" + _resStr + "-ndir" + buff;

  // precomputed basis
  _basisCache = 
    new CSphBasisEvalCache<float,float>(numDirections, order, shBasis->GetAnalytic() );
}

// --------------------------------------------------------------------------
//  CBrdfRepManager::CleanUp()
// --------------------------------------------------------------------------
/*
CBrdfRepManager::CleanUp()
{
  /// delete precomputed basis
  SAFE_DELETE(_basisCache);

  /// delete the BRDF representation
  for(TBrdfMap::iterator i=_brdfMap.begin(); i!=_brdfMap.end(); i++)
    i->second.DestroyRep();
  _brdfMap.clear();
}
*/


// --------------------------------------------------------------------------
//  CBrdfRepManager::~CBrdfRepManager()
// --------------------------------------------------------------------------
CBrdfRepManager::~CBrdfRepManager()
{
  /// delete precomputed basis
  delete _basisCache;

  /// delete the BRDF representation
  for(TBrdfMap::iterator i=_brdfMap.begin(); i!=_brdfMap.end(); i++)
    i->second.DestroyRep();
}

// --------------------------------------------------------------------------
//  CBrdfRepManager::SetUpBRDFRep()
// --------------------------------------------------------------------------
bool
CBrdfRepManager::SetUpBrdfRep(CSurface *s)
{
  if(!s) return true; // error

  // const string &brdfSign = s->GetSignature();

  // reset the timer
  CTimer timer;
  timer.Reset();

  // representation for 's'
  const CBrdfRepData* repData = 0;

  // unique file name for this BRDF
  string repName = CreateBrdfRepFileName(s);

  // try to look it up from already loaded BRDF reps
  if ( (repData=FindBrdfRep(repName)) != 0) {
    // we found it
    assert(repData->adapted);
    s->SetSHRep(repData->adapted);
    return false; // ok
  }

  // try to load the BRDF representation from a file
  if(_loadBrdfRep) {
    STATUS << " - loading BRDF rep from file..." << std::endl;
    timer.Start();
    if ( (repData=LoadBrdfRep(repName)) != 0) {
      timer.Stop();
      STATUS << " - done in " << timer.UserTime() << " secs." << std::endl;
      assert(repData->adapted);
      s->SetSHRep(repData->adapted);
      return false; // ok
    }
  }
  else {
    STATUS 
      << " - skipping BRDF loading (RadianceCache.BRDF.loadBrdfRep=OFF)" 
      << std::endl;
  }
  
  // brdf could not be loaded, compute it
  STATUS << " - computing BRDF representation..." << std::endl;
  timer.Start();
  if ( (repData=ComputeBrdfRep(s)) == 0 ) {
    STATUS << " - failed!!!" << std::endl;
    return true; // error
  }
  assert(repData->adapted);
  s->SetSHRep(repData->adapted);
  timer.Stop();
  STATUS 
    << " - BRDF precomputation done in " << timer.UserTime() << " secs." 
    << std::endl;
  
  if(_saveBrdfRep) {
    STATUS << " - saving BRDF rep to file..." << std::endl;
    timer.Reset();
    timer.Start();
    if( SaveBrdfRep(repData) == true ) {
      STATUS << " - failed!!!" << std::endl;		
      return false; // saving failed, but representation was set up ok
    }
    else {
      timer.Stop();
      STATUS << " - done in " << timer.UserTime() << " secs." << std::endl;
      return false; // ok
    }
  }
  else {
    STATUS << " - skipping BRDF saving  (RadianceCache.BRDF.saveBrdfRep=OFF)" << std::endl;
    return false; // ok
  }

  FATAL << "Unreachable code." << std::endl;
  FATAL_EXIT;

}

// --------------------------------------------------------------------------
//  CBrdfRepManager::CreateBRDFRepFileName()
// --------------------------------------------------------------------------
string
CBrdfRepManager::CreateBrdfRepFileName(const CSurface *surface) const
{
  if(!surface) return string("");

  const char* shBasisStr = _shBasis->GetTypeString();

  return 
    string(shBasisStr) + "-" + _surfIndependentSignature + "-" + 
    surface->GetSignature() + ".brdfman";
}

// --------------------------------------------------------------------------
//  CBrdfRepManager::FindBrdfRep()
// --------------------------------------------------------------------------
const CBrdfRepManager::CBrdfRepData*
CBrdfRepManager::FindBrdfRep(const string& repName) const
{
  TBrdfMap::const_iterator it = _brdfMap.find(repName);
  if(it==_brdfMap.end())
    return 0;
  else
    return &(it->second);
}

// --------------------------------------------------------------------------
//  CBrdfRepManager::LoadBRDFRep()
// --------------------------------------------------------------------------
const CBrdfRepManager::CBrdfRepData*
CBrdfRepManager::LoadBrdfRep(const string& repName)
{
  char *errReason="";

  // try to open the file name
  FILE *fin = fopen(repName.c_str(), "rt");

  // The adapted representation
  CSphBRDF<float,float> *adaptedShRep = 0;
  CSphBRDF<float,float> *shRep = 0;

  if(!fin) {
    errReason = "can't open input file";
    goto err_return;
  }

  // load the BRDF rep from the file
  shRep = new CSphBRDF<float,float>(fin, _adaptiveDeletion);
  fclose(fin);
  
  if(shRep) {
    if(!shRep->OK()) {
      errReason = "could not load BRDF coefficients";
      goto err_return;
    }
  }
  else {
    errReason = "not enough memory to allocate BRDF rep";
    goto err_return;
  }

  // adapt the BRDF representation to a given order and maximum error 
  adaptedShRep = 
    shRep->Adapt(_order, _maxError, _adaptiveOrder, _adaptiveDeletion);
  
  // the contents of the file are not enough to meet the required accuracy
  if(!adaptedShRep) {
    delete shRep;
    errReason = "file contents need upgrade";
    goto err_return;
  }

  // save the original and adapted BRDF rep to the internal database
  _brdfMap[repName] = CBrdfRepData(shRep,adaptedShRep,repName);

  // and return the adapted BRDF
  return FindBrdfRep(repName);
  
err_return:
  STATUS 
    << " - could not load BRDF rep. from file (" 
    << errReason << ")." << std::endl;
  return 0;
}


// --------------------------------------------------------------------------
//  CBrdfRepManager::SaveBrdfRep()
// --------------------------------------------------------------------------
bool
CBrdfRepManager::SaveBrdfRep(const CBrdfRepData* repData)
{
  // error description
  string errReason;
  // file for writing
  FILE *f = 0;

  // it was not found -> nothing will be saved
  if(!repData || !repData->raw) {
    errReason = "invalid argument";
    goto err_return;
  }

  // open the file name
  f = fopen(repData->repName.c_str(),"wt");
  if(!f) {
    errReason = "could not open '"  + repData->repName + "' for writing" ;
    goto err_return;
  }

  if ( repData->raw->SaveToFile(f) ) {
    errReason = "could not write data";
    goto err_return;
  }

  fclose(f);
  return false; // ok

err_return:
  if(f) fclose(f);
  STATUS 
    << " - could not save BRDF rep. to file (" 
    << errReason << ")." << std::endl;
  return true; // error
}

// --------------------------------------------------------------------------
//  CBrdfRepManager::ComputeBrdfRep()
// --------------------------------------------------------------------------
const CBrdfRepManager::CBrdfRepData*
CBrdfRepManager::ComputeBrdfRep(const CSurface *s)
{
  if(!s) return 0;

  // create a proxy for 's'
  CBRDFSurfaceWrapper wrapper(s);

  // compute non-adaptive representation
  CSphBRDF<float,float> *rawShRep = new CSphBRDF<float,float>(
    wrapper,           // brdf
    _channels,          // num of color channels
    _order,             // max allowed sh-expansion order
    _xRes,              // texture resolution
    _yRes,              // texture resolution
    _maxError,          // max. allowd error
    false,              // adaptive length of sh-expansion
    false,              // delete sh-evectors with too high error
    *_basisCache);      // precomputed directions
    
  assert(rawShRep);

  // adapt the representation to a given maxError
  CSphBRDF<float,float> *adaptedShRep = 
    rawShRep->Adapt(_order, _maxError, _adaptiveOrder, _adaptiveDeletion);

  assert(adaptedShRep);

  const string& repName = CreateBrdfRepFileName(s);

  // save the original and adapted BRDF rep to the internal database
  _brdfMap[repName] = CBrdfRepData(rawShRep,adaptedShRep,repName);

  // and return the adapted BRDF
  return FindBrdfRep(repName);
}


__END_GOLEM_SOURCE
