// ===================================================================
// $Id: brdfman.h,v 1.3 2005/11/02 12:58:42 xkrivanj Exp $
//
// brdfman.h
//    Manages the basis function representation of BRDFs.
//
// Initial coding by Jaroslav Krivanek - December 2004.

#ifndef __BRDFMAN_H__
#define __BRDFMAN_H__

// GOLEM headers
#include "configh.h"

// standard headers
#include <list>
#include <map>

// forward declaration
class CSurface;
template<class TV,class TA> class CSphBRDF;
template<class TV,class TA> class CSphBasisEvalCache;
template<class TV,class TA> class CSphBasis;

__BEGIN_GOLEM_HEADER

class CBrdfRepManager {

protected: //data

  /// A pair of BRDF reps.
  /**
     One representation is unadapted and the other is adapted to the
     maximum error. 
     They might point to the same representation, in case no adaptation
     is needed.
    */

  /// encapsulated storage for adaptive and non-adaptive BRDF representation
  class CBrdfRepData {

  public: // data

    /// file name for storing this BRDF
    string   repName;

    /// raw BRDF representation (non-adaptive)
    CSphBRDF<float,float> *raw;

    /// adaptive BRDF representation
    CSphBRDF<float,float> *adapted;

  public: // methods

    /// default constructor
    CBrdfRepData():raw(0),adapted(0){}

    CBrdfRepData(
      CSphBRDF<float,float> *r, 
      CSphBRDF<float,float> *a, 
      const string& s)
      :raw(r),adapted(a?a:r),repName(s) {}

    /// delete the BRDF representation
    void DestroyRep();
  };

  /// type for the map from the BRDF signature and BRDF representation
  typedef map<string,CBrdfRepData>  TBrdfMap;

  // the actual map from the BRDF signature to the BRDF representation
  TBrdfMap _brdfMap;

  /// (hemi)spherical basis
  const CSphBasis<float,float> *_shBasis;

  /// number of color channels
  int _channels;

  /// order of the BRDF representation
  int _order;

  /// reslution of the BRDF texture
  int _xRes, _yRes;

  /// maximum allowed BRDF representation error
  float _maxError;

  /// adapt BRDF order to the representation error?
  bool _adaptiveOrder;

  /// delete directions can't be specified with a given maximum error
  bool _adaptiveDeletion;

  /// load the BRDF representation from a file?
  bool _loadBrdfRep;

  /// save BRDF representation to a file after it has been computed?
  bool _saveBrdfRep;

  /// _order converted to a string
  string _orderStr;

  /// number of channels converted to a string
  string _channelsStr;

  /// resolution of the texture converted to a sting
  string _resStr;

  /// the part of brdf-rep signature wich is independent of the BRDF
  string _surfIndependentSignature;

  /// precomputed basis functions
  CSphBasisEvalCache<float,float>  *_basisCache;

public: // methods

  CBrdfRepManager(
    const CSphBasis<float,float> *shBasis,
    int   channels,
    int   order,
    int   xRes,
    int   yRes,
    float maxError,
    bool  adaptiveOrder,
    bool  adaptiveDeletion,
    int   numDirections,
    bool  loadBrdfRep,
    bool  saveBrdfRep);

  virtual ~CBrdfRepManager();

  ///
  /**
    @return NULL on error, BRDF representation on success
   */
  virtual bool SetUpBrdfRep(CSurface *);

  virtual const CBrdfRepData* FindBrdfRep(const string& repName) const;

  virtual const CBrdfRepData* LoadBrdfRep(const string& repName);

  virtual const CBrdfRepData* ComputeBrdfRep(const CSurface *);

  virtual bool  SaveBrdfRep(const CBrdfRepData*);

  virtual string CreateBrdfRepFileName(const CSurface *surface) const;
};

__END_GOLEM_HEADER

#endif //  __BRDFMAN_H__
