// ===================================================================
// $Id: cachmode.h,v 1.1 2005/11/02 12:58:48 xkrivanj Exp $
//
// cachmode.h
//      Caching mode for radiance and irradiance caching.
//
// Initial coding by Jaroslav Krivanek, March 2005.

#ifndef __CACHMODE_H__
#define __CACHMODE_H__

// GOLEM headers
#include "configh.h"

// standard headers
#include <string>

__BEGIN_GOLEM_HEADER

// forward declarations
class CEnvironment;

/// Mode of caching for irradiance or radiance cache.
class CCachingMode {

public:

  enum EMode {

    /// Shall we look for samples in the cache?
		EE_LookIntoCache    = 1,
    /// Shall we insert samples into the cache?
    EE_InsertIntoCache  = 2,
    /// Shall we sample the hemisphere if nothing found in the cache? 
    EE_SampleHemisphere = 4,

    /// Normal caching behaviour. If something in the cache, return it. 
    /// Otherwise sample hemisphere and insert.
    EE_Normal       = EE_LookIntoCache|EE_InsertIntoCache|EE_SampleHemisphere,
    /// Do no use caching, always sample the hemisphere.
    EE_AlwaysSample = EE_SampleHemisphere,
    /// Always sample the hemisphere & insert the irradiance into the cache.
    EE_AlwaysInsert = EE_InsertIntoCache|EE_SampleHemisphere,
    /// Never sample the hemisphere. Read from cache. If nothing found, return 
    /// black.
    EE_AlwaysRead   = EE_LookIntoCache,
    /// Special value - use the same caching mode for higher order and 
    /// primary rays. Only applicable to higher order rays.
    EE_Same         = 10000,
    /// Unrecognized string value in SetFromString() method.
    EE_Error        = 50000
  };

  /// Decode the caching mode from a string.
  /**
    The following string values, corresponding to the values of EMode
    are recognized:
    "normal", "always_sample", "always_insert", "always_read" and "same"
    If any other string is passed in 's', EE_Error is returned.   
   */
  static EMode SetFromString(const std::string& s);

  /// Report an error in a string parsed by SetFromString()
  static void ReportError(
    const std::string& name,
    const std::string& s,
    bool sameAllowed);

  /// Read the caching mode from environment
  /**
     @return true is error occured (icorrect syntax in environment). false
       if everything went ok.
    */
  static bool ReadCachingMode(
    const CEnvironment& env,
    const std::string& prefix,
    EMode &presampleModePrim,
    EMode &presampleModeHi,
    EMode &renderModePrim,
    EMode &renderModeHi);

  inline static bool LookIntoCache(EMode m)   {return m & EE_LookIntoCache;}
  inline static bool InsertIntoCache(EMode m) {return m & EE_InsertIntoCache;}
  inline static bool SampleHemisphere(EMode m){return m & EE_SampleHemisphere;}
};

__END_GOLEM_HEADER

#endif //__CACHMODE_H__
