// ==================================================================
// $Id: diffsampl.cpp,v 1.5 2005/11/02 12:59:01 xkrivanj Exp $
//
// diffsampl.cpp
//    Support routines for sampling diffuse hemisphere.
//    Used for irradiance caching.
//
// Initial coding by Jaroslav Krivanek, Oct 2003

// GOLEM headers
#include "config.h"
//#include "blocker.h"
#include "contrib.h"
#include "diffsampl.h"
#include "incrad.h"
#include "rnd.h"
#include "surface.h"
#include "basmacr.h"

// #include <cfloat>

__BEGIN_GOLEM_SOURCE

/////////////////////////////////////////////////////////////////////////////
//
//   CDiffuseSampler
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CDiffuseSampler::CDiffuseSampler()
// --------------------------------------------------------------------------
CDiffuseSampler::CDiffuseSampler(int                depth, 
                                 int                numRays, 
                                 float             infRayLength, 
                                 float             minAllowedHitDist,
                                 CIncidentRadiance* incRad, 
                                 const void*        data)
{
  (void)data;
  assert(depth>=0);
  assert(numRays>0);
  assert(incRad!=0);
  
  _depth             = depth;
  _numRays           = numRays;
  _incRad            = incRad;
  _infRayLength      = infRayLength;
  _minAllowedHitDist = minAllowedHitDist;
}

/////////////////////////////////////////////////////////////////////////////
//
//   CDiffuseSamplerStratAbstr
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CDiffuseSampler::_computeMN()
// --------------------------------------------------------------------------
void
CDiffuseSamplerStratAbstr::_computeMN(int numRays, float NoverM)
{
  if (numRays==1) {
	   M = N  = 1;
     invM = invN = 1.0f;
  }
  else {
	   M = (int)(sqrtf((float)numRays/NoverM));
     if (M < 1) M = 1;
     N = (int)((float)numRays/(float)M);
     invN = 1.0f/N;
     invM = 1.0f/M;
  }
  invRayNum = 1.0f/(M*N);
}


/////////////////////////////////////////////////////////////////////////////
//
//   CDiffuseSamplerWard
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CDiffuseSampler::CDiffuseSamplerWard()
// --------------------------------------------------------------------------
CDiffuseSamplerWard::CDiffuseSamplerWard(int                depth, 
                                         int                numRays,
                                         float              infRayLength, 
                                         float              minAllowedHitDist,
                                         CIncidentRadiance* incRad, 
                                         const void*        data)
: CDiffuseSamplerStratAbstr(depth, numRays, infRayLength, minAllowedHitDist, 
                            incRad,data)
{
  v_base_vect_canonical = 0;
  u_vect_canonical = 0;
  v_vect_canonical = 0;
  
  sin_mul_cos2 = 0;
  sin_diff     = 0;
  tan_theta    = 0;
  sin_theta    = 0;
  cos_theta    = 0;

  cos_phi = 0;
  sin_phi = 0;
  
  _Li = 0;
  _ir = 0;
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerWard::_allocAndPrecomputePhi()
// --------------------------------------------------------------------------
void
CDiffuseSamplerWard::_allocAndPrecomputePhi()
{
  int i,k;

  M_PI_div_MN       =  M_PI / (M*N);
  M_PI_mul_2_div_N  =  M_PI*2.0 / N; 

  cos_theta              = new float[M];
  sin_theta              = new float[M];
  tan_theta              = new float[M];
  sin_diff               = new float[M];
  sin_mul_cos2           = new float[M + 1];

  // nothing is precalculated in these arrays, they are here so we don't
  // have to allocate them everytime we sample the diffuse surface
  _Li  = new CValues[M * N];
  for(i=0; i<M*N; i++)
    _Li[i].SetRepresentation(CStandardValues::GetDefaultColorType());
  _ir = new float[M * N];

  // -----------------------
  //  precompute phi arrays
  // -----------------------
  v_base_vect_canonical = new CVector3D[N];
  u_vect_canonical      = new CVector3D[N];
  v_vect_canonical      = new CVector3D[N];
  cos_phi = new float[N];
  sin_phi = new float[N];

  for (k = 0; k < N; k++) {
    float phi = 2.0 * M_PI * ((float)k + 0.5) / (float)N;
    cos_phi[k] = cos(phi);
    sin_phi[k] = sin(phi);
    
    u_vect_canonical[k] = CVector3D(cos_phi[k], sin_phi[k], 0);
    
    float phiBase = phi + 0.5*M_PI;
    if (phiBase >= 2.0*M_PI) {
      phiBase -= 2.0*M_PI;
    }
    v_base_vect_canonical[k] = CVector3D(cos(phiBase), sin(phiBase), 0);
    float wv = 2.0*M_PI*float(k)/float(N) + 0.5*M_PI;
    if (wv >= 2.0 * M_PI) { wv -= 2.0*M_PI; }
    v_vect_canonical[k] = CVector3D(cos(wv), sin(wv), 0);
  }
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerWard::~CDiffuseSamplerWard()
// --------------------------------------------------------------------------
CDiffuseSamplerWard::~CDiffuseSamplerWard()
{
  // precomputed data
  delete[] v_base_vect_canonical;
  delete[] u_vect_canonical;
  delete[] v_vect_canonical;
  
  delete[] sin_mul_cos2;
  delete[] sin_diff;
  delete[] tan_theta;
  delete[] sin_theta;
  delete[] cos_theta;

  delete[] cos_phi;
  delete[] sin_phi;
  
  // buffers
  delete [] _Li;
  delete [] _ir;
}


/////////////////////////////////////////////////////////////////////////////
//
//   CDiffuseSamplerCosWard
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CDiffuseSamplerCosWard::CDiffuseSamplerCosWard()
// --------------------------------------------------------------------------
CDiffuseSamplerCosWard::CDiffuseSamplerCosWard(int                depth, 
                                               int                numRays,
                                               float              infRayLength, 
                                               float              minAllowedHitDist,
                                               CIncidentRadiance* incRad, 
                                               const void*        data)
: CDiffuseSamplerWard(depth, numRays, infRayLength, minAllowedHitDist, incRad,
                      data)
{
  _computeMN(numRays,M_PI);
  
  // alloc buffers and precompute phi-related data
  _allocAndPrecomputePhi();
 
  // fill theta arrays
  float* sin_theta_grid  = new float[M + 1];
  float* cos_theta_grid  = new float[M + 1];

  int j;
  for(j = 0; j <= M; j++) {
    sin_theta_grid[j]  = sqrt((float)j / M);
    cos_theta_grid[j]  = sqrt(1.0 - (float)j / M);
    sin_mul_cos2[j]    = (1.0 - (float)j / M) * sin_theta_grid[j];
  }
  
  for (j = 0; j < M; j++) {
    float sqr_sin_theta = (j+0.5) / M;
    sin_theta[j] = sqrt(sqr_sin_theta);
    cos_theta[j] = sqrt(1.0-sqr_sin_theta);
    tan_theta[j] = sin_theta[j] / cos_theta[j];

#if 1
    // According to Ward
    sin_diff [j] = (sin_theta_grid[j+1] - sin_theta_grid[j]); 
#else
    // According to <JK/PG> -> this formula follows from the gradients for 
    // spherical harmonics (SCCG05) if the harmonics are replaced by cosines.
    // It gives the same result as Ward's formula.
    sin_diff [j] = 
      (cos_theta_grid[j]-cos_theta_grid[j+1])*cos_theta[j]/sin_theta[j]; 
#endif
  }

  delete[] sin_theta_grid;
  delete[] cos_theta_grid;
}

// --------------------------------------------------------------------------
//  lintp()
// --------------------------------------------------------------------------
// linear interpolation
static inline float lintp(float a, float b, float t)
{
  // return a*(1-t) + b*t;
  return  a + (b-a)*t;
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerCosWard::ComputeIrradiance()
// --------------------------------------------------------------------------

bool 
CDiffuseSamplerCosWard::ComputeIrradiance(const CVector3D &pt,
                                          const CVector3D& normal,
                                          const CVector3D& unormal,
                                          const CObject3D* object,
                                          const void* termination,
                                          const float &footprintArea,
                                          CHitPointInfo& outHitInfo,
                                          CValues&       result,
                                          float&         harmMeanDistance,
                                          float&         minHitDist,
                                          float&         stdDev,
                                          CVector3D*       rotGrad,
                                          CVector3D*       transGrad,
                                          const void*      data)
{
  DIFF_SAMPLER_SETUP;
  // reciprocal hit distance
  float invHitDist;
  // reciprocal of infinity
  float invInfty = float(1.0)/CLimits::Infinity;
  CVector3D raydir;
  int numRays=0;

  // for computing variance estimate
  CValues LVar(CStandardValues::GetDefaultColorType());
  int     numRaysVar = 0;

  CVector3D uN = (DotProd(unormal,normal)>=0) ? unormal : -unormal;

  for (int k = 0; k < N; k++) {
    for (int j = 0; j < M; j++) {
      float phi        = 2.0 * M_PI * ( (float)k + RND::rand01f() ) * invN;
      float sin2_theta = ((float)j + RND::rand01f()) * invM;   
      float sin_theta  = sqrt(sin2_theta);
      float cos_theta  = sqrt(1.0 - sin2_theta);
      float cos_phi    = cos(phi);
      float sin_phi    = sin(phi);
      contrib.prob     =  cos_theta * M_1_PI;
      raydir_canonical.SetValue( cos_phi*sin_theta, 
                                 sin_phi*sin_theta, 
                                 cos_theta);
      // compute the direction to be traced
      raydir = ToGlobalFrame(raydir_canonical, X, Y, Z);

      // if the ray is below horizon (w.r.t the geometric normal), skip it
#if 0
      if( DotProd(raydir, uN)<=0 ) {
        _ir[idx(j,k)] = -1;
        continue;
      }
#endif
      numRays++;
      // initialize the ray
      contrib.ray.Init(pt, raydir, REFLECTED_RAY, termination, object, 0, 
        true /*,&lastCell*/);
      // unset,if it was set from previous iteration
      contrib.invalid = false;
      contrib.diffuseness = 1;
      // cast the ray an compute higher order light contribution
      if(_incRad->IncidentRadiance(outHitInfo,&contrib,_depth,footprintArea)){
        hitDist = outHitInfo.GetT();
        // check the min dist and possibly bail out
        CHECK_HIT_DIST;
        // compute reciprocal of the hit distance
        invHitDist = float(1.0)/hitDist;
        // update harmonic mean distance
        distancesInv += invHitDist;
        // if ray goes to infty, there is no contrib to distancesInv (1/infty)
        numHits++;
      }
      else
        invHitDist = invInfty;
      // index for saving incident radiance and the hit distance
      int index = idx(j,k);   
      // save reciprocal hit distance for gradient computation
      _ir[index] = invHitDist;
      // even or odd ray?
      int oddRay = (j+k)&1;
      // increment the number of even/odd rays
      numRaysVar += oddRay;
      // save incident radiance for gradient computation
      if(contrib.invalid)
        _Li[index].setz();
      else {
        _Li[index] = contrib.color;
        // update total irradiance
        irradiance += contrib.color;
        // update total even/orr irradiance
        if(oddRay) LVar += contrib.color;
      }
    } // for t
  } // for p
  
  // divide by the number of rays & set the reults
  if(numRays) 
  {
    irradiance *= M_PI/(float)(numRays);
    result = irradiance;
  }
  else {
    result.setz();
    result.Set(1,100);
    PRINTVAR(numRays);
  }

  // estimate result variance
  float L  = irradiance.Luminance();
  if(numRaysVar) {
    float La = LVar.Luminance()*M_PI/numRaysVar;
    stdDev = fabs(L-La);
  }
  else
    stdDev = 0;

  // harmonic mean distance is the inverse of mean inverses 
  if(numHits)
	   harmMeanDistance = (float)numHits / distancesInv;
  else
	   harmMeanDistance = _infRayLength;

  // gradients
  if(rotGrad) {
	   assert(transGrad);
     // compute translational and rotational gradients
     _computeGradients(rotGrad,transGrad,channels);
     // transform the gradients to the global coord. frame
     _gradToGlobalFrame(rotGrad,transGrad,X,Y,Z,channels);
  }
 
  return true;
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerCosWard::_computeGradients()
// --------------------------------------------------------------------------
void
CDiffuseSamplerCosWard::_computeGradients(CVector3D *urotGrad, 
                                         CVector3D *utransGrad,
                                         int channels)
{
  int i,j,k;
  // set gradients to zero
  _gradSetz(urotGrad,utransGrad,channels);

  CValues 
    rotGrad    ( _Li[0].GetType() ), 
    transGradU ( _Li[0].GetType() ), 
    transGradV ( _Li[0].GetType() );

  const float *rotGradData    = rotGrad.data();
  const float *transGradUData = transGradU.data();
  const float *transGradVData = transGradV.data();
  
  // phi intervals
  for (k = 0; k < N ; k++)
  {
    rotGrad.setz();
    transGradU.setz();
    transGradV.setz();
    
    // theta intervals
    for(j = 0; j<M; j++) {
      // index of this cell
      int idxjk  = idx(j,k);
      // skip this cell - it was under the horizon
      if(_ir[idxjk]==-1) continue;
      // rotational gradient
      rotGrad -= (_Li[idxjk] * tan_theta[j]);
      // u - translational gradient
      if (j>0) {
        int idxj1k = idx(j-1,k);
        // skip this cell - it was under the horizon
        if(_ir[idxj1k]==-1) continue;
        float ir = Max( _ir[idxjk], _ir[idxj1k]);
        transGradU += (_Li[idxjk] - _Li[idxj1k]) * ( sin_mul_cos2[j] * ir );
      }
      // v - translational gradient
      int idxjk1 = (k==0)?idx(j, N-1):idx(j,k-1);
      // skip this cell - it was under the horizon
      if(_ir[idxjk1]==-1) continue;
      float ir = Max(_ir[idxjk], _ir[idxjk1]);
      transGradV += (_Li[idxjk] - _Li[idxjk1]) * (sin_diff[j] * ir);
    }
    // accumulate rotational gradients
    for(i=0; i<channels; i++) {
      urotGrad[i]   += rotGradData[i] * v_base_vect_canonical[k];		
      utransGrad[i] += 
        M_PI_mul_2_div_N * transGradUData[i] * u_vect_canonical[k]
        + transGradVData[i] * v_vect_canonical[k];
    }
  } // for k
  
  for(i=0; i<channels; i++) {
    urotGrad[i] *= M_PI_div_MN;
  }
   return;
}

/////////////////////////////////////////////////////////////////////////////
//
//   CDiffuseSamplerUniformDisplace
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CDiffuseSamplerUniformDisplace::CDiffuseSamplerUniformDisplace()
// --------------------------------------------------------------------------
CDiffuseSamplerUniformDisplace::
CDiffuseSamplerUniformDisplace(int                depth, 
                               int                numRays, 
                               float              infRayLength, 
                               float              minAllowedHitDist,
                               CIncidentRadiance* incRad, 
                               const void*        data)
: CDiffuseSamplerStratAbstr(depth, numRays, infRayLength, minAllowedHitDist, 
                            incRad,data)
{
  _computeMN(numRays,4);
  
  _r      = new float    [M*N];
  _dir    = new CVector3D[M*N];
  _Li     = new CValues  [M*N];
  _normal = new CVector3D[M*N];
  _dot    = new float    [M*N];
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerUniformDisplace::~CDiffuseSamplerUniformDisplace()
// --------------------------------------------------------------------------
CDiffuseSamplerUniformDisplace::~CDiffuseSamplerUniformDisplace() 
{
  delete [] _r;
  delete [] _dir;
  delete [] _Li;
  delete [] _normal;
  delete [] _dot;
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerUniformDisplace::ComputeIrradiance()
// --------------------------------------------------------------------------
bool 
CDiffuseSamplerUniformDisplace::ComputeIrradiance(const CVector3D &pt,
                                                  const CVector3D& normal,
                                                  const CVector3D& unormal,
                                                  const CObject3D* object,
                                                  const void*      termination,
                                                  const float &footprintArea,
                                                  CHitPointInfo&   outHitInfo,
                                                  CValues&         result,
                                                  float&           harmMeanDistance,
                                                  float&           minHitDist,
                                                  float&         stdDev,
                                                  CVector3D*       rotGrad,
                                                  CVector3D*       transGrad,
                                                  const void*      data)
{
  DIFF_SAMPLER_SETUP;
  
  assert(rotGrad==0); assert(transGrad==0);
  
  float TWO_PI_div_MN = 2.0f*M_PI/(float)(M*N);
  
  int i=0;
  float r1 = 0.0, r2 = 0.0;
  for(int ir1 = M; ir1-- ; r1 += invM, r2 = 0 ) {
	   for(int ir2 = N ; ir2-- ; r2 += invN , i++) {
       // generate random tuple
       CVector2D v(r1+RND::rand0bf(invM), r2+RND::rand0bf(invN));
       // sample the hemisphere in canonical coordinates
       float cos_theta = 1.0f - v.xx;
       float sin_theta = sqrtf(1.0f - cos_theta * cos_theta);
       float phi       = 2.0f * M_PI * v.yy;
       raydir_canonical.SetValue( cos(phi)*sin_theta, 
                                  sin(phi)*sin_theta, 
                                  cos_theta );
       contrib.prob = 0.5f * M_1_PI;
       // compute the direction to be traced
       _dir[i] = ToGlobalFrame(raydir_canonical, X, Y, Z);   
       // initialize the ray
       contrib.ray.Init(pt, _dir[i], REFLECTED_RAY, termination, object, 0, true);
       // unset,if it was set from previous iteration
       contrib.invalid = false;
       contrib.diffuseness = 1;
       // cast the ray an compute higher order light contribution
       if(_incRad->IncidentRadiance(outHitInfo,&contrib,_depth,footprintArea)){
         hitDist = outHitInfo.GetT();
         // check the min dist and possibly bail out
         CHECK_HIT_DIST;
         // update harmonic mean distance
         distancesInv += 1.0f / hitDist;
         // if ray goes to infty, there is no contrib to distancesInv (1/infty)
         numHits++;
         // save the hit normal and dot product
         _normal[i] = outHitInfo.GetNormal();
         _dot[i]    = outHitInfo.GetDotProd();
       }
       else
         hitDist = CLimits::Infinity;
       // save the hit distance
       _r[i] = hitDist;
       
       if(contrib.invalid) {
         _Li[i].Set(0);
       }
       else {
         _Li[i] = contrib.color;
         // update total irradiance
         contrib.color *= raydir_canonical.z;
         irradiance += contrib.color;
       }
       
     } // for t
  } // for p
  
  // divide by the number of rays
  irradiance *= TWO_PI_div_MN;
  // set the reults
  result = irradiance;
  
  // harmonic mean distance is the inverse of mean inverses 
  if(numHits)
	   harmMeanDistance = (float)numHits / distancesInv;
  else
	   harmMeanDistance = _infRayLength;
  
  ///////////////////////////////////////////////////////////////////////
  //            GRADIENTS
  ///////////////////////////////////////////////////////////////////////
  
  if( !rotGrad ) return true;  // no gradients
  
  int ch;
  
  // reset the derivative vectors
  setz(rotGrad,   channels);
  setz(transGrad, channels);
  
  // displacement in the tangent plane transformed to global frame
  float displacement = 0.01;
  
  CVector3D xdisplaced = displacement * X;
  CVector3D ydisplaced = displacement * Y;
  
  CVector3D xrayDir, yrayDir;
  float     xmul, ymul, xmag, ymag;
  float     xdistratio, ydistratio, xdotratio, ydotratio;
  
  for(i=0; i<M*N; i++) {
    
	   // ray direction in the shifted hemisphere
	   if(_r[i] < CLimits::Infinity) {
       
       xrayDir = (_r[i] * _dir[i] - xdisplaced);
       yrayDir = (_r[i] * _dir[i] - ydisplaced);
       
       xmag = Magnitude(xrayDir);
       ymag = Magnitude(yrayDir);
       
       if( xmag < CLimits::Small || ymag < CLimits::Small )
         continue;
       
       xrayDir /= xmag;
       yrayDir /= ymag;
       
       /// calculate the change of the solid angle
       /// mul = cos(theta_) * r_old^2 / r_new^2
       xdistratio = sqr(_r[i] / xmag);
       ydistratio = sqr(_r[i] / ymag);
       xdotratio = DotProd(xrayDir,_normal[i]) / _dot[i];
       ydotratio = DotProd(yrayDir,_normal[i]) / _dot[i];
       xmul = xdotratio > 0 ? xdotratio * xdistratio : 0;
       ymul = ydotratio > 0 ? ydotratio * ydistratio : 0;
     }
     else { 
       // ray goes to infinity
       xrayDir = yrayDir = _dir[i];
       xmul    = ymul    = 1;
     }
     
     xmul *= DotProd(normal,xrayDir);
     ymul *= DotProd(normal,yrayDir);
     
     const float *Li = _Li[i].data();
     for(ch=0; ch < channels; ch++) {
       transGrad[ch].x += xmul * Li[ch];
       transGrad[ch].y += ymul * Li[ch];
     }
  }
  
  // at this point we have the projection coefficients of Li in transGrad 
  // convert them to differences and transform them to global frame
  const float *E = result.data();
  float m = 1.0 / displacement;
  for(ch=0; ch<channels; ch++) {
	   transGrad[ch].x = ( TWO_PI_div_MN * transGrad[ch].x - E[ch]) * m;
     transGrad[ch].y = ( TWO_PI_div_MN * transGrad[ch].y - E[ch]) * m;
     transGrad[ch] = ToGlobalFrame(transGrad[ch], X,Y,Z);
  }
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//
//   CDiffuseSamplerCosDisplace
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
// CDiffuseSamplerCosDisplace::CDiffuseSamplerCosDisplace()
// --------------------------------------------------------------------------
CDiffuseSamplerCosDisplace::
CDiffuseSamplerCosDisplace(int                depth, 
                           int                numRays, 
                           float              infRayLength, 
                           float              minAllowedHitDist,
                           CIncidentRadiance* incRad, 
                           const void*        data)
: CDiffuseSamplerStratAbstr(depth, numRays, infRayLength, minAllowedHitDist, 
                            incRad,data)
{
  _computeMN(numRays,M_PI);
  
  _r          = new float     [M*N];
  _dir        = new CVector3D [M*N];
  _Li         = new CValues   [M*N];
  _normal     = new CVector3D [M*N];
  _dot        = new float     [M*N];
  _icos_theta = new float     [M*N];
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerCosDisplace::~CDiffuseSamplerCosDisplace()
// --------------------------------------------------------------------------
CDiffuseSamplerCosDisplace::~CDiffuseSamplerCosDisplace() 
{
  delete [] _r;
  delete [] _dir;
  delete [] _Li;
  delete [] _normal;
  delete [] _dot;
  delete [] _icos_theta;
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerCosDisplace::ComputeIrradiance()
// --------------------------------------------------------------------------
bool 
CDiffuseSamplerCosDisplace::ComputeIrradiance(const CVector3D &pt,
                                              const CVector3D& normal,
                                              const CVector3D& unormal,
                                              const CObject3D* object,
                                              const void*      termination,
                                              const float &footprintArea,
                                              CHitPointInfo&   outHitInfo,
                                              CValues&         result,
                                              float&           harmMeanDistance,
                                              float&           minHitDist,
                                              float&         stdDev,
                                              CVector3D*       rotGrad,
                                              CVector3D*       transGrad,
                                              const void*      data)
{
  DIFF_SAMPLER_SETUP;
  
  assert(rotGrad==0); assert(transGrad==0);
  
  float PI_div_MN = M_PI/(float)(M*N);
  
  int i=0;
  for (int p = 0; p < N; p++) {
    for (int t = 0; t < M; t++, i++) {	
      float phi        = 2.0 * M_PI * ( (float)p + RND::rand01f() ) * invN;
      float sin2_theta = ((float)t + RND::rand01f()) * invM;
      float sin_theta  = sqrt(sin2_theta);
      float cos_theta  = sqrt(1.0 - sin2_theta);
      float cos_phi    = cos(phi);
      float sin_phi    = sin(phi);
      contrib.prob     =  cos_theta * M_1_PI;
      raydir_canonical.SetValue( cos_phi*sin_theta, sin_phi*sin_theta, cos_theta);
      // save cosine
      _icos_theta[i] = 1.0f/cos_theta;
      // compute the direction to be traced
      _dir[i] = ToGlobalFrame(raydir_canonical, X, Y, Z);   
      // initialize the ray
      contrib.ray.Init(pt, _dir[i], REFLECTED_RAY, termination, object, 0, true);
      // unset,if it was set from previous iteration
      contrib.invalid = false;
      contrib.diffuseness = 1;
      // cast the ray an compute higher order light contribution
      if(_incRad->IncidentRadiance(outHitInfo,&contrib,_depth,footprintArea)){
        hitDist = outHitInfo.GetT();
        // check the min dist and possibly bail out
        CHECK_HIT_DIST;
        // update harmonic mean distance
        distancesInv += 1.0f / hitDist;
        // if ray goes to infty, there is no contrib to distancesInv (1/infty)
        numHits++;
        // save the hit normal and dot product
        _normal[i] = outHitInfo.GetNormal();
        _dot[i]    = outHitInfo.GetDotProd();
      }
      else
        hitDist = CLimits::Infinity;
      // save the hit distance
      _r[i] = hitDist;
      
      if(contrib.invalid) {
        _Li[i].Set(0);
      }
      else {
        _Li[i] = contrib.color;
        // update total irradiance
        irradiance += contrib.color;
      }
      
    } // for t
  } // for p
  
  // divide by the number of rays
  irradiance *= PI_div_MN;
  // set the reults
  result = irradiance;
  
  // harmonic mean distance is the inverse of mean inverses 
  if(numHits)
	   harmMeanDistance = (float)numHits / distancesInv;
  else
	   harmMeanDistance = _infRayLength;
  
  ///////////////////////////////////////////////////////////////////////
  //            GRADIENTS
  ///////////////////////////////////////////////////////////////////////
  
  if( !rotGrad )  return true;  // no gradients
  
  int ch;
  
  // reset the derivative vectors
  setz(rotGrad,   channels);
  setz(transGrad, channels);
  
  // displacement in the tangent plane transformed to global frame
  float displacement = 0.01;
  
  CVector3D xdisplaced = displacement * X;
  CVector3D ydisplaced = displacement * Y;
  
  CVector3D xrayDir, yrayDir;
  float     xmul, ymul, xmag, ymag;
  float     xdistratio, ydistratio, xdotratio, ydotratio;
  
  for(i=0; i<M*N; i++) {
    
	   // ray direction in the shifted hemisphere
	   if(_r[i] < CLimits::Infinity) {
       
       xrayDir = (_r[i] * _dir[i] - xdisplaced);
       yrayDir = (_r[i] * _dir[i] - ydisplaced);
       
       xmag = Magnitude(xrayDir);
       ymag = Magnitude(yrayDir);
       
       if( xmag < CLimits::Small || ymag < CLimits::Small )
         continue;
       
       xrayDir /= xmag;
       yrayDir /= ymag;
       
       /// calculate the change of the solid angle
       /// mul = cos(theta_) * r_old^2 / r_new^2
       xdistratio = sqr(_r[i] / xmag);
       ydistratio = sqr(_r[i] / ymag);
       xdotratio = DotProd(xrayDir,_normal[i]) / _dot[i];
       ydotratio = DotProd(yrayDir,_normal[i]) / _dot[i];
       // distance and orientation
       xmul = xdotratio > 0 ? xdotratio * xdistratio : 0;
       ymul = ydotratio > 0 ? ydotratio * ydistratio : 0;
       // distance only
       // xmul = xdistratio;
       // ymul = ydistratio;
       // orientation only 
       // xmul = xdotratio;
       // ymul = ydotratio;
     }
     else { 
       // ray goes to infinity
       xrayDir = yrayDir = _dir[i];
       xmul    = ymul    = 1;
     }
     
     xmul *= DotProd(normal,xrayDir) * _icos_theta[i];
     ymul *= DotProd(normal,yrayDir) * _icos_theta[i];
     
     const float *Li = _Li[i].data();
     for(ch=0; ch < channels; ch++) {
       transGrad[ch].x += xmul * Li[ch];
       transGrad[ch].y += ymul * Li[ch];
     }
  }
  
  // at this point we have the projection coefficients of Li in transGrad 
  // convert them to differences and transform them to global frame
  const float *E = result.data();
  float m = 1.0 / displacement;
  for(ch=0; ch<channels; ch++) {
	   transGrad[ch].x = ( PI_div_MN * transGrad[ch].x - E[ch]) * m;
     transGrad[ch].y = ( PI_div_MN * transGrad[ch].y - E[ch]) * m;
     transGrad[ch] = ToGlobalFrame(transGrad[ch], X,Y,Z);
  }
  
  return true;
}


/////////////////////////////////////////////////////////////////////////////
//
//   CDiffuseSamplerUniformDeriv
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CDiffuseSamplerUniformDeriv::CDiffuseSamplerUniformDeriv()
// --------------------------------------------------------------------------
CDiffuseSamplerUniformDeriv::
CDiffuseSamplerUniformDeriv(int                depth, 
                            int                numRays, 
                            float              infRayLength, 
                            float              minAllowedHitDist,
                            CIncidentRadiance* incRad, 
                            const void*        data)
: CDiffuseSamplerStratAbstr(depth, numRays, infRayLength, minAllowedHitDist,
                            incRad,data)
{
  _computeMN(numRays,4); 
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerUniformDeriv::~CDiffuseSamplerUniformDeriv()
// --------------------------------------------------------------------------
CDiffuseSamplerUniformDeriv::~CDiffuseSamplerUniformDeriv() 
{
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerUniformDeriv::ComputeIrradiance()
// --------------------------------------------------------------------------
bool 
CDiffuseSamplerUniformDeriv::ComputeIrradiance(const CVector3D &pt,
                                               const CVector3D& normal,
                                               const CVector3D& unormal,
                                               const CObject3D* object,
                                               const void*      termination,
                                               const float &footprintArea,
                                               CHitPointInfo&   outHitInfo,
                                               CValues&         result,
                                               float&           harmMeanDistance,
                                               float&           minHitDist,
                                               float&         stdDev,
                                               CVector3D*       rotGrad,
                                               CVector3D*       transGrad,
                                               const void*      data)
{
  DIFF_SAMPLER_SETUP;
  
  assert(rotGrad==0); assert(transGrad==0);
  
  // reset gradients
  if(rotGrad) {
	   assert(transGrad);
     setz(rotGrad,   channels);
     setz(transGrad, channels);
  }
  
  float TWO_PI_div_MN = 2.0f*M_PI/(float)(M*N);
  bool  inf;
  int   i = 0, ch;
  float r1 = 0.0, r2 = 0.0;
  for(int ir1 = M; ir1-- ; r1 += invM, r2 = 0 ) {
	   for(int ir2 = N ; ir2-- ; r2 += invN , i++) {
       // generate random tuple
       CVector2D v(r1+RND::rand0bf(invM), r2+RND::rand0bf(invN));
       // sample the hemisphere uniformly in canonical coordinates
       float cos_theta  = v.xx;
       float sin_theta  = sqrtf(1.0f-cos_theta*cos_theta);
       float phi        = 2.0f * M_PI * v.yy;
       float cos_phi    = cosf(phi);
       float sin_phi    = sinf(phi);
       raydir_canonical =  CVector3D(  cos_phi*sin_theta, sin_phi*sin_theta, cos_theta );
       contrib.prob = 0.5f*M_1_PI;
       // compute the direction to be traced
       CVector3D raydir = ToGlobalFrame(raydir_canonical, X, Y, Z);   
       // initialize the ray
       contrib.ray.Init(pt, raydir, REFLECTED_RAY, termination, object, 0, true);
       // unset,if it was set from previous iteration
       contrib.invalid = false;
       contrib.diffuseness = 1;
       // cast the ray an compute higher order light contribution
       if(_incRad->IncidentRadiance(outHitInfo,&contrib,_depth,footprintArea)){
         hitDist = outHitInfo.GetT();
         // check the min dist and possibly bail out
         CHECK_HIT_DIST;
         // update harmonic mean distance
         distancesInv += 1.0f / hitDist;
         // if ray goes to infty, there is no contrib to distancesInv (1/infty)
         numHits++;
         inf = false;
       }
       else {
         hitDist = CLimits::Infinity;
         inf = true;
       }
       
       if(!contrib.invalid) {
         // update total irradiance
         irradiance += (contrib.color * cos_theta);
         
         if(transGrad && !inf) {
           ////////////////////////////////////
           //  GRADIENTS
           ////////////////////////////////////
           float r  = hitDist;
           float ir = 1.0f/hitDist;
           // (I) d cos(theta)/dx
           float dTheta_dx = -cos_theta * cos_phi * ir;
           float dTheta_dy = -cos_theta * sin_phi * ir;
           
           // (II) dOmega/dx * cos(theta)
           float cos_xi = -outHitInfo.GetDotProd();
           // PRINTVAR(cos_xi);
           CVector3D nl  = ToLocalFrame(outHitInfo.GetNormal(),X,Y,Z);
#if 1
           // optimized calculation od dOmega_dx a dOmega_dy
           float dOmega_dx = 
             ( nl.x + 3.0f * raydir_canonical.x * cos_xi ) / (r * cos_xi);
           float dOmega_dy = 
             ( nl.y + 3.0f * raydir_canonical.y * cos_xi ) / (r * cos_xi);
#else
           // verbatim calculation of dOmega_dx a dOmega_dy
           // size of the area seen by the solid angle around this ray
           float dA = (r*r) / cos_xi;
           // d/dx ( cos_xi / r^2)
           float ddx = 
             (r*nl.x + 3.0f * (raydir_canonical.x*r) * cos_xi ) / sqr(sqr(r));
           float ddy = 
             (r*nl.y + 3.0f * (raydir_canonical.y*r) * cos_xi ) / sqr(sqr(r));
           float dOmega_dx = ddx * dA;
           float dOmega_dy = ddy * dA;
#endif           
           // (I) + (II)
           float dx = (cos_theta * dOmega_dx - sin_theta * dTheta_dx);
           float dy = (cos_theta * dOmega_dy - sin_theta * dTheta_dy);
           
           const float *Lk = contrib.color.data();
           for(ch=0; ch<channels; ch++) {
             transGrad[ch].x += Lk[ch] * dx;
             transGrad[ch].y += Lk[ch] * dy;
           }
         }
       }
     } // for t
  } // for p
  
  // divide by the number of rays
  irradiance *= TWO_PI_div_MN;
  // set the reults
  result = irradiance;
  
  // harmonic mean distance is the inverse of mean inverses 
  if(numHits)
	   harmMeanDistance = (float)numHits / distancesInv;
  else
	   harmMeanDistance = _infRayLength;
  
  ///////////////////////////////////////////////////////////////////////
  //            GRADIENTS
  ///////////////////////////////////////////////////////////////////////
  
  if( !rotGrad ) return true;  // no gradients
  
  // normalize and transform to the global frame
  for(ch=0; ch<channels; ch++)
	   transGrad[ch] = TWO_PI_div_MN * ToGlobalFrame(transGrad[ch], X,Y,Z);
  
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//
//   CDiffuseSamplerCosDeriv
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CDiffuseSamplerCosDeriv::CDiffuseSamplerCosDeriv()
// --------------------------------------------------------------------------
CDiffuseSamplerCosDeriv::
CDiffuseSamplerCosDeriv(int                depth, 
                        int                numRays, 
                        float              infRayLength, 
                        float              minAllowedHitDist,
                        CIncidentRadiance* incRad, 
                        const void*        data)
: CDiffuseSamplerStratAbstr(depth, numRays, infRayLength, minAllowedHitDist, 
                            incRad,data)
{
  _computeMN(numRays,M_PI);
  
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerCosDeriv::~CDiffuseSamplerCosDeriv()
// --------------------------------------------------------------------------
CDiffuseSamplerCosDeriv::~CDiffuseSamplerCosDeriv() 
{
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerCosDeriv::ComputeIrradiance()
// --------------------------------------------------------------------------
bool 
CDiffuseSamplerCosDeriv::ComputeIrradiance(const CVector3D &pt,
                                           const CVector3D& normal,
                                           const CVector3D& unormal,
                                           const CObject3D* object,
                                           const void*      termination,
                                           const float &footprintArea,
                                           CHitPointInfo&   outHitInfo,
                                           CValues&         result,
                                           float&           harmMeanDistance,
                                           float&           minHitDist,
                                           float&         stdDev,
                                           CVector3D*       rotGrad,
                                           CVector3D*       transGrad,
                                           const void*      data)
{
  DIFF_SAMPLER_SETUP;
 
  // reset gradients
  if(rotGrad) {
	   assert(transGrad);
     setz(rotGrad,   channels);
     setz(transGrad, channels);
  }
  
  float PI_div_MN = M_PI/(float)(M*N);
  bool  inf;
  int   i = 0, ch;
  
  for (int p = 0; p < N; p++) {
    for (int t = 0; t < M; t++) {	
      float phi        = 2.0 * M_PI * ( (float)p + RND::rand01f() ) * invN;
      float sin2_theta = ((float)t + RND::rand01f()) * invM;
      float sin_theta  = sqrt(sin2_theta);
      float cos_theta  = sqrt(1.0 - sin2_theta);
      float cos_phi    = cos(phi);
      float sin_phi    = sin(phi);
      contrib.prob     =  cos_theta * M_1_PI;
      raydir_canonical.SetValue( cos_phi*sin_theta, sin_phi*sin_theta, cos_theta);
      contrib.prob = cos_theta*M_1_PI;
      // compute the direction to be traced
      CVector3D raydir = ToGlobalFrame(raydir_canonical, X, Y, Z);   
      // initialize the ray
      contrib.ray.Init(pt, raydir, REFLECTED_RAY, termination, object, 0, true);
      // unset,if it was set from previous iteration
      contrib.invalid = false;
      contrib.diffuseness = 1;
      // cast the ray an compute higher order light contribution
      if(_incRad->IncidentRadiance(outHitInfo,&contrib,_depth,footprintArea)){
        hitDist = outHitInfo.GetT();
        // check the min dist and possibly bail out
        CHECK_HIT_DIST;
        // update harmonic mean distance
        distancesInv += 1.0f / hitDist;
        // if ray goes to infty, there is no contrib to distancesInv (1/infty)
        numHits++;
        inf = false;
      }
      else {
        hitDist = CLimits::Infinity;
        inf = true;
      }
      
      if(!contrib.invalid) {
        // update total irradiance
        irradiance += contrib.color;
       
        if(transGrad && !inf) {
          ////////////////////////////////////
          //  GRADIENTS
          ////////////////////////////////////
          float cos_xi  = -outHitInfo.GetDotProd();
          if(cos_xi > 0.07) 
          {
            float r  = hitDist;
            float ir = 1.0f/hitDist;
#if 0 // unoptimized
            // (I) d cos(theta)/dx * Li
            float dTheta_dx = -cos_theta * cos_phi * ir;
            float dTheta_dy = -cos_theta * sin_phi * ir;
            
            // (II) dLi/dx * cos(theta)
            CVector3D nl  = ToLocalFrame(outHitInfo.GetNormal(),X,Y,Z);
            float dOmega_dx = 
              ( nl.x + 3.0f * raydir_canonical.x * cos_xi ) / (r * cos_xi);
            float dOmega_dy = 
              ( nl.y + 3.0f * raydir_canonical.y * cos_xi ) / (r * cos_xi);
            
            // (I) + (II)
            float dx = (cos_theta*dOmega_dx - sin_theta*dTheta_dx) / cos_theta; 
            float dy = (cos_theta*dOmega_dy - sin_theta*dTheta_dy) / cos_theta; 
#else // optimized			   
            const CVector3D &n  = outHitInfo.GetNormal();
            
            float dOmega_dx = 
              ( DotProd(n,X) + 3.0f * raydir_canonical.x * cos_xi ) / cos_xi;
            float dOmega_dy = 
              ( DotProd(n,Y) + 3.0f * raydir_canonical.y * cos_xi ) / cos_xi;
            
            float dx = (dOmega_dx + sin_theta * cos_phi) * ir; 
            float dy = (dOmega_dy + sin_theta * sin_phi) * ir; 
#endif
            const float *Lk = contrib.color.data();
            for(ch=0; ch<channels; ch++) {
              transGrad[ch].x += Lk[ch] * dx;
              transGrad[ch].y += Lk[ch] * dy;
            }
          }          
        }
      }
    } // for t
  } // for p
  
  // divide by the number of rays
  irradiance *= PI_div_MN;
  // set the reults
  result = irradiance;
  
  // harmonic mean distance is the inverse of mean inverses 
  if(numHits)
	   harmMeanDistance = (float)numHits / distancesInv;
  else
	   harmMeanDistance = _infRayLength;
  
  ///////////////////////////////////////////////////////////////////////
  //            GRADIENTS
  ///////////////////////////////////////////////////////////////////////
  
  if( !rotGrad ) return true;  // no gradients
  
  // normalize and transform to the global frame
  for(ch=0; ch<channels; ch++) {
	   transGrad[ch] = PI_div_MN * ToGlobalFrame(transGrad[ch], X,Y,Z);
  }
  
  return true;
}


/*
bool 
CDiffuseSamplerStratifiedBlocker::ComputeIrradiance(const CVector3D &pt,
const CVector3D& normal,
const CObject3D* object,
const void* termination,
const float &footprintArea,
CHitPointInfo& outHitInfo,
CValues&       result,
float&         harmMeanDistance,
const void*          data)
{
// number of rays hitting any surface
int numHits = 0;
// sum of inverse distances to the intersected objects 
float distancesInv = 0;
// accumulated irradiance
CValues irradiance(result.GetType());
// contribution item used to gather IncidentRadiance()
CContributionItem contrib;
// set the color type
contrib.color.SetRepresentation(result.GetType());
// set the interaction type
contrib.interaction = EE_REFLECTED_DIFFUSE;
// we are going to make the contribution go through a diffuse relection, thus:
contrib.diffuseness = 1;
// and this is a reflected ray
contrib.flags = CContributionItem::EE_ReflectedRay;
// build the coordinate system
CVector3D U = ArbitraryNormal(normal);     // U is perpendicular to N
CVector3D V = CrossProd(normal,U);         // orthonormal base with N and U
// outgoing ray direction in canonical coords (Z = up = surface normal)
CVector3D canDir;

  float r1 = 0.0, r2 = 0.0;
  
    const CDirBlocker *blocker = (const CDirBlocker*)data;
    assert(blocker);
    
      for(int ir1 = N; ir1-- ; r1 += invN, r2 = 0 ) {
      for(int ir2 = M ; ir2-- ; r2 += invM ) {
      // sample the hemisphere in canonical coordinates
      RND::hemisph_cos(canDir,
      CVector2D(r1 + RND::rand0bf(invN),
      r2 + RND::rand0bf(invM)));
      // transform from local coordinates
      CVector3D rayDir = canDir.x * U + canDir.y * V + canDir.z * normal ;
      // rejection test based on blocker
      if(blocker->Blocked(rayDir))
			   continue;
         // initialize the ray
         contrib.ray.Init(pt, rayDir, REFLECTED_RAY, termination, object, 0, true);
         // unset,if it was set from previous iteration
         contrib.invalid = false;
         // cast the ray an compute higher order light contribution
         if( _incRad -> IncidentRadiance( outHitInfo, &contrib, _depth ) ) {
         // update harmonic mean distance
         distancesInv += 1.0f / outHitInfo.GetT();
         // if ray goes to infty, there is no contrib to distancesInv (1/infty)
		       numHits++; 
           }
           // the harmonic mean is not increased by the rays that go to infinity, since 
           // this would make it far too large
           
             if( ! contrib.invalid ) {
             irradiance += (contrib.color * M_PI ) ;
             }
             }
             }
             // divide by the number of rays
             irradiance *= (invRayNum);
             // set the reults
             result = irradiance;
             // harmonic mean distance is the inverse of mean inverses 
             if(numHits)
             harmMeanDistance = (float)numHits / distancesInv;
             else
             harmMeanDistance = _infRayLength;
             
               return true;
               }
               */
               
               
// --------------------------------------------------------------------------
// CreateDiffuseSampler()
// --------------------------------------------------------------------------
CDiffuseSampler* 
CreateDiffuseSampler(CDiffuseSampler::EDiffuseSamplerType samplerType, 
                     int depth, 
                     int numRays, 
                     float infRayLength, 
                     float minAllowedHitDist,
                     CIncidentRadiance* incRad, 
                     const void*  data)
{
  switch(samplerType) {
  case CDiffuseSampler::EE_DiffuseSamplerCosWard:
    return new CDiffuseSamplerCosWard(depth,numRays,infRayLength,
      minAllowedHitDist,incRad,data);
  case CDiffuseSampler::EE_DiffuseSamplerUniformWard:
    return new CDiffuseSamplerUniformWard(depth,numRays,infRayLength,
      minAllowedHitDist,incRad,data);
  case CDiffuseSampler::EE_DiffuseSamplerUniformDisplace:
    return new CDiffuseSamplerUniformDisplace(depth,numRays,infRayLength,
      minAllowedHitDist,incRad,data);
  case CDiffuseSampler::EE_DiffuseSamplerUniformDeriv:
    return new CDiffuseSamplerUniformDeriv(depth,numRays,infRayLength,
      minAllowedHitDist,incRad,data);
  case CDiffuseSampler::EE_DiffuseSamplerCosDisplace:
    return new CDiffuseSamplerCosDisplace(depth,numRays,infRayLength,
      minAllowedHitDist,incRad,data);
  case CDiffuseSampler::EE_DiffuseSamplerCosDeriv:
    return new CDiffuseSamplerCosDeriv(depth,numRays,infRayLength,
      minAllowedHitDist,incRad,data);
  case CDiffuseSampler::EE_DiffuseSamplePlaneUniform:
    return new CDiffuseSamplerPlaneUniform(depth,numRays,infRayLength,
      minAllowedHitDist,incRad,data);
  default:
    FATAL << "CreateDiffuseSampler(): Incorrect sampler type specified, "
      "returning NULL." << std::endl;
    FATAL_EXIT;
    return 0;
  }
}

__END_GOLEM_SOURCE
