// ===================================================================
// $Id: diffsampl.h,v 1.4 2005/11/02 12:58:42 xkrivanj Exp $
//
// diffsampl.h
//     Header file for the CDiffuseSampler class - sampling of the 
//     diffuse hemisphere.
//
// Class: CIncidentRadiance
//
// Initial coding by Jaroslav Krivanek, Oct 2003.

#ifndef __DIFFSAMPL_H__
#define __DIFFSAMPL_H__

// GOLEM headers
#include "configh.h"
#include "color.h"
#include "hminfo.h"

__BEGIN_GOLEM_HEADER

// forward declarations
class CIncidentRadiance;
class CContributionItem;

#define CDiffuseSamplerStratifiedBlocker  CDiffuseSamplerCosWard

inline void setz(CVector3D* v, int n)   
{ for(int i=0; i<n; i++) v[i].SetValue(0); }

// since the followin setup code is identical in each diffuse sampler, 
// I put it as a macro

#define DIFF_SAMPLER_SETUP \
  /* number of rays hitting any surface */ \
  int numHits = 0;                          \
  /* sum of inverse distances to the intersected objects */ \
  float distancesInv = 0, hitDist; \
  /* accumulated irradiance */ \
  CValues irradiance(result.GetType()); \
  /* contribution item used to gather IncidentRadiance() */ \
  CContributionItem contrib; \
  /* set the color type */ \
  contrib.color.SetRepresentation(result.GetType()); \
  /* set the interaction type */ \
  contrib.interaction = EE_REFLECTED_DIFFUSE; \
  /* contribution go through a diffuse relection, thus: */ \
  contrib.diffuseness = 1; \
  /* and this is a reflected ray */ \
  contrib.flags = CContributionItem::EE_ReflectedRay; \
  /* local frame */ \
  const CVector3D &Z = normal; \
  CVector3D X,Y;                 \
  normal.RightHandedBase(X,Y);   \
  /* ray direction in local coordinate frame */  \
  CVector3D  raydir_canonical; \
  /* minimum of all hit distances */ \
  minHitDist=MAXFLOAT; \
  /* number of color channels */ \
  int channels = result.GetCount();

#if 0
#define CHECK_HIT_DIST if(hitDist<minHitDist)minHitDist=hitDist;
#else
#define CHECK_HIT_DIST \
  if(hitDist<minHitDist)minHitDist=hitDist; \
  if(minHitDist < _minAllowedHitDist) {    \
    DEBUG << "IC:(minHitDist=" << minHitDist << ") < (_minAllowedHitDist=" << _minAllowedHitDist << ")" << std::endl; \
    /* seems that we were dangerously close to something */ \
    harmMeanDistance = -minHitDist;   \
    return false;  \
  }
#endif


/// An object for computing irradiance and irradiance gradients at a point.
/**
This class exists mainly as an interface for the (pure) virtual method 
ComputeIrradiance() and also as a storage of some precomputed data
of a particular implementation of the diffuse sampler. The purpose is to
compute irradiance at a point by sampling the hemisphere of incoming directions.
The method ComputeIrradiance() samples the hemisphere and for each generated 
direction, it calls CIncidentRadiance::IncidentRadiance() method to get the
incoming radiance. An instance of CIncidentRadiance interface is passed to
the constructor CDiffuseSampler::CDiffuseSampler() and assures the independence
of the diffuse sampler from the way the incoming radiance is computed.
It can be a simple look-up in an environment map or a call to a recursive 
ray tracer. For irradiance caching, the CIncidentRadiance::IncidentRadiance() 
casts a ray and calls the rendering application's ShadeObject2() virtual method 
to compute the incoming radiance.

The diffuse sampler is used to compute irradiance at a point. In particular,
it can be used to precompute diffuse illumination for 
scene vertices, or to bake diffuse illumination into textures. One notable
application of CDiffuseSampler is for Ward's irradiance caching 
(the 'primary method' in [Ward88]).
In irradiance caching, the use of CDiffuseSampler is as follows.
First, in the construction phase, the 
constructor CICAlgo::CICAlgo() creates
one instance of a class inherited from CDiffuseSampler for each 
recursion level according to the environment (CEnvironment).
It also creates one instance of CICache for each level,
passing the respective CDiffuseSampler to it's constructor. Second,
during rendering, the application queries 
CICAlgo::IndirectDiffuseLighting() or 
CICAlgo::Irradiance() to compute indirect diffuse lighting
using irradiance caching. The call is passed on the the 
CICache::GetIrradiance() method. Here the cache is queried for
interpolation (CRCOctree::GetIrradiance()). If the query fails (i.e.
nothing usable in the cache), the instance of CDiffuseSampler is called
(CDiffuseSampler::ComputeIrradiance()) to compute irradiance and irradiance 
gradients and they are stored in the cache.
*/
class CDiffuseSampler
{
  
protected: // data
  
  /// recursion depth of this diffuse sampler
  int _depth;
  /// number of rays to sample the diffuse hemisphere
  int _numRays;
  /// what length should be assigned to a ray that does not hit anything
  float _infRayLength;
  /// minimum allowed hit distance during hemisphere sampling
  float _minAllowedHitDist;
  /// pointer an interface implementing IncidentRadiance() method
  CIncidentRadiance *_incRad;
  
public: // types

  /// Diffuse sampler type.
  enum EDiffuseSamplerType {
    /// Stratified cos-sampling with gradients a la Ward.
    EE_DiffuseSamplerCosWard = 1,
    /// Stratified uniform sampling with gradients a la Ward.
    EE_DiffuseSamplerUniformWard = 2,
    /// uniform hemisphere sampling, gradients by hemisphere displacement
    EE_DiffuseSamplerUniformDisplace = 3,
    /// uniform hemisphere sampling, gradients by differentiation
    EE_DiffuseSamplerUniformDeriv = 4,
    /// cosine sampling, gradients by hemisphere displacement
    EE_DiffuseSamplerCosDisplace = 5,
    /// cosine sampling, gradients by differentiation
    EE_DiffuseSamplerCosDeriv = 6,
    /// plane sampling - simulation of hardware rasterization
    EE_DiffuseSamplePlaneUniform = 7,
  };
  
public: // methods
  
  /// The one and only constructor.
  /**
  Initialize the object (hey man, what did you expect!).
  
    @param depth   (in) 
      Recursion depth for which this is going to be used.
    @param numRays (in) 
      Number of rays to be shot.
    @param infRayLength (in) 
      Harmonic mean distance to be set if all rays go to infinity.
    @param incRad  (in) 
      Pointer to a particular implementation of the CIncidentRadiance interface.
      incRad->IncidentRadiance() is called by GetIrradiance() to compute the 
      incoming radiance from a given direction.
    @param data    (in) 
      Additional per-instance data.
  */
  CDiffuseSampler(
    int                 depth, 
    int                 numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance*  incRad, 
    const void*         data=0);
  
  /// Virtual destructor.
  virtual ~CDiffuseSampler() {}
  
  /// Compute the radiance at a given point and given normal.
  /**
    Actually a wrapper for the virtual function of the same name.
  
    @param inHitInfo  (in)  
      Hitpoint at which we are to compute the irradiance.
    @param outHitInfo (tmp) 
      This MUST be used to shoot the rays for hemisphere sampling.
    @param result           (out) 
      Set the resulting irradiance here. Undefined on fail.
    @param harmMeanDistance (out) 
      Harmonic mean of the distances of the hit points.
    @param mitHitDist (out) 
      Minimum hit point ditance.
    @param stdDev           (out) 
      Estimate of the standard deviation of the computed irradiance.
    @param rotGrad          (out)
      Rotational gradient for each color band.
    @param transGrad        (out)
      Translational gradient for each color band.
    
    @return  'true' if the irradiance has been succesfully computed and 
     the result stored in 'result'. Otherwise, 'false' is returned and 
     'result' is undefined.
  */
  inline bool ComputeIrradiance(
    const CHitPointInfo& inHitInfo,
    const float &footprintArea,
    CHitPointInfo& outHitInfo,
    CValues&       result,
    float&         harmMeanDistance,
    float&         minHitDist,
    float&         stdDev,
    CVector3D*     rotGrad,
    CVector3D*     transGrad,
    const void*    data=0);
  
  /// Compute the radiance at a given point and given normal.
  /**
    This is actually the function which is overloaded in the specific 
    implementations of diffuse hemisphere sampling.
    
     It's the same thing as the otherComputeIrradiance() method, but this one 
     does not need the incoming hitInfo (const CHitPointInfo& inHitInfo).
     The reason is that sometimes we may want to compute irradiance without 
     having any ray hitting that point.
	 */
  virtual bool ComputeIrradiance(
    const CVector3D & pt,
    const CVector3D & N,
    const CVector3D & uN,
    const CObject3D * object,
    const void      * termination,
    const float    & footprintArea,
    CHitPointInfo   & outHitInfo,
    CValues         & result,
    float          & harmMeanDistance,
    float          & minHitDist,
    float          &stdDev,
    CVector3D       * rotGrad,
    CVector3D       * transGrad,
    const void      * data=0) = 0;
  
  /// Returns the number of rays to be shot.
  int GetNumRays() const { return _numRays; }
  
  /// Returns the sampler's recursion depth.
  int GetDepth() const { return _depth; }
  
  /// Returns true is the implementation of diffuse sample can compute 
  /// irradiance gradients.
  virtual bool CanComputeGradients() const = 0;

protected: // methods

  /// Helper function to transform gradients to global coordinate frame.
  inline void _gradToGlobalFrame(
    CVector3D *rotGrad, 
    CVector3D *transGrad,
    const CVector3D &X,
    const CVector3D &Y,
    const CVector3D &Z,
    const int channels);

  /// Helper function to set gradients to zero.
  inline void _gradSetz(
    CVector3D *rotGrad,
    CVector3D *transGrad,
    const int channels);
};

// --------------------------------------------------------------------------
//  CDiffuseSampler::ComputeIrradiance()
// --------------------------------------------------------------------------
inline bool 
CDiffuseSampler::ComputeIrradiance(const CHitPointInfo& inHitInfo,
                                   const float &footprintArea,
		                               CHitPointInfo& outHitInfo,
                                     CValues&       result,
                                     float&         harmMeanDistance,
                                     float&         minHitDist,
                                     float          &stdDev,
                                     CVector3D*       rotGrad,
                                     CVector3D*       transGrad,
                                     const void*          data)
{
  return ComputeIrradiance(
    inHitInfo.GetPoint(),
    inHitInfo.GetNormal(),
    inHitInfo.GetGeometricNormal(),
    inHitInfo.GetObject(),
    inHitInfo.GetRay()->GetTermination(),
    footprintArea,
    outHitInfo,
    result,
    harmMeanDistance,
    minHitDist,
    stdDev,
    rotGrad,
    transGrad,
    data);
}

// --------------------------------------------------------------------------
//  CDiffuseSampler::_gradSetz()
// --------------------------------------------------------------------------
inline void 
CDiffuseSampler::_gradSetz(CVector3D *rotGrad,
                           CVector3D *transGrad,
                           const int channels)
{
  assert(rotGrad);
  assert(transGrad);
  for(int i=0; i<channels; i++) {	
    rotGrad[i]  .SetValue(0);
    transGrad[i].SetValue(0);
  }
}

// --------------------------------------------------------------------------
//  CDiffuseSampler::_gradToGlobalFrame()
// --------------------------------------------------------------------------
inline void 
CDiffuseSampler::_gradToGlobalFrame(CVector3D *rotGrad, 
                                    CVector3D *transGrad,
                                    const CVector3D &X,
                                    const CVector3D &Y,
                                    const CVector3D &Z,
                                    const int channels)
{
  assert(rotGrad);
  assert(transGrad);
  for(int i=0; i<channels; i++) {
    rotGrad[i]   = ToGlobalFrame(rotGrad[i],   X,Y,Z);
    transGrad[i] = ToGlobalFrame(transGrad[i], X,Y,Z);
  }
}

/// Base class for all samplers that divided hemisphere into strata.
class CDiffuseSamplerStratAbstr : public CDiffuseSampler {
  
protected: // data
  
  /// Number of divisions for theta.
  int   M;
  /// Number of divisions for phi.
  int   N;
  /// invM = 1/M;
  float invM;
  /// invN = 1/N;
  float invN;
  /// invRayNum = 1/(M*N)
  float invRayNum;
  
protected: // methods
  
  /// compute M and N surch that M*N = numRays and N/M = NoverM
  void _computeMN(int numRays, float NoverM);
  
public: // methods
  
  CDiffuseSamplerStratAbstr(int                depth, 
    int                numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const void*        data=0)
  : CDiffuseSampler(depth,numRays,infRayLength,minAllowedHitDist,incRad,data)
  {}
  
  virtual ~CDiffuseSamplerStratAbstr() {}
  
};


/// Common ancestor for Ward-like diffuse samplers.
/*
  Ward-like diffuse sampler computes the gradients as described
  in the 'Irradiance Gradients' paper. 
  There are two Ward samplers; cos and uniform. Cos is exactly
  from the paper, uniform is slightly modified.
 */
class CDiffuseSamplerWard : public CDiffuseSamplerStratAbstr
{
protected: // data
  
  float      M_PI_div_MN , M_PI_mul_2_div_N;
  
  float *    tan_theta;
  float *    sin_theta;
  float *    cos_theta;
  float *    sin_diff;
  float *    sin_mul_cos2;

  float *    cos_phi;
  float *    sin_phi;
  
  CVector3D* v_base_vect_canonical;
  CVector3D* u_vect_canonical;
  CVector3D* v_vect_canonical;
  
  /// incident radiance for each hemisphere cell
  CValues *_Li;
  /// reciprocal of hit distance for each hemisphere cell
  float  *_ir;
  
protected: // methods

  /// Common part of constructor for Ward samplers. 
  /// M and N must be correctly initialized before calling this.
  void _allocAndPrecomputePhi();
  
  /// index to the 2d-array
  inline int idx(int j, int k) const { return j+k*M; }

public: // methods
  
  CDiffuseSamplerWard(
    int                depth, 
    int                numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const void*        data=0);
  
  virtual ~CDiffuseSamplerWard();
  
  virtual bool CanComputeGradients() const {return true;}
};


/// Stratified sampling with irradiance gradients according to Ward.
/**
  Stratified sampling of the diffuse hemisphere is implemented exactly 
  according to Ward's papers on irradiance caching and irradiance gradients.

  Some comments on Ward's sampling: this is a cosine sampling,
  the code is actually nearly the same as if RND::hemisph_cos() was 
  used to generate the direction. hemisph_cos() uses acos() to get theta 
  but Ward uses asin() to get theta.
  This looks strange, but it DOES NOT MATTER whether acos() or asin() is 
  used:
  A = asin ( sqrt( X ) ) 
  sin^2(A) = X
  1-sin^2(A) = 1 - X
  cos^2(A) = 1 - X
  A  = acos ( sqrt( 1 - X ) )
  And since X is UNIFORMLY distributed in [0;1) the result is the same, only
  the rays are sent in a different order.
  We keep the asin() version.
*/
class CDiffuseSamplerCosWard : public CDiffuseSamplerWard
{
protected: // methods
   
  /// after spawning the rays - Ljk and rjk must be filled
  void _computeGradients(CVector3D *rotGrad,CVector3D* transGrad,int channels);
  
public: // methods
 
  CDiffuseSamplerCosWard(
    int                depth, 
    int                numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const void*        data=0);
  
  virtual ~CDiffuseSamplerCosWard() {}
  
  virtual bool ComputeIrradiance(
    const CVector3D &pt,
    const CVector3D& N,
    const CVector3D& uN,
    const CObject3D* object,
    const void* termination,
    const float &footprintArea,
    CHitPointInfo& outHitInfo,
    CValues&       result,
    float&         harmMeanDistance,
    float&         minHitDist,
    float          &stdDev,
    CVector3D*       rotGrad,
    CVector3D*       transGrad,
    const void*          data=0);
};



/// The same as 'Ward' but uses uniform hemisphere sampling.
/**
  Gradients are computed using the hemisphere division as described by Ward 
  and inplemented by CDiffuseSamplerWard, but the sampling is uniform.
*/
class CDiffuseSamplerUniformWard : public CDiffuseSamplerWard
{  
protected: // methods
  
  /// after spawning the rays - Ljk and rjk must be filled
  void _computeGradients(CVector3D *rotGrad,CVector3D* transGrad,int channels);
  
public: // methods
  
  CDiffuseSamplerUniformWard(
    int                depth, 
    int                numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const void*        data=0);
  
  virtual ~CDiffuseSamplerUniformWard() {}
  
  virtual bool ComputeIrradiance(
    const CVector3D &pt,
    const CVector3D& N,
    const CVector3D& uN,
    const CObject3D* object,
    const void* termination,
    const float &footprintArea,
    CHitPointInfo& outHitInfo,
    CValues&       result,
    float&         harmMeanDistance,
    float&         minHitDist,
    float          &stdDev,
    CVector3D*       rotGrad,
    CVector3D*       transGrad,
    const void*          data=0);
};



/// Uniform hemisphere sampling, gradients computed by hemisphere 
/// displacement
class CDiffuseSamplerUniformDisplace : public CDiffuseSamplerStratAbstr
{   
  
protected: // data
  
  /// array od sample distances
  float     *_r;
  /// array of sample directions (in global frame)
  CVector3D *_dir;
  /// array of incoming radiances
  CValues   *_Li;
  /// array of normals
  CVector3D *_normal;
  /// array of  dot products
  float     *_dot;
  
protected: // methods
  
  
public: // methods
  
  CDiffuseSamplerUniformDisplace(
    int                depth, 
    int                numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const void*        data=0);
  
  virtual ~CDiffuseSamplerUniformDisplace();
  
  virtual bool ComputeIrradiance(
    const CVector3D &pt,
    const CVector3D& N,
    const CVector3D& uN,
    const CObject3D* object,
    const void* termination,
    const float &footprintArea,
    CHitPointInfo& outHitInfo,
    CValues&       result,
    float&         harmMeanDistance,
    float&         minHitDist,
    float          &stdDev,
    CVector3D*       rotGrad,
    CVector3D*       transGrad,
    const void*          data=0);
  
  virtual bool CanComputeGradients() const {return true;}
};


/// Cosine sampling, gradients computed by hemisphere displacement
class CDiffuseSamplerCosDisplace : public CDiffuseSamplerStratAbstr
{   
  
protected: // data
  
  /// array od sample distances
  float     *_r;
  /// array of sample directions (in global frame)
  CVector3D *_dir;
  /// array of incoming radiances
  CValues   *_Li;
  /// array of normals
  CVector3D *_normal;
  /// array of  dot products
  float     *_dot;
  /// array of  cosine of light direction
  float     *_icos_theta;
  
protected: // methods
  
  
public: // methods
  
  CDiffuseSamplerCosDisplace(int                depth, 
    int                numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const void*        data=0);
  
  virtual ~CDiffuseSamplerCosDisplace();
  
  virtual bool ComputeIrradiance(
    const CVector3D &pt,
    const CVector3D& N,
    const CVector3D& uN,
    const CObject3D* object,
    const void* termination,
    const float &footprintArea,
    CHitPointInfo& outHitInfo,
    CValues&       result,
    float&         harmMeanDistance,
    float&         minHitDist,
    float          &stdDev,
    CVector3D*       rotGrad,
    CVector3D*       transGrad,
    const void*          data=0);
  
  virtual bool CanComputeGradients() const {return true;}
};


/// Uniform hemisphere sampling, gradients computed by differentiation
class CDiffuseSamplerUniformDeriv : public CDiffuseSamplerStratAbstr
{   	
protected: // data
  
protected: // methods
  
public: // methods
  
  CDiffuseSamplerUniformDeriv(int                depth, 
    int                numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const void*        data=0);
  
  virtual ~CDiffuseSamplerUniformDeriv();
  
  virtual bool ComputeIrradiance(const CVector3D &pt,
    const CVector3D& N,
    const CVector3D& uN,
    const CObject3D* object,
    const void* termination,
    const float &footprintArea,
    CHitPointInfo& outHitInfo,
    CValues&       result,
    float&         harmMeanDistance,
    float&         minHitDist,
    float          &stdDev,
    CVector3D*       rotGrad,
    CVector3D*       transGrad,
    const void*          data=0);
  
  virtual bool CanComputeGradients() const {return true;}
};


/// Cosine sampling, gradients computed by differentiation
class CDiffuseSamplerCosDeriv : public CDiffuseSamplerStratAbstr
{   	
protected: // data
  
protected: // methods
  
public: // methods
  
  CDiffuseSamplerCosDeriv(
    int                depth, 
    int                numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const void*        data=0);
  
  virtual ~CDiffuseSamplerCosDeriv();
  
  virtual bool ComputeIrradiance(
    const CVector3D &pt,
    const CVector3D& N,
    const CVector3D& uN,
    const CObject3D* object,
    const void*      termination,
    const float &footprintArea,
    CHitPointInfo&   outHitInfo,
    CValues&         result,
    float&           harmMeanDistance,
    float&           minHitDist,
    float          &stdDev,
    CVector3D*       rotGrad,
    CVector3D*       transGrad,
    const void*      data=0);
  
  virtual bool CanComputeGradients() const {return true;}
};

/// Uniform plane sampling.
/**
  This sampler simulates positioning a perspective camera at the 
  hemisphere center and rasterizing the scene with that camera.

  This is used to simulate Pascal's GPU-based implementation
  of irradiance caching.
 */
class CDiffuseSamplerPlaneUniform : public CDiffuseSampler
{   
 
protected: // data

  /// sqrt(_numRays)
  int _samplingRes;
  /// distance between the sampling center and the plane
  float _samplingNear;

  /// distance to hit
  float *_rjk;
  /// incoming radiance
  CValues *_Ljk;
  /// array of precomputed ray directions
  CVector3D *_directions;
  /// array of per-pixel weight for computing irradiance
  float *_irradWeights;
  /// array of per-pixel weight for computing translational gradient
  float *_transGradWeights;
  /// array of per-pixel weight for computing rotational gradient
  CVector3D *_rotGradWeights;

protected: // methods

  /// return an array index given image location
  int _getIdx(int j, int k) { return _samplingRes*k + j; }

  /// evaluate irradiance gradients
  bool _computeGradients(CVector3D *rotGrad,CVector3D *transGrad,int channels);
  
public: // methods
  
  CDiffuseSamplerPlaneUniform(
    int                depth, 
    int                numRays, 
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance *incRad, 
    const void        *data=0);
  
  virtual ~CDiffuseSamplerPlaneUniform();
  
  virtual bool ComputeIrradiance(
    const CVector3D  &pt,
    const CVector3D  &N,
    const CVector3D  &uN,
    const CObject3D  *object,
    const void       *termination,
    const float &footprintArea,
    CHitPointInfo    &outHitInfo,
    CValues          &result,
    float            &harmMeanDistance,
    float&         minHitDist,
    float          &stdDev,
    CVector3D        *rotGrad,
    CVector3D        *transGrad,
    const void       *data=0);
  
  virtual bool CanComputeGradients() const {return true;}
};


/// Sampling of the diffuse hemisphere implemented using stratified sampling 
/// and blocking rays.
/**
  This particular implementation of CDiffuseSampler divides the hemisphere 
  into strata (in preprocess) and the uses jittering within each stratum 
  (in ComputeIrradiance()) in the same ways as CDiffuseSamplerStratified
  do. 

  Moreover each ray is subjected to a 'blocker test': ComputeIrradiance()
  gets a pointer to an implementation of a 'CDirBlocker' interface. 
  For every generated direction,  CDirBlocker::Blocked() method is called
  to see if the ray should be kept or discarded.
  
  A particular implementation of CDirBlocker might test the direction against 
  a list of CPlanar's.    
  This is usefull e.g. for smooth geometry, where the shading normal is not 
  aligned with the geometric normal. In this case the polygon itself can 
  block some of the rays. With this technique, we keep the rays from 
  penetrating the polygon and bringing incorrect contribution.
*/
/*
class CDiffuseSamplerStratifiedBlocker : public CDiffuseSamplerStratified
{
public: // methods

  CDiffuseSamplerStratifiedBlocker(int                depth, 
  int                numRays, 
  float              infRayLength, 
  CIncidentRadiance* incRad, 
  const void*        data=0) :
  CDiffuseSamplerStratified(depth,numRays,infRayLength,incRad,data) 
  { FATAL << "Nechci blocker"; FATAL_EXIT; }
  
    virtual ~CDiffuseSamplerStratifiedBlocker() {}
    
      virtual bool ComputeIrradiance(const CVector3D &pt,
      const CVector3D& N,
      const CObject3D* object,
      const void* termination,
      const float &footprintArea,
      CHitPointInfo& outHitInfo,
      CValues&       result,
      float&         harmMeanDistance,
      const void*          data=0);
      };
*/


/// Creates an instance of the given type of a diffuse sampler.
CDiffuseSampler* 
CreateDiffuseSampler(CDiffuseSampler::EDiffuseSamplerType samplerType, 
                     int depth, 
                     int numRays, 
                     float infRayLength, 
                     float minAllowedHitDist,
                     CIncidentRadiance* incRad, 
                     const void*  data=0);

__END_GOLEM_HEADER


#endif // __DIFFSAMPL_H__
