// ==================================================================
// $Id: diffsampl2.cpp,v 1.4 2005/11/02 12:59:01 xkrivanj Exp $
//
// diffsampl2.cpp
//    Support routines for sampling diffuse hemisphere.
//    Exist to limit the size of diffsampl.cpp.
//    Used for irradiance caching.
//
// Initial coding by Jaroslav Krivanek, May 2004

// GOLEM headers
#include "config.h"
// #include "blocker.h"
#include "contrib.h"
#include "diffsampl.h"
#include "incrad.h"
#include "rnd.h"
#include "surface.h"

__BEGIN_GOLEM_SOURCE

/////////////////////////////////////////////////////////////////////////////
// Notes:
//   (theta-  - theta+) * cos(theta)  \approx  (sin(theta-) - sin(theta+))
//   (tj. rozdil uhlu nasobeny kosinem je skoro stejny jako rozdil sinu)
//   
//
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
//
//   CDiffuseSamplerUniformWard
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CDiffuseSamplerUniformWard::CDiffuseSamplerUniformWard()
// --------------------------------------------------------------------------
CDiffuseSamplerUniformWard::
CDiffuseSamplerUniformWard(int                depth, 
                           int                numRays,
                           float              infRayLength, 
                           float              minAllowedHitDist,
                           CIncidentRadiance* incRad, 
                           const void*        data)
: CDiffuseSamplerWard(depth, numRays, infRayLength, minAllowedHitDist, 
                      incRad,data)
{
  int j;

  // M ... theta divisions
  // N ... phi divisions

  _computeMN(numRays,4);

  // alloc buffers and precompute phi-related data
  _allocAndPrecomputePhi();
  
  // -------------------------
  //  precompute theta arrays
  // -------------------------

  for (j = 0; j < M; j++) {
    cos_theta[j] = 1.0 - ((float)j+0.5)*invM;
    sin_theta[j] = sqrtf(1.0f - sqr(cos_theta[j]));
    tan_theta[j] = sin_theta[j] / cos_theta[j];
    // The formula for sin_diff below is correct even if at the first sight 
    // it might not seem so. 
    // Actually, sin_diff should be simply constant for uniform sampling
    // and equal to M_PI/(2.0*M). The multiplication by cos_theta[j] stems from
    // the formula for the wall displacement for uniform sampling:
    //    sin_diff * ( L_j,k * cos_j,k  - L_j,k-1 * cos_j,k-1 ) / Min(r)
    // But because c_j,k and cos_j,k-1 are the same numbers, we can premultiply 
    // by cos_j,k here, and that's exactly what we do.
    sin_diff[j] =  1.0f/(M * sin_theta[j]); // * cos_theta[j];
    // sin_diff[j] =  1.0f/M;
    // sin_diff[j] =  M_PI/(2.0*M) * cos_theta[j];
  }


  float* sin_theta_grid  = new float[M + 1]; 
  for(j = 0; j <= M; j++) {
    float cos_theta_grid = 1.0 - (float)j*invM;
    sin_theta_grid[j]    = sqrtf(1.0f - sqr(cos_theta_grid));
    sin_mul_cos2[j]      =  sin_theta_grid[j] * cos_theta_grid; // * cos_theta_grid;
  }
  
  delete[] sin_theta_grid;
 
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerUniformWard::_computeGradients()
// --------------------------------------------------------------------------
void
CDiffuseSamplerUniformWard::_computeGradients(CVector3D *urotGrad, 
                                              CVector3D* utransGrad,
                                              int channels)
{
  int i;
  for(i = 0; i < channels; i++) {	
    urotGrad[i]  .SetValue(0);
    utransGrad[i].SetValue(0);
  }
  
  int j,k;
  
  CValues rotGrad(_Li[0].GetType()),
    transGradU(_Li[0].GetType()),
    transGradV(_Li[0].GetType());
  
  // phi intervals
  for (k = 0; k < N ; k++)
  {
    rotGrad.Set(0.0);
    transGradU.Set(0.0);
    transGradV.Set(0.0);
    
    // theta intervals
    for(j = 0; j<M ; j++) {
      // rotational gradient
      rotGrad -= (_Li[idx(j,k)] * sin_theta[j]);

      float fff = cos_theta[j];

#if 1
      // u - translational gradient
      if (j>0)
        transGradU += (_Li[idx(j,k)] - _Li[idx(j-1,k)] ) * fff *
        ( sin_mul_cos2[j] * Max( _ir[idx(j,k)] , _ir[idx(j-1,k)]) );
#endif
#if 1
      // v - translational gradient
      if (k == 0)
        transGradV += (_Li[idx(j, 0)] - _Li[idx(j, N-1)]) * fff *
        (sin_diff[j] * Max(_ir[idx(j, 0)] , _ir[idx(j,N-1)])) ;
      else
        transGradV += (_Li[idx(j, k)] - _Li[idx(j, k-1)]) * fff *
        (sin_diff[j] * Max(_ir[idx(j,k)] , _ir[idx(j,k-1)]));
#endif
    }

    // accumulate rotational gradients
    for(i=0; i<channels; i++) {
      urotGrad[i] += rotGrad.Get(i) * v_base_vect_canonical[k];		
      utransGrad[i] += 
        M_PI_mul_2_div_N * transGradU.Get(i) * u_vect_canonical[k]
        + transGradV.Get(i) * v_vect_canonical[k];
    }
  } // for k
  
  for(i=0; i<channels; i++)
    urotGrad[i] *= (2.0 * M_PI_div_MN);
  
  return;
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerUniformWard::ComputeIrradiance()
// --------------------------------------------------------------------------
bool 
CDiffuseSamplerUniformWard::ComputeIrradiance(const CVector3D &pt,
                                              const CVector3D& normal,
                                              const CVector3D& unormal,
                                              const CObject3D* object,
                                              const void* termination,
                                              const float &footprintArea,
                                              CHitPointInfo& outHitInfo,
                                              CValues&       result,
                                              float&         harmMeanDistance,
                                              float&         minHitDist,
                                              float&         stdDev,
                                              CVector3D*       rotGrad,
                                              CVector3D*       transGrad,
                                              const void*          data)
{
  DIFF_SAMPLER_SETUP;
  // reciprocal hit distance
  float invHitDist;
  // reciprocal of infinity
  float invInfty = float(1.0)/CLimits::Infinity;

  float TWO_PI_div_MN = 2.0f*M_PI/(float)(M*N);
  
  for (int k = 0; k < N; k++) {
    for (int j = 0; j < M; j++) {
      float r1 = RND::rand01f();
      float r2 = RND::rand01f();
       float cos_theta = 1.0f - ((float)j + r1) * invM;
       float sin_theta = sqrtf(1.0f - cos_theta * cos_theta);
       float phi        = 2.0 * M_PI * ( (float)k + r2 ) * invN;
       contrib.prob = 0.5f * M_1_PI;
       raydir_canonical.SetValue( cos(phi)*sin_theta, 
                                  sin(phi)*sin_theta, 
                                  cos_theta );
      // compute the direction to be traced
      CVector3D raydir  = ToGlobalFrame(raydir_canonical, X, Y, Z);   
      // initialize the ray
      contrib.ray.Init(pt, raydir, REFLECTED_RAY, termination, object, 0, 
        true);
      // unset,if it was set from previous iteration
      contrib.invalid = false;
      contrib.diffuseness = 1;
      // cast the ray an compute higher order light contribution
      if(_incRad->IncidentRadiance(outHitInfo,&contrib,_depth,footprintArea)){
        hitDist = outHitInfo.GetT();
        // check the min dist and possibly bail out
        CHECK_HIT_DIST;
        // compute reciprocal of the hit distance
        invHitDist = float(1.0)/hitDist;
        // update harmonic mean distance
        distancesInv += invHitDist;
        // if ray goes to infty, there is no contrib to distancesInv (1/infty)
        numHits++;
      }
      else
        invHitDist = invInfty;
      // index for saving incident radiance and the hit distance
      int index = idx(j,k);   
      // save reciprocal hit distance for gradient computation
      _ir[index] = invHitDist;
      // save incident radiance for gradient computation
      if(contrib.invalid)
        _Li[index].setz();
      else {
        _Li[index] = contrib.color;
        // update total irradiance
        contrib.color *= cos_theta;
        irradiance += contrib.color;
      }
      
    } // for t
  } // for p
  
  // divide by the number of rays
  irradiance *= TWO_PI_div_MN;
  // set the results
  result = irradiance;
  
  // harmonic mean distance is the inverse of mean inverses 
  if(numHits)
	   harmMeanDistance = (float)numHits / distancesInv;
  else
	   harmMeanDistance = _infRayLength;
  
  // gradients
  if(rotGrad) {
	   assert(transGrad);
     // compute translational and rotational gradients
     _computeGradients(rotGrad,transGrad,channels);
     // transform the gradients to the global coord. frame
     _gradToGlobalFrame(rotGrad,transGrad,X,Y,Z,channels);
  }

  return true;
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerPlaneUniform::CDiffuseSamplerPlaneUniform()
// --------------------------------------------------------------------------
CDiffuseSamplerPlaneUniform::
CDiffuseSamplerPlaneUniform(int                depth, 
                            int                numRays, 
                            float              infRayLength, 
                            float              minAllowedHitDist,
                            CIncidentRadiance *incRad, 
                            const void        *data)
: CDiffuseSampler(depth, numRays, infRayLength, minAllowedHitDist,incRad,data)
{
  assert(numRays>0);

  _samplingRes = (int)sqrt((double)numRays);
  _numRays = sqr(_samplingRes);

  // array allocation
  _Ljk = new CValues[_numRays];
  assert(_Ljk);
  _rjk = new float[_numRays];
  assert(_rjk);
  _directions = new CVector3D[_numRays];
  assert(_directions);
  _irradWeights=new float[_numRays];
  assert(_irradWeights);
  _transGradWeights= new float[_numRays];
  assert(_transGradWeights);
  _rotGradWeights = new CVector3D[_numRays];
  assert(_rotGradWeights);

  // Direction setup code

  _samplingNear = 0.01f;
  float sqrSamplingNear = _samplingNear*_samplingNear;
  unsigned int imgsize  = _samplingRes *_samplingRes;

  // Grid size (doubled compared to standard hemicube : like this we account 
  // for more than 80% of whole hemisphere)
  float size  = 1.0f * _samplingNear; 
  // Side length of a pixel (=distance between two pixel centers)
  float pixelSize = 2.0f * size/_samplingRes;
  // Pixel area for pixel form factor computation
  float pixelArea = pixelSize*pixelSize;
  float offset    = pixelSize*0.5f;

  // compensation for the unsampled part of the hemisphere
  // const float compensation = 1.203f;
  const float compensation = 1.15f;

  for(int k=0; k<_samplingRes; k++)
  {
    float y  = -size+offset+k*pixelSize;
    for(int j=0; j<_samplingRes; j++)
    {
      float x  = -size+offset+j*pixelSize;
      int i = _getIdx(j,k);
      // ray direction for current pixel in the local frame
      _directions[i] = Normalize(CVector3D(-x, y, _samplingNear));
      // cos(theta) for this direction (in the local frame)
      float cos_theta=_directions[i].z;
      // dist(center, pixel)
      float dist = _samplingNear/cos_theta;
      // cos(theta) / dist^2(center, pixel)
      float cos_theta_d2 = cos_theta/(dist*dist);
      // solid angle of the pixel
      float solidAngle = pixelArea * cos_theta_d2;
      // weight for computing irradiance
      _irradWeights[i] = solidAngle * cos_theta * compensation;
      // weights for translational gradient
      _transGradWeights[i] = cos_theta_d2 * pixelSize * dist;
      // precomputed weights for computing rotational gradient
      _rotGradWeights[i] = Normalize(CrossProd(CVector3D(0,0,1),_directions[i]));
      _rotGradWeights[i] *= -2.0 * sqrt(1.0f-cos_theta) * solidAngle;
    }
  }
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerPlaneUniform::~CDiffuseSamplerPlaneUniform()
// --------------------------------------------------------------------------
CDiffuseSamplerPlaneUniform::~CDiffuseSamplerPlaneUniform()
{
  delete [] _Ljk;
  delete [] _rjk;
  delete [] _directions;
  delete [] _irradWeights;
  delete [] _transGradWeights;
  delete [] _rotGradWeights;
}


// --------------------------------------------------------------------------
//  CDiffuseSamplerPlaneUniform::ComputeIrradiance()
// --------------------------------------------------------------------------
bool 
CDiffuseSamplerPlaneUniform::ComputeIrradiance(const CVector3D &pt,
                                               const CVector3D &normal,
                                               const CVector3D& unormal,
                                               const CObject3D *object,
                                               const void *termination,
                                               const float &footprintArea,
                                               CHitPointInfo &outHitInfo,
                                               CValues       &result,
                                               float         &harmMeanDistance,
                                               float&         minHitDist,
                                               float&         stdDev,
                                               CVector3D     *rotGrad,
                                               CVector3D     *transGrad,
                                               const void    *data)
{
  DIFF_SAMPLER_SETUP;
 
  for (int k=0; k<_samplingRes; k++) {
    for (int j=0; j<_samplingRes; j++) {
      int i = _getIdx(j,k);
      // compute the direction to be traced
      CVector3D raydir = ToGlobalFrame(_directions[i], X, Y, Z);   
      // initialize the ray
      contrib.ray.Init(pt, raydir, REFLECTED_RAY, termination, object, 0, 
        true);
      // unset,if it was set from previous iteration
      contrib.invalid = false;
      contrib.diffuseness = 1;
      // cast the ray an compute higher order light contribution
      if(_incRad->IncidentRadiance(outHitInfo,&contrib,_depth,footprintArea)){
        hitDist = outHitInfo.GetT();
        // check the min dist and possibly bail out
        CHECK_HIT_DIST;
        // update harmonic mean distance
        distancesInv += 1.0f / hitDist;
        // if ray goes to infty, there is no contrib to distancesInv (1/infty)
        numHits++;
      }
      else
        hitDist = CLimits::Infinity;
      // save hit distance for gradient computation
      _rjk[i] = hitDist;
      // save incoming radiance    
      if(contrib.invalid) 
        _Ljk[i].setz();
      else {
        _Ljk[i] = contrib.color;
        irradiance += _irradWeights[i] * _Ljk[i];
      }
    } // for j
  } // for k

  // set the reults
  result = irradiance;
   
  // harmonic mean distance is the inverse of mean inverses 
  if(numHits)
	   harmMeanDistance = (float)numHits / distancesInv;
  else
	   harmMeanDistance = _infRayLength;
  
  // gradients
  if(rotGrad) {
	   assert(transGrad);
     // compute translational and rotational gradients
     _computeGradients(rotGrad,transGrad,channels);
     // transform the gradients to the global coord. frame
     for(int i=0; i<channels; i++) {
       rotGrad[i]   = ToGlobalFrame(rotGrad[i],   X,Y,Z);
       transGrad[i] = ToGlobalFrame(transGrad[i], X,Y,Z);
     }
  }

  return true; // ok
}

// --------------------------------------------------------------------------
//  CDiffuseSamplerPlaneUniform::_computeGradients()
// --------------------------------------------------------------------------
bool 
CDiffuseSamplerPlaneUniform::_computeGradients(CVector3D *rotGrad,
                                               CVector3D *transGrad,
                                               int channels)
{
  assert(rotGrad);
  assert(transGrad);

  // set gradients to zero
  _gradSetz(rotGrad,transGrad,channels);

  // ----------------------------
  // 1) Rotational gradient
  // ----------------------------
#if 1
  for(int i=0; i<_numRays; i++) {
    // Contribution of this sample to the record rotational gradient
    for(int ch=0; ch<channels; ch++) 
      rotGrad[ch] += _rotGradWeights[i] * _Ljk[i].Get(ch);
  }
#endif

  // ----------------------------
  // 2) Translationnal gradient
  // ----------------------------
  CValues xGrad(_Ljk[0].GetType());
  CValues yGrad(_Ljk[0].GetType());
  xGrad.setz();
  yGrad.setz();

  for(int j=1; j<_samplingRes; j++) {
    for(int k=1; k<_samplingRes; k++) {
      int idxjk  = _getIdx(j,k);
      int idxj1k = _getIdx(j-1,k);
      int idxjk1 = _getIdx(j,k-1);

      float wx = _transGradWeights[idxjk] / min(_rjk[idxjk],_rjk[idxj1k]);
      float wy = _transGradWeights[idxjk] / min(_rjk[idxjk],_rjk[idxjk1]);

      xGrad += (_Ljk[idxjk] - _Ljk[idxj1k]) * wx;
      yGrad += (_Ljk[idxjk] - _Ljk[idxjk1]) * wy;
    }
  }

  for(int ch=0; ch<channels; ch++) {
    transGrad[ch].x = - xGrad.Get(ch); // flipped w.r.t local x-axis
    transGrad[ch].y =   yGrad.Get(ch);
    transGrad[ch].z = 0;
  }

  return true; // ok
}

__END_GOLEM_SOURCE
