// ===================================================================
// $Id: icache.cpp,v 1.4 2005/11/29 16:04:05 xkrivanj Exp $
//
// icache.cpp
//     Implementation of the CICache class.
//     Encapsulation of the irradiance caching and diffuse hemisphere sampling.
//
// Initial coding by Jaroslav Krivanek, Oct 2003.

// GOLEM headers
#include "config.h"
#include "basmacr.h"
#include "icache.h"
#include "diffsampl.h"
#include "rnd.h"
#include "polygon.h"

#include "pharr.h"  // only because of Clamp() function

__BEGIN_GOLEM_SOURCE

// forward declarations
class CHitPointInfo;

// static member initialization
template<class TRecord>
const float CCachingBase<TRecord>::aRatio = 0.17/0.5;

// --------------------------------------------------------------------------
//  CCachingBase::_neighborClamping()
// --------------------------------------------------------------------------
template<class TRecord>
void
CCachingBase<TRecord>::_neighborClamping(TRecord* newrec)
{
  static CRecordContribArray<TRecord> carray;
  carray.clear();

  bool clampedDownByNeighbor = false;
  float unclampedR  = newrec->GetUnclampedR();
  const CVector3D& P = newrec->GetPoint();
  const CVector3D& N = newrec->GetNormal();

  CRCOctree<record_type>::template FindInInfluenceArea<false>(P,N,
    newrec->GetValidDomain(),carray,-1/*not used*/);

  static const bool fwdTest = true;
  static const bool revTest = true;

  // reduce this record's radius based on neighbor radii
  if(_clampByNeighbors) {
    for(unsigned i=0; i<carray.size(); i++) {
      const TRecord* rec_i = carray[i].rec;
      float sqrDist = SqrDistance(rec_i->GetPoint(), newrec->GetPoint());
      if( sqr(unclampedR) <= sqrDist + sqr(rec_i->GetUnclampedR()) ) continue;
      float distLimit = sqrt(sqrDist) + rec_i->GetUnclampedR();
      distLimit /= DotProd(rec_i->GetN(), newrec->GetN());
      if(unclampedR > distLimit ) {
        if(!(fwdTest && rec_i->Behind(*newrec)) && 
          !(revTest && newrec->Behind(*rec_i)) )
        {
          unclampedR = distLimit;
          clampedDownByNeighbor = true;
        }
      }
    }
    newrec->SetUnclampedR(unclampedR);
    newrec->SetRClamp(unclampedR);
  }

  // reduce the radii of this record's neighbors
  if(_clampNeighbors) {
    const float sqrUnclampedR = sqr(unclampedR);
    for(unsigned i=0; i<carray.size(); i++) {
      TRecord* rec_i = carray[i].rec;
      float sqrDist = SqrMagnitude(rec_i->GetPoint() - newrec->GetPoint());
      if( sqr(rec_i->GetUnclampedR()) <= sqrDist + sqrUnclampedR ) continue;
      float distLimit = sqrt(sqrDist) + unclampedR;
      distLimit /= DotProd(newrec->GetN(),rec_i->GetN());
      if( rec_i->GetUnclampedR() > distLimit ) {
        if(!(fwdTest && newrec->Behind(*rec_i)) && 
          !(revTest && rec_i->Behind(*newrec)) ) 
        {
          rec_i->SetUnclampedR(distLimit);
          float R = Clamp(distLimit, rec_i->GetRMinus(), rec_i->GetRPlus() );
          DecreaseRadius(rec_i,R,true/*true->reinsert into tree*/);
        }
      }
    }
  }
}

// --------------------------------------------------------------------------
//  CCachingBase::_displaceCenter()
// --------------------------------------------------------------------------
template<class TRecord>
bool 
CCachingBase<TRecord>::_displaceCenter(CVector3D &modifiedP,
                              const CObject3D *object,
                              const float pixelSize)
{
  if(!CCachingMode::InsertIntoCache(GetCachingMode()))
    return false;
  if(_centerDisplaceDist<=0)
    return false;

  // is the object a polygon?
  EObjectType type = object->ID();
  if(type == EE_Triangle || type == EE_TriangleInt ||
    type == EE_Polygon  || type == EE_PolygonInt )
  {
    // Move by the current pixel size, but not more than _centerDisplaceDist.
    float edgeMinDist = std::min<float>(0.5*pixelSize,_centerDisplaceDist);
    const CPolygon *polygon = (const CPolygon*)object;
    return polygon->MovePointOffEdge(modifiedP,edgeMinDist);
  }
  else
    return false;
}

// --------------------------------------------------------------------------
//  CICache::_visualizationProfile()
// --------------------------------------------------------------------------
template<class TRecord>
inline float 
CCachingBase<TRecord>::VisualizationProfile(float sqrNormDist)
{
  return sqrNormDist>4 ? 0 : exp( -sqrNormDist );
}

// --------------------------------------------------------------------------
//  CICache::VisualizeNearestRecord()
// --------------------------------------------------------------------------
template<class TRecord>
float
CCachingBase<TRecord>::VisualizeNearestRecords(const CVector3D  &P,
                                               const CVector3D  &N,
                                               const float size) const
{
  float size2 = sqr(size);
#if 0
  // nearest neighbor - based visualization
#if 0 // stack based
  const CIrradianceRecord* rec = FindNearest(P,N,size*2);
#else // recursive - actually faster
  const CIrradianceRecord* rec = FindNearestRecursive(P,N,size*2);
#endif
  if(!rec) return 0;
  float sqrNormDist = SqrDistance(P,rec->GetPoint()) / size2;
  return VisualizationProfile(sqrNormDist);
#else

  // range based visualization
  static growvector<const TRecord*> carray;
  carray.clear();
  FindInRange(P,N,size*2,carray);

  float sum = 0;
  for(unsigned i=0; i<carray.size(); i++) {
    const TRecord* rec = carray[i];
    float sqrNormDist = SqrDistance(rec->GetPoint(), P) / size2;
    sum += VisualizationProfile(sqrNormDist);
  }
  return sum;

#endif
}

// --------------------------------------------------------------------------
//  CICache::InterpolateIrradiance()
// --------------------------------------------------------------------------
template<bool useStoredA,bool jitterWeight>
CICache::EQueryResult
CICache::InterpolateIrradiance(const CVector3D  &P,
                               const CVector3D  &N,
                               CValues          &E,
                               const float     a) const
{
  _carray.clear();
  if (!GetContributions<useStoredA,jitterWeight>(P,N,_carray,a))
    return EE_Failed; // nothing found
  // extrapolate irradiance contributions
  ExtrapolateE(_carray.begin(),_carray.end(), P, N);

  if(_carray.size()==1)
    E.SetData(_carray[0].GetE().data());
  else 
  {
    float sumWeight=0;
    CTriCol sumE((float)0.0);
    _carray.Sum(sumWeight,sumE);
    sumWeight = 1.0f/sumWeight;
    sumE *= sumWeight;

#if 0
    /// Introduce noise proportional to the variance of record contributions.
    /// The effect of doing this cancels out by supersampling.
    /// Ward's jittering in the weight function is actually a better idea.
    /// Code left here for possible futher testing.
    float ELum = sumE.Mean();

    // compute standard deviation
    float stdDev = 0;
    for(unsigned i=0; i<_carray.size(); i++) {
      float wi    = _carray[i].w;
      float EiLum = _carray[i].E.Mean();
      stdDev += sqr(EiLum-ELum)*wi;
    }
    stdDev = sqrt(stdDev*sumWeight);

    // add a random multilply of the standard deviation
    stdDev /= (1e-6 + ELum);
    float m = (1 + 1.0*RND::rand11f()*stdDev);
    sumE *= max<float>(0,m);
#endif

    E.SetData(sumE.data());
  }

#if 0 // visualize record positions
  float vizdist = 0.1;
  float qqq=0;
  for(unsigned i=0; i<_carray.size(); i++) {
    qqq += exp( -SqrDistance(P,_carray[i].rec->GetPoint())/sqr(vizdist*_carray[i].rec->GetR()));
  }
  // float qqq = 100 * VisualizeNearestRecord(point,normal, 0.1 );
  for(int ch=0; ch<3; ch++) E.Add(ch,qqq);
#endif

  return EE_Interpolated;
}

template CICache::EQueryResult CICache::InterpolateIrradiance<true,true>(
  const CVector3D  &P, const CVector3D  &N,
  CValues          &E, const float     a) const;
template CICache::EQueryResult CICache::InterpolateIrradiance<false,true>(
  const CVector3D  &P, const CVector3D  &N,
  CValues          &E, const float     a) const;
template CICache::EQueryResult CICache::InterpolateIrradiance<true,false>(
  const CVector3D  &P, const CVector3D  &N,
  CValues          &E, const float     a) const;
template CICache::EQueryResult CICache::InterpolateIrradiance<false,false>(
  const CVector3D  &P, const CVector3D  &N,
  CValues          &E, const float     a) const;


// --------------------------------------------------------------------------
//  CICache::AddIrradiance()
// --------------------------------------------------------------------------
CICache::EQueryResult
CICache::AddIrradiance(const CVector3D &P,
                       const CVector3D &N,
                       const CVector3D &uN,
                       const CObject3D *object,
                       const void      *termination,
                       const float    &footprintArea,
                       CHitPointInfo   &outHitInfo,
                       const float    &aparam,
                       const CValues   &directIllum,
                       CValues         &E,
                       CIrradianceRecord **newRecRet)
{
  assert(footprintArea>0);
  assert(aparam>0);

  // number of color channels
  int channels = E.GetCount();

  if(newRecRet) *newRecRet = 0;

  if( ! CCachingMode::SampleHemisphere(GetCachingMode()) )
    return EE_Failed; // nothing valid stored in E

  float a = _determineNewRecA(P,N,aparam);

  float harmMeanDist, minHitDist, stdDev;
  CVector3D *rotGrad=0, *transGrad=0;

  if(_useGradients) {
    rotGrad   = new CVector3D[channels];
    transGrad = new CVector3D[channels];
  }

  // get pixel size
  float pixelSize  = sqrt(footprintArea);

  // ------------------------------------------
  // move the P half a pixel away from the polygon edge
  // ------------------------------------------
  CVector3D modifiedP = P;
  bool  posModified = _displaceCenter(modifiedP,object,pixelSize);

  // -----------------------
  //  sample hemisphere
  // -----------------------
  if( !_diffSampler->ComputeIrradiance(modifiedP, _useGeometricNormal ? uN : N,
    uN, object, termination, footprintArea, outHitInfo, E, harmMeanDist, 
    minHitDist, stdDev, rotGrad, transGrad) ) 
  {
    delete [] rotGrad; delete [] transGrad;
    return EE_Failed; // nothing valid stored in E
  }

  if( !CCachingMode::InsertIntoCache(GetCachingMode()) ) {
    delete [] rotGrad; delete [] transGrad;
    return EE_NewRecord; // E contains a valid irradiance value
  }

  // Radius assignment (minHitDist vs. HMD)
  float unclampedR = _setupUnclampedR(minHitDist,harmMeanDist);

  // reduce unclampedR if gradient large
  if(_useGradients && _clampRadiusByGradients) {
    // maximum ratio between upper bound and the actual gradient
    const float gradRatio = _useMinHitDist ? aRatio : 1.0;
    // const float gradRatio =  1.0;
    // maximum gradient over all channels
    static float zeros[] = {0,0,0,0,0,0,0,0};
    bool _clampWithDirectIllum = false;
    const float *direct = _clampWithDirectIllum ? directIllum.data() : zeros;
    const float *Ev = E.data();
    float sqrGradMag = 0;
    for(int ch=0; ch<channels; ch++) {
      float g = SqrMagnitude(transGrad[ch])/sqr(Ev[ch]+direct[ch]);
      if(g>sqrGradMag) sqrGradMag = g;
    }
    if ( sqrGradMag * sqr(unclampedR) > sqr(gradRatio) ) {
      unclampedR = gradRatio/sqrt(sqrGradMag);
    }
  }

  // clamp by minDist/maxDist
  float Rminus,Rplus;
  _setRminusRplus(Rminus,Rplus, pixelSize);
  float R = Clamp(unclampedR, Rminus, Rplus);

  if(_useGradients) {
    // Rectify the gradients if radius enlarged.
    // Greg Ward does this gradient capping by the upper bound derived
    // from the split sphere model (i.e. by E/R), but this test works
    // better.
    if(unclampedR < R) {
      float m = unclampedR / R;
      for(int i=channels; i--; )
        transGrad[i] *= m;
    }

    // If the center was displaced, modify irradiance by the translational 
    // gradient.
    if(posModified) {
      CVector3D diff = P - modifiedP;
      for(int ch=0; ch<channels; ch++)
        E.Add( ch, DotProd(diff, transGrad[ch]) );
    }

    // If geometric normal used for hemisphere sampling, the resulting
    // irradiance must be rotated with gradients.
    if( _useGeometricNormal) {
      CVector3D nnorm = CrossProd(N, uN);
      for(int ch=0; ch<channels; ch++) {
        E.Add( ch, DotProd(nnorm, rotGrad[ch]) );
      }
    }

  } // if(_useGradients)


  // -------------------
  // rotation multiplier
  // -------------------
  float rMult = _rotationMultiplier();
  // increase the multiplicator for the weight error if rotation gradient high
#if 0 // still to be decided whether to use or not
  if(rotGrad) {
     // compute magnitude of the rotation gradient
    const float *Ev = E.data();
    float rotGradMag = 0;
    for(int ch=0; ch<channels; ch++)
      rotGradMag += Magnitude(rotGrad[ch])/(Ev[ch]+1e-3);
    rotGradMag /= 3;
    if(rotGradMag>1) rMult *= rotGradMag;
  }
#endif

  // Create new record.
  CIrradianceRecord *newrec = new CIrradianceRecord(P,N,E,a,
    R,unclampedR,Rminus,Rplus,rMult,stdDev,rotGrad,transGrad);

  // --------------------
  // radius clamping by the presence of overlapping records
  // (this is a heuristic to rectify some problems near corners)
  // --------------------
  if(_clampNeighbors || _clampByNeighbors)
    _neighborClamping(newrec);

  // in case R was changed in _neigborClamping()
  newrec->UnsetChanged();

  // -------------------
  //  insert into cache
  // -------------------
  InsertRecord(newrec);

  if(newRecRet)
    *newRecRet = newrec;

  // visualize record positions
  if( _visualizeNewRecPos )
    E = _visualNewRecColor; 

  return EE_NewRecord; // result contains a valid irradiance value
}


// --------------------------------------------------------------------------
//  CICache::RandomizeIrradiances()
// --------------------------------------------------------------------------
void
CICache::RandomizeIrradiances()
{
  CIrradianceList::iterator it = GetRecordList().begin();
  for( ; it!=GetRecordList().end(); it++) {
    CValues &E = (CValues&)(*it)->GetE();
    float m = 2*E.MeanValue();
    if(m==0)
      E.Set(100);
    else {
      E.setz();
      E.Set(RND::randi()%3, m);
      /*
      for(int i=0; i<E.GetCount(); i++)
        E.Set(i, RandomValue(0,m));
      */
    }
  }
}


template class CCachingBase<CRadianceRecord>;
template class CCachingBase<CIrradianceRecord>;

__END_GOLEM_SOURCE
