// ===================================================================
// $Id: icache.h,v 1.4 2005/11/02 12:58:44 xkrivanj Exp $
//
// icache.h
//     Header file for the CICache class.
//     Encapsulation of the irradiance caching and diffuse hemisphere sampling.
//
// Class: CICache
//
// Initial coding by Jaroslav Krivanek, Oct 2003.

#ifndef __ICACHE_H__
#define __ICACHE_H__

// GOLEM headers
#include "configh.h"
#include "cachmode.h"
#include "rcoctree.h"
#include "hminfo.h"

__BEGIN_GOLEM_HEADER
  
class CDiffuseSampler;
class CHitPointInfo;

// --------------------------------------------------------------------------
//  class CCachingInterface
// --------------------------------------------------------------------------
/// Common ancestor for radiance and irradiance caching.
template<class TRecord>
class CCachingBase : public CRCOctree<TRecord>
{
public:
  typedef TRecord record_type;

  /// Set the caching mode.
  inline void SetCachingMode(CCachingMode::EMode m);
  
  /// Return the caching mode.
  CCachingMode::EMode GetCachingMode(void) const { return _cachingMode; }

  CCachingBase(
    // octree parameters
    const CAABox &sceneBox,
    float octreeScale,
    // common radiance/irradiance caching parameters
    float minDist,
    float maxDist,
    float centerDisplaceDist,
    float rotWeightMultiplier,
    float nearestRecordAMul,
    bool  useGradients,
    bool  cacheLastResult,
    bool  clampInScreenSpace,
    bool  useMinHitDist,
    bool  useGeometricNormal,
    bool  clampRadiusByGradients,
    bool  clampByNeighbors,
    bool  clampNeighbors,
    int   depth) :
  CRCOctree<TRecord>(sceneBox, octreeScale),
  _minDist(minDist),
    _maxDist(maxDist),
    _useGradients(useGradients),
    _cacheLastResult(cacheLastResult),
    _rotWeightMultiplier(rotWeightMultiplier),
    _nearestRecordAMul(nearestRecordAMul),
    _centerDisplaceDist(centerDisplaceDist),
    _clampInScreenSpace(clampInScreenSpace),
    _useMinHitDist(useMinHitDist),
    _useGeometricNormal(useGeometricNormal),
    _clampRadiusByGradients(clampRadiusByGradients),
    _clampByNeighbors(clampByNeighbors),
    _clampNeighbors(clampNeighbors),
    _depth(depth),
    _jitterWeight(false),
    _visualRecColor(CStandardValues::GetDefaultColorType()),
    _useStoredA(false),
    _useNearestRecordA(false)
  {
    assert(minDist>0);
    assert(maxDist>minDist);
    assert(depth>=0);

    // visualization - no visualization by default
    _visualizeNewRecPos = false;
    _visualizeRecPos    = false;
    _visualRecSize      = 1;
    _visualRecColor.Set(1); 
  }

  /// Set parameters for record visualization.
  inline void SetRecPosVisualParams(float s,const CValues& c)
  { _visualRecSize = s; _visualRecColor = c; }

  /// Set parameters for newly added record visualization.
  inline void SetNewRecPosVisualParams(const CValues& c)
  { _visualNewRecColor = c; }

  inline void SetVisualizeRecPos(bool v) { _visualizeRecPos=v; }
  inline bool GetVisualizeRecPos() const { return _visualizeRecPos; }

  inline void SetVisualizeNewRecPos(bool v) { _visualizeNewRecPos=v; }
  inline bool GetVisualizeNewRecPos() const { return _visualizeNewRecPos; }

  inline bool GetJitterWeight() const { return _jitterWeight; }
  inline void SetJitterWeight(bool jw) { _jitterWeight=jw; }

  inline bool GetUseStoredA() const { return _useStoredA; }
  inline void SetUseStoredA(bool b) {_useStoredA = true; }

  inline bool GetUseNearestRecordA() { return _useNearestRecordA; }
  inline void SetUseNearestRecordA(bool b) {_useNearestRecordA=b;}

  /// Return the weight based on distance of nearby records according to the
  /// VisualizationProfile (@see VisualizationProfile())
  float VisualizeNearestRecords(
    const CVector3D  &P,
    const CVector3D  &N,
    const float size) const;

  /// Returns the visualization intenzity based on the normalized distance.
  static inline float VisualizationProfile(float sqrNormDist);

  /// Return value of the query methods.
  enum EQueryResult {
    /// Cache query failed.
    EE_Failed = 0,
    /// Cache query succeeded. Interpolation used.
    EE_Interpolated,
    /// Cache query succeeded. New record added.
    EE_NewRecord
  };

protected:

  bool _displaceCenter(
    CVector3D &modifiedP,
    const CObject3D *object,
    const float pixelSize);

  inline float _setupUnclampedR(
    const float minHitDist,
    const float harmMeanDist) const;

  inline void _setRminusRplus(
    float &Rminus,
    float &Rplus,
    const float pixelSize) const;

  inline float
    _determineNewRecA(
    const CVector3D &P, 
    const CVector3D &N,
    float approxError);

  /// Helper function for AddIrradiance()
  void _neighborClamping(TRecord* newrec);

  /// Return rotation multiplier
  inline float _rotationMultiplier() const
  {
    return _useMinHitDist ? _rotWeightMultiplier/aRatio : _rotWeightMultiplier;
  }

protected:

  /// minimum harmonic mean distance for a record
  const float           _minDist;
  /// maximum harmonic mean distance for a record
  const float           _maxDist;
  /// use irradiance gradients for interpolation?
  const bool             _useGradients;
  /// clamp the harmonic mean distance of a new record in screen space
  /// (as opposed to object space)
  const bool _clampInScreenSpace;
  /// use minimum of hit distances for the caching criterion 
  /// (as opposed to usig the harmonic mean of hit distances)
  const bool _useMinHitDist;
  /// Use geometric normal instead of shading normal for hemisphere sampling.
  /// This does not work and is left only for the sake of further experiments.
  const bool _useGeometricNormal;
  /// Clamp the radius of a record by the translational gradient.
  /// The idea comes from Greg Ward and Radiance code.
  const bool _clampRadiusByGradients;
  /// Clamp radius of the new record based on neighbors.
  const bool _clampByNeighbors;
  /// Clamp radius of neighbors based on the new record's radius.
  const bool _clampNeighbors;
  /// Cache the result of last query.
  const bool _cacheLastResult;
  /// use weight jittering
  bool _jitterWeight;
  /// use the value of 'a' stored with the record
  bool _useStoredA;
  /// when a new record is added, use nearest record's approximation
  /// error multiplied by 'nearestRecordAMul' instead fof the given value.
  bool _useNearestRecordA;
  /// @see _useNearestRecordA
  float _nearestRecordAMul;

  /// distance of hemisphere center displacement from the polygon edge
  const float      _centerDisplaceDist;
  /// Multiplier for the rotation error in evaluting weight of a record.
  const float     _rotWeightMultiplier;
  /// recursion depth of this irradiance cache
  const int        _depth;

  /// visualize newly added record positions
  bool   _visualizeNewRecPos;
  /// visualize record positions
  bool   _visualizeRecPos;
  /// size of the visualized record position in pixels
  float  _visualRecSize;
  /// color of the visualized record position
  CValues _visualRecColor;
  /// color of the visualized newly added record position
  CValues _visualNewRecColor;

  /// caching mode
  CCachingMode::EMode _cachingMode;

  /// Ratio between the typical value of approximation error for 
  /// R=harmonic mean distance a R = minimum hit distance.
  /**
    All heuristics in AddIrradiance() method are based on using minimum hit
    distance for R and approximation error of a=0.5.
    However, if harmonic mean distance is used for R instead of the minimum hit distance, 
    then the typical value  of approximation error is (say) a=0.17.
    This multiplicative factor compensates for this discrepancy and is used
    several times throughout AddIrradiance().
   */
  static const float aRatio;

};

// --------------------------------------------------------------------------
//  CCachingBase::SetCachingMode()
// --------------------------------------------------------------------------
template<class TRecord>
inline void 
CCachingBase<TRecord>::SetCachingMode(CCachingMode::EMode m)
{
  assert(m!=CCachingMode::EE_Same);
  assert(m!=CCachingMode::EE_Error);
	_cachingMode = m;
}

// --------------------------------------------------------------------------
//  CCachingBase::_setupUnclampedR()
// --------------------------------------------------------------------------
template<class TRecord>
inline float 
CCachingBase<TRecord>::_setupUnclampedR(const float minHitDist,
                               const float harmMeanDist) const
{

  if(_useMinHitDist) {
    // Use minimum hit distance for R. If harmonic mean distance
    // divided by aRatio (see above) is less than minimum distance, use
    // harmMeanDist/aRatio. The rationale is that the typical
    // value of a for nim hit dist is 0.5. Typical value of a for
    // HMD is 0.15.
    return std::min(minHitDist,harmMeanDist*aRatio);
    // unclampedR = minHitDist;
  }
  else {
    // Use harmonic mean distance for R.
    return harmMeanDist;
  }
}

// --------------------------------------------------------------------------
//  CCachingBase::_setRminusRplus()
// --------------------------------------------------------------------------
template<class TRecord>
inline void 
CCachingBase<TRecord>::_setRminusRplus(float &Rminus,
                                       float &Rplus,
                                       const float pixelSize) const
{
  if(_clampInScreenSpace) {
    if(pixelSize<=0) {
      FATAL 
        << "CCachingBase::_setRminusRplus(): Invalid pixelSize=" << pixelSize 
        << std::endl;
      FATAL_EXIT;
    }
    // Clamp in screen space.
    // Division by 0.5 because we want minDist/maxDist
    // work for the standard value of a=0.5.
    Rminus = _minDist * pixelSize/0.5;
    Rplus  = _maxDist * pixelSize/0.5;
    // Compensate for different typical value of 'a' for minDist/HMD.
    /*
    if(_useMinHitDist==false) {
      Rminus /= aRatio;
      Rplus  /= aRatio;
    }
    */
    // I have made the value of mindist and maxDist independent of 'a'
    // (by dividing them by 'a' just after they are read from the environment)
    // so the division here is not needed.
  }
  else {
    // clamp in object space
    Rminus = _minDist;
    Rplus  = _maxDist;
  }
}

// --------------------------------------------------------------------------
//  CCachingBase::_determineNewRecA()
// --------------------------------------------------------------------------
template<class TRecord>
inline float
CCachingBase<TRecord>::_determineNewRecA(const CVector3D &P, 
                                         const CVector3D &N,
                                         float approxError)
{
  // assign the value of 'a' of the closest record to the new rec
  if(!_useNearestRecordA)
    return approxError;
  // find the nearest record
  const TRecord *nearestRec = FindNearest(P,N);
  return min<float>(nearestRec->GetA()*_nearestRecordAMul, approxError);
}

// --------------------------------------------------------------------------
//  class CICache
// --------------------------------------------------------------------------
/// Encapsulation of the irradiance caching and diffuse hemisphere sampling.
class CICache : public CCachingBase<CIrradianceRecord>
{
protected: // data
  /// diffuse sampler to be used for the 'primary method' (i.e. when nothing is found in the cache)
  mutable CDiffuseSampler* _diffSampler;

private:
  /// just to avoid allocating every time GetIrradiance is called
  mutable CICContribArray _carray;

public:    // methods
  
  /// Constructor. 
  /**
  'diffSampler' is basically an implementation of the 'primary method', i.e.
  sampling the diffuse hemisphere.
  */
  CICache::CICache (
    // Octree parameters
    const CAABox &sceneBox,
    float octreeScale,
    // common radiance/irradiance caching parameters
    float minDist,
    float maxDist,
    float centerDisplaceDist,
    float rotWeightMultiplier,
    float nearestRecordAMul,
    bool  useGradients,
    bool  cacheLastResult,
    bool  clampInScreenSpace,
    bool  useMinHitDist,
    bool  useGeometricNormal,
    bool  clampRadiusByGradients,
    bool  clampByNeighbors,
    bool  clampNeighbors,
    int   depth,
    // specific irradiance caching parameters
    CDiffuseSampler* diffSampler) : 
    CCachingBase<CIrradianceRecord>(
    // Octree parameters
    sceneBox,
    octreeScale,
    // common radiance/irradiance caching parameters
    minDist,
    maxDist,
    centerDisplaceDist,
    rotWeightMultiplier,
    nearestRecordAMul,
    useGradients,
    cacheLastResult,
    clampInScreenSpace,
    useMinHitDist,
    useGeometricNormal,
    clampRadiusByGradients,
    clampByNeighbors,
    clampNeighbors,
    depth)
  {
    assert( diffSampler != 0);
    _diffSampler = diffSampler;
    SetCachingMode(CCachingMode::EE_Normal);
  }

  /// Interpolate irradiance at a given point.
  /**
     @param useStoredA (template parameter)
       If true, the value of approximation error is used for weight evaluation.
       If false, the value 'a' passed to the function will be used for weight 
       evaluation.
     @param jitterWeight (template parameter)
       If true, weight evaluation will be jittered, otherwise, weight will
       be deterministic.
     @param P (in) 
       Query point.
     @param N (in) 
       Surface (shading) normal at P.
     @param E (out) 
       The interpolated irradiance at P.
     @param a (in) 
       Approximation error to use. 

     @return 'false' if no usable irradiance record is found in the cache
     (the primary method - i.e. hemisphere sampling - is required).
     'true' if at least one record was found and irradiance data has been set 
     to 'irradiance' parameter.
  */
  template<bool useStoredA,bool jitterWeight>
  EQueryResult InterpolateIrradiance(
    const CVector3D  &P,
    const CVector3D  &N,
    CValues          &E,
    const float     a) const;

  EQueryResult InterpolateIrradianceDispatch(
    const CVector3D  &P,
    const CVector3D  &N,
    CValues          &E,
    const float     a) const
  {
    if(_useStoredA) {
      if(_jitterWeight) return InterpolateIrradiance<true,true>(P, N, E, a);
      else return InterpolateIrradiance<true,false>(P, N, E, a);
    }
    else {
      if(_jitterWeight) return InterpolateIrradiance<false,true>(P, N, E, a);
      else return InterpolateIrradiance<false,false>(P, N, E, a);
    }
  }

  /// Return irradiance at a given point.
  /**
    @param inHitInfo (in) represents the point at which we want to 
    compute the irradiance
  
    @param outHitInfo (in,out) this the hitinfo that HAS to be passed 
    to the diffuse sampler if the primary method is used.
    
    @param result (out)  store the irradiance here. 
      
    @return 'true' if irradiance has been evaluated and irradiance value 
    stored at 'result'. 'false' if it didn't work
    ('result' is undefined in this case).
  */
  inline EQueryResult GetIrradiance(
    const CHitPointInfo &inHitInfo,
    const float        &footprintArea,
    CHitPointInfo       &outHitInfo,
    const float        &a,
    const CValues       &directIllum,
    CValues             &E);

  /// Return irradiance at a given point.
  inline EQueryResult GetIrradiance(
    const CVector3D  &P,
    const CVector3D  &N,
    const CVector3D  &uN,
    const CObject3D* object,
    const void*      termination,
    const float     &footprintArea,
    CHitPointInfo    &outHitInfo,
    const float     &a,
    const CValues    &directIllum,
    CValues          &E);

  /// Computes and adds new irradiance record at the specified position.
  /**
     This method perform all the various heuristics for setting record radius
     and capping the gradients.
    */
  EQueryResult AddIrradiance(
    const CVector3D  &P,
    const CVector3D  &N,
    const CVector3D  &uN,
    const CObject3D* object,
    const void*      termination,
    const float     &footprintArea,
    CHitPointInfo    &outHitInfo,
    const float     &a,
    const CValues    &directIllum,
    CValues          &result,
    CIrradianceRecord **newRec);

  /// Set all irradiances to a random value (to visualize cache contents)
  void RandomizeIrradiances();

  CDiffuseSampler* GetSampler() {return _diffSampler;}
  const CDiffuseSampler* GetSampler() const {return _diffSampler;}
};


// --------------------------------------------------------------------------
//  CICache::GetIrradiance()
// --------------------------------------------------------------------------
inline CICache::EQueryResult
CICache::GetIrradiance(const CHitPointInfo &inHitInfo,
                       const float        &footprintArea,
                       CHitPointInfo       &outHitInfo,
                       const float        &a,
                       const CValues       &directIllum,
                       CValues             &E)
{
  return GetIrradiance(
    inHitInfo.GetPoint(),
    inHitInfo.GetNormal(),
    inHitInfo.GetGeometricNormal(),
    inHitInfo.GetObject(),
    inHitInfo.GetRay()->GetTermination(),
    footprintArea,
    outHitInfo,
    a,
    directIllum,
    E);
}

// --------------------------------------------------------------------------
//  CICache::GetIrradiance()
// --------------------------------------------------------------------------
inline CICache::EQueryResult
CICache::GetIrradiance(const CVector3D  &P,
                       const CVector3D  &N,
                       const CVector3D  &uN,
                       const CObject3D  *object,
                       const void       *termination,
                       const float     &footprintArea,
                       CHitPointInfo    &outHitInfo,
                       const float     &a,
                       const CValues    &directIllum,
                       CValues          &E)
{
  // return value
  EQueryResult retval = EE_Failed;

  // look up the cache
  if(  CCachingMode::LookIntoCache(GetCachingMode()) ) {
    if(_jitterWeight)
      retval=InterpolateIrradiance<false,true>(P, N, E, a);
    else
      retval=InterpolateIrradiance<false,false>(P, N, E, a);
  }

  if(retval==EE_Failed) {
    // no irradiance value found in the cache or no caching -- sample
    // hemisphere
    retval = AddIrradiance(P,N,uN,object,termination,footprintArea,
      outHitInfo,a,directIllum,E,0);
  }

  if(_visualizeRecPos) {
    float w = VisualizeNearestRecords(P,N,sqrtf(footprintArea)*_visualRecSize);
    E += _visualRecColor*w;
  }

  return retval;
}


__END_GOLEM_HEADER

#endif // __ICACHE_H__
