// ===================================================================
// $Id: icadapt.cpp,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// icadapt.cpp
//   Perceptually-based adaptive quality setting for irradiance cache.
//
// Initial coding by Jaroslav Krivanek (March, 2005).

// GOLEM headers
#include "config.h"
#include "basstr.h"
#include "hinflite.h"

#include "icadapt.h"
#include "icalgo.h"
#include "rcalgo.h"
#include "icache.h"
#include "icontrib.h"
#include "icsplat.h"
#include "jkimage.h"
#include "matrix4.h"
#include "timer.h"
#include "vdm.h"

#include "fileio.h"
#include "diffsampl.h"

// SPHAR headers
#include "shvec.h"
#include "shbrdf.h"
#include "shrot.h"

// standard headers
#include <assert.h>

#define USE_SPLATTING

__BEGIN_GOLEM_SOURCE

// #define USE_TONE_MAPPING

/*
TODO:
  Zkusit modifikovane perceptualni kriterium:
    misto: jak se prispevek recordu lisi od prumeru pouzit
      jak se od sebe lisi odhad pkud se record pouzije nebo ne
 */

// --------------------------------------------------------------------------
//  CAdaptiveCaching::CAdaptiveCaching()
// --------------------------------------------------------------------------
CAdaptiveCaching::CAdaptiveCaching(EConvergenceAlgo ca,
                                   CICAlgo *icAlgo,
                                   float maxErrorIC,
                                   CRCAlgo *rcAlgo,
                                   float maxErrorRC,
                                   bool adaptive,
                                   float mappingCoeff,
                                   float mappingGamma) :
_adaptive(adaptive),
_tmpVec1(0),
_tmpVec2(0),
_hdrFileType("pic"),
_icAlgo(icAlgo),
_rcAlgo(rcAlgo),
_maxErrorIC(maxErrorIC),
_maxErrorRC(maxErrorRC),
_jitterWeightIC(icAlgo?icAlgo->GetJitterWeight():false),
_jitterWeightRC(rcAlgo?rcAlgo->GetJitterWeight():false),
_tmpDirectIllum(CStandardValues::GetDefaultColorType())
{
#ifdef USE_TONE_MAPPING
  // minimum delta L is one pixel
  _linScale = mappingCoeff;
  _gamma = 1.0/mappingGamma;
  _gammaCorrect = (mappingGamma != 1);
  _minDeltaL = 0.5/255.0;
#else
  _linScale = 1;
  _gamma = 1;
  _gammaCorrect = false;
  _minDeltaL = 0.5/(255*mappingCoeff);
#endif

  _convergenceAlgo = ca;
  ResetStats();

  // default setting
  _saveDebugImages = false;
  _useThresholdElevation = true;
  _useMasking = false;

  _sdiStartRecords = 0;
  _debugImageIndex = 0;
  _debugImageSaveFrequency = 100;

  tmpE.SetRepresentation(CStandardValues::GetDefaultColorType());

  // _pixContribImage = 0;
  // _pixContribImageLen = 0;

  if(_rcAlgo) {
    if(_rcAlgo->GetNumChannels() != 3) {
      FATAL 
        << "This implementation only works with three color channels." 
        << std::endl;
      FATAL_EXIT;
    }
    _tmpVec1 = new CCoefVec<float>(3,_rcAlgo->GetOrder());
    assert(_tmpVec1);
    _tmpVec2 = new CCoefVec<float>(3,_rcAlgo->GetOrder());
    assert(_tmpVec2);
  }
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::~CAdaptiveCaching()
// --------------------------------------------------------------------------
CAdaptiveCaching::~CAdaptiveCaching()
{
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::CreateFileName()
// --------------------------------------------------------------------------
string 
CAdaptiveCaching::CreateFileName(const string &postfix)
{
  string s1,s2;
  PrefixExtensionByName(GetImgBaseName(),postfix,s1);
  ChangeFilenameExtension(s1,_hdrFileType,s2);
  return s2;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::IndirectIlluminationOverlap()
// --------------------------------------------------------------------------
void 
CAdaptiveCaching::IndirectIlluminationOverlap(
  float              *resultImage,
  const CHitInfoLite *hitInfoImage,
  int                 imgWidth,
  int                 imgHeight,
  CHitPointInfo      &outHitInfo,
  CMatrix4           &projMat,
  CMatrix4           &modelviewMat)
{
  // turn off copying approximation error from the nearest record
  // (because AdaptiveCaching handles this in its own way)
  bool icUseNearestRecA = false, rcUseNearestRecA = false;
  if(_icAlgo) {
    icUseNearestRecA = _icAlgo->GetICache(0)->GetUseNearestRecordA();
    _icAlgo->GetICache(0)->SetUseNearestRecordA(false);
  }
  if(_rcAlgo) {
    rcUseNearestRecA = _rcAlgo->GetRCache(0)->GetUseNearestRecordA();
    _rcAlgo->GetRCache(0)->SetUseNearestRecordA(false);
  }

  // compute the viewer position
  CVector3D at, up;
  GluUnLookAt(modelviewMat, _viewPos, at, up);

  switch(_convergenceAlgo) 
  {
  case EE_ConvergenceAlgoNonAdaptive:
    IndirectIlluminationNonAdaptive(resultImage,
      hitInfoImage,imgWidth,imgHeight,outHitInfo,projMat,modelviewMat);
    break;
  case EE_ConvergenceAlgoPerPixel:
    IndirectIlluminationPerPixel(resultImage,
      hitInfoImage,imgWidth,imgHeight,outHitInfo,projMat,modelviewMat);
    break;
  case EE_ConvergenceAlgoOnePass:
    IndirectIlluminationOverlapOnePass(resultImage,
      hitInfoImage,imgWidth,imgHeight,outHitInfo,projMat,modelviewMat);
    break;
  case EE_ConvergenceAlgoList:
    IndirectIlluminationOverlapConverge(resultImage,
      hitInfoImage,imgWidth,imgHeight,outHitInfo,projMat,modelviewMat);
    break;
  case EE_ConvergenceAlgoTree:
    IndirectIlluminationOverlapConvergeTree(resultImage,
      hitInfoImage,imgWidth,imgHeight,outHitInfo,projMat,modelviewMat);
    break;
  case EE_ConvergenceAlgoBBox:
    if(_icAlgo) {
      IndirectIlluminationOverlapConvergeBBox<CIrradianceRecord>(resultImage,
        hitInfoImage,imgWidth,imgHeight,outHitInfo,projMat,modelviewMat);
    }
    if(_rcAlgo) {
      IndirectIlluminationOverlapConvergeBBox<CRadianceRecord>(resultImage,
        hitInfoImage,imgWidth,imgHeight,outHitInfo,projMat,modelviewMat);
    }
    break;
  case EE_ConvergenceAlgoBBoxIrreg:
    if(_icAlgo) {
      IndirectIlluminationOverlapConvergeBBoxIrreg<CIrradianceRecord>(
        resultImage,hitInfoImage,imgWidth*imgHeight,outHitInfo);
    }
    if(_rcAlgo) {
      IndirectIlluminationOverlapConvergeBBoxIrreg<CRadianceRecord>(
      resultImage,hitInfoImage,imgWidth*imgHeight,outHitInfo);
    }
    break;
  default:
    FATAL << "Unexpected convergence algo type." << std::endl;
    FATAL_EXIT;
  }

  // restore the original settings
  if(_icAlgo) {
    _icAlgo->GetICache(0)->SetUseNearestRecordA(icUseNearestRecA);
  }
  if(_rcAlgo) {
    _rcAlgo->GetRCache(0)->SetUseNearestRecordA(rcUseNearestRecA);
  }

}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_updateStatDecreaseA()
// --------------------------------------------------------------------------
inline void 
CAdaptiveCaching::_updateStatDecreaseA(float newA)
{ 
  _numDecreaseA++;
  if( newA < _minA )
    _minA = newA;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_updateStatMinNewRecA()
// --------------------------------------------------------------------------
inline void 
CAdaptiveCaching::_updateStatMinNewRecA(float newRecA)
{
  if( newRecA < _minNewRecA )
    _minNewRecA = newRecA;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::ResetStats()
// --------------------------------------------------------------------------
void 
CAdaptiveCaching::ResetStats()
{
  _numBrdfRetrieval  = 0;
  _numProcessOverlap = 0;
  _numDecreaseA      = 0;
  _numICPixels       = 0;
  _numContribs       = 0;
  _maxNumContribs    = 0;
  _minNewRecA        = MAXFLOAT;
  _minA              = MAXFLOAT;
  _numFilledPixels   = 0;
  _numRecFilling     = 0;
  _totalTime         = 0;
  _addRecTotalTime   = 0;
  _numRec            = 0;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::DoReport()
// --------------------------------------------------------------------------
bool 
CAdaptiveCaching::DoReport(ostream &app)
{
  app
    << "\n"
    << "+--------------------------------------------------------------------------+\n"
    << "|        Adaptive Irradiance Caching Statistics                            |\n"
    << "+--------------------------------------------------------------------------+\n"
    << "  TotalTime:                          " << _totalTime << std::endl
    << "  AddRecTotalTime:                    " << _addRecTotalTime  << std::endl
    << "  NumRec:                             " << _numRec  << std::endl
    << "| Num Process Overlap Pixels:         " << GetNumProcessOverlap();
  if(_numICPixels>0) {
    app 
      << " (=" << 100.0f*(float)GetNumProcessOverlap()/_numICPixels
      << "% of IC pixels).\n";
  }
  else 
    app << "\n";
  app
    << "| Num BRDF Retrieval:                 " << _numBrdfRetrieval <<"\n"
    << "| Num Decrease Approx Error:          " << GetNumDecreaseA() << "\n"
    << "| Num IC Pixels:                      " << _numICPixels << "\n"
    << "| Num contributions:                  " << _numContribs ;
  if(_numICPixels)
    app << " (=" << (float)_numContribs/(float)_numICPixels << "contrib/pix)\n";
  else
    app << "\n";
  app
    << "| Max num contributions:              " << _maxNumContribs << "\n"
    << "| Minimum approx error for a new rec: " << _minNewRecA << "\n"
    << "| Minimum approx error:               " << _minA << "\n"
    << "| Num interpolated pixels:            " << _numFilledPixels << "\n"
    << "| Num records in hole filling:        " << _numRecFilling << std::endl;
     
    // << "| Image resolution: " << imgWidth << " x " << imgHeight << " = " 
    // << (long)imgWidth*(long)imgHeight << " pixels.\n";

    return false; // ok
}


// --------------------------------------------------------------------------
//  CAdaptiveCaching::_lumFromRGBTonemap()
// --------------------------------------------------------------------------
inline float 
CAdaptiveCaching::_lumFromRGBTonemap(const float *c)
{
  CTriCol cc(c);
#ifdef USE_TONE_MAPPING
  cc *= _linScale;
  if(_gammaCorrect)
    for(int ch=0; ch<3; ch++)
      cc[ch] = pow(cc[ch],_gamma);
  cc.Clamp(0,1);
#endif
  return CVDM::LumFromRGB(cc.data());
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_processPixelOverlap()
// --------------------------------------------------------------------------
template<class TRecord,class TContribArray>
int
CAdaptiveCaching::_processPixelOverlap(const CHitInfoLite &hit,
                                       const float *directIllum,
                                       const float  thresholdElevation,
                                       TContribArray &carray,
                                       CRCOctree<TRecord> *octree)
{
  const unsigned MAX_CONTRIBS=1000;
  if(carray.size()>MAX_CONTRIBS) {
    FATAL << "Too many contributions. Reimplement this code." << std::endl;
    FATAL_EXIT;
  }
  // update stats
  _numProcessOverlap ++;
  int numDecreaseABefore = _numDecreaseA;
  // threshold contrast for this pixel
  float mdot = 1.0/(fabs(GetDotForThreshold<TRecord>(hit))+1e-5);
  if(mdot>4) { mdot = 4;}
  const float thresholdContrast = 
    GetMaxError<TRecord>() * thresholdElevation * mdot ;

  // weighted average
  float  sumWeight=0;
  CTriCol sumE(0.0);
  // SumContributions(carray,sumWeight,sumE);
  // active records (those, that haven't been elliminated yet)
  bool  active[MAX_CONTRIBS];
  // number of contributions in the contribution array
  unsigned arrayLen  = carray.size();
  // number of active records
  unsigned numActive = 0;
  // array of luminances
  float LEi[MAX_CONTRIBS];
  // loop counter
  unsigned i;
  // average standard deviation
  float avgStdDev = 0;
  // compute the luminances, set all records active
  typename TContribArray::iterator it = carray.begin();
  for(i=0; it!=carray.end(); it++, i++) {
    if(!IsInvalid(*it)) {
      active[i] = true;
      numActive++;
      CTriCol Ei = GetE(*it) + directIllum;
      LEi[i] = _lumFromRGBTonemap(Ei.data());
      sumWeight += W(*it);
      sumE += W(*it) * GetE(*it);
      avgStdDev += Rec(*it)->GetStdDev();
    }
    else
      active[i] = false;
  }
  avgStdDev /= numActive;

  // repeat until there are at least two records active
  while(numActive >= 2) {

    // compute irradiance as the weighted sum
    float   iWeight = 1.0f/sumWeight;
    CTriCol E = sumE*iWeight + directIllum;
    // compute luminace channel of irradiance
    float LE = _lumFromRGBTonemap(E.data());

    // find the index of the errorneous contribution with minimum weight
    for(it=carray.begin(), i=0; it!=carray.end(); it++, i++) {
      if(!active[i]) continue;
      // find the luminance contrast
#if 1
      // difference between irradiance if the i-it record is and is not used
      float iw   = 1.0/(sumWeight-W(*it));
      CTriCol EEi = (sumE-GetE(*it) * W(*it))*iw + directIllum;
      float L2 = _lumFromRGBTonemap(EEi.data());
#else
      // difference between the total irradiance and the irradiance of the
      // i-th record
#### float L2 = CVDM::LumContrast( LE, LEi[i] );
#endif

      float stdDev = std::max(Rec(*it)->GetStdDev(),avgStdDev);
      stdDev *= _linScale;
      float deltaL = max(thresholdContrast * LE, _minDeltaL);
      deltaL += stdDev;
      // does the contrast exceed a threshold?
      if( fabsf(LE-L2) > deltaL )
        break; // we have found a visible difference
    }
  
    // no errorneous contributions - we're done
    if( i==arrayLen )
      break;

    // there is a visible error in the overlap 
    // - find and remove the record with minimum weight
    float minWeight = MAXFLOAT;
    unsigned idxMinWeight = 0;
    typename TContribArray::iterator iterMinWeight;
    for(it=carray.begin(), i=0; it!=carray.end(); it++, i++) {
      if(active[i] && W(*it)<minWeight) {
        minWeight = W(*it);
        idxMinWeight = i;
        iterMinWeight = it;
      }
    }

    // the record to be excluded
    typename TContribArray::reference item = *iterMinWeight;
    // find the value of 'a' that will just exclude the record
    float newA = Rec(item)->FindJustExclusionA(hit.P,hit.N);
    // decrease 'a' for the record
    octree->DecreaseApproxError(Rec(item), newA, true);
    // update stats 
    _updateStatDecreaseA(newA);

    // update the sums
    sumWeight -= W(item);
    sumE      -= GetE(item)*W(item);
    // the contribution is not active anymore
    active[idxMinWeight] = false;
    numActive--;
  }

  // return the number of rectified records
  return _numDecreaseA - numDecreaseABefore;
}

template int 
CAdaptiveCaching::_processPixelOverlap<CIrradianceRecord,CICContribArray>(
  const CHitInfoLite &hit,const float *directIllum,
  const float  thresholdElevation,CICContribArray &carray,
  CRCOctree<CIrradianceRecord> *octree);

template int 
CAdaptiveCaching::_processPixelOverlap<CIrradianceRecord,CPixICContribPtrList>(
  const CHitInfoLite &hit,const float *directIllum,
  const float thresholdElevation,CPixICContribPtrList &carray,
  CRCOctree<CIrradianceRecord> *octree);

template int 
CAdaptiveCaching::_processPixelOverlap<CIrradianceRecord,CPixICContribList>(
  const CHitInfoLite &hit,const float *directIllum,
  const float thresholdElevation,CPixICContribList &carray,
  CRCOctree<CIrradianceRecord> *octree);

template int 
CAdaptiveCaching::_processPixelOverlap<CRadianceRecord,CPixRCContribList>(
  const CHitInfoLite &hit,const float *directIllum,
  const float thresholdElevation,CPixRCContribList &carray,
  CRCOctree<CRadianceRecord> *octree);

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_determineNewRecA()
// --------------------------------------------------------------------------
template<class TRecord>
inline float
CAdaptiveCaching::_determineNewRecA(const CHitInfoLite &hit, 
                                    float approxError,
                                    CRCOctree<TRecord> *octree)
{
  float newRecA;
  // find the nearest record
  const TRecord *nearestRec = octree->FindNearest(hit.P,hit.N);
  // assign the value of 'a' of the closest record to the new rec
  if(nearestRec) 
    newRecA = min<float>(nearestRec->GetA()*1.4, approxError);
  else
    newRecA = approxError;
  // update stats
  _updateStatMinNewRecA(newRecA);
  return newRecA;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_addRec()
// --------------------------------------------------------------------------
template<class TRecord>
TRecord*
CAdaptiveCaching::_addRec(const CHitInfoLite &hit,
                          CHitPointInfo      &outHitInfo)
{
  FATAL << "Only specialized version of _addRec() can be called.\n";
  FATAL_EXIT;
  return 0; // dummy return to satyisfy the compiler
}

template<>
CIrradianceRecord*
CAdaptiveCaching::_addRec<CIrradianceRecord>(const CHitInfoLite &hit,
                                             CHitPointInfo      &outHitInfo)
{
  // sanity checking
  assert(_icAlgo);
  if(!hit.IsSet(CHitInfoLite::EE_IrrCaching)) {
    FATAL << "Adding irradiance record for hit point without EE_IrrCaching.\n";
    FATAL_EXIT;
  }

  float approxError = _icAlgo->_approxErrorCurrent[0];
  CICache *icache   = _icAlgo->GetICache(0);
  float newRecA = _determineNewRecA(hit,approxError,icache);
  // create the new rec
  CIrradianceRecord *newRec=0;
  _tmpDirectIllum.SetData(hit.directIllum.data());
  icache->AddIrradiance(hit.P,hit.N,hit.uN,hit.GetObject(),hit.GetTermination(),
    hit.footprintArea,outHitInfo,newRecA,_tmpDirectIllum,tmpE,&newRec);
  return newRec;
}

template<>
CRadianceRecord*
CAdaptiveCaching::_addRec<CRadianceRecord>(const CHitInfoLite &hit,
                                           CHitPointInfo      &outHitInfo)
{
  // sanity checking
  assert(_rcAlgo);
  if(!hit.IsSet(CHitInfoLite::EE_RadCaching)) {
    FATAL << "Adding radiance record for hit point without EE_RadCaching.\n";
    FATAL_EXIT;
  }

  float approxError = _rcAlgo->GetApproxErrorCurrent(0);
  CRCache *rcache   = _rcAlgo->GetRCache(0);
  float newRecA = _determineNewRecA(hit,approxError,rcache);
  // create the new rec
  CRadianceRecord *newRec=0;

  int desiredOrder = Max(
    hit.shRep->GetStats().order99Quantile,
    hit.shRep->GetOrderUV(hit.brdf_u,hit.brdf_v));

  rcache->AddRadiance(hit.P,hit.U,hit.V,hit.N,hit.uN,hit.GetObject(),
    hit.GetTermination(),hit.footprintArea,outHitInfo,newRecA,
    desiredOrder, 0, 0, &newRec);
  return newRec;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::IndirectIlluminationOverlap()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::IndirectIlluminationOverlapOnePass(
  float              *resultImage,
  const CHitInfoLite *hitInfoImage,
  int                 imgWidth,
  int                 imgHeight,
  CHitPointInfo      &outHitInfo,
  CMatrix4           &projMat,
  CMatrix4           &modelviewMat)
{
  // ------------------------------------------
  // 1) SETUP
  // ------------------------------------------
  STATUS << "- adaptive caching, convergence algo: one_pass" << std::endl;
  float approxError = _icAlgo->_approxErrorCurrent[0];
  // allocate weight buffer
  float *weightImage = new float[imgWidth*imgHeight];
  assert(weightImage);
  // init the splatter
  CICSplatterRegular splatter(projMat,modelviewMat,imgWidth,imgHeight,hitInfoImage,
    CHitInfoLite::EE_IrrCaching);
  splatter.SetWeightImage(weightImage);
  splatter.SetIrradImage(resultImage);
  // temporary to store the answer from GetIrradiance()
  CValues E(CStandardValues::GetDefaultColorType());
  // list of contributions for a given pixel
  CICContribArray carray;
  // irradiance cache interface for direct illumination
  CICache *icache = _icAlgo->GetICache(0);
  // number of image pixels
  int nPixels = imgWidth * imgHeight;
  // pixel list for image traversal
  CPixelList pixelList;
  FillPixelList(pixelList,splatter);
  const CPixelList::TPixelArray &pixels      = pixelList.pixels;
  const CPixelList::TPixelLevelArray &levels = pixelList.levels;
  // threshold elevation data
  const float *thrElevImage=0;
  jkimage<float> auxElevationImage;
  CVDMRamasubramanian *vdm=0;
  // auxiliary canvas for saving debug images
  float *auxImage = 0;
  // auxiliary weight canvas for saving debug images
  float *auxWeightImage = 0;
  // auxiliary splatter for saving debug images
  CICSplatterRegular *auxSplatter = 0;
  if(_saveDebugImages) {
    auxImage       = new float[imgWidth*imgHeight*3];
    assert(auxImage);
    auxWeightImage = new float[imgWidth*imgHeight];
    assert(auxWeightImage);
    auxSplatter = new CICSplatterRegular(projMat,modelviewMat,imgWidth,imgHeight,
      hitInfoImage,CHitInfoLite::EE_IrrCaching);
    assert(auxSplatter);
    auxSplatter->SetWeightImage(auxWeightImage);
    auxSplatter->SetIrradImage(auxImage);
  }
  // turn off smooth weight for the convergence
  bool smoothWardWeight = CIrradianceRecord::smoothWardWeight;
  CIrradianceRecord::smoothWardWeight = false;

  CTimer timerTotal, timer;

  STATUS << "- starting overlap IC filling..." << std::endl;
  timerTotal.Start();

  // ------------------------------------------
  // 2) INITIAL CACHE FILLING
  // ------------------------------------------
  STATUS << "  - initial irradiance cache filling...";
  timer.Restart();

  // we need smooth weighting for the initial pass
  // (because of the perceptual metrics which is sensitive to
  // high frequency)
  CIrradianceRecord::smoothWardWeight = true;
  // fill holes in the indirect illum image
  CValues avgE(CStandardValues::GetDefaultColorType());

  int numAddedRec,pass=0;
  do {
    // numAddedRec = FillHoles(pixelList,approxError,splatter,outHitInfo,avgE);
    numAddedRec = FillHolesICQuery(pixelList,approxError,splatter,outHitInfo,avgE);
    STATUS<<"    - pass "<<pass<<", # added recs "<<numAddedRec<<std::endl;
    pass ++;
  } while(0 /*numAddedRec>0*/ );

  STATUS
    << "done in " << timer.UserTime() << " sec. / added " 
    << numAddedRec << " records." << std::endl;

  // ------------------------------------------
  // 3) SPATIAL COMPONENT OF THE PERCEPTUAL METRIC
  // ------------------------------------------

  if(_useThresholdElevation) {

#if 0 // use the splatted irradiance for spatial component precomputation
    // normalize the splat buffer
    splatter.NormalizeBuffer();
    // multiply by the kdImage and add directIllumImage
    shv<float>::mulvvip(nPixels*3, resultImage, kdImage);
    shv<float>::addvvip(nPixels*3, resultImage, directIllumImage);
#else // use the average irradiance for spatial component precomputation
    const float *avgIrrad = avgE.data();
    float *resultImagePtr = resultImage;
    const CHitInfoLite *hitPtr = hitInfoImage;
    for(int k=0; k<nPixels; k++, resultImagePtr+=3, hitPtr++) {
      for(int ch=0; ch<3; ch++)
        resultImagePtr[ch]=hitPtr->directIllum[ch]+hitPtr->kd[ch]*avgIrrad[ch];
    }
#endif
    if(_saveVDMInOut) {
      WriteImageTiff(imgWidth,imgHeight,3,resultImage,"vdm-input.tif");
    }
    // convert to luminances image
    auxElevationImage.resize(imgWidth,imgHeight);
    for(int i=0; i<nPixels; i++) {
      auxElevationImage[i]=CVDM::LumFromRGB(resultImage+3*i);
    }
    // build the threshold elevation map
    vdm = new CVDMRamasubramanian(auxElevationImage,_useMasking,true,
      _saveVDMDebugImages);
    assert(vdm);
    thrElevImage = vdm->GetElevationImage().data();
    // release memory allocated for the luminance image
    auxElevationImage.resize(0,0);
    if(_saveVDMInOut)
      WriteImageTiff(imgWidth,imgHeight,1,thrElevImage,"vdm-output.tif");
  }
  else {
    // no threshold elevation
    // fill the threshold elevation image with ones
    auxElevationImage.resize(imgWidth,imgHeight);
    auxElevationImage.set(1);
    thrElevImage = auxElevationImage.data();
  }

  // revert to normal weighting
  CIrradianceRecord::smoothWardWeight = false;

  // ------------------------------------------
  // 2) IMAGE TRAVERSAL
  // ------------------------------------------

  // number of records rectified in one pass of a level
  int numRectified = 0;

  unsigned level;
  for(level=0; level<levels.size(); level++) {

    int pass = 0;
    do { // while (numRectified>0)
      // reset the number of rectified pixels
      numRectified = 0;
      // reset the timer
      timer.Reset(); timer.Start();

      // traverse all pixel in the current subdivision level
      for(int p=levels[level].first; p<levels[level].second; p++)
      {
        //////////////////////////////////////////////
        // index of the current image position
        TPixelIndex idx = pixels[p];
        // current hit point info
        const CHitInfoLite &hit = hitInfoImage[idx];

        carray.clear();
        if(icache->GetContributions<true,false>(hit.P,hit.N,carray,-1/*not used*/)) {
          if(carray.size()>1) {
            FATAL << "sem se asi musi dat extrapolateEValid()" << std::endl;
            FATAL_EXIT;
            numRectified += _processPixelOverlap(hit,hit.directIllum.data(),
              thrElevImage[idx], carray, icache);
          }
        }
        else {
          _addRec<CIrradianceRecord>(hit,outHitInfo);
        }
        ///////////////////////////////////////////////
      }

      timer.Stop();

      if(_saveDebugImages) {
        // save the refined image into a file
        CreateFinalImage(*auxSplatter,icache,auxImage,auxWeightImage);
        char imgName[50];
        sprintf(imgName,"resultImage-level%03d-pass%03d.tif",level,pass);
        WriteImageTiff(imgWidth,imgHeight,3,auxImage,imgName);

        CreateApproxErrImage(*auxSplatter,icache,auxImage,auxWeightImage);
        sprintf(imgName,"approxImage-level%03d-pass%03d.tif",level,pass);
        WriteImageTiff(imgWidth,imgHeight,3,auxImage,imgName);
      }

      char report[500];
      sprintf(report,"  - level #%2d (pass %2d): numRectified = %4d, "
        "time = %.3f", level, pass, numRectified, timer.UserTime());
      STATUS << report << std::endl;

      // next pass of this level
      pass ++;

      // numRectified=0;

    } while( numRectified>100 );

  }

  // ------------------------------------------
  // 2) FINISH-UP
  // ------------------------------------------
  // icache->RandomizeIrradiances();

  // clear the weight buffer and the irradiance buffer
  STATUS 
    << "  - splatting cache contents (" << icache->GetRecordList().size()
    << " records)...";
  timer.Reset(); timer.Start();

  // increase 'a' a bit
  icache->MulApproxError(1.05);

  shvf::setz(nPixels,  weightImage);
  shvf::setz(nPixels*3,resultImage);
  // splat cache contents
  splatter.SetDrawCenter(false);
  
  CIrradianceRecord::smoothWardWeight = smoothWardWeight;

  if(_jitterWeightIC)
    splatter.Splat<true>(icache->GetRecordList());
  else
    splatter.Splat<false>(icache->GetRecordList());

  timer.Stop();
  STATUS << "done in " << timer.UserTime() << " sec." << std::endl;

  STATUS << "  - filling holes...";
  timer.Reset(); timer.Start();

  // fill holes in the indirect illum image
  FillHolesPost(pixelList,splatter,outHitInfo);

  STATUS 
    << "done in " << timer.UserTime() << " sec. / added " 
    << _numRecFilling << " records." << std::endl;


  if( _saveDebugImages ) {
    // save the refined image into a file
    CreateFinalImage(*auxSplatter,icache,auxImage,auxWeightImage);
    char imgName[50];
    sprintf(imgName,"resultImage-level%03d-pass%03d-holesfilled.tif",level,0);
    WriteImageTiff(imgWidth,imgHeight,3,auxImage,imgName);

    CreateApproxErrImage(*auxSplatter,icache,auxImage,auxWeightImage);
    sprintf(imgName,"approxImage-level%03d-pass%03d-holesfilled.tif",level,0);
    WriteImageTiff(imgWidth,imgHeight,3,auxImage,imgName);
  }

  // normalize the splat buffer
  splatter.NormalizeBuffer();
  // multiply by the kd
  MulByKd(nPixels,resultImage,hitInfoImage);

  OverlapStats(splatter);

  timerTotal.Stop();
  STATUS 
    << "- overlap IC filling finished in " << timerTotal.UserTime() 
    << " sec." << std::endl;

  // delete allocated buffers
  delete [] weightImage;
  delete vdm;

  if(_saveDebugImages) {
    delete [] auxImage;
    delete [] auxWeightImage;
    delete auxSplatter;
  }
}


// --------------------------------------------------------------------------
//  CAdaptiveCaching::_fillSinglePixel()
// --------------------------------------------------------------------------
inline bool
CAdaptiveCaching::_fillSinglePixel(int x,int y, 
                                   int imgWidth, int imgHeight,
                                   float *irradImg,
                                   float *weightImg)
{
  int   idx;
  int   numContribs = 0;
  float weight = 0;
  float E[3]   = {0,0,0};

  if(x>0) {
    idx = y*imgWidth + (x-1);
    if(weightImg[idx]==0)
      return false;
    else {
      numContribs ++;
      weight += weightImg[idx];
      for(int ch=0; ch<3; ch++) E[ch] += irradImg[3*idx+ch];
    }
  }

  if(x<imgWidth-1) {
    idx = y*imgWidth + (x+1);
    if(weightImg[idx]==0)
      return false;
    else {
      numContribs ++;
      weight += weightImg[idx];
      for(int ch=0; ch<3; ch++) E[ch] += irradImg[3*idx+ch];
    }
  }

  if(y>0) {
    idx = (y-1)*imgWidth + x;
    if(weightImg[idx]==0)
      return false;
    else {
      numContribs ++;
      weight += weightImg[idx];
      for(int ch=0; ch<3; ch++) E[ch] += irradImg[3*idx+ch];
    }
  }

  if(y<imgHeight-1) {
    idx = (y+1)*imgWidth + x;
    if(weightImg[idx]==0)
      return false;
    else {
      numContribs ++;
      weight += weightImg[idx];
      for(int ch=0; ch<3; ch++) E[ch] += irradImg[3*idx+ch];
    }
  }

  if(numContribs) {
    float m = 1.0f/(float)numContribs;
    idx = y*imgWidth + x;
    weightImg[idx] = weight * m;
    for(int ch=0; ch<3; ch++)
      irradImg[3*idx+ch] = E[ch]*m;
    // update stats
    _numFilledPixels ++;
    return true;
  }
  else
    return false;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::FillHolesPost()
// --------------------------------------------------------------------------
int
CAdaptiveCaching::FillHolesPost(const CPixelList &pixelList,
                                CICSplatterRegular      &splatter,
                                CHitPointInfo    &outHitInfo)
{
  FATAL << "Throw this away, please!" << std::endl;
  FATAL_EXIT;
  return 0;
/*
  float *weightImg = splatter.GetWeightImage();
  const CHitInfoLite *hitInfoImg = splatter.GetHitInfoImage();
  float approxError = _icAlgo->_approxErrorCurrent[0];
  CICache *icache = _icAlgo->GetICache(0);
  CValues E(CStandardValues::GetDefaultColorType());
  float newRecA;
  int numRecBefore = icache->GetNumRecords();

  // for all pixels in the pixel list
  CPixelList::TPixelArray::const_iterator i;
  for(i=pixelList.pixels.begin(); i!= pixelList.pixels.end(); i++) {
    // index of the current image position
    TPixelIndex idx = *i;
    if(weightImg[idx]>0) continue;

    // detect if all neigbors have a valid contribution
    // if(_fillSinglePixel(x,y,imgWidth,imgHeight,irradImg,weightImg))
    // continue;

    // new records has to be added and splatted

    // current hit point info
    const CHitInfoLite &hit = hitInfoImg[idx];
    // find the closes record
    const CIrradianceRecord *closestRec=icache->FindNearest(hit.P,hit.N);
    // assign the value of 'a' of the closest record to the new rec
    if(closestRec) 
      newRecA = std::min<float>(closestRec->GetA()*1.05f, approxError);
    else
      newRecA = approxError;
    // create the new rec
    CIrradianceRecord *newRec=0;
    icache->AddIrradiance(hit.P,hit.N,hit.uN,hit.GetObject(),hit.GetTermination(),
      hit.footprintArea,outHitInfo,newRecA,E,&newRec);
    // update stats
    _updateStatMinNewRecA(newRecA);
    _numRecFilling++;
    // splat the new record
    if(newRec) {
      if(_jitterWeightIC)
        splatter.Splat<true>(newRec);
      else
        splatter.Splat<false>(newRec);
    }

    // @@@ tag this pixel for reconstruction using interpolation
  }
  return icache->GetNumRecords() - numRecBefore;
  */
}


// --------------------------------------------------------------------------
//  CAdaptiveCaching::FillHolesICQuery()
// --------------------------------------------------------------------------
int
CAdaptiveCaching::FillHolesICQuery(const CPixelList &pixelList,
                                   float newRecA,
                                   CICSplatterRegular      &splatter,
                                   CHitPointInfo    &outHitInfo,
                                   CValues          &avgE)
{
  FATAL << "Throw this away, please!" << std::endl;
  FATAL_EXIT;
  return 0;
/*
  const CHitInfoLite *hitInfoImg = splatter.GetHitInfoImage();
  CICache *icache = _icAlgo->GetICache(0);
  CValues E(CStandardValues::GetDefaultColorType());
  avgE.setz();
  unsigned long numPixels=0;
  int numRecBefore = icache->GetNumRecords();

  // for all pixels in the pixel list
  CPixelList::TPixelArray::const_iterator i;
  for(i=pixelList.pixels.begin(); i!= pixelList.pixels.end(); i++) {
    // index of the current image position
    TPixelIndex idx = (*i);
    // new record has to be added and splatted
    const CHitInfoLite &hit = hitInfoImg[idx];
    icache->GetIrradiance(hit.P,hit.N,hit.uN,hit.GetObject(),hit.GetTermination(),
      hit.footprintArea,outHitInfo,newRecA,E);
    // splat the new record
    avgE      += E;
    numPixels ++;
  }
  if(numPixels>0)
    avgE *= 1.0f/numPixels;

  return icache->GetNumRecords() - numRecBefore;
*/
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::FillHoles()
// --------------------------------------------------------------------------
int
CAdaptiveCaching::FillHoles(const CPixelList &pixelList,
                            float newRecA,
                            CICSplatterRegular      &splatter,
                            CHitPointInfo    &outHitInfo,
                            CValues          &avgE)
{
  FATAL << "Throw this away, please!" << std::endl;
  FATAL_EXIT;
  return 0;
/*

  float *weightImg = splatter.GetWeightImage();
  const CHitInfoLite *hitInfoImg = splatter.GetHitInfoImage();
  CICache *icache = _icAlgo->GetICache(0);
  CValues E(CStandardValues::GetDefaultColorType());
  avgE.setz();
  unsigned long numPixels=0;
  int numRecBefore = icache->GetNumRecords();

  // for all pixels in the pixel list
  CPixelList::TPixelArray::const_iterator i;
  for(i=pixelList.pixels.begin(); i!= pixelList.pixels.end(); i++) {
    // index of the current image position
    TPixelIndex idx = *i;
    if(weightImg[idx]>0) continue;
    // new record has to be added and splatted
    const CHitInfoLite &hit = hitInfoImg[idx];
    CIrradianceRecord *newRec=0;
    icache->AddIrradiance(hit.P,hit.N,hit.uN,hit.GetObject(),hit.GetTermination(),
      hit.footprintArea,outHitInfo,newRecA,E,&newRec);
    // splat the new record
    if(newRec) {
      unsigned long numCovered = splatter.Splat<false>(newRec);
      avgE      += newRec->GetIrradiance()*numCovered;
      numPixels += numCovered;
    }
    // @@@ tag this pixel for reconstruction using interpolation
  }
  if(numPixels>0)
    avgE *= 1.0f/numPixels;

  return icache->GetNumRecords() - numRecBefore;
  */
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::CreateFinalImage()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::CreateFinalImage(CICSplatterRegular               &splatter,
                                   CICache *icache,
                                   float              *auxImage,
                                   float              *weightImage)
{
  int nPixels = splatter.GetWidth()*splatter.GetHeight();
  // clear the weight buffer and the irradiance buffer
  shvf::setz(nPixels, weightImage);
  shvf::setz(nPixels*3, auxImage);
  // splat cache contents
  splatter.SetIrradImage(auxImage);
  if(_jitterWeightIC)
    splatter.Splat<true>(icache->GetRecordList());
  else
    splatter.Splat<false>(icache->GetRecordList());
  // normalize the splat buffer
  splatter.NormalizeBuffer();
  // multiply by the kd
  MulByKd(nPixels,auxImage,splatter.GetHitInfoImage());
}


// --------------------------------------------------------------------------
//  CAdaptiveCaching::CreateApproxErrImage()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::CreateApproxErrImage(CICSplatterRegular               &splatter,
                                       CICache *icache,
                                       float              *auxImage,
                                       float              *weightImage)
{
  int nPixels = splatter.GetWidth()*splatter.GetHeight();
  // clear the weight buffer and the irradiance buffer
  shvf::setz(nPixels, weightImage);
  shvf::setz(nPixels*3, auxImage);
  // splat cache contents
  splatter.SetIrradImage(auxImage);
  splatter.SplatA(icache->GetRecordList());
  // normalize the splat buffer
  splatter.NormalizeBuffer();
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::OverlapStats()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::OverlapStats(CICSplatterRegular &splatter)
{
  const CHitInfoLite *hitInfoImage = splatter.GetHitInfoImage();
  const int           imgWidth  = splatter.GetWidth();
  const int           imgHeight = splatter.GetHeight();

  // irradiance cache interface for direct illumination
  CICache *icache = _icAlgo->GetICache(0);
  // list of contributions for a given pixel
  CICContribArray carray;
  // reset stats
  _numICPixels       = 0;
  _numContribs       = 0;
  _maxNumContribs    = 0;
  // loop over image
  for(int y=0; y<imgHeight; y++) {
    for(int x=0; x<imgWidth; x++) {
      int idx = y*imgWidth + x;
      // current hit point info
      const CHitInfoLite &hit = hitInfoImage[idx];
      if(!splatter.IsHandled(hit)) continue;

      // get the contributions
      carray.clear();
      if ( icache->GetContributions<true,false>(hit.P,hit.N,carray,-1/*not used*/)) {
        _numICPixels ++;
        _numContribs += carray.size();
        if((unsigned int)carray.size()>_maxNumContribs) _maxNumContribs = carray.size();
      }
    }
  }
}


// --------------------------------------------------------------------------
//  CAdaptiveCaching::FillPixelList()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::FillPixelList(CPixelList& list,const CICSplatterRegular &splatter)
{
  const CHitInfoLite *hitInfoImage = splatter.GetHitInfoImage();
  int imgWidth  = splatter.GetWidth();
  int imgHeight = splatter.GetHeight();

  // using pascal's image traversal algorithm

  int step = pow2ge(Max(imgWidth,imgHeight));
  // Effective step for x
  int effectivestepx = step; 
  // Starting x coordinates
  int startx=0; 
  // Row number in the algorithm
  int rownumber=0,  evenstart=0; 
  // subdivision level
  int level = 0;
  // corrent pixel coords
  int x,y;
  // index to the first and the last pixel of the hierarchical level
  pair<int,int> levelidx;

  // This algorithm is executed until the steps are less than 1 pixel wide. 
  // It means the image has been fully explored
  while(step>=1) {

    // index to the first pixel in this level
    levelidx.first = list.pixels.size();

    // Loop over the y values
    for(y=0, rownumber=0; y < imgHeight; y += step, rownumber++)
    {	
      // If rownumber is even, we have to shift the starting point to avoid
      // reprocessing already processed pixels
      if (rownumber%2==0)
        startx = evenstart;
      else // Otherwise the starting point is at the beginning of the line 
        startx = 0;        
      effectivestepx = step+startx;

      // Loop over x values
      for(x=startx; x < imgWidth; x += effectivestepx )
      {
        ///////////////////////////////////////////
        if(splatter.IsHandled(hitInfoImage[y*imgWidth+x])) // yes to radiance caching
          list.pixels.push_back( y*imgWidth+x );
        ///////////////////////////////////////////
      }
    }

    // Divide the steps to go to the next resolution
    step /= 2;
    // On the even row numbers, the starting point must be shifted by stepx/2 
    // to avoid reprocessing pixels
    evenstart = step;

    // index to the last pixel in this level
    levelidx.second = list.pixels.size();

    // update the level ist
    list.levels.push_back(levelidx);
  }
}

#include "kdtmpl.h"

// --------------------------------------------------------------------------
//  CAdaptiveCaching::FillPixelList()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::FillPixelList(CLinearPixelList& list, 
                                const CICSplatterRegular &splatter)
                                
{
  const CHitInfoLite *hitInfoImage = splatter.GetHitInfoImage();
  int imgWidth = splatter.GetWidth();
  int imgHeight = splatter.GetHeight();

  // Using pascal's image traversal algorithm.
  int step = pow2ge(Max(imgWidth,imgHeight));
  int effectivestepx = step; 
  int startx=0; 
  int rownumber=0,  evenstart=0; 
  int level = 0;
  int x,y;

  while(step>=1) {
    for(y=0, rownumber=0; y < imgHeight; y += step, rownumber++) {	
      if (rownumber%2==0) startx = evenstart;
      else startx = 0;        
      effectivestepx = step+startx;
      for(x=startx; x < imgWidth; x += effectivestepx )  {
        //////////////////////////////////////////////
        unsigned idx = y*imgWidth+x;
        if(splatter.IsHandled(hitInfoImage[idx])) // yes to radiance caching
          list.push(idx);
        //////////////////////////////////////////////
      }
    }
    step /= 2;
    evenstart = step;
  }
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::IndirectIlluminationNonAdaptive()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::IndirectIlluminationNonAdaptive(
  float              *resultImage,
  const CHitInfoLite *hitInfoImage,
  int                 imgWidth,
  int                 imgHeight,
  CHitPointInfo      &outHitInfo,
  CMatrix4           &projMat,
  CMatrix4           &modelviewMat)
{
  // --------------------------
  //  SETUP
  // --------------------------
  CTimer timer, timerTotal, addRecTimer;
  timerTotal.Restart();
  float addRecTotalTime = 0;
  STATUS << "- adaptive caching, convergence algo: none"  << std::endl;
  STATUS << "- initializing...";
  timer.Restart();
  // number of pixels
  unsigned nPixels = imgWidth*imgHeight;
  // splatter only to be passed to fill pixel list
  CICSplatterRegular splatter(projMat,modelviewMat,imgWidth,imgHeight,hitInfoImage,
    CHitInfoLite::EE_IrrCaching);
  // pixel list for image traversal
  CPixelList pixelList;
  FillPixelList(pixelList,splatter);
  // IC interface
  CICache *icache = _icAlgo->GetICache(0);
  // pixel contribution list
  static CICContribArray pixContribs;
  // turn off smooth weight for the convergence
  bool smoothWardWeight = CIrradianceRecord::smoothWardWeight;
  CIrradianceRecord::smoothWardWeight = false;
  unsigned numRecordsBeforePass, passPixels;
  unsigned numRecordsBefore = icache->GetNumRecords();
  timer.Stop();
  STATUS << "done in " << timer.UserTime() << "s" << std::endl;

  // --------------------------
  //  PIXEL TRAVERSAL
  // --------------------------
  
  // presampling pass
  STATUS << "- presampling pass..." << std::endl;
  addRecTimer.Reset();
  timer.Restart();
  numRecordsBeforePass = icache->GetNumRecords();
  passPixels = pixelList.size();

  _icAlgo->SetPresampleMode();

  CPixelList::TPixelArray::const_iterator i;
  for(i=pixelList.pixels.begin(); i!= pixelList.pixels.end(); i++) {
    _printPercentage(pixelList.pixels.end()-i,passPixels,1000);
    TPixelIndex idx = (*i);
    const CHitInfoLite &hit = hitInfoImage[idx];
    // query the tree for contributions
    pixContribs.clear();
    icache->GetContributions<true,false>(hit.P,hit.N,pixContribs,-1);
    if(pixContribs.size()==0) {
      // add new record
      addRecTimer.Start();
      CIrradianceRecord *rec = _addRec<CIrradianceRecord>(hit,outHitInfo);
      addRecTimer.Stop();
    }
  }

  timer.Stop();
  STATUS
    << "- presampling pass finished in " << timer.UserTime() << " s / added " 
    << icache->GetNumRecords()-numRecordsBeforePass << " recs in " 
    << addRecTimer.UserTime() << "s" << std::endl;
  addRecTotalTime += addRecTimer.UserTime();

  // rendering pass
  STATUS << "- rendering pass..." << std::endl;
  addRecTimer.Reset();
  timer.Restart();
  numRecordsBeforePass = icache->GetNumRecords();
  passPixels = pixelList.size();

  _icAlgo->SetRenderMode();
  CIrradianceRecord::smoothWardWeight = smoothWardWeight;

  for(i=pixelList.pixels.begin(); i!= pixelList.pixels.end(); i++) {
    _printPercentage(pixelList.pixels.end()-i,passPixels);
    TPixelIndex idx = (*i);
    const CHitInfoLite &hit = hitInfoImage[idx];
    // query the tree for contributions
    pixContribs.clear();
    if(_jitterWeightIC)
      icache->GetContributions<true,true>(hit.P,hit.N,pixContribs,-1);
    else
      icache->GetContributions<true,false>(hit.P,hit.N,pixContribs,-1);
    if(!pixContribs.WeightedAverage(resultImage+3*idx)) {
      // add new record
      addRecTimer.Start();
      CIrradianceRecord *rec = _addRec<CIrradianceRecord>(hit,outHitInfo);
      addRecTimer.Stop();
      rec->GetIrradiance().CopyData(resultImage+3*idx);
    }
  }

  timer.Stop();
  STATUS
    << "- rendering pass finished in " << timer.UserTime() << " s / added " 
    << icache->GetNumRecords()-numRecordsBeforePass << " recs in " 
    << addRecTimer.UserTime() << "s" << std::endl;
  addRecTotalTime += addRecTimer.UserTime();

  // multiply by the kdImage
  MulByKd(nPixels,resultImage,hitInfoImage);
  // print final stats
  timerTotal.Stop();
  _printFinalStats(timerTotal.UserTime(),addRecTotalTime,
    icache->GetNumRecords()-numRecordsBefore);
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::IndirectIlluminationPerPixel()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::IndirectIlluminationPerPixel(
  float              *resultImage,
  const CHitInfoLite *hitInfoImage,
  int                 imgWidth,
  int                 imgHeight,
  CHitPointInfo      &outHitInfo,
  CMatrix4           &projMat,
  CMatrix4           &modelviewMat)
{
  // --------------------------
  //  SETUP
  // --------------------------
  CTimer timer, timerTotal, addRecTimer;
  timerTotal.Restart();
  STATUS << "- adaptive caching, convergence algo: per pixel sampling" << std::endl;
  STATUS << "- initializing...";
  timer.Restart();
  // splatter only to be passed to fill pixel list
  CICSplatterRegular splatter(projMat,modelviewMat,imgWidth,imgHeight,hitInfoImage,
    CHitInfoLite::EE_IrrCaching);
  // pixel list for image traversal
  CPixelList pixelList;
  FillPixelList(pixelList,splatter);
  // diffuse sampler
  CICache *icache = _icAlgo->GetICache(0);
  CDiffuseSampler* diffSampler = icache->GetSampler();
  unsigned numRecordsBefore = icache->GetNumRecords();
  
  float hmDist,minDist,stdDev;
  timer.Stop();
  STATUS << "done in " << timer.UserTime() << "s" << std::endl;

  // --------------------------
  //  PIXEL TRAVERSAL
  // --------------------------
  
  STATUS << "- image sampling..." << std::endl;
  addRecTimer.Reset();
  unsigned passPixels = pixelList.size();

  CPixelList::TPixelArray::const_iterator i;
  for(i=pixelList.pixels.begin(); i!= pixelList.pixels.end(); i++) {
    _printPercentage(pixelList.pixels.end()-i,passPixels);
    TPixelIndex idx = (*i);
    const CHitInfoLite &hit = hitInfoImage[idx];
    // sample hemisphere
    addRecTimer.Start();
    bool b = diffSampler->ComputeIrradiance(hit.P,hit.N,hit.uN,hit.GetObject(),
      hit.GetTermination(),hit.footprintArea,outHitInfo,tmpE,hmDist,minDist,
      stdDev,0,0);
    addRecTimer.Stop();
    tmpE.CopyData(resultImage+3*idx);
  }

  // multiply by the kdImage
  MulByKd(imgWidth*imgHeight,resultImage,hitInfoImage);
  // print final stats
  timerTotal.Stop();
  addRecTimer.Stop();
  _printFinalStats(timerTotal.UserTime(),addRecTimer.UserTime(),
    icache->GetNumRecords()-numRecordsBefore);
}

/*
// --------------------------------------------------------------------------
//  CAdaptiveCaching::_initPixContribImage()
// --------------------------------------------------------------------------
CPixICContribList*
CAdaptiveCaching::_initPixContribImage(unsigned count)
{
  if(count==0) { 
    delete [] _pixContribImage;
    _pixContribImage = 0;
    _pixContribImageLen = 0;
  }
  else {
    if(_pixContribImageLen<count) {
      delete [] _pixContribImage;
      _pixContribImage = new CPixICContribList[count];
      _pixContribImageLen = count;
    }
    else {
      for(unsigned i=0; i<count; i++)
        _pixContribImage[i].clear();
    }
  }
  return _pixContribImage;
}
*/

// --------------------------------------------------------------------------
//  _extrapolateRadianceValid()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_extrapolateRadianceValid(CPixContribList<TRecord> &l,
                                            const CHitInfoLite& hit)
{
  FATAL 
    << "Only specialized version of _extrapolateRadianceValid() can be called."
    << std::endl;
  FATAL_EXIT;
}

template<>
void
CAdaptiveCaching::_extrapolateRadianceValid<CIrradianceRecord>(
  CPixICContribList &l,
  const CHitInfoLite& hit)
{
  CTriCol E;
  for(CPixICContribList::iterator it=l.begin(); it!=l.end(); ++it) {
    CICContrib &item = *it;
    if(!item.IsInvalid() && !item.IsEValid()) {
      item.rec->ExtrapolateIrradiance(hit.P,hit.N,E.data());
      E *= hit.kd;
      item.SetE(E);
    }
  }
}

template<>
void
CAdaptiveCaching::_extrapolateRadianceValid<CRadianceRecord>(
  CPixRCContribList &l,
  const CHitInfoLite& hit)
{
  static const CTriCol black(0.0);
  CTriCol E;

  const CCoefVec<float> *inVec;
  const CCoefVec<float>* shbrdf=0;
  CRCache *rcache;
  CPixRCContribList::iterator it=l.begin();
  for( ; it!=l.end(); ++it) {
    CRCContrib &item = *it;
    if(!item.IsInvalid() && !item.IsEValid()) {
      const CRadianceRecord *rec = item.rec;
      // retrieve the BRDF & possible rotate it
      if(!shbrdf) {
        // float u,v;
        // hit.shRep->GetUV(CVector3D(0,0,1), u, v);
        // shbrdf = hit.shRep->GetCoeffsUV(u,v);
        shbrdf = hit.shRep->GetCoeffsUV(hit.brdf_u,hit.brdf_v);

        _numBrdfRetrieval++;
        if(!shbrdf) {
          FATAL << "Could not get BRDF coefficients.\n" << std::endl;
          FATAL_EXIT;
        }
        if(shbrdf->GetOrder() <= 0) {
          // extrapolate all records with black
          item.SetE(black);
          ++it;
          for( ; it!=l.end(); ++it)
            if(!it->IsInvalid() && !it->IsEValid()) it->SetE(black);
          return;
        }
        rcache = _rcAlgo->GetRCache(0);
        if(!rcache->LocalFrame()) { // global frame -> rotate the BRDF
          int order=min(rec->GetCoeffs().GetOrder(),shbrdf->GetOrder());
          _tmpVec2->_setOrder(order);
          // allign the BRDF with the global frame
          zxzxzRotObject.ToGlobalFrame<float>(hit.U,hit.V,hit.N,order);
          for(int ch=0; ch<3; ch++) {
            zxzxzRotObject.TransformVector(_tmpVec2->GetChannel(ch),
              shbrdf->GetChannel(ch),order,rotGenerator);
          }
          _tmpVec2->ZeroPadding();
          shbrdf = _tmpVec2;
        }
      }

      int order=min(rec->GetCoeffs().GetOrder(),shbrdf->GetOrder());
      _tmpVec1->_setOrder(order);
      _tmpVec2->_setOrder(order);

      // apply gradients
      if(rec->UseGradients()) {
        rcache->ExtrapolateRadiance(hit.P - rec->GetPoint(),rec,*_tmpVec1);
        inVec = _tmpVec1;
      }
      else {
        inVec = &rec->GetCoeffs();
      }
      assert(rec->GetCoeffs().GetNumChannels()==3);
      if(rcache->LocalFrame()) {
        // rotate
        rcache->RotateAndAdd<true>(rec,inVec,hit.U,hit.V,hit.N,0,*_tmpVec2);
        inVec = _tmpVec2;
      }
      // dot
      int veclen = inVec->GetVecLen();
      for(unsigned ch=0; ch<3; ch++)
        E[ch] = shvf::dot4(veclen,inVec->GetChannel(ch),shbrdf->GetChannel(ch));
      item.SetE(E);
    }
  }
}

__END_GOLEM_SOURCE
