// ===================================================================
// $Id: icadapt.h,v 1.2 2005/11/29 16:04:04 xkrivanj Exp $
//
// icadapt.h
//   Class managing perceptually-based adaptive quality setting
//   for irradiance cache.
//
// Class:
//
// REPLACEMENT_STRING
//
// Initial coding by Jaroslav Krivanek (March, 2005).

#ifndef __ICADAPT_H__
#define __ICADAPT_H__

// GOLEM headers
#include "configh.h"
#include "color.h"
#include "vector3.h"
#include "point2.h"
#include "icrecord.h"
#include "icontrib.h"
#include "icsplatr.h"
#include "poolque.h"
#include "icalgo.h"
#include "rcalgo.h"

// standard headers
#include <vector>

__BEGIN_GOLEM_HEADER

// #define _DEBUG_ICADAPT

// forward declarations
class CIrradianceRecordID;
class CIrradianceRecord;
class CObject3D;
class CICAlgo;
class CRCAlgo;
class CHitPointInfo;
class CHitInfoLite;
class CMatrix4;
class CICSplatterRegular;
class CICache;
class CRCache;
class CValues;
class CLinearPixelList;
class CTimer;
class CVDMRamasubramanian;
class CICSplatterIrregular;

/// Hierarchical list of pixels
class CPixelList {

public:
  /// type for the array of pixels
  typedef vector< TPixelIndex >  TPixelArray;

  /// type for the list of levels in the hierarchical screen subdivision
  typedef vector< pair<int,int> > TPixelLevelArray;

  /// the array of pixels itself
  TPixelArray pixels;

  /// list of levels in the hierarchical screen subdivision
  TPixelLevelArray levels;

  /// return the length of the pixel array
  TPixelArray::size_type size() const { return pixels.size(); }
};

/// List of pixels to process with a map indicating
/// presence of a pixel in the list.
class CLinearPixelList: public poolqueue<TPixelIndex>
{
private:
  /// _dirtyMap[i] == true if i-th pixel is in the list
  bool *_dirtyMap;
  /// superior class in the hierarchy
  typedef poolqueue<TPixelIndex> super;

  /// base hit array pointer -> for adding pixels by
  /// pointer to hit info
  const CHitInfoLite *_baseHitPtr;

public:

  /// constructor
  CLinearPixelList(unsigned numPixels,const CHitInfoLite *baseHitPtr)
    : poolqueue<TPixelIndex>(numPixels)
  {
    assert(numPixels>0);
    _dirtyMap = new bool[numPixels];
    assert(_dirtyMap);
    memset(_dirtyMap,0,numPixels*sizeof(_dirtyMap[0]));
    _baseHitPtr = baseHitPtr;
  }

  /// desctructor
  ~CLinearPixelList() {delete _dirtyMap;}

  /// push a new pixel in the list
  void push(TPixelIndex idx) {
    if(!_dirtyMap[idx]) {
      _dirtyMap[idx]=true;
      super::push(idx);
    }
  }

  // push a new pixel via pointer to base hitinfo
  void push(const CHitInfoLite *hitptr)
  {
    push(TPixelIndex(hitptr - _baseHitPtr));
  }

  /// remove front
  void pop() {
    _dirtyMap[front()]=false;
    super::pop();
  }

  /// erase all pixels in the list
  void erase() {
    super::erase();
    memset(_dirtyMap,0,capacity()*sizeof(_dirtyMap[0]));
  }

  /// return the dirty map
  const bool* GetDirtyMap() const { return _dirtyMap; }
};

class CAdaptiveCaching {

public: // types

  /// which convergence algorithm?
  enum  EConvergenceAlgo {
    EE_ConvergenceAlgoNonAdaptive,
    EE_ConvergenceAlgoPerPixel,
    EE_ConvergenceAlgoOnePass,
    EE_ConvergenceAlgoList,
    EE_ConvergenceAlgoTree,
    EE_ConvergenceAlgoBBox,
    EE_ConvergenceAlgoBBoxIrreg
  };

public: // methods

  CAdaptiveCaching(EConvergenceAlgo ca, 
    CICAlgo *icAlgo, float maxErrorIC,
    CRCAlgo *rcAlgo, float maxErrorRC,
    bool adaptive,float mappingCoeff,float mappingGamma);

  ~CAdaptiveCaching();

  /// Irradiance caching with adaptive refinement based on the visual
  /// difference in the record overlap areas: dispatching method for
  // the particular convergence algo.
  void IndirectIlluminationOverlap(
    float              *resultImage,
    const CHitInfoLite *hitInfoImage,
    int                 imgWidth,
    int                 imgHeight,
    CHitPointInfo      &outHitInfo,
    CMatrix4           &projMatrix,
    CMatrix4           &modelviewMatrix);

  void IndirectIlluminationPerPixel(
    float              *resultImage,
    const CHitInfoLite *hitInfoImage,
    int                 imgWidth,
    int                 imgHeight,
    CHitPointInfo      &outHitInfo,
    CMatrix4           &projMat,
    CMatrix4           &modelviewMat);

  void IndirectIlluminationNonAdaptive(
    float              *resultImage,
    const CHitInfoLite *hitInfoImage,
    int                 imgWidth,
    int                 imgHeight,
    CHitPointInfo      &outHitInfo,
    CMatrix4           &projMat,
    CMatrix4           &modelviewMat);

  void IndirectIlluminationOverlapOnePass(
    float              *resultImage,
    const CHitInfoLite *hitInfoImage,
    int                 imgWidth,
    int                 imgHeight,
    CHitPointInfo      &outHitInfo,
    CMatrix4           &projMatrix,
    CMatrix4           &modelviewMatrix);

  void IndirectIlluminationOverlapConverge(
    float              *resultImage,
    const CHitInfoLite *hitInfoImage,
    int                 imgWidth,
    int                 imgHeight,
    CHitPointInfo      &outHitInfo,
    CMatrix4           &projMatrix,
    CMatrix4           &modelviewMatrix);

  void IndirectIlluminationOverlapConvergeTree(
    float              *resultImage,
    const CHitInfoLite *hitInfoImage,
    int                 imgWidth,
    int                 imgHeight,
    CHitPointInfo      &outHitInfo,
    CMatrix4           &projMatrix,
    CMatrix4           &modelviewMatrix);

  template<class TRecord>
  void IndirectIlluminationOverlapConvergeBBox(
    float              *resultImage,
    const CHitInfoLite *hitInfoImage,
    int                 imgWidth,
    int                 imgHeight,
    CHitPointInfo      &outHitInfo,
    CMatrix4           &projMatrix,
    CMatrix4           &modelviewMatrix);

  template<class TRecord>
  void IndirectIlluminationOverlapConvergeBBoxIrreg(
    float              *resultImage,
    const CHitInfoLite *hitPoints,
    unsigned            nPixels,
    CHitPointInfo      &outHitInfo);

  /// Fill the pixel list with the pixels that need irradiance caching
  void FillPixelList(CPixelList& list,const CICSplatterRegular &splatter);

  /// Fill the pixel list with the pixels that need irradiance caching
  void FillPixelList(CLinearPixelList& list,const CICSplatterRegular &splatter);

  /// Fill the overlap statistics.
  void OverlapStats(CICSplatterRegular &splatter);

  /// Splat the cache contents, multiply by kd and add direct illumination.
  void CreateFinalImage(
    CICSplatterRegular               &splatter,
    CICache *icache,
    float              *auxImage,
    float              *weightImage);

  /// Splat the value of 'a' for each record in the cache.
  void CreateApproxErrImage(
    CICSplatterRegular               &splatter,
    CICache *icache,
    float              *auxImage,
    float              *weightImage);

  /// Fill every image pixel with a valid contribution.
  /**
    This is to be used once after the adaptive step.
    All pixels in the given pixel list are checked for a valid contribution
    (weight>0), if there is no valid contribution, a record is added to the
    cache and splatted.
    The value of 'a' for new records is copied from the neares neighbor
    in the cache.

    @return Number of added records.
  */
  int FillHolesPost(
    const CPixelList &pixelList,
    CICSplatterRegular &splatter, 
    CHitPointInfo &outHitInfo);

  /// Fill every image pixel with a valid contribution.
  /**
    This is a simpler version of FillHolesPost. 
    
    All pixels in the given pixel list are checked for a valid contribution
    (weight>0), if there is no valid contribution, a record is added to the
    cache and splatted.
    The value of 'a' for new records given as a parameter (unlike in the 
    FillHolesPost() method, where it is copied from the closest record.

    @return The number of added records.
   */
  int FillHoles(
    const CPixelList &pixelList,
    float newRecA,
    CICSplatterRegular &splatter, 
    CHitPointInfo &outHitInfo,
    CValues          &avgE);

  /// The same as FillHoles(), but splatting is not used (cache query is
  /// performed for each pixel instead).
  int FillHolesICQuery(
    const CPixelList &pixelList,
    float newRecA,
    CICSplatterRegular      &splatter,
    CHitPointInfo    &outHitInfo,
    CValues          &avgE);

  /// Do what the name says.
  CVDMRamasubramanian* CreateThresholdElavationImage(
    float *resultImage,
    const CHitInfoLite *hitImage,
    unsigned imgWidth,
    unsigned imgHeight,
    const float *avgIrrad) const;

  /// statistics
  //@{
  /// number of time a query with overlap has been processed
  unsigned long GetNumProcessOverlap() const { return _numProcessOverlap; }
  /// number of times a record's approx error has been decreased
  unsigned long GetNumDecreaseA() const { return _numDecreaseA; }
  /// number of pixels that use irradiance caching
  unsigned long GetNumICPixels() const { return _numICPixels; }
  /// total number of contributions
  unsigned long GetNumContribs() const { return _numContribs; }
  /// maximum number of contributions per pixel
  unsigned int  GetMaxNumContribs() const { return _maxNumContribs; }

  /// set all stat counters to zero
  void ResetStats();
  //@}

  /// make a report 
  bool DoReport(ostream &o);

  /// return the handle flag depending on the template argument
  template<class TRecord>
    static inline unsigned GetHandleFlag();

  /// return the number of records depending on the template argument
  template<class TRecord>
    inline unsigned long GetNumRecords(unsigned level=0) const;

  /// return the record list depending on the template argument
  template<class TRecord>
    inline CRecordList<TRecord>& GetRecordList();

  /// return the record octree depending on the template argument
  template<class TRecord>
    inline CRCOctree<TRecord>* GetOctree();

  /// return the max error depending on the template argument
  template<class TRecord>
    inline float GetMaxError();

  /// return the color for record position visualization 
  /// depending on the template argument
  template<class TRecord>
    static inline const float* GetRecPosCol();

  /// Return 1 for irradianc caching and hit.dot for radiance caching.
  /// The value is used to elevate the threshold near silhouettes in
  /// radiance caching.
  template<class TRecord>
    static inline float GetDotForThreshold(const CHitInfoLite &hit);

  /// Return "ic" or "rc" depending on the template argument
  template<class TRecord>
    inline static const char* GetCacheAbbrev();

protected: // methods

  class CProjPosAndBBox {
  public:
    CICSplatterRegular::TBBox bbox;
    CVector3D   projPos;
  };

  /// Multiply irradiance buffer by 'kd' stored in hit info
  friend void MulByKd(
    unsigned numPixes,
    float *irradPtr,
    const CHitInfoLite *hitInfoPtr);

  friend void AddDirectIllum(
    unsigned numPixes,
    float *irradPtr,
    const CHitInfoLite *hitInfoPtr);

  template<class TRecord>
    void _extrapolateRadianceValid(
    CPixContribList<TRecord> &l,
    const CHitInfoLite& hit);

  /// print out the percentage of pixels processed
  void _printPercentage(unsigned left,unsigned total,unsigned freq=20000);

  /// print out final statistics on convergence
  void _printFinalStats(
    float totalTime,
    float addRecTotalTime,
    unsigned numRec);

  // create the final image from records
  void _finalNormalization(CTimer &timer,CICSplatterRegular &splatter);

  template<class TRecord>
    inline float _determineNewRecA(
    const CHitInfoLite &hit, 
    float approxError,
    CRCOctree<TRecord> *octree);

  /// add new record at the specified position
  template<class TRecord>
  TRecord* _addRec(
    const CHitInfoLite &hit,
    CHitPointInfo &outHitInfo);

  /// return the number of rectified records

  // The (ir)radiance values in the array are supposed to be already
  // multiplied by the albedo, or dotted with the brdf coeffs.
  // Therefore they present the outgoing radiance due to indirect illumination.
  template<class TRecord,class TContribArray>
  int _processPixelOverlap(
    const CHitInfoLite &hit,
    const float *directIllum,
    const float  thresholdElevation,
    TContribArray &carray,
    CRCOctree<TRecord> *octree);

  /// Compute the indirect irradiance for an isolated pixel using interpolation
  /**
     All four neighbors of the given pixel are checked for a valid contribution.
     If the test succeeds, the indirect irradiance is interpolated from
     the neigbour.
     @return 'true' if the interpolation was successful
    */
  inline bool _fillSinglePixel(
    int x,int y, 
    int imgWidth, int imgHeight,
    float *irradImg,
    float *weightImg);

  /// Check if the contributions in the list are still valid.
  /**
    If and invalid contribution is found, it is set to invalid state
    and the pointer to it is physically removed from the pixel
    contribtion list.
    @return The number of valid contributions in the list.
  */
  inline unsigned _updatePixelContribPtrList(
    CPixICContribPtrList &list,
    const CHitInfoLite &hit);

  /// Find the first invalid contribution in 'clist'.
  /// Returns clist.end() if all contributions are valid.
  inline CPixICContribPtrList::iterator _findFirstInvalidContribPtr(
    CPixICContribPtrList &clist,
    const CHitInfoLite &hit);

  /// Set the records in the list to unchanged state.
  template<class TRecord>
    void _unsetChanged(CRecordList<TRecord> &recList);

  /// Add item to the pixel contribution list.
  /// Return the iterator to the newly added item.
  inline void _addContribPtr(
    CPixICContribPtrList &clist,
    const TPixICContribPtr &item,
    const CHitInfoLite &hit);

  /// Add record contributions to the contribution buffer.
  template<class T>
    int _rasterizeContribPtrs(
    const CIrradianceRecord *rec, 
    CICSplatterRegular &splatter,
    CLinearPixelList &pixelList,
    T &addList,
    const CICSplatterRegular::TBBox& r);

  /// Splat contributions of a record
  template<class T>
    int _splatContribPtrs(
    const CIrradianceRecord* rec,
    CICSplatterRegular &splatter,
    CLinearPixelList &pixelList,
    T &addList);

  /// update the statistics concerning decreasing the value of 'a'
  inline void _updateStatDecreaseA(float newA);
  /// update statistics concerning the value of 'a' for a new record
  inline void _updateStatMinNewRecA(float newRecA);

  /// Remove and delete invalid contribution for all the records in the list.
  void _removeInvalidRecordContribs(
    CIrradianceList &recList,
    CICSplatterRegular &splatter,
    CLinearPixelList &pixelList);

  inline void _removeInvalidRecordContribsVerbose(
    CTimer &timer,
    CIrradianceList &recList,
    CICSplatterRegular &splatter,
    CLinearPixelList &pixelList);

  /// Debugging function. Verify that all pixels have a valid contribution 
  /// by querying the octree.
  void _checkValidContribsCode(CICSplatterRegular &splatter);

  /// Displatching function for _checkValidContribsCode()
  void _checkValidContribs(CICSplatterRegular &splatter);

  /// initializes some stuff before debug image saving
  template<class TRecord>
  inline void _initSaveDebugImages();

  /// calls _doSaveDebugImagesCode() is _saveDebugImages is true
  template<class TRecord>
  inline void _doSaveDebugImages(
    CTimer &timer,
    CICSplatterRegular &splatter,
    CPixContribList<TRecord> *contribImage,
    bool checkFrequency=true);

  /// saves debug images
  template<class TRecord>
  void _doSaveDebugImagesCode(
    CICSplatterRegular &splatter,
    CPixContribList<TRecord> *contribImage,
    bool checkFrequency=true);

  /////////////////////////////////////////////////////
  /// Methods for full convergence with octree queries
  /////////////////////////////////////////////////////

  /// Destroy the contributions for each record in the list.
  template<class TRecord,class Tdata>
  static void _destroyRecordData(const CRecordList<TRecord> &recList);
  /// Destroy the contribution list for each record in the list.
  template<class Tlist>
  static void _destroyRecordList(const CIrradianceList &recList);

  template<class TRecord>
    void _scheduleChangedRecordPixelsVerbose(
    CTimer &timer,
    CRecordList<TRecord> &recList,
    CICSplatterRegular &splatter,
    CLinearPixelList &pixelList);

  /// schedule pixels in changed record's bboxes for processing
  template<class TRecord>
    void _scheduleChangedRecordPixels(
    CRecordList<TRecord> &recList,
    CICSplatterRegular &splatter,
    CLinearPixelList &pixelList);

  int _splatContribCount(
    const CIrradianceRecord* rec,
    CICSplatterRegular &splatter,
    CLinearPixelList &pixelList,
    CProjPosAndBBox &recdata);

  inline int _rasterizeContribCount(
    const CIrradianceRecord *rec, 
    CICSplatterRegular &splatter,
    CLinearPixelList &pixelList,
    const CICSplatterRegular::TBBox& r);

  ////////////////////////////////////////////////

  ////////////////////////////////////////////////////
  // Methods for full convergence without record contribution lists.
  ////////////////////////////////////////////////////

  /// Check if the contributions in the list are still valid.
  /**
    If and invalid contribution is found, it is set to invalid state,
    but not physically removed from the list.
    @return The number of valid contributions in the list.
  */
  template<class TRecord>
  inline unsigned _updatePixelContribList(
    CPixContribList<TRecord> &list,
    const CHitInfoLite       &hit);

  template<class TRecord>
  int _splatContrib(
    const TRecord* rec,
    CICSplatterRegular &splatter,
    CPixContribList<TRecord> *contribImage,
    CLinearPixelList &pixelList,
    CProjPosAndBBox &recdata);

  template<class TRecord>
  int _rasterizeContrib(
    const TRecord *rec, 
    CICSplatterRegular &splatter,
    CPixContribList<TRecord> *contribImage,
    CLinearPixelList &pixelList,
    const CICSplatterRegular::TBBox& r);

  template<class TRecord>
  inline void _addContrib(
    CPixContribList<TRecord>      &clist,
    const CRecordContrib<TRecord> &item,
    const CHitInfoLite            &hit);

  template<class TRecord>
  inline typename CPixContribList<TRecord>::iterator _findFirstInvalidContrib(
    CPixContribList<TRecord>  &clist,
    const CHitInfoLite        &hit);

  template<class TRecord>
    void _finalNormalizationBBox(
    CTimer &timer,
    CPixContribList<TRecord> *contribImage,
    CICSplatterRegular &splatter,
    float *resultImage);

  //////////////////////////////////////////////////////
  // Methods for full convergence with bounding spheres
  //////////////////////////////////////////////////////
  template<class TRecord>
    void _scheduleChangedRecordPointsVerbose(
    CTimer &timer,
    CRecordList<TRecord> &recList,
    CICSplatterIrregular &splatter,
    CLinearPixelList &pointList);

  template<class TRecord>
  void _scheduleChangedRecordPoints(
    CRecordList<TRecord> &recList,
    CICSplatterIrregular &splatter,
    CLinearPixelList &pixelList);

  template<class TRecord>
  int _splatContribIrreg(
    const TRecord *rec,
    CPixContribList<TRecord> *contribImage,
    CICSplatterIrregular &splatter,
    CLinearPixelList &pixelList);

  /// Add contributions dues to weight jittering.
  template<class TRecord>
    unsigned _addJitter(
    CPixContribList<TRecord> *contribImage,
    CICSplatterIrregular &splatter);

  /// Reevaluate weight in the contribution array.
  /** 
    This must be done before the final normalization if smoothWardWeight is
    used. (Because during the convergence algo, the tmooth weight is never
    used).
  */
  template<bool jitter,class TRecord>
    void _reevaluateWeights(
    CPixContribList<TRecord> *contribImage,
    CICSplatterIrregular &splatter);

  template<class TRecord>
  void _renderFromContribs(
    float *resultImage,
    CPixContribList<TRecord> *contribImage,
    const CHitInfoLite *hitPoints,
    unsigned numHitPoints,
    bool updatePixelList);

  template<class TRecord>
    void _finalNormalization(
    CTimer &timer,
    CPixContribList<TRecord> *contribImage,
    CICSplatterIrregular &splatter,
    float *resultImage);

  inline unsigned _setupIndices(
    const CHitInfoLite *hitPoints,
    const unsigned numPoints,
    const unsigned handleFlag);

  inline float _lumFromRGBTonemap(const float *c);

public: //data

  /// Create filename from _imgBaseName, postfix and extension
  string CreateFileName(const string &postfix);

  /// Set/Get the flag for saving debug images during refinement
  //@{
  void EnableSaveDebugImages()    { _saveDebugImages = true; }
  void DisableSaveDebugImages()   { _saveDebugImages = false; }
  void SetSaveDebugImages(bool b) { _saveDebugImages = b; }
  bool GetSaveDebugImages() const { return _saveDebugImages; }
  //@}


  /// Set/Get the flag for saving debug images during threshold elevation
  /// model evaluation
  //@{
  void EnableSaveVDMDebugImages()    { _saveVDMDebugImages = true; }
  void DisableSaveVDMDebugImages()   { _saveVDMDebugImages = false; }
  void SetSaveVDMDebugImages(bool b) { _saveVDMDebugImages = b; }
  bool GetSaveVDMDebugImages() const { return _saveVDMDebugImages; }
  //@}

  /// Set/Get the flag for saving VDM input and output images.
  //@{
  void EnableSaveVDMInOut()    { _saveVDMInOut = true; }
  void DisableSaveVDMInOut()   { _saveVDMInOut = false; }
  void SetSaveVDMInOut(bool b) { _saveVDMInOut = b; }
  bool GetSaveVDMInOut() const { return _saveVDMInOut; }
  //@}

  /// Set/Get the flag for using threshold elevation
  //@{
  void EnableUseThresholdElevation()    { _useThresholdElevation = true; }
  void DisableUseThresholdElevation()   { _useThresholdElevation = false; }
  void SetUseThresholdElevation(bool b) { _useThresholdElevation = b; }
  bool GetUseThresholdElevation() const { return _useThresholdElevation; }
  //@}

  /// Set/Get the flag for using masking in the threshold elevation model
  //@{
  void EnableUseMasking()      { _useMasking = true; }
  void DisableUseMasking()     { _useMasking = false; }
  void SetUseMasking(bool b)   { _useMasking = b; }
  bool GetUseMasking() const   { return _useMasking; }
  //@}

  /// Set/Get the frequency for saving debug images
  //@{
  unsigned GetDebugImageSaveFrequency()const {return _debugImageSaveFrequency;}
  void SetDebugImageSaveFrequency(unsigned f) {_debugImageSaveFrequency=f;}
  //@}

  /// Set/Get base name for debug images
  //@{
  const string &GetImgBaseName() const {return _imgBaseName;}
  void SetImgBaseName(const string& name) { _imgBaseName=name; }
  //@}

  /// Set/Get type (extension) for debug images
  //@{
  const string GetHdrFileType() const {return _hdrFileType; }
  void SetHdrFileType(const string& type) { _hdrFileType = type; }
  //@}

  inline const CValues& GetRecE(const CIrradianceRecord* rec) const
  { return rec->GetE(); }

  inline const CValues& GetRecE(const CRadianceRecord* rec) const
  { 
    static CValues c(0,0,0);
    DEBUG << "GetE(const CRadianceRecord* rec)" << std::endl;
    // it should be computed by dotting with cosine, but it just return
    // the contant term
    c.Set(0,rec->GetCoeffs().GetCoeff(0,0));
    c.Set(1,rec->GetCoeffs().GetCoeff(1,0));
    c.Set(2,rec->GetCoeffs().GetCoeff(2,0));
    return c;
  }

protected: //data

  // temporary vectors for radiance extrapolation
  CCoefVec<float> *_tmpVec1;
  CCoefVec<float> *_tmpVec2;

  /// Pre-allocated data buffers
  //@{
  /// List of contributions for each valid pixel (point).
  // CPixICContribList *_pixContribImage;
  /// Currently allocated length of _pixContribImage.
  // unsigned _pixContribImageLen;
  /// Reallocate/clear _pixContribImage
  // CPixICContribList* _initPixContribImage(unsigned count);
  //@}

  CICAlgo *_icAlgo;
  CRCAlgo *_rcAlgo;

  /// number of records at the beginning of debug image saving
  unsigned _sdiStartRecords;
  /// index of the last save debug image
  unsigned _debugImageIndex;
  /// debug images are save every i-th new record
  unsigned _debugImageSaveFrequency;

  /// base name for debug images
  string _imgBaseName;

  /// type (extension) for debug images
  string _hdrFileType;

  /// save debug images during refinement
  bool _saveDebugImages;

  /// save debug images during threshold elevation model evaluation
  bool _saveVDMDebugImages;

  /// save VDM input and output
  bool _saveVDMInOut;

  /// use threshold elevation according to Ramasubramanian '99
  bool _useThresholdElevation;

  /// use masking in the threshold elevation model (effective only if
  /// _useThresholdElevation==true
  bool _useMasking;

  /// use weight jittering in the image generation pass
  bool _jitterWeightIC;
  bool _jitterWeightRC;

  /// lower limit for delta L in luminance comparison
  float _minDeltaL;

  /// gama for tone mapping
  float _gamma;

  // true if _gamma!=1 and tone mapping is on
  bool _gammaCorrect; 

  /// linear scaling factor
  float _linScale;

  /// maximum allowed visible error
  /**
    The exact meaning as of now (04/04/05) is the maximum
    allowed Luminance contrast.
    It will surely be something else.
  */
  float _maxErrorIC;
  float _maxErrorRC;

  // type of convergence algorithm passed to constructor
  EConvergenceAlgo _convergenceAlgo;

  // whether or not to refine
  bool _adaptive;

  /// view position
  CVector3D _viewPos;

  /// statistics
  //@{
  unsigned long _numBrdfRetrieval;
  /// number of time a query with overlap has been processed
  unsigned long _numProcessOverlap;
  /// number of times a record's approx error has been decreased
  unsigned long _numDecreaseA;
  /// number of pixels that use irradiance caching
  unsigned long _numICPixels;
  /// total number of contributions
  unsigned long _numContribs;
  /// maximum number of contributions per pixel
  unsigned int   _maxNumContribs;
  /// minimum 'a' assigned to a new record
  float _minNewRecA;
  /// minimum 'a' used for any record (taking into account decreasing 'a')
  float _minA;
  /// number of single pixels filled by interpolation
  unsigned long _numFilledPixels;
  /// number of records added in the hole-filling stage
  unsigned long _numRecFilling;

  /// total time
  float _totalTime;
  /// time for adding new records
  float _addRecTotalTime;
  /// number of records
  unsigned long _numRec;

  //@}

  /// temporary color value
  CValues tmpE;
  /// temporary color value
  CValues _tmpDirectIllum;
};

// --------------------------------------------------------------------------
//  erase_to_end()
// --------------------------------------------------------------------------
/// erase from where to end()
template<class listT, class itT>
inline itT 
erase_to_end(listT &l, itT where)
{
  // DEBUG << "erase_to_end - general" << std::endl;
  return l.erase(where,l.end());
}

/// erase from where to end() - specialized version for growvector
template<class T>
inline typename growvector<T>::iterator
erase_to_end(growvector<T> &l, typename growvector<T>::iterator where)
{
  // DEBUG << "erase_to_end - growvector" << std::endl;
  return l.erase_to_end(where);
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_printPercentage()
// --------------------------------------------------------------------------
inline void 
CAdaptiveCaching::_printPercentage(unsigned left,unsigned total,unsigned freq)
{
  static unsigned pixProcessed=0;
  if(++pixProcessed>freq) {
    pixProcessed = 0;
    STATUS 
      << "   - " << ((left<=total)?((float)(total-left)/total*100):0)
      <<  "%    \r";
  }
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::MulByKd()
// --------------------------------------------------------------------------
inline void
MulByKd(unsigned numPixels,float *irradPtr,const CHitInfoLite *hitPtr)
{
  for(unsigned i=0; i<numPixels; i++,irradPtr+=3,hitPtr++) {
    for(unsigned ch=0; ch<3; ch++)
      irradPtr[ch] *= hitPtr->kd[ch];
  }
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::MulByKd()
// --------------------------------------------------------------------------
inline void
AddDirectIllum(unsigned numPixels,float *irradPtr,const CHitInfoLite *hitPtr)
{
  for(unsigned i=0; i<numPixels; i++,irradPtr+=3,hitPtr++) {
    for(unsigned ch=0; ch<3; ch++)
      irradPtr[ch] += hitPtr->directIllum[ch];
  }
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::GetHandleFlag()
// --------------------------------------------------------------------------
template<class TRecord>
inline unsigned
CAdaptiveCaching::GetHandleFlag()
{ 
  FATAL << "Only specialized version of GetHandleFlag() can be called.\n";
  FATAL_EXIT;
  return 0;
}

template<>
inline unsigned
CAdaptiveCaching::GetHandleFlag<CIrradianceRecord>()
{ 
  return CHitInfoLite::EE_IrrCaching;
}

template<>
inline unsigned
CAdaptiveCaching::GetHandleFlag<CRadianceRecord>()
{ 
  return CHitInfoLite::EE_RadCaching;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::GetNumRecords()
// --------------------------------------------------------------------------
template<class TRecord>
inline unsigned long
CAdaptiveCaching::GetNumRecords(unsigned level) const
{ 
  FATAL << "Only specialized version of GetNumRecords() can be called.\n";
  FATAL_EXIT;
  return 0;
}

template<>
inline unsigned long
CAdaptiveCaching::GetNumRecords<CIrradianceRecord>(unsigned level) const
{ 
  assert(_icAlgo);
  return _icAlgo->GetNumRecords(level);
}

template<>
inline unsigned long
CAdaptiveCaching::GetNumRecords<CRadianceRecord>(unsigned level) const
{ 
  assert(_rcAlgo);
  return _rcAlgo->GetNumRecords(level);
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::GetRecordList()
// --------------------------------------------------------------------------
template<class TRecord>
inline CRecordList<TRecord>& 
CAdaptiveCaching::GetRecordList()
{
  FATAL << "Only specialized version of GetRecordList() can be called.\n";
  FATAL_EXIT;
  return _icAlgo->GetICache(0)->GetRecordList(); // dummy return to satisfy the compiler
}

template<>
inline CIrradianceList& 
CAdaptiveCaching::GetRecordList<CIrradianceRecord>()
{
  return _icAlgo->GetICache(0)->GetRecordList();
}

template<>
inline CRadianceList& 
CAdaptiveCaching::GetRecordList<CRadianceRecord>()
{
  return _rcAlgo->GetRCache(0)->GetRecordList();
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::GetRecordList()
// --------------------------------------------------------------------------
template<class TRecord>
inline CRCOctree<TRecord>*
CAdaptiveCaching::GetOctree()
{
  FATAL << "Only specialized version of GetOctree() can be called.\n";
  FATAL_EXIT;
  return 0; // dummy return to satisfy the compiler
}

template<>
inline CRCOctree<CIrradianceRecord>*
CAdaptiveCaching::GetOctree<CIrradianceRecord>()
{
  return _icAlgo->GetICache(0);
}

template<>
inline CRCOctree<CRadianceRecord>*
CAdaptiveCaching::GetOctree<CRadianceRecord>()
{
  return _rcAlgo->GetRCache(0);
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::GetMaxError()
// --------------------------------------------------------------------------
template<class TRecord>
inline float
CAdaptiveCaching::GetMaxError()
{
  FATAL << "Only specialized version of GetMaxError() can be called.\n";
  FATAL_EXIT;
  return 0; // dummy return to satisfy the compiler
}

template<>
inline float
CAdaptiveCaching::GetMaxError<CIrradianceRecord>()
{
  return _maxErrorIC;
}

template<>
inline float
CAdaptiveCaching::GetMaxError<CRadianceRecord>()
{
  return _maxErrorRC;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::GetRecPosCol()
// --------------------------------------------------------------------------
template<class TRecord>
inline const float*
CAdaptiveCaching::GetRecPosCol()
{
  FATAL << "Only specialized version of GetRecPosCol() can be called.\n";
  FATAL_EXIT;
  static float dummycol[]={0,0,0};
  return dummycol;
}

template<>
inline const float*
CAdaptiveCaching::GetRecPosCol<CIrradianceRecord>()
{
  static float iccol[]={0,1,0};
  return iccol;
}

template<>
inline const float*
CAdaptiveCaching::GetRecPosCol<CRadianceRecord>()
{
  static float rccol[]={1,1,0};
  return rccol;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::GetDotForThreshold()
// --------------------------------------------------------------------------
template<class TRecord>
inline float
CAdaptiveCaching::GetDotForThreshold(const CHitInfoLite &hit)
{
  FATAL << "Only specialized version of GetDotForThreshold() can be called.\n";
  FATAL_EXIT;
  return 1; // dummy return to satisfy the compiler
}

template<>
inline float
CAdaptiveCaching::GetDotForThreshold<CIrradianceRecord>(const CHitInfoLite &hit)
{
  return 1;
}

template<>
inline float
CAdaptiveCaching::GetDotForThreshold<CRadianceRecord>(const CHitInfoLite &hit)
{
  return hit.GetDot();
}


// --------------------------------------------------------------------------
//  CAdaptiveCaching::GetCacheAbbrev()
// --------------------------------------------------------------------------
template<class TRecord>
inline const char*
CAdaptiveCaching::GetCacheAbbrev()
{
  FATAL << "Only specialized version of GetCacheAbbrev() can be called.\n";
  FATAL_EXIT;
  return 0; // dummy return to satisfy the compiler
}

template<>
inline const char*
CAdaptiveCaching::GetCacheAbbrev<CIrradianceRecord>()
{
  return "ic";
}

template<>
inline const char*
CAdaptiveCaching::GetCacheAbbrev<CRadianceRecord>()
{
  return "rc";
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_updatePixelContribList()
// --------------------------------------------------------------------------
template<class TRecord>
inline unsigned
CAdaptiveCaching::_updatePixelContribList(CPixContribList<TRecord> &clist,
                                          const CHitInfoLite       &hit)
{
  unsigned numValid = 0;
  typename CPixContribList<TRecord>::iterator it = clist.begin();
  for ( ; it != clist.end(); it++ ) {
    CRecordContrib<TRecord> &item = *it;
    if ( item.changeClock < item.rec->GetChangeClock() ) {
      if( item.rec->ComputeWeight(hit.P,hit.N,item.w) )
        item.changeClock = item.rec->GetChangeClock();
      else {
        item.SetInvalid();
        clist.UnsetProcessPixelDone();
        continue;
      }
    }
    numValid++;
  }
  return numValid;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_findFirstInvalidContrib()
// --------------------------------------------------------------------------
template<class TRecord>
inline typename CPixContribList<TRecord>::iterator
CAdaptiveCaching::_findFirstInvalidContrib(CPixContribList<TRecord> &clist,
                                           const CHitInfoLite       &hit)
{
  typename CPixContribList<TRecord>::iterator it = clist.begin();
  for ( ; it != clist.end(); it++ ) {
    CRecordContrib<TRecord> &item = *it;
    if ( item.IsInvalid() )
      break; // invalid contribution
    if ( item.ChangeClock() < item.Rec()->GetChangeClock() ) {
      // 'R' or 'a' has changed, the whole weight has to be re-evaluated
      if( item.Rec()->ComputeWeight(hit.P,hit.N,item.W()) ) {
        // update the contribution clock
        item.ChangeClock() = item.Rec()->GetChangeClock();
      }
      else {
        // invalid contribution
        break;
      }
    }
  }
  return it;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_addContrib
// --------------------------------------------------------------------------
template<class TRecord>
inline void
CAdaptiveCaching::_addContrib(CPixContribList<TRecord>      &clist,
                              const CRecordContrib<TRecord> &newItem,
                              const CHitInfoLite            &hit)
{
  // add the contribution to the list
  // try to replace an invalid item in the list
  typename CPixContribList<TRecord>::iterator it = _findFirstInvalidContrib(clist,hit);
  if(it!=clist.end()) 
  {
    // replace the invalid contribution
    (*it) . SetInvalid();
    (*it) = newItem;
  }
  else
  {
    // add new item
    clist.push_back(newItem);
  }
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_initSaveDebugImages()
// --------------------------------------------------------------------------
template<class TRecord>
inline void
CAdaptiveCaching::_initSaveDebugImages()
{
  _sdiStartRecords = GetNumRecords<TRecord>();
  _debugImageIndex = 0;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_doSaveDebugImages()
// --------------------------------------------------------------------------
template<class TRecord>
inline void
CAdaptiveCaching::_doSaveDebugImages(CTimer &timer,
                                     CICSplatterRegular &splatter,
                                     CPixContribList<TRecord> *contribImage,
                                     bool checkFrequency)
{
  if(_saveDebugImages) {
    timer.Stop();
    _doSaveDebugImagesCode<TRecord>(splatter,contribImage,checkFrequency);
    timer.Start();
  }
}

__END_GOLEM_HEADER

#endif //  __ICADAPT_H__
