// ===================================================================
// $Id: icadapt2.cpp,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// icadapt2.cpp
//   Perceptually-based adaptive quality setting for irradiance cache.
//
// Initial coding by Jaroslav Krivanek (August, 2005).

// GOLEM headers
#include "config.h"

#include "icadapt.h"
#include "timer.h"
#include "icalgo.h"
#include "icache.h"

#include "fileio.h"
#include "vdm.h"
#include "jkimage.h"

// SPHAR headers
#include "shvec.h"

// standard headers
#include <algorithm>
#include <assert.h>

__BEGIN_GOLEM_SOURCE

#define REMOVE_INVALID_FROM_PIXEL_LIST

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_unsetChanged()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_unsetChanged(CRecordList<TRecord> &recList)
{
  typename CRecordList<TRecord>::const_iterator it = recList.begin();
  for( ; it!=recList.end(); it++)
    (*it)->UnsetChanged();
}
template void CAdaptiveCaching::_unsetChanged<CIrradianceRecord>(CIrradianceList &l);
template void CAdaptiveCaching::_unsetChanged<CRadianceRecord>(CRadianceList &l);

// --------------------------------------------------------------------------
//  _checkHitInfo()
// --------------------------------------------------------------------------
inline static void _checkHitInfo(const CHitInfoLite &hit)
{
#ifdef _DEBUG_ICADAPT
  if( !hit.object ) {
    FATAL << "Irradiance caching where should not happen." << std::endl;
    FATAL_EXIT;
  }
#endif
}

// --------------------------------------------------------------------------
//  _checkCleanPixels()
// --------------------------------------------------------------------------
inline static void _checkCleanPixels(const CLinearPixelList &pixelList)
{
#ifdef _DEBUG_ICADAPT
  unsigned nPixels = pixelList.GetWidth() * pixelList.GetHeight();
  for(unsigned idx=0; idx<nPixels; idx++) {
    if(pixelList.GetDirtyMap()[idx]) {
      FATAL << "Pixel left dirty after convergence pass." << std::endl;
      PRINTVAR(idx);
      PRINTVAR(pixelList.GetDirtyMap()[idx]);
      FATAL_EXIT;
    }
  }
#endif
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_checkValidContribsCode()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::_checkValidContribsCode(CICSplatterRegular &splatter)
{
  // verify that all pixels have a valid contribution through the octree queries
  STATUS << "- verifying all pixels have a contribution...";
  CTimer timer;
  timer.Restart();
  CICache *icache = _icAlgo->GetICache(0);
  const CHitInfoLite *hitInfoImage = splatter.GetHitInfoImage();
  CPixICContribPtrList *contribPtrImage = splatter.GetContribPtrImage();
  unsigned nPixels = splatter.GetWidth() * splatter.GetHeight();
  icache->SetCachingMode(CCachingMode::EE_AlwaysRead);
  for(unsigned idx=0; idx<nPixels; idx++) {
    const CHitInfoLite &hit = hitInfoImage[idx];
    if(!splatter.IsHandled(hit)) continue;
    if( !icache->InterpolateIrradiance<true,false>(hit.P,hit.N,tmpE,-1) ) {
      FATAL << "No countribution for pixel " << idx << std::endl;
      CPixICContribPtrList::iterator it=contribPtrImage[idx].begin();
      for( ; it!=contribPtrImage[idx].end(); it++) {
        PRINTVAR(W(*it) * Rec(*it)->GetA());
        PRINTVAR(W(*it));
        PRINTVAR(IsInvalid(*it));
        PRINTVAR(ChangeClock(*it));
        PRINTVAR(Rec(*it)->GetChangeClock());
        bool bw = Rec(*it)->ComputeWeight(hit.P,hit.N,W(*it));
        PRINTVAR(W(*it));
        PRINTVAR(bw);
        DEBUG << "===================" << std::endl;
      }
      FATAL_EXIT;
    }
  }
  timer.Stop();
  STATUS << "done in " << timer.UserTime() << "s" << std::endl;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_checkValidContribs()
// --------------------------------------------------------------------------
inline void 
CAdaptiveCaching::_checkValidContribs(CICSplatterRegular &splatter)
{
#ifdef _DEBUG_ICADAPT
  _checkValidContribsCode(splatter);
#endif
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_destroyRecordContribPtrs()
// --------------------------------------------------------------------------
template<class Tlist>
void
CAdaptiveCaching::_destroyRecordList(const CIrradianceList &recList)
{
  CIrradianceList::const_iterator it;
  for(it = recList.begin(); it!=recList.end(); it++) {
    CIrradianceRecord *rec = *it;
    if(!rec->GetData()) continue;
    // cast the record data
    Tlist* list = (Tlist*)rec->GetData(); 
    // delete the allocated contributions
    list->clear();
    // delete the list itself
    delete list;
    rec->SetData(0);
  }
}
template void CAdaptiveCaching::_destroyRecordList<CRecContribList>(
  const CIrradianceList &recList);
template void CAdaptiveCaching::_destroyRecordList<CRecICContribPtrList>(
  const CIrradianceList &recList);

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_destroyRecordData()
// --------------------------------------------------------------------------
template<class TRecord,class Tdata>
void
CAdaptiveCaching::_destroyRecordData(const CRecordList<TRecord> &recList)
{
  typename CRecordList<TRecord>::const_iterator it;
  for(it = recList.begin(); it!=recList.end(); it++) {
    const TRecord *rec = *it;
    if(!rec->GetData()) continue;
    // cast the record data
    Tdata *recdata = (Tdata*)rec->GetData(); 
    // delete the bbox
    delete recdata;
    rec->SetData(0);
  }
}

template void 
CAdaptiveCaching::_destroyRecordData<CIrradianceRecord,CAdaptiveCaching::CProjPosAndBBox>
(const CIrradianceList &recList);
template void
CAdaptiveCaching::_destroyRecordData<CIrradianceRecord,float>
(const CIrradianceList &recList);

template void 
CAdaptiveCaching::_destroyRecordData<CRadianceRecord,CAdaptiveCaching::CProjPosAndBBox>
(const CRadianceList &recList);
template void
CAdaptiveCaching::_destroyRecordData<CRadianceRecord,float>
(const CRadianceList &recList);


// --------------------------------------------------------------------------
//  CAdaptiveCaching::_updatePixelContribPtrList()
// --------------------------------------------------------------------------
inline unsigned
CAdaptiveCaching::_updatePixelContribPtrList(CPixICContribPtrList &clist,
                                             const CHitInfoLite &hit)
{
  CPixICContribPtrList::iterator src_it  = clist.begin();
  unsigned numValid = 0;

  CPixICContribPtrList::iterator dest_it  = src_it;
  // first without copying
  for ( ; src_it != clist.end(); src_it++ ) {
    TPixICContribPtr item = *src_it;
    if ( item->ChangeClock() < item->Rec()->GetChangeClock() ) {
      if( item->Rec()->ComputeWeight(hit.P,hit.N,item->W()) ) {
        item->ChangeClock() = item->Rec()->GetChangeClock();
      }
      else {
        item->SetInvalid();
        dest_it = src_it++;
        goto update_with_copy;
      }
    }
  }
  return clist.size();
update_with_copy:

  // then with copying
  while ( src_it != clist.end() ) {
    TPixICContribPtr item = *src_it;
    if ( item->ChangeClock() < item->Rec()->GetChangeClock() ) {
      // 'R' or 'a' has changed, the whole weight has to be re-evaluated
      if( item->Rec()->ComputeWeight(hit.P,hit.N,item->W()) ) {
        // update the contribution clock
        item->ChangeClock() = item->Rec()->GetChangeClock();
      }
      else {
        // remove from the list and set the contribution as invalid
        item->SetInvalid();
        src_it++;
        continue;
      }
    }
    *dest_it++ = *src_it++;
  }

  // erase all at once
  erase_to_end(clist,dest_it);
  return clist.size();
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_findFirstInvalidContribPtr()
// --------------------------------------------------------------------------
inline CPixICContribPtrList::iterator
CAdaptiveCaching::_findFirstInvalidContribPtr(CPixICContribPtrList &clist,
                                              const CHitInfoLite &hit)
{
  CPixICContribPtrList::iterator it = clist.begin();
  while ( it != clist.end() ) {
    TPixICContribPtr item = *it;
    if ( item->IsInvalid() )
      break; // invalid contribution
    if ( item->ChangeClock() < item->Rec()->GetChangeClock() ) {
      // 'R' or 'a' has changed, the whole weight has to be re-evaluated
      if( item->Rec()->ComputeWeight(hit.P,hit.N,item->W()) ) {
        // update the contribution clock
        item->ChangeClock() = item->Rec()->GetChangeClock();
      }
      else {
        // invalid contribution
        break;
      }
    }
    it++;
  }
  return it;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_addContribPtr
// --------------------------------------------------------------------------
inline void
CAdaptiveCaching::_addContribPtr(CPixICContribPtrList &clist,
                                 const TPixICContribPtr &item,
                                 const CHitInfoLite &hit)
{
  // add the contribution to the list
  // try to replace an invalid item in the list
  CPixICContribPtrList::iterator it = _findFirstInvalidContribPtr(clist,hit);
  if(it!=clist.end()) 
  {
    // replace the invalid contribution
    (*it) -> SetInvalid();
    (*it) = item;
  }
  else
  {
    // add new item
    clist.push_back(item);
  }
}

// --------------------------------------------------------------------------
//   CAdaptiveCaching::_rasterizeContribPtrs()
// --------------------------------------------------------------------------
template<class T>
int 
CAdaptiveCaching::_rasterizeContribPtrs(const CIrradianceRecord *rec, 
                                        CICSplatterRegular &splatter,
                                        CLinearPixelList &pixelList,
                                        T &addList,
                                        const CICSplatterRegular::TBBox& r)
{
  assert(r.xmin>=0); assert(r.ymin>=0); 
  assert(r.xmax<splatter.GetWidth()); assert(r.ymax<splatter.GetHeight());

  int numPixels=0;
  for(int y=r.ymin; y <= r.ymax; y++) {
    int idx = splatter.Idx(r.xmin,y);
    const CHitInfoLite *hitptr = splatter.GetHitInfoImage() + idx;
    CPixICContribPtrList *contribptrptr = splatter.GetContribPtrImage() + idx;
    for(int x=r.xmin; x <= r.xmax; x++, hitptr++, contribptrptr++, idx++) {
      float weight;
      if(splatter.IsHandled(hitptr) && rec->ComputeWeight(hitptr->P,hitptr->N,weight)){
        // add the contribution to the additional list
        addList.push_back(CICContrib(rec,weight,idx));
        // add item to the pixel contribution list
        _addContribPtr(*contribptrptr,&addList.back(),*hitptr);
        // schedule pixel for processing
        if(contribptrptr->size()>1)
          pixelList.push(idx);
        // increase counter
        numPixels++;
      }
    }
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::_splatContribPtrs()
// --------------------------------------------------------------------------
template<class T>
int 
CAdaptiveCaching::_splatContribPtrs(const CIrradianceRecord* rec,
                                    CICSplatterRegular &splatter,
                                    CLinearPixelList &pixelList,
                                    T &addList)
{
  if(splatter.CheckBackFacing(rec)) return 0;
  // number of affected pixels
  int numPixels = 0;
  // Compute the record position in camera coordinates
  CVector3D projPos = splatter.Modelview(rec->GetPoint());
  // compute the screen bounding box
  CICSplatterRegular::TBBox bbox;
  if( splatter.ComputeBBox(rec,rec->GetA(),projPos,bbox) ) {
    // rasterize the record
    numPixels += _rasterizeContribPtrs<T>(rec,splatter,pixelList,addList,bbox);
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_removeInvalidRecordContribs()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::_removeInvalidRecordContribs(CIrradianceList &recList,
                                               CICSplatterRegular &splatter,
                                               CLinearPixelList &pixelList)
{
  const CHitInfoLite *hitInfoImage = splatter.GetHitInfoImage();
  CPixICContribPtrList *contribPtrImage = splatter.GetContribPtrImage();

  unsigned invalidRemoved = 0;
  unsigned changeRemoved = 0;

#ifdef _DEBUG_ICADAPT
  // DEBUG << "check before..." << std::endl;
  unsigned nPixels = splatter.GetWidth() * splatter.GetHeight();
  // is there a pixel without recorded contribution?
  for(unsigned idx=0; idx<nPixels; idx++) {
    const CHitInfoLite &hit = hitInfoImage[idx];
    if(hit.object && contribImage[idx].empty()) {
      FATAL << "check before..." << std::endl;
      FATAL << "pixel " << idx << "has no recorded contributions " << std::endl;
      FATAL_EXIT;
    }
  }
#endif

  CIrradianceList::iterator rec_it=recList.begin();
  for( ; rec_it!=recList.end(); rec_it++) {
    CIrradianceRecord *rec = *rec_it;
    // if record hasn't changed since the last pass, not worth worrying about it
    if(!rec->Changed()) continue;
    rec->UnsetChanged();
    // cast the record data
    CRecContribList* clist = (CRecContribList*)rec->GetData(); 
    if(!clist) continue;
    // check the validity of the record contibutions
    CRecContribList::iterator iter = clist->begin();
    while ( iter!=clist->end() ) {
      CICContrib &item = *iter;
      if(item.IsInvalid()) {
        // It has already been removed from the pixel contribution list
        // (in _updatePixelContributionList())
        iter = clist->erase(iter);
        invalidRemoved++;
        continue;
      }
      else if( item.ChangeClock() < item.Rec()->GetChangeClock() ) {
        // re-evaluate weight
        TPixelIndex idx = item.GetPixelIdx();
        const CHitInfoLite &hit = hitInfoImage[idx];
        // 'R' has changed, the whole weight has to be re-evaluated
        if( item.Rec()->ComputeWeight(hit.P,hit.N,item.W()) ) {
          // update the contribution clock
          item.ChangeClock() = item.Rec()->GetChangeClock();
        }
        else {
          // schedule pixel for processing
          pixelList.push(idx);
          // remove from the pixel list
          CPixICContribPtrList &pixContribs = contribPtrImage[idx];
          CPixICContribPtrList::iterator pix_iter = 
            find(pixContribs.begin(),pixContribs.end(), &(*iter));
          if(pix_iter != pixContribs.end()) {
            pixContribs.erase(pix_iter);
          }
          else {
            FATAL << "Contribution not found in its pixel list." << std::endl;
            PRINTVAR(idx);
            FATAL_EXIT;
          }
          // remove from the record list
          iter = clist->erase(iter);
          // update stats
          changeRemoved ++;
          continue;
        }
      }
      // advance to the next contribution in the record contribution list
      iter++;
    }
  }

#ifdef _DEBUG_ICADAPT
  DEBUG << std::endl;
  PRINTVAR(invalidRemoved);
  PRINTVAR(changeRemoved);

  unsigned numEmptyScheduled=0;
  // DEBUG << "check after..." << std::endl;
  // is there a pixel without recorded contribution?
  for(unsigned idx=0; idx<nPixels; idx++) {
    const CHitInfoLite &hit = hitInfoImage[idx];
    if(hit.object && contribImage[idx].empty()) {
      if(!pixelList.GetDirtyMap()[idx]) {
        FATAL << "check after..." << std::endl;
        FATAL << "pixel " << idx << "has no recorded contribs " << std::endl;
        FATAL_EXIT;
      }
      else {
        numEmptyScheduled++;
      }
    }
  }
  PRINTVAR(numEmptyScheduled);
#endif
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_finalNormalization()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::_finalNormalization(CTimer &timer,
                                      CICSplatterRegular &splatter)
{
  typedef CIrradianceRecord TRecord;

  STATUS << "- starting normalization...";
  timer.Restart();
  unsigned nPixels = splatter.GetWidth() * splatter.GetHeight();
  // alloc weight image if not in the splatter
  float *weightImage = 0;
  if(!splatter.GetWeightImage()) {
    weightImage = new float[nPixels];
    assert(weightImage);
    splatter.SetWeightImage(weightImage);
  }
  splatter.Clear(CICSplatterRegular::EE_Irradiance|CICSplatterRegular::EE_Weight);
  // splat records
  if(_jitterWeightIC)
    splatter.Splat<true>(_icAlgo->GetICache(0)->GetRecordList());
  else
    splatter.Splat<false>(_icAlgo->GetICache(0)->GetRecordList());
  splatter.NormalizeBuffer();
  // delete allocated weight image
  if(weightImage) {
    delete [] weightImage;
    splatter.SetWeightImage(0);
  }
  // visualize record positions
  if(_icAlgo->_visualizeNewRecPos) {
    splatter.SetRecPosColor(GetRecPosCol<TRecord>());
    splatter.SplatRecordPos(_icAlgo->GetICache(0)->GetRecordList());
  }
  timer.Stop();
  STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_removeInvalidRecordContribsVerbose()
// --------------------------------------------------------------------------
inline void
CAdaptiveCaching::_removeInvalidRecordContribsVerbose(
  CTimer &timer,
  CIrradianceList &recList,
  CICSplatterRegular &splatter,
  CLinearPixelList &pixelList)
{
  STATUS << "   - removing invalid contribs...";
  timer.Restart();
  _removeInvalidRecordContribs(_icAlgo->GetICache(0)->GetRecordList(),
    splatter,pixelList);
  timer.Stop();
  STATUS 
    << "done in " << timer.UserTime() << " s / found " << pixelList.size()
    << " pix to process." << std::endl;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::IndirectIlluminationOverlapConverge()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::IndirectIlluminationOverlapConverge(
  float              *resultImage,
  const CHitInfoLite *hitInfoImage,
  int                 imgWidth,
  int                 imgHeight,
  CHitPointInfo      &outHitInfo,
  CMatrix4           &projMat,
  CMatrix4           &modelviewMat)
{
  // --------------------------
  //  SETUP
  // --------------------------
  CTimer timer, timer1, timerTotal, addRecTimer;
  float addRecTotalTime=0;
  timerTotal.Restart();
  STATUS << "- adaptive caching, convergence algo: list"  << std::endl;
  STATUS << "- initializing...";
  timer.Restart();
  // number of pixels
  unsigned nPixels = imgWidth*imgHeight;
  // canvas for contributions
  CPixICContribPtrList *contribImage = new CPixICContribPtrList[nPixels];
  // init the splatter
  CICSplatterRegular splatter(projMat,modelviewMat,imgWidth,imgHeight,hitInfoImage,
    CHitInfoLite::EE_IrrCaching);
  splatter.SetIrradImage(resultImage);
  splatter.SetContribPtrImage(contribImage);
  // allocate memory for pixel contributions
  for(unsigned idx=0; idx<nPixels; idx++)
    if(splatter.IsHandled(hitInfoImage[idx]))
      contribImage[idx].reserve(6);
  // pixel list for image traversal
  CLinearPixelList pixelList(imgWidth*imgHeight,hitInfoImage);
  FillPixelList(pixelList,splatter);
  // IC interface
  CICache *icache = _icAlgo->GetICache(0);
  // turn off smooth weight for the convergence
  bool smoothWardWeight = CIrradianceRecord::smoothWardWeight;
  CIrradianceRecord::smoothWardWeight = false;
  unsigned numRecordsBeforePass,passPixels,nContribs=0;
  unsigned numRecordsBefore = icache->GetNumRecords();
  // threshold elevation related stuff
  CVDMRamasubramanian *vdm = 0;
  jkimage<float> auxElevationImage;
  const float *thrElevImage=0;
  CValues avgE(CStandardValues::GetDefaultColorType());
  avgE.setz();

  timer.Stop();
  STATUS << "done in " << timer.UserTime() << "s" << std::endl;

  // --------------------------------------
  //    SPLAT PREVIOUS CACHE CONTENTS
  // --------------------------------------
 if(icache->GetNumRecords()>0) {
    STATUS 
      << "- splatting existing cache contents ("
      << icache->GetNumRecords() << " recs)...";
    timer.Restart();
    CIrradianceList &recList = icache->GetRecordList();
    for(CIrradianceList::iterator it=recList.begin(); it!=recList.end(); it++) {
      CIrradianceRecord *rec= *it;
      // create the contribution list for this record
      CRecContribList *clist = new CRecContribList;
      rec->SetData(clist);
      // splat new record
      unsigned thisContribs = 
        _splatContribPtrs(rec,splatter,pixelList,*clist);
      nContribs += thisContribs;
      // update average irradiance
      avgE += (rec->GetE()*thisContribs);
    }
    timer.Stop();
    STATUS << "done in " << timer.UserTime() << "sec." << std::endl;
  }

  static const unsigned maxContribsToCleanUp = 3000000;

  if(_useThresholdElevation) {
    // --------------------------------------
    //  INITIAL PASS FOR THRESHOLD ELEVATION
    // --------------------------------------
    STATUS << "- initial pass for threshold elevation..." << std::endl;
    addRecTimer.Reset();
    timer.Restart();
    numRecordsBeforePass = icache->GetNumRecords();
    passPixels = pixelList.size();
    // CLinearPixelList::const_iterator pixIter = pixelList.begin();
    while( !pixelList.empty() ) {
      _printPercentage(pixelList.size(),passPixels);
      unsigned idx = pixelList.front();
      pixelList.pop();
      const CHitInfoLite &hit = hitInfoImage[idx];
      // list of contributions for this pixel
      CPixICContribPtrList &pixContribs = contribImage[idx];
      // we consider all contributions are valid in this first pass, since no
      // refinement is done and we disregard neighbor clamping here
      unsigned numValid = pixContribs.size();
      if(numValid==0) {
        // add new record
        addRecTimer.Start();
        CIrradianceRecord *rec = _addRec<CIrradianceRecord>(hit,outHitInfo);
        addRecTimer.Stop();
        // create the contribution list for this record
        CRecContribList *clist = new CRecContribList;
        rec->SetData(clist);
        // splat new record
        unsigned thisContribs = 
          _splatContribPtrs(rec,splatter,pixelList,*clist);
        nContribs += thisContribs;
        // update average irradiance
        avgE += (rec->GetE()*thisContribs);
      }
    }
    if(nContribs) avgE /= nContribs;
    FillPixelList(pixelList,splatter);
    // check if all pixels are clean
    _checkCleanPixels(pixelList);
    // for all records, remove invalid contributions
    _removeInvalidRecordContribsVerbose(timer1,icache->GetRecordList(),
      splatter,pixelList);
    timer.Stop();
    STATUS
      << "- initial pass finished in " << timer.UserTime() << " s / added " 
      << icache->GetNumRecords()-numRecordsBeforePass << " recs in " 
      << addRecTimer.UserTime() << "s" << std::endl;
    addRecTotalTime += addRecTimer.UserTime();

    // -----------------------------------
    //  create threshold elevation image
    // -----------------------------------
    vdm = CreateThresholdElavationImage(resultImage,hitInfoImage,
      imgWidth,imgHeight,avgE.data());
    assert(vdm);
    thrElevImage = vdm->GetElevationImage().data();

    timerTotal.Stop();
    if(_saveVDMInOut) {
      // save the VDM input
      WriteImageByExt(imgWidth,imgHeight,3,resultImage,
        CreateFileName("-vdm-input"));
      // save the VDM output
      WriteImageByExt(imgWidth,imgHeight,1,thrElevImage,
        CreateFileName("-vdm-output"));
    }
    timerTotal.Start();
  }
  else 
  {
    auxElevationImage.resize(imgWidth,imgHeight);
    auxElevationImage.set(1);
    thrElevImage = auxElevationImage.data();
  }

  // --------------------------
  //  PIXEL TRAVERSAL
  // --------------------------
  
  // set records to unchanged state
  _unsetChanged(icache->GetRecordList());

  for(unsigned passCount=0; !pixelList.empty(); passCount++) {

    addRecTimer.Reset();
    STATUS << "- pixel traversal pass " << passCount  << "..." << std::endl;
    timer.Restart();
    numRecordsBeforePass = icache->GetNumRecords();
    nContribs = 0;
    passPixels = pixelList.size();

    while( !pixelList.empty() ) {
      _printPercentage(pixelList.size(),passPixels);
      unsigned idx = pixelList.front();
      pixelList.pop();
      // current hit point info
      const CHitInfoLite &hit = hitInfoImage[idx];
      // debugging check
      _checkHitInfo(hit);
      // list of contributions for this pixel
      CPixICContribPtrList &pixContribs = contribImage[idx];   
      // remove contributions that aren't valid anymore
      unsigned numValid = _updatePixelContribPtrList(pixContribs,hit);
      switch(numValid) {
      case 0:
        {
          // add new record
          addRecTimer.Start();
          CIrradianceRecord *rec = _addRec<CIrradianceRecord>(hit,outHitInfo);
          addRecTimer.Stop();
          // create the contribution list for this record
          CRecContribList *clist = new CRecContribList;
          rec->SetData(clist);
          // splat new record
          nContribs += _splatContribPtrs(rec,splatter,pixelList,*clist);
          // clean up invalid contribs
          if(nContribs >= maxContribsToCleanUp ) {
            _removeInvalidRecordContribsVerbose(timer1,icache->GetRecordList(),
              splatter,pixelList);
            nContribs = 0;
          }
        }
        break;
      case 1:
        break;
      default: // 2 and more
          // _extrapolateEValid(pixContribs.begin(),pixContribs.end(),hit.P,hit.N);
        FATAL << "extrapolate & nasobeni kd tu chybi ." << std::endl;
        FATAL_EXIT;
          _processPixelOverlap(hit,hit.directIllum.data(), 
            thrElevImage[idx], pixContribs,icache);
        break;
      }
    }
    //// pixel traversal finished for this pass ////
    // check if all pixels are clean
    _checkCleanPixels(pixelList);
    // for all records, remove invalid contributions
    _removeInvalidRecordContribsVerbose(timer1,icache->GetRecordList(),splatter,pixelList);
    // print pass stats
    timer.Stop();
    STATUS
      << "- pass finished in " << timer.UserTime() << " s / added " 
      << icache->GetNumRecords()-numRecordsBeforePass << " recs in " 
      << addRecTimer.UserTime() << "s" << std::endl;
    addRecTotalTime += addRecTimer.UserTime();
  }

  // check if allpixels have a valid contribution
  _checkValidContribs(splatter);
  // do the final normalization
  CIrradianceRecord::smoothWardWeight = smoothWardWeight;
  _finalNormalization(timer,splatter);
  // multiply by the kdImage
  MulByKd(nPixels,resultImage,hitInfoImage);
  // print final stats
  timerTotal.Stop();
  _printFinalStats(timerTotal.UserTime(),addRecTotalTime,
    icache->GetNumRecords()-numRecordsBefore);
  // delete the contributions of all the records
  _destroyRecordList<CRecContribList>( icache->GetRecordList() );
  // Delete pixel contribution lists
  delete [] contribImage;
  delete vdm;
}


__END_GOLEM_SOURCE
