// ===================================================================
// $Id: icadapt3.cpp,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// icadapt3.cpp
//   Perceptually-based adaptive quality setting for irradiance cache.
//
// Initial coding by Jaroslav Krivanek (August, 2005).

// GOLEM headers
#include "config.h"

#include "icadapt.h"
#include "timer.h"
#include "icalgo.h"
#include "icache.h"

#include "fileio.h"
#include "vdm.h"
#include "jkimage.h"

// LIBSPHAR headers
#include "shvec.h"

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  CAdaptiveCaching::IndirectIlluminationOverlapConvergeTree()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::IndirectIlluminationOverlapConvergeTree(
  float              *resultImage,
  const CHitInfoLite *hitInfoImage,
  int                 imgWidth,
  int                 imgHeight,
  CHitPointInfo      &outHitInfo,
  CMatrix4           &projMat,
  CMatrix4           &modelviewMat)
{
  typedef CIrradianceRecord TRecord;
  // --------------------------
  //  SETUP
  // --------------------------
  CTimer timer, timer1, timerTotal, addRecTimer;
  float addRecTotalTime=0;
  timerTotal.Restart();
  STATUS << "- adaptive caching, convergence algo: tree"  << std::endl;
  STATUS << "- initializing...";
  timer.Restart();
  // number of pixels
  unsigned nPixels = imgWidth*imgHeight;
  // canvas for contribution count
  TPixContribCount *contribCountImage = new TPixContribCount[nPixels];
  shv<TPixContribCount>::setz(nPixels,contribCountImage);
  // init the splatter
  CICSplatterRegular splatter(projMat,modelviewMat,imgWidth,imgHeight,hitInfoImage,
    CHitInfoLite::EE_IrrCaching);
  splatter.SetIrradImage(resultImage);
  splatter.SetContribCountImage(contribCountImage);
  // pixel list for image traversal
  CLinearPixelList pixelList(imgWidth*imgHeight,hitInfoImage);
  FillPixelList(pixelList,splatter);
  // IC interface
  CICache *icache = _icAlgo->GetICache(0);
  // turn off smooth weight for the convergence
  bool smoothWardWeight = CIrradianceRecord::smoothWardWeight;
  CIrradianceRecord::smoothWardWeight = false;
  unsigned numRecordsBeforePass,passPixels,nContribs=0;
  unsigned numRecordsBefore = icache->GetNumRecords();
  // threshold elevation related stuff
  CVDMRamasubramanian *vdm = 0;
  jkimage<float> auxElevationImage;
  const float *thrElevImage=0;
  CValues avgE(CStandardValues::GetDefaultColorType());
  avgE.setz();

  timer.Stop();
  STATUS << "done in " << timer.UserTime() << "s" << std::endl;

  // --------------------------------------
  //    SPLAT PREVIOUS CACHE CONTENTS
  // --------------------------------------
  if(icache->GetNumRecords()>0) {
    STATUS 
      << "- splatting existing cache contents ("
      << icache->GetNumRecords() << " recs)...";
    timer.Restart();
    CIrradianceList &recList = icache->GetRecordList();
    for(CIrradianceList::iterator it=recList.begin(); it!=recList.end(); it++) {
      CIrradianceRecord *rec= *it;
        // create the bounding box for this records
        CProjPosAndBBox *recdata = new CProjPosAndBBox;
        rec->SetData(recdata);
        // splat new record & set 
        unsigned thisContribs = 
          _splatContribCount(rec,splatter,pixelList,*recdata);
        nContribs += thisContribs;
        // update average irradiance
        avgE += (rec->GetE()*thisContribs);
    }
    timer.Stop();
    STATUS << "done in " << timer.UserTime() << "sec." << std::endl;
  }

  if(_useThresholdElevation) {
    // --------------------------------------
    //  INITIAL PASS FOR THRESHOLD ELEVATION
    // --------------------------------------
    STATUS << "- initial pass for threshold elevation..." << std::endl;
    addRecTimer.Reset();
    timer.Restart();
    numRecordsBeforePass = icache->GetNumRecords();
    passPixels = pixelList.size();
    // CLinearPixelList::const_iterator pixIter = pixelList.begin();
    while( !pixelList.empty() ) {
      _printPercentage(pixelList.size(),passPixels);
      unsigned idx = pixelList.front();
      pixelList.pop();
      const CHitInfoLite &hit = hitInfoImage[idx];
      // we consider all contributions are valid in this first pass, since no
      // refinement is done and we disregard neighbor clamping here
      unsigned numValid = contribCountImage[idx];
      if(numValid==0) {
        // add new record
        addRecTimer.Start();
        CIrradianceRecord *rec = _addRec<CIrradianceRecord>(hit,outHitInfo);
        addRecTimer.Stop();
        // create the bounding box for this records
        CProjPosAndBBox *recdata = new CProjPosAndBBox;
        rec->SetData(recdata);
        // splat new record & set 
        unsigned thisContribs = 
          _splatContribCount(rec,splatter,pixelList,*recdata);
        nContribs += thisContribs;
        // update average irradiance
        avgE += (rec->GetE()*thisContribs);
      }
    }
    if(nContribs) avgE /= nContribs;
    FillPixelList(pixelList,splatter);
    timer.Stop();
    STATUS
      << "- initial pass finished in " << timer.UserTime() << " s / added " 
      << icache->GetNumRecords()-numRecordsBeforePass << " recs in " 
      << addRecTimer.UserTime() << "s" << std::endl;
    addRecTotalTime += addRecTimer.UserTime();

    // -----------------------------------
    //  create threshold elevation image
    // -----------------------------------
    vdm = CreateThresholdElavationImage(resultImage,hitInfoImage,
      imgWidth,imgHeight,avgE.data());
    assert(vdm);
    thrElevImage = vdm->GetElevationImage().data();
    timerTotal.Stop();
    if(_saveVDMInOut) {
      // save the VDM input
      WriteImageByExt(imgWidth,imgHeight,3,resultImage,
        CreateFileName("-vdm-input"));
      // save the VDM output
      WriteImageByExt(imgWidth,imgHeight,1,thrElevImage,
        CreateFileName("-vdm-output"));
    }
    timerTotal.Start();
  }
 else 
  {
    auxElevationImage.resize(imgWidth,imgHeight);
    auxElevationImage.set(1);
    thrElevImage = auxElevationImage.data();
  }

  // --------------------------
  //  PIXEL TRAVERSAL
  // --------------------------
  
  // set records to unchanged state
  _unsetChanged(icache->GetRecordList());

  for(unsigned passCount=0; !pixelList.empty(); passCount++) {

    addRecTimer.Reset();
    STATUS << "- pixel traversal pass " << passCount  << "..." << std::endl;
    timer.Restart();
    numRecordsBeforePass = icache->GetNumRecords();
    nContribs = 0;
    passPixels = pixelList.size();

    while( !pixelList.empty() ) {
      _printPercentage(pixelList.size(),passPixels);
      unsigned idx = pixelList.front();
      pixelList.pop();
      // current hit point info
      const CHitInfoLite &hit = hitInfoImage[idx];
      // debugging check
      // _checkHitInfo(hit);

      // query the tree for contributions
      static CICContribArray pixContribs;
      pixContribs.clear();
      icache->GetContributions<true,false>(hit.P,hit.N,pixContribs,-1);
      contribCountImage[idx] = pixContribs.size();

      switch(pixContribs.size()) {
      case 0:
        {
          // add new record
          addRecTimer.Start();
          CIrradianceRecord *rec = _addRec<CIrradianceRecord>(hit,outHitInfo);
          addRecTimer.Stop();
          // create the bounding box for this records
          CProjPosAndBBox *recdata = new CProjPosAndBBox;
          rec->SetData(recdata);
          // splat new record & set 
          nContribs += _splatContribCount(rec,splatter,pixelList,*recdata);
        }
        break;
      case 1:
        break;
      default: // 2 and more
        // _extrapolateRadianceValid(pixContribs,hit);
        // EtrapolateE(
        FATAL << "Tady se to musi extrapolovat a vynasobit kd." << std::endl;
        FATAL_EXIT;
        _processPixelOverlap(hit,hit.directIllum.data(), 
          thrElevImage[idx], pixContribs,icache);
        break;
      }
    }
    // pixel traversal finished for this pass

    // check if all pixels are clean
    // _checkCleanPixels(pixelList);

    // check if all pixels have a valid contribution
    _checkValidContribs(splatter);

    // schedule pixels in changed record's bboxes for processing
    _scheduleChangedRecordPixelsVerbose<TRecord>(timer1,
      icache->GetRecordList(),splatter,pixelList);

    timer.Stop();
    STATUS
      << "- pass finished in " << timer.UserTime() << " s / added " 
      << icache->GetNumRecords()-numRecordsBeforePass << " recs in " 
      << addRecTimer.UserTime() << "s" << std::endl;
    addRecTotalTime += addRecTimer.UserTime();
  }

  // check if all pixels have a valid contribution
  _checkValidContribs(splatter);
  // do the final normalization
  CIrradianceRecord::smoothWardWeight = smoothWardWeight;
  _finalNormalization(timer,splatter);
  // multiply by kd
  MulByKd(nPixels,resultImage,hitInfoImage);
  // print final stats
  timerTotal.Stop();
  _printFinalStats(timerTotal.UserTime(),addRecTotalTime,
    icache->GetNumRecords()-numRecordsBefore);
  // Delete the contributions of all the records.
  _destroyRecordData<CIrradianceRecord,CProjPosAndBBox>( icache->GetRecordList() );
  // Delete pixel contributions
  delete [] contribCountImage;
  delete vdm;
}


// --------------------------------------------------------------------------
//  CAdaptiveCaching::_scheduleChangedRecordPixelsVerbose()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_scheduleChangedRecordPixelsVerbose(
  CTimer &timer,
  CRecordList<TRecord> &recList,
  CICSplatterRegular &splatter,
  CLinearPixelList &pixelList)
{
  STATUS << "   - scheduling from bboxes...";
  timer.Restart();
  _scheduleChangedRecordPixels(GetRecordList<TRecord>(),splatter,pixelList);
  timer.Stop();
  STATUS 
    << "done in " << timer.UserTime() << " s / " << pixelList.size()
    << " pix to process." << std::endl;
}

template
void
CAdaptiveCaching::_scheduleChangedRecordPixelsVerbose(
  CTimer &timer,
  CRecordList<CIrradianceRecord> &recList,
  CICSplatterRegular &splatter,
  CLinearPixelList &pixelList);

template
void
CAdaptiveCaching::_scheduleChangedRecordPixelsVerbose(
  CTimer &timer,
  CRecordList<CRadianceRecord> &recList,
  CICSplatterRegular &splatter,
  CLinearPixelList &pixelList);

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_scheduleChangedRecordPixels()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_scheduleChangedRecordPixels(CRecordList<TRecord> &recList,
                                               CICSplatterRegular &splatter,
                                               CLinearPixelList &pixelList)
{
  const CHitInfoLite *hitInfoImage = splatter.GetHitInfoImage();

  unsigned invalidRemoved = 0;
  unsigned changeRemoved = 0;

  typename CRecordList<TRecord>::iterator rec_it=recList.begin();
  for( ; rec_it!=recList.end(); rec_it++) {
    TRecord *rec = *rec_it;
    // if record hasn't changed since the last pass, not worth worrying about it
    if(!rec->Changed()) continue;
    rec->UnsetChanged();
    // cast the record data
    CProjPosAndBBox* recdata = (CProjPosAndBBox*)rec->GetData(); 
    if(!recdata) {
      FATAL << "Record with no bbox found!" << std::endl;
      FATAL_EXIT;
    }
    CICSplatterRegular::TBBox &r=recdata->bbox;
    // schedule all pixels within the old box for processing
    for(int y=r.ymin; y<=r.ymax; y++) {
      int idx = splatter.Idx(r.xmin,y);
      const CHitInfoLite *hitptr = splatter.GetHitInfoImage()+idx;
      for(int x=r.xmin; x <= r.xmax; x++, hitptr++, idx++) {
        if(splatter.IsHandled(hitptr))
          pixelList.push(idx);
      }
    }
    // compute the new bounding box
    if(!splatter.ComputeBBox(rec,rec->GetA(),recdata->projPos,r)) {
      FATAL << "- record outside screen area" << std::endl;
    }
  }
}

// --------------------------------------------------------------------------
//   CAdaptiveCaching::_rasterizeContribCount()
// --------------------------------------------------------------------------
inline int
CAdaptiveCaching::_rasterizeContribCount(const CIrradianceRecord *rec, 
                                         CICSplatterRegular &splatter,
                                         CLinearPixelList &pixelList,
                                         const CICSplatterRegular::TBBox& r)
{
  assert(r.xmin>=0); assert(r.ymin>=0); 
  assert(r.xmax<splatter.GetWidth()); assert(r.ymax<splatter.GetHeight());

  float weight;
  int numPixels=0;
  for(int y=r.ymin; y <= r.ymax; y++) {
    int idx = splatter.Idx(r.xmin,y);
    const CHitInfoLite *hitptr = splatter.GetHitInfoImage() + idx;
    TPixContribCount *contribcntptr = splatter.GetContribCountImage() + idx;
    for(int x=r.xmin; x <= r.xmax; x++, hitptr++, contribcntptr++, idx++) {
      if(splatter.IsHandled(hitptr) && rec->ComputeWeight(hitptr->P,hitptr->N,weight)){
        contribcntptr[0]++;
        // schedule pixel for processing
        if( contribcntptr[0]>1 )
          pixelList.push(idx);
        // increase counter
        numPixels++;
      }
    }
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::_splatContribCount()
// --------------------------------------------------------------------------
int 
CAdaptiveCaching::_splatContribCount(const CIrradianceRecord* rec,
                                CICSplatterRegular &splatter,
                                CLinearPixelList &pixelList,
                                CProjPosAndBBox &recdata)
{
  if(splatter.CheckBackFacing(rec)) return 0;
  // number of affected pixels
  int numPixels;
  // Compute the record position in camera coordinates
  recdata.projPos = splatter.Modelview(rec->GetPoint());
  // compute the screen bounding box
  if( splatter.ComputeBBox(rec,rec->GetA(),recdata.projPos,recdata.bbox) ) {
    // rasterize the record
    numPixels = _rasterizeContribCount(rec,splatter,pixelList,recdata.bbox);
  }
  else {
    FATAL << "Record bbox outside screen area." << std::endl;
    FATAL_EXIT;
  }
  return numPixels;
}


__END_GOLEM_SOURCE

