// ===================================================================
// $Id: icadapt4.cpp,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// icadapt4.cpp
//   Perceptually-based adaptive quality setting for irradiance cache.
//
// Initial coding by Jaroslav Krivanek (August, 2005).

// GOLEM headers
#include "config.h"

#include "icadapt.h"
#include "timer.h"
#include "icalgo.h"
#include "icache.h"

#include "fileio.h"
#include "vdm.h"
#include "jkimage.h"

// LIBSPHAR headers
#include "shvec.h"

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_finalNormalizationBBox()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_finalNormalizationBBox(CTimer &timer,
                                          CPixContribList<TRecord> *contribImage,
                                          CICSplatterRegular &splatter,
                                          float *resultImage)
{
  FATAL 
    << "Only specialized version of _finalNormalizationBBox() can be called."
    << std::endl;
  FATAL_EXIT;
  return;
}

template<>
void
CAdaptiveCaching::_finalNormalizationBBox<CIrradianceRecord>(
  CTimer &timer,
  CPixICContribList *contribImage,
  CICSplatterRegular &splatter,
  float *resultImage)
{
  typedef CIrradianceRecord TRecord;

  STATUS << "- starting normalization...";
  timer.Restart();
  unsigned nPixels = splatter.GetWidth() * splatter.GetHeight();
  // alloc weight image if not in the splatter
  float *weightImage = 0;
  if(!splatter.GetWeightImage()) {
    weightImage = new float[nPixels];
    assert(weightImage);
    splatter.SetWeightImage(weightImage);
  }
  splatter.Clear(CICSplatterRegular::EE_Irradiance|CICSplatterRegular::EE_Weight);
  // splat records
  if(_jitterWeightIC)
    splatter.Splat<true>(GetRecordList<CIrradianceRecord>());
  else
    splatter.Splat<false>(GetRecordList<CIrradianceRecord>());
  splatter.NormalizeBuffer();
  // delete allocated weight image
  if(weightImage) {
    delete [] weightImage;
    splatter.SetWeightImage(0);
  }
  // visualize record positions
  if(_icAlgo->_visualizeNewRecPos) {
    splatter.SetRecPosColor(GetRecPosCol<TRecord>());
    splatter.SplatRecordPos(_icAlgo->GetICache(0)->GetRecordList());
  }

  // multiply by the kdImage
  const CHitInfoLite *hitPoints = splatter.GetHitInfoImage();
  unsigned numHitPoints = splatter.GetNumHitPoints();
  for(unsigned idx=0; idx<numHitPoints; idx++) {
    const CHitInfoLite &hit = hitPoints[idx];
    if(hitPoints[idx].IsSet(CHitInfoLite::EE_IrrCaching)) {
      float *res = resultImage + 3*idx;
      const float *irrad = splatter.GetIrradImage() + 3*idx;
      CTriCol c = hit.kd;
      c *= irrad;
      c.CopyData(res);
    }
  }

  timer.Stop();
  STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
}

// specialization for radiance caching
template<>
void
CAdaptiveCaching::_finalNormalizationBBox<CRadianceRecord>(
  CTimer &timer,
  CPixRCContribList *contribImage,
  CICSplatterRegular &splatter,
  float *resultImage)
{
  STATUS << "- starting RC normalization...";
  timer.Restart();

  // jittering and smooth weight is nor suppotred here

  _renderFromContribs(resultImage,contribImage,splatter.GetHitInfoImage(),
    splatter.GetNumHitPoints(),false);

  timer.Stop();
  STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
}


// --------------------------------------------------------------------------
//   CAdaptiveCaching::_rasterizeContrib()
// --------------------------------------------------------------------------
template<class TRecord>
inline int
CAdaptiveCaching::_rasterizeContrib(const TRecord *rec, 
                                    CICSplatterRegular &splatter,
                                    CPixContribList<TRecord> *contribImage,
                                    CLinearPixelList &pixelList,
                                    const CICSplatterRegular::TBBox& r)
{
  assert(r.xmin>=0); assert(r.ymin>=0); 
  assert(r.xmax<splatter.GetWidth()); assert(r.ymax<splatter.GetHeight());

  const CHitInfoLite *hitInfoImage = splatter.GetHitInfoImage();
  // CPixICContribList *contribImage = splatter.GetContribImage();

  int numPixels=0;
  for(int y=r.ymin; y <= r.ymax; y++) {
    int idx = splatter.Idx(r.xmin,y);
    const CHitInfoLite *hitptr = hitInfoImage + idx;
    CPixContribList<TRecord> *contribptr = contribImage + idx;
    for(int x=r.xmin; x <= r.xmax; x++, hitptr++, contribptr++, idx++) {
      float weight;
      if(splatter.IsHandled(hitptr) && rec->ComputeWeight(hitptr->P,hitptr->N,weight)){
        // add item to the pixel contribution list
        _addContrib(*contribptr,CRecordContrib<TRecord>(rec,weight),
          *hitptr);
        contribptr->UnsetProcessPixelDone();
        // schedule pixel for processing
        if(contribptr->size()>1)
          pixelList.push(idx);
        // increase counter
        numPixels++;
      }
    }
  }
  return numPixels;
}


// --------------------------------------------------------------------------
//   CICSplatterRegular::_splatContrib()
// --------------------------------------------------------------------------
template<class TRecord>
int 
CAdaptiveCaching::_splatContrib(const TRecord* rec,
                                CICSplatterRegular &splatter,
                                CPixContribList<TRecord> *contribImage,
                                CLinearPixelList &pixelList,
                                CProjPosAndBBox &recdata)
{
  if(splatter.CheckBackFacing(rec)) return 0;
  // number of affected pixels
  int numPixels = 0;
  // Compute the record position in camera coordinates
  recdata.projPos = splatter.Modelview(rec->GetPoint());
  // compute the screen bounding box
  if( splatter.ComputeBBox(rec,rec->GetA(),recdata.projPos,recdata.bbox) ) {
    // rasterize the record
    numPixels += _rasterizeContrib(rec,splatter,contribImage,pixelList,
      recdata.bbox);
  }
  return numPixels;
}

template
int 
CAdaptiveCaching::_splatContrib(const CIrradianceRecord* rec,
                                CICSplatterRegular &splatter,
                                CPixICContribList *contribImage,
                                CLinearPixelList &pixelList,
                                CProjPosAndBBox &recdata);

template
int 
CAdaptiveCaching::_splatContrib(const CRadianceRecord* rec,
                                CICSplatterRegular &splatter,
                                CPixRCContribList *contribImage,
                                CLinearPixelList &pixelList,
                                CProjPosAndBBox &recdata);


// --------------------------------------------------------------------------
//  CAdaptiveCaching::_printFinalStats()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::_printFinalStats(float totalTime,
                                   float addRecTotalTime,
                                   unsigned numRec)
{
  _totalTime = totalTime;
  _addRecTotalTime = addRecTotalTime;
  _numRec = numRec;

  float overhead = totalTime - addRecTotalTime;
  STATUS 
    << "- convergence finished" << std::endl
    << "   numRecords        " << numRec << std::endl
    << "   totalTime         " << totalTime ;
  if(numRec>0)
    STATUS << " (= " << totalTime/numRec*1000 << " ms/rec).";
  STATUS << std::endl;
  STATUS << "   addRecTotalTime   " << addRecTotalTime;
  if(numRec>0)
    STATUS << " (= " << addRecTotalTime/numRec*1000 << " ms/rec).";
  STATUS << std::endl;
  STATUS << "   overhead          " << overhead;
  if(numRec>0)
    STATUS << " (= " << overhead/numRec*1000 << " ms/rec).";
  STATUS << std::endl;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::CreateThresholdElevationImage()
// --------------------------------------------------------------------------
CVDMRamasubramanian*
CAdaptiveCaching::CreateThresholdElavationImage(
  float *resultImage,
  const CHitInfoLite *hitImage,
  unsigned imgWidth,
  unsigned imgHeight,
  const float *avgIrrad) const
{
  // create the input for the threshold elevation model
  unsigned nPixels = imgWidth*imgHeight;
  float *riPtr = resultImage;
  const CHitInfoLite *hiPtr = hitImage;
  for(unsigned k=0; k<nPixels; k++, riPtr+=3, hiPtr++) {
    for(unsigned ch=0; ch<3; ch++)
      riPtr[ch]=hiPtr->directIllum[ch] + hiPtr->kd[ch]*avgIrrad[ch];
  }
  // convert to luminances image
  jkimage<float> auxImage(imgWidth,imgHeight);
  for(unsigned i=0; i<nPixels; i++)
    auxImage[i]=CVDM::LumFromRGB(resultImage+3*i);
  // build the threshold elevation map
  return new CVDMRamasubramanian(auxImage,_useMasking,true,_saveVDMDebugImages);
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_doSaveDebugImagesCode()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_doSaveDebugImagesCode(
  CICSplatterRegular &splatter,
  CPixContribList<TRecord> *contribImage,
  bool checkFrequency)
{
  int addedRecords = GetNumRecords<TRecord>()- _sdiStartRecords;
  if(checkFrequency && addedRecords%_debugImageSaveFrequency != 0)
    return;

  STATUS << "- saving debug image #" << _debugImageIndex << std::endl;

  const char *icrc = GetCacheAbbrev<TRecord>();

  char fname[200];
  unsigned imgWidth = splatter.GetWidth();
  unsigned imgHeight=splatter.GetHeight();
#if 0
  // save record positions
  splatter.SetRecPosColor(GetRecPosCol<TRecord>());
  splatter.Clear(CICSplatterRegular::EE_Weight|CICSplatterRegular::EE_Irradiance);
  splatter.SplatRecordPos(GetRecordList<TRecord>());
  sprintf(fname,"recpos-%s%06u.tif",icrc,_debugImageIndex);
  WriteImageTiff(imgWidth,imgHeight,3,splatter.GetIrradImage(),fname);
#endif
#if 0
  // save approximation error
  splatter.SetRecPosColor(GetRecPosCol<TRecord>());
  splatter.Clear(CICSplatterRegular::EE_Weight|CICSplatterRegular::EE_Irradiance);
  splatter.SplatA(GetRecordList<TRecord>());
  splatter.NormalizeBuffer();
  sprintf(fname,"approx-%s%06u.tif",icrc,_debugImageIndex);
  WriteImageTiff(imgWidth,imgHeight,3,splatter.GetIrradImage(),fname);
#endif
#if 1
  // save the image itself
  splatter.Clear(CICSplatterRegular::EE_Weight|CICSplatterRegular::EE_Irradiance);
  // splatter.Splat<false>(GetRecordList<TRecord>());
  // splatter.NormalizeBuffer();
  // MulByKd(imgWidth*imgHeight,splatter.GetIrradImage(),splatter.GetHitInfoImage());
  _renderFromContribs(splatter.GetIrradImage(),contribImage,
    splatter.GetHitInfoImage(),splatter.GetNumHitPoints(),true);

  splatter.SetRecPosColor(GetRecPosCol<TRecord>());
  splatter.SplatRecordPos(GetRecordList<TRecord>());

  sprintf(fname,"irrad-%s%06u.tif",icrc,_debugImageIndex);
  WriteImageTiff(imgWidth,imgHeight,3,splatter.GetIrradImage(),fname);
#endif
  _debugImageIndex++;
}

template void
CAdaptiveCaching::_doSaveDebugImagesCode<CIrradianceRecord>(
  CICSplatterRegular &splatter,CPixICContribList *contribImage,
  bool checkFrequency);

template void
CAdaptiveCaching::_doSaveDebugImagesCode<CRadianceRecord>(
  CICSplatterRegular &splatter,CPixRCContribList *contribImage,
  bool checkFrequency);

// --------------------------------------------------------------------------
//  CAdaptiveCaching::IndirectIlluminationOverlapConvergeBBox()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::IndirectIlluminationOverlapConvergeBBox(
  float              *resultImage,
  const CHitInfoLite *hitInfoImage,
  int                 imgWidth,
  int                 imgHeight,
  CHitPointInfo      &outHitInfo,
  CMatrix4           &projMat,
  CMatrix4           &modelviewMat)
{
  // --------------------------
  //  SETUP
  // --------------------------
  CTimer timer, timer1, timerTotal, addRecTimer;
  float addRecTotalTime=0;
  timerTotal.Restart();
  STATUS << "- adaptive caching, convergence algo: bbox" << std::endl;
  STATUS << "- initializing...";
  timer.Restart();
  // number of pixels
  unsigned nPixels = imgWidth*imgHeight;
  // set-up indices
  for(unsigned idx=0; idx<nPixels; idx++)
    hitInfoImage[idx].dataIdx = idx;  // 1-to-1 mapping
  // canvas for contributions
  CPixContribList<TRecord> *contribImage = 
    new CPixContribList<TRecord>[nPixels];
  assert(contribImage);
  float *weightImage=new float[nPixels]; 
  assert(weightImage);
  float *irradImage=new float[nPixels*3];
  assert(nPixels);
  // init the splatter
  CICSplatterRegular splatter(projMat,modelviewMat,imgWidth,imgHeight,
    hitInfoImage,GetHandleFlag<TRecord>());
  splatter.SetIrradImage(irradImage);
  splatter.SetWeightImage(weightImage);
  // splatter.SetContribImage(contribImage);

  // pixel list for image traversal
  CLinearPixelList pixelList(imgWidth*imgHeight,hitInfoImage);
  FillPixelList(pixelList,splatter);
  // IC interface
  // CICache *icache = _icAlgo->GetICache(0);
  // allocate memory for pixel contributions
  for(unsigned idx=0; idx<nPixels; idx++)
    if(splatter.IsHandled(hitInfoImage[idx])) contribImage[idx].reserve(6);
  // turn off smooth weight for the convergence
  bool smoothWardWeight = TRecord::smoothWardWeight;
  TRecord::smoothWardWeight = false;
  unsigned numRecordsBeforePass,passPixels,nContribs=0;
  unsigned numRecordsBefore = GetNumRecords<TRecord>();
  // threshold elevation related stuff
  CVDMRamasubramanian *vdm = 0;
  jkimage<float> auxElevationImage;
  const float *thrElevImage=0;
  // debug image saving
  _initSaveDebugImages<TRecord>();
  CValues avgE(CStandardValues::GetDefaultColorType());
  avgE.setz();
  timer.Stop();
  STATUS << "done in " << timer.UserTime() << "s" << std::endl;

  // --------------------------------------
  //    SPLAT PREVIOUS CACHE CONTENTS
  // --------------------------------------
 if(GetNumRecords<TRecord>()>0) {
    STATUS 
      << "- splatting existing cache contents ("
      << GetNumRecords<TRecord>() << " recs)...";
    timer.Restart();
    CRecordList<TRecord> &recList = GetRecordList<TRecord>();
    typename CRecordList<TRecord>::iterator it=recList.begin();
    for( ; it!=recList.end(); it++) {
      TRecord *rec= *it;
      // create the bounding box for this records
      CProjPosAndBBox *recdata = new CProjPosAndBBox;
      rec->SetData(recdata);
      // splat new record
      unsigned thisContribs = _splatContrib(rec,splatter,contribImage,
        pixelList,*recdata);
      nContribs += thisContribs;
      // update average irradiance
      avgE += (GetRecE(rec)*thisContribs);
    }
    timer.Stop();
    STATUS << "done in " << timer.UserTime() << "sec." << std::endl;
 }

  _doSaveDebugImages<TRecord>(timerTotal,splatter,contribImage,false);

  if(_useThresholdElevation) {
    // --------------------------------------
    //  INITIAL PASS FOR THRESHOLD ELEVATION
    // --------------------------------------
    STATUS << "- initial pass for threshold elevation..." << std::endl;
    addRecTimer.Reset();
    timer.Restart();
    numRecordsBeforePass = GetNumRecords<TRecord>();
    passPixels = pixelList.size();
    // CLinearPixelList::const_iterator pixIter = pixelList.begin();
    while( !pixelList.empty() ) {
      _printPercentage(pixelList.size(),passPixels);
      unsigned idx = pixelList.front();
      pixelList.pop();
      const CHitInfoLite &hit = hitInfoImage[idx];
      // list of contributions for this pixel
      CPixContribList<TRecord> &pixContribs = contribImage[idx];
      // we consider all contributions are valid in this first pass, since no
      // refinement is done and we disregard neighbor clamping here
      unsigned numValid = pixContribs.size();
      if(numValid==0 && !pixContribs.AddRecFailed()) {
        // add new record
        addRecTimer.Start();
        TRecord *rec = _addRec<TRecord>(hit,outHitInfo);
        addRecTimer.Stop();
        if(rec) {
          // create the bounding box for this records
          CProjPosAndBBox *recdata = new CProjPosAndBBox;
          rec->SetData(recdata);
          // splat new record
          unsigned thisContribs = 
            _splatContrib(rec,splatter,contribImage,pixelList,*recdata);
          nContribs += thisContribs;
          // update average irradiance
          avgE += (GetRecE(rec)*thisContribs);
          // save debug images
          _doSaveDebugImages<TRecord>(timerTotal,splatter,contribImage);
        }
        else {
          pixContribs.SetAddRecFailed();
        }
      }
    }
    if(nContribs) avgE /= nContribs;
    FillPixelList(pixelList,splatter);
    // save debug images
    _doSaveDebugImages<TRecord>(timerTotal,splatter,contribImage,false);
    timer.Stop();
    STATUS
      << "- initial pass finished in " << timer.UserTime() << " s / added " 
      << GetNumRecords<TRecord>()-numRecordsBeforePass << " recs in " 
      << addRecTimer.UserTime() << "s" << std::endl;
    addRecTotalTime += addRecTimer.UserTime();

    // -----------------------------------
    //  create threshold elevation image
    // -----------------------------------
    vdm = CreateThresholdElavationImage(resultImage,hitInfoImage,imgWidth,
      imgHeight,avgE.data());
    assert(vdm);
    thrElevImage = vdm->GetElevationImage().data();

    timerTotal.Stop();
    if(_saveVDMInOut) {
      // save the VDM input
      WriteImageByExt(imgWidth,imgHeight,3,resultImage,
        CreateFileName("-vdm-input"));
      // save the VDM output
      WriteImageByExt(imgWidth,imgHeight,1,thrElevImage,
        CreateFileName("-vdm-output"));
    }
    timerTotal.Start();
  }
  else {
    auxElevationImage.resize(imgWidth,imgHeight);
    auxElevationImage.set(1);
    thrElevImage = auxElevationImage.data();
  }

  // --------------------------
  //  PIXEL TRAVERSAL
  // --------------------------
  
  // set records to unchanged state
  _unsetChanged(GetRecordList<TRecord>());

  for(unsigned passCount=0; !pixelList.empty(); passCount++) {

    addRecTimer.Reset();
    STATUS << "- pixel traversal pass " << passCount  << "..." << std::endl;
    timer.Restart();
    numRecordsBeforePass = GetNumRecords<TRecord>();
    nContribs = 0;
    passPixels = pixelList.size();

    while( !pixelList.empty() ) {
      _printPercentage(pixelList.size(),passPixels);
      unsigned idx = pixelList.front();
      pixelList.pop();
      // current hit point info
      const CHitInfoLite &hit = hitInfoImage[idx];
      // debugging check
      // _checkHitInfo(hit);
      // list of contributions for this pixel
      CPixContribList<TRecord> &pixContribs = contribImage[idx];
  
      // remove contributions that aren't valid anymore
      unsigned numValid = _updatePixelContribList(pixContribs,hit);

      switch(numValid) {
      case 0:
        if(!pixContribs.AddRecFailed() && _adaptive) {
          // add new record
          addRecTimer.Start();
          TRecord *rec = _addRec<TRecord>(hit,outHitInfo);
          addRecTimer.Stop();
          if(rec) {
            // create the bounding box for this records
            CProjPosAndBBox *recdata = new CProjPosAndBBox;
            rec->SetData(recdata);
            // splat new record
            nContribs += 
              _splatContrib(rec,splatter,contribImage,pixelList,*recdata);
            // save debug images
            _doSaveDebugImages<TRecord>(timerTotal,splatter,contribImage);
          }
          else {
            pixContribs.SetAddRecFailed();
          }
        }
        break;
      case 1:
        break;
      default: // 2 and more
        if(!pixContribs.ProcessPixelDone())
        {
          _extrapolateRadianceValid(pixContribs,hit);
          _processPixelOverlap(hit,hit.directIllum.data(), 
            thrElevImage[idx], pixContribs,GetOctree<TRecord>());
          pixContribs.SetProcessPixelDone();
        }
        break;
      }
    }
    // pixel traversal finished for this pass

    // check if all pixels are clean
    // _checkCleanPixels(pixelList);

    // schedule pixels in changed record's bboxes for processing
    _scheduleChangedRecordPixelsVerbose(timer1,GetRecordList<TRecord>(),
      splatter,pixelList);

    // save debug images
    _doSaveDebugImages<TRecord>(timerTotal,splatter,contribImage,false);

    timer.Stop();
    STATUS
      << "- pass finished in " << timer.UserTime() << " s / added " 
      << GetNumRecords<TRecord>()-numRecordsBeforePass << " recs in " 
      << addRecTimer.UserTime() << "s" << std::endl;
    addRecTotalTime += addRecTimer.UserTime();
  }

  // check if allpixels have a valid contribution
  _checkValidContribs(splatter);
  // do the final normalization
  CIrradianceRecord::smoothWardWeight = smoothWardWeight;
  _finalNormalizationBBox(timer,contribImage,splatter,resultImage);
  // print final stats
  timerTotal.Stop();
  _printFinalStats(timerTotal.UserTime(),addRecTotalTime,
    GetNumRecords<TRecord>()-numRecordsBefore);
  // Delete the record data.
  _destroyRecordData<TRecord,CProjPosAndBBox>( GetRecordList<TRecord>() );
  // Delete pixel contribution lists
  delete [] contribImage;
  delete [] weightImage;
  delete [] irradImage;
  delete vdm;
}


template
void
CAdaptiveCaching::IndirectIlluminationOverlapConvergeBBox<CIrradianceRecord>(
  float              *resultImage,
  const CHitInfoLite *hitInfoImage,
  int                 imgWidth,
  int                 imgHeight,
  CHitPointInfo      &outHitInfo,
  CMatrix4           &projMat,
  CMatrix4           &modelviewMat);

template
void
CAdaptiveCaching::IndirectIlluminationOverlapConvergeBBox<CRadianceRecord>(
  float              *resultImage,
  const CHitInfoLite *hitInfoImage,
  int                 imgWidth,
  int                 imgHeight,
  CHitPointInfo      &outHitInfo,
  CMatrix4           &projMat,
  CMatrix4           &modelviewMat);

__END_GOLEM_SOURCE
