// ===================================================================
// $Id: icadapt5.cpp,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// icadapt5.cpp
//   Perceptually-based adaptive quality setting for irradiance cache.
//
// Initial coding by Jaroslav Krivanek (August, 2005).

// GOLEM headers
#include "config.h"

#include "icadapt.h"
#include "timer.h"
#include "icalgo.h"
#include "icache.h"

#include "icsplati.h"

#include "fileio.h"
// #include "vdm.h"
// #include "jkimage.h"

// LIBSPHAR headers
#include "shvec.h"

// standard headers
#include <cfloat>

__BEGIN_GOLEM_SOURCE

// static inline bool _weird(double f) { return !_finite(f) || _isnan(f); }
// static inline bool _weird(const CVector3D &v) { return _weird(v.x) || _weird(v.y) || _weird(v.z); }

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_addJitter()
// --------------------------------------------------------------------------
template<class TRecord>
unsigned
CAdaptiveCaching::_addJitter(CPixContribList<TRecord> *contribImage,
                             CICSplatterIrregular &splatter)
{
  unsigned numPixels=0;
  CRecordList<TRecord> &list = GetRecordList<TRecord>();
  typename CRecordList<TRecord>::const_iterator recit=list.begin();

  for( ; recit!=list.end(); recit++) {
    TRecord *rec = *recit;
    // let the splatter locate points within the bounding sphere
    const CICSplatterIrregular::CNearestPointArray &nearestPoints
      = splatter.FindInSphere(rec->GetP(),
      rec->GetValidDomain()*TRecord::jitterAmount);
    CICSplatterIrregular::CNearestPointArray::const_iterator it = 
      nearestPoints.begin();
    for( ; it != nearestPoints.end(); ++it) {
      float weight;
      const CHitInfoLite &hit = *(it->pt);
      if(rec->template ComputeWeightTemplate<false>(hit.P,hit.N,weight))
        continue; // this contribution is already there
      if(!rec->template ComputeWeightTemplateDistPassed<true>(hit.P,hit.N,it->d2,weight))
        continue; // no new contribution
      // add new contribution
      unsigned idx = hit.dataIdx;
      CPixContribList<TRecord> &pixContribs = contribImage[idx];
      // add item to the pixel contribution list
      _addContrib(pixContribs,CRecordContrib<TRecord>(rec,weight),hit);
      pixContribs.UnsetProcessPixelDone();
      numPixels++;
    }
  }
  return numPixels;
}

template unsigned CAdaptiveCaching::_addJitter(
  CPixContribList<CIrradianceRecord> *contribImage,
  CICSplatterIrregular &splatter);

template unsigned CAdaptiveCaching::_addJitter(
  CPixContribList<CRadianceRecord> *contribImage,
  CICSplatterIrregular &splatter);

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_reevaluateWeights()
// --------------------------------------------------------------------------
template<bool jitter,class TRecord>
void
CAdaptiveCaching::_reevaluateWeights(CPixContribList<TRecord> *contribImage,
                                     CICSplatterIrregular &splatter)
{
  unsigned handleFlag = GetHandleFlag<TRecord>();
  const CHitInfoLite *hitPoints = splatter.GetHitInfoImage();
  unsigned numHitPoints = splatter.GetNumHitPoints();
  for(unsigned idx=0; idx<numHitPoints; idx++) {
    const CHitInfoLite &hit = hitPoints[idx];
    if(hitPoints[idx].IsSet(handleFlag)) {
      CPixContribList<TRecord> &pixContribs = contribImage[hit.dataIdx];
      typename CPixContribList<TRecord>::iterator it=pixContribs.begin();
      for( ; it!=pixContribs.end(); it++) {
        if(IsInvalid(*it)) continue;
        if(!it->rec->template ComputeWeightTemplate<jitter>(hit.P,hit.N,it->w)) {
          FATAL 
            << "Inconsistency: weight is zero after copnvergence finished."
            << std::endl;
          FATAL_EXIT;
        }
      }
    }
  }
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_renderFromContribs()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_renderFromContribs(float *resultImage,
                                      CPixContribList<TRecord> *contribImage,
                                      const CHitInfoLite *hitPoints,
                                      unsigned numHitPoints,
                                      bool updatePixelList)
{
  const unsigned handleFlag = GetHandleFlag<TRecord>();
  for(unsigned idx=0; idx<numHitPoints; idx++) {
    const CHitInfoLite &hit = hitPoints[idx];
    if(hit.IsSet(handleFlag)) {
      float *res = resultImage + 3*idx;
      CPixContribList<TRecord> &pixContribs = contribImage[hit.dataIdx];
      if(updatePixelList)_updatePixelContribList(pixContribs,hit);
      _extrapolateRadianceValid(pixContribs,hit);
      float sumWeight = 0;
      CTriCol sumE(0.0);
      SumValidContributions(pixContribs,sumWeight,sumE);
      if(sumWeight == 0 ) {
        sumE.setz();
        sumE.CopyData(res);
      }
      else {
        sumE /= sumWeight;
        sumE.CopyData(res);
      }
    }
  }
}

template
void
CAdaptiveCaching::_renderFromContribs(float *resultImage,
                                      CPixContribList<CIrradianceRecord> *contribImage,
                                      const CHitInfoLite *hitPoints,
                                      unsigned numHitPoints,
                                      bool updatePixelList);

template
void
CAdaptiveCaching::_renderFromContribs(float *resultImage,
                                      CPixContribList<CRadianceRecord> *contribImage,
                                      const CHitInfoLite *hitPoints,
                                      unsigned numHitPoints,
                                      bool updatePixelList);


// --------------------------------------------------------------------------
//  CAdaptiveCaching::_finalNormalization()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_finalNormalization(CTimer &timer,
                                      CPixContribList<TRecord> *contribImage,
                                      CICSplatterIrregular &splatter,
                                      float *resultImage)
{
  FATAL 
    << "Only specialized version of _finalNormalization() can be called."
    << std::endl;
  FATAL_EXIT;
  return ;
}

// specializetion for irradiance caching
template<>
void
CAdaptiveCaching::_finalNormalization<CIrradianceRecord>(
  CTimer &timer,
  CPixICContribList *contribImage,
  CICSplatterIrregular &splatter,
  float *resultImage)
{
  STATUS << "- starting IC normalization...";
  timer.Restart();
  unsigned nPixels = splatter.GetNumPixels();
  // alloc weight image if not in the splatter
  float *weightImage = 0;
  if(!splatter.GetWeightImage()) {
    weightImage = new float[nPixels];
    assert(weightImage);
    splatter.SetWeightImage(weightImage);
  }
  float *irradImage = 0;
  if(!splatter.GetIrradImage()) {
    irradImage = new float[nPixels*3];
    assert(irradImage);
    splatter.SetIrradImage(irradImage);
  }
  splatter.Clear(CICSplatter::EE_Irradiance|CICSplatter::EE_Weight);
  // splat records
  if(_jitterWeightIC)
    splatter.Splat<true>(_icAlgo->GetICache(0)->GetRecordList());
  else
    splatter.Splat<false>(_icAlgo->GetICache(0)->GetRecordList());
  splatter.NormalizeBuffer();
  // visualize record positions
  // if(_icAlgo->_visualizeNewRecPos)
    // splatter.SplatRecordPos(_icAlgo->GetICache(0)->GetRecordList());

  // multiply by the kdImage
  const CHitInfoLite *hitPoints = splatter.GetHitInfoImage();
  unsigned numHitPoints = splatter.GetNumHitPoints();
  for(unsigned idx=0; idx<numHitPoints; idx++) {
    const CHitInfoLite &hit = hitPoints[idx];
    if(hitPoints[idx].IsSet(CHitInfoLite::EE_IrrCaching)) {
      float *res = resultImage + 3*idx;
      const float *irrad = splatter.GetIrradImage() + 3*hit.dataIdx;
      CTriCol c = hit.kd;
      c *= irrad;
      c.CopyData(res);
    }
  }

  // delete allocated weight image
  if(weightImage) {
    delete [] weightImage;
    splatter.SetWeightImage(0);
  }
  if(irradImage) {
    delete [] irradImage;
    splatter.SetIrradImage(0);
  }

  timer.Stop();
  STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
}

// specialization for radiance caching
template<>
void
CAdaptiveCaching::_finalNormalization<CRadianceRecord>(
  CTimer &timer,
  CPixRCContribList *contribImage,
  CICSplatterIrregular &splatter,
  float *resultImage)
{
  STATUS << "- starting RC normalization...";
  timer.Restart();

  // reevaluate weights if smooth weight should be used for final
  // normalization
  if(CRadianceRecord::smoothWardWeight) {
    if(_jitterWeightRC)
      _reevaluateWeights<true>(contribImage,splatter);
    else
      _reevaluateWeights<false>(contribImage,splatter);
  }

  if(_jitterWeightRC)
    _addJitter(contribImage,splatter);

  _renderFromContribs(resultImage,contribImage,splatter.GetHitInfoImage(),
    splatter.GetNumHitPoints(),false);

  timer.Stop();
  STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
}

// --------------------------------------------------------------------------
//   CAdaptiveCaching::_splatContribIrreg()
// --------------------------------------------------------------------------
template<class TRecord>
int
CAdaptiveCaching::_splatContribIrreg(const TRecord *rec,
                                     CPixContribList<TRecord> *contribImage,
                                     CICSplatterIrregular &splatter,
                                     CLinearPixelList &pixelList)
{
  const CHitInfoLite *hitPoints = splatter.GetHitPoints();

  int numPixels=0;

  // let the splatter locate points within the bounding sphere
  const CICSplatterIrregular::CNearestPointArray &nearestPoints
    = splatter.FindInSphere(rec->GetP(),rec->GetValidDomain());
  CICSplatterIrregular::CNearestPointArray::const_iterator it = 
    nearestPoints.begin();
  for( ; it != nearestPoints.end(); ++it) {
    float weight;
    const CHitInfoLite &hit = *(it->pt);
    if(rec->template ComputeWeightTemplateDistPassed<false>(hit.P,hit.N,it->d2,weight)) {
      unsigned idx = hit.dataIdx;
      CPixContribList<TRecord> &pixContribs = contribImage[idx];
      // add item to the pixel contribution list
      _addContrib(pixContribs,CRecordContrib<TRecord>(rec,weight),hit);
      pixContribs.UnsetProcessPixelDone();
      // schedule pixel for processing
      if(pixContribs.size()>1)
        pixelList.push(&hit);
      // increase counter
      numPixels++;
    }
  }
  return numPixels;
}

template int 
CAdaptiveCaching::_splatContribIrreg<CIrradianceRecord>(
const CIrradianceRecord *rec,
 CPixContribList<CIrradianceRecord> *contribImage,
 CICSplatterIrregular &splatter,
 CLinearPixelList &pixelList);

template int
CAdaptiveCaching::_splatContribIrreg<CRadianceRecord>(
const CRadianceRecord *rec,
 CPixContribList<CRadianceRecord> *contribImage,
 CICSplatterIrregular &splatter,
 CLinearPixelList &pixelList);

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_scheduleChangedRecordPointsVerbose()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_scheduleChangedRecordPointsVerbose(
  CTimer &timer,
  CRecordList<TRecord> &recList,
  CICSplatterIrregular &splatter,
  CLinearPixelList &pointList)
{
  STATUS << "   - scheduling from bspheres...";
  timer.Restart();
  _scheduleChangedRecordPoints(recList,splatter,pointList);
  timer.Stop();
  STATUS 
    << "done in " << timer.UserTime() << " s / " << pointList.size()
    << " pts to process." << std::endl;
}

template
void
CAdaptiveCaching::_scheduleChangedRecordPointsVerbose<CIrradianceRecord>(
  CTimer &timer,
  CRecordList<CIrradianceRecord> &recList,
  CICSplatterIrregular &splatter,
  CLinearPixelList &pointList);

template
void
CAdaptiveCaching::_scheduleChangedRecordPointsVerbose<CRadianceRecord>(
  CTimer &timer,
  CRecordList<CRadianceRecord> &recList,
  CICSplatterIrregular &splatter,
  CLinearPixelList &pointList);

// --------------------------------------------------------------------------
//  CAdaptiveCaching::_scheduleChangedRecordPoints()
// --------------------------------------------------------------------------
template<class TRecord>
void
CAdaptiveCaching::_scheduleChangedRecordPoints(CRecordList<TRecord> &recList,
                                               CICSplatterIrregular &splatter,
                                               CLinearPixelList &pixelList)
{
  const CHitInfoLite *hitPoints = splatter.GetHitPoints();

  typename CRecordList<TRecord>::iterator rec_it=recList.begin();
  for( ; rec_it!=recList.end(); rec_it++) {
    TRecord *rec = *rec_it;
    // if record hasn't changed since the last pass, not worth worrying about it
    if(!rec->Changed()) continue;
    rec->UnsetChanged();
    // cast the record data
    float* oldValidDomain = (float*)rec->GetData(); 
    if(!oldValidDomain) {
      FATAL << "Record with no oldValidDomain found!" << std::endl;
      FATAL_EXIT;
    }
    // let the splatter locate points within the bounding sphere
    const CICSplatterIrregular::CNearestPointArray &nearestPoints
      = splatter.FindInSphere(rec->GetP(),(*oldValidDomain));
    // schedule all pixels within the old bounding sphere for processing
    CICSplatterIrregular::CNearestPointArray::const_iterator it = 
      nearestPoints.begin();
    for( ; it != nearestPoints.end(); ++it)
      pixelList.push(it->pt);
    // compute the new bounding sphere
    *oldValidDomain = rec->GetValidDomain();
  }
}

template
void
CAdaptiveCaching::_scheduleChangedRecordPoints<CIrradianceRecord>(
  CRecordList<CIrradianceRecord> &recList,
  CICSplatterIrregular &splatter,
  CLinearPixelList &pointList);

template
void
CAdaptiveCaching::_scheduleChangedRecordPoints<CRadianceRecord>(
  CRecordList<CRadianceRecord> &recList,
  CICSplatterIrregular &splatter,
  CLinearPixelList &pointList);


// --------------------------------------------------------------------------
//  CAdaptiveCaching::_setupIndices()
// --------------------------------------------------------------------------
inline unsigned
CAdaptiveCaching::_setupIndices(const CHitInfoLite *hitPoints,
                                const unsigned numHitPoints,
                                const unsigned handleFlag)
{
  // count the number of valid pixels
  unsigned numValidPoints = 0;
  for(unsigned idx=0; idx<numHitPoints; idx++) {
    /*
    if(hitPoints[idx].IsSet(GetHandleFlag<CRadianceRecord>())) {
      PRINTVAR(hitPoints[idx].GetP());
    }
    */

    if(hitPoints[idx].IsSet(handleFlag)) {
      hitPoints[idx].dataIdx = numValidPoints;
      numValidPoints++;
    }
  }
  return numValidPoints;
}

// --------------------------------------------------------------------------
//  CAdaptiveCaching::IndirectIlluminationOverlapConvergeBBoxIrreg()
// --------------------------------------------------------------------------
/**
  - resultImage corresponds one-to-one to hitPoints.
  - hitPoints are arbitrarily located in space, some might
    have the handleFlag set, some might not.
  - only the pixels of 'resultImage' whose corresponding hitInfo
    has the handleFlag set, can be modified. All other pixel mush be left
    intact.
  - the convergence algorithm should allocate the internal buffer only
    for the pixel having the handleFlag set.
  - irradiance and radiance caching are EXCLUSIVE for each shading point
  - it is the responsability of the caller to designate some shading points
    for irradiance caching, others for radiance caching.
*/
template<class TRecord>
void
CAdaptiveCaching::IndirectIlluminationOverlapConvergeBBoxIrreg(
  float              *resultImage,
  const CHitInfoLite *hitPoints,
  unsigned            numHitPoints,
  CHitPointInfo      &outHitInfo)
{
  if(numHitPoints<1) return;
  // --------------------------
  //  SETUP
  // --------------------------
  // setup timers
  CTimer timer, timer1, timerTotal, addRecTimer;
  float addRecTotalTime=0;
  timerTotal.Restart();
  STATUS << "- adaptive caching, convergence algo: bbox irregular" << std::endl;
  STATUS << "- initializing...";
  timer.Restart();
  // count the number of valid pixels
  unsigned handleFlag = GetHandleFlag<TRecord>();
  unsigned numValidPoints = _setupIndices(hitPoints,numHitPoints,handleFlag);
  if(numValidPoints<1) return;  
  PRINTVAR(numValidPoints);
  // init splatter
  CICSplatterIrregular splatter(hitPoints,numHitPoints,numValidPoints,handleFlag);
  // pixel list for image traversal
  CLinearPixelList pixelList(numHitPoints,hitPoints);
  splatter.FillPointList(pixelList);
  // alloc canvas for contributions
  CPixContribList<TRecord> *contribImage = 
    new CPixContribList<TRecord>[numValidPoints];
  // turn off smooth weight for the convergence
  bool smoothWardWeight = TRecord::smoothWardWeight;
  TRecord::smoothWardWeight = false;
  unsigned numRecordsBeforePass,passPixels,nContribs=0;
  unsigned numRecordsBefore = GetNumRecords<TRecord>();
  // debug image saving
  _initSaveDebugImages<TRecord>();
  timer.Stop();
  STATUS << "done in " << timer.UserTime() << "s" << std::endl;

  // --------------------------------------
  //    SPLAT PREVIOUS CACHE CONTENTS
  // --------------------------------------
 if(GetNumRecords<TRecord>()>0) {
    STATUS 
      << "- splatting existing cache contents ("
      << GetNumRecords<TRecord>() << " recs)...";
    timer.Restart();
    CRecordList<TRecord> &recList = GetRecordList<TRecord>();
    typename CRecordList<TRecord>::iterator it=recList.begin();
    for( ; it!=recList.end(); it++) {
      TRecord *rec= *it;
      // create the bounding box for this records
      float *recdata = new float;
      *recdata = rec->GetValidDomain();
      rec->SetData(recdata);
      // splat new record
      unsigned thisContribs = _splatContribIrreg(rec,contribImage,splatter,pixelList);
      // save debug images
      // _doSaveDebugImages(timerTotal,splatter,kdImage);
      nContribs += thisContribs;
      // update average irradiance
      // avgE += (GetRecE(rec)*thisContribs);
    }
    timer.Stop();
    STATUS << "done in " << timer.UserTime() << "sec." << std::endl;
 }

  // _doSaveDebugImages<TRecord>(timerTotal,splatter,contribImage,false);

  // --------------------------
  //  PIXEL TRAVERSAL
  // --------------------------
  
  // set records to unchanged state
  _unsetChanged(GetRecordList<TRecord>());

  for(unsigned passCount=0; !pixelList.empty(); passCount++) {

    addRecTimer.Reset();
    STATUS << "- point traversal pass " << passCount  << "..." << std::endl;
    timer.Restart();
    numRecordsBeforePass = GetNumRecords<TRecord>();
    nContribs = 0;
    passPixels = pixelList.size();

    while( !pixelList.empty() ) {
      _printPercentage(pixelList.size(),passPixels);
      unsigned idx = pixelList.front();
      pixelList.pop();
      // current hit point info
      const CHitInfoLite &hit = hitPoints[idx];
      // debugging check
      // _checkHitInfo(hit);
      // list of contributions for this pixel
      CPixContribList<TRecord> &pixContribs = contribImage[hit.dataIdx];
  
      // remove contributions that aren't valid anymore
      unsigned numValid = _updatePixelContribList(pixContribs,hit);

      switch(numValid) {
      case 0:
        if(!pixContribs.AddRecFailed()) {
          // add new record
          addRecTimer.Start();
          TRecord *rec = _addRec<TRecord>(hit,outHitInfo);
          addRecTimer.Stop();
          if(rec) {
            // create the bounding box for this records
            float *recdata = new float;
            *recdata = rec->GetValidDomain();
            rec->SetData(recdata);
            // splat new record
            nContribs += _splatContribIrreg(rec,contribImage,splatter,pixelList);
            // save debug images
            // _doSaveDebugImages(timerTotal,splatter,kdImage);
          }
          else {
            pixContribs.SetAddRecFailed();
          }
        }
        break;
      case 1:
        break;
      default: // 2 and more
        if(_adaptive && !pixContribs.ProcessPixelDone())
        {
          _extrapolateRadianceValid(pixContribs,hit);
          _processPixelOverlap(hit,hit.directIllum.data(), 
            1/*thrElevImage[idx]*/,pixContribs,GetOctree<TRecord>());
          pixContribs.SetProcessPixelDone();
        }
        break;
      }
    }
    // pixel traversal finished for this pass

    // schedule pixels in changed record's bboxes for processing
    _scheduleChangedRecordPointsVerbose(timer1,GetRecordList<TRecord>(),
      splatter,pixelList);

    timer.Stop();
    STATUS
      << "- pass finished in " << timer.UserTime() << " s / added " 
      << GetNumRecords<TRecord>()-numRecordsBeforePass << " recs in " 
      << addRecTimer.UserTime() << "s" << std::endl;
    addRecTotalTime += addRecTimer.UserTime();
  }

  // do the final normalization
  TRecord::smoothWardWeight = smoothWardWeight;

  // For irradiance caching, the contribution list can be deleted before
  // the final normalization.
  if(GetHandleFlag<TRecord>()==CHitInfoLite::EE_IrrCaching) {
    SAFE_DELETE_ARRAY(contribImage);
    //_icAlgo->GetICache(0)->RandomizeIrradiances();
  }

  PRINTVAR(splatter.treeTimer.UserTime());
  _finalNormalization(timer,contribImage,splatter,resultImage);
  PRINTVAR(splatter.treeTimer.UserTime());

  // print final stats
  timerTotal.Stop();
  _printFinalStats(timerTotal.UserTime(),addRecTotalTime,
    GetNumRecords<TRecord>()-numRecordsBefore);
  // Delete the record data.
  _destroyRecordData<TRecord,float>( GetRecordList<TRecord>() );

  // For radiance caching, the contribution list has to be deleted after
  // the final normalization.
  if(GetHandleFlag<TRecord>()==CHitInfoLite::EE_RadCaching) {
    SAFE_DELETE_ARRAY(contribImage);
  }
}

template
void
CAdaptiveCaching::IndirectIlluminationOverlapConvergeBBoxIrreg<CIrradianceRecord>(
  float              *resultImage,
  const CHitInfoLite *hitPoints,
  unsigned            numHitPoints,
  CHitPointInfo      &outHitInfo);

template
void
CAdaptiveCaching::IndirectIlluminationOverlapConvergeBBoxIrreg<CRadianceRecord>(
  float              *resultImage,
  const CHitInfoLite *hitPoints,
  unsigned            numHitPoints,
  CHitPointInfo      &outHitInfo);

// ==========================================================================
// ==========================================================================
// ==========================================================================
// ==========================================================================
// ==========================================================================
// ==========================================================================
// ==========================================================================
// ==========================================================================
// ==========================================================================
// ==========================================================================
// ==========================================================================
/*
class CPixAccumContribs {
public:
  /// accumulated incident radiance
  CTriCol   Li;
  /// accumulated weight
  float     w;
  /// number of contributions
  unsigned  n;

  Add(const float* aLi, const float aw)
  { Li += aLi; w+=aw; n++; }
  setz() { Li.setz(); w=0; n=0;}
};

// --------------------------------------------------------------------------
//   CAdaptiveCaching::_splatIncidentRayIrreg()
// --------------------------------------------------------------------------
unsigned 
CAdaptiveCaching::_splatIncidentRayIrreg(const float *Li,
                                         float hitDist,
                                         CPixAccumContribs *contribImage,
                                         CICSplatterIrregular &splatter,
                                         CLinearPixelList &pixelList)
{
  const CHitInfoLite *hitPoints = splatter.GetHitPoints();

  int numPixels=0;

  // let the splatter locate points within the bounding sphere
  const CICSplatterIrregular::CNearestPointArray &nearestPoints
    = splatter.FindInSphere(rec->GetP(),rec->GetValidDomain());
  CICSplatterIrregular::CNearestPointArray::const_iterator it = 
    nearestPoints.begin();
  for( ; it != nearestPoints.end(); ++it) {
    float weight;
    const CHitInfoLite &hit = *(it->pt);
    if(rec->template ComputeWeightTemplateDistPassed<false>(hit.P,hit.N,it->d2,weight)) {
      unsigned idx = hit.dataIdx;
      CPixContribList<TRecord> &pixContribs = contribImage[idx];
      // add item to the pixel contribution list
      _addContrib(pixContribs,CRecordContrib<TRecord>(rec,weight),hit);
      pixContribs.UnsetProcessPixelDone();
      // schedule pixel for processing
      if(pixContribs.size()>1)
        pixelList.push(&hit);
      // increase counter
      numPixels++;
    }
  }
  return numPixels;
}





}

// --------------------------------------------------------------------------
//   CAdaptiveCaching::IndirectIlluminationIncidentRadianceSplatting()
// --------------------------------------------------------------------------
void
CAdaptiveCaching::IndirectIlluminationIncidentRadianceSplatting(
  float              *resultImage,
  const CHitInfoLite *hitPoints,
  unsigned            numHitPoints,
  CHitPointInfo      &outHitInfo)
{
  if(numHitPoints<1) return;
  // --------------------------
  //  SETUP
  // --------------------------
  // setup timers
  CTimer timer, timer1, timerTotal, addRecTimer;
  float addRecTotalTime=0;
  timerTotal.Restart();
  STATUS << "- incident radiance splatting" << std::endl;
  STATUS << "- initializing...";
  timer.Restart();
  // count the number of valid pixels
  unsigned handleFlag = CHitInfoLite::EE_IrrCaching;
  unsigned numValidPoints = _setupIndices(hitPoints,numHitPoints,handleFlag);
  if(numValidPoints<1) return;  
  PRINTVAR(numValidPoints);
  // init splatter
  CICSplatterIrregular splatter(hitPoints,numHitPoints,numValidPoints,handleFlag);
  // pixel list for image traversal
  CLinearPixelList pixelList(numHitPoints,hitPoints);
  splatter.FillPointList(pixelList);
  // alloc canvas for contributions
  CPixAccumContribs *contribImage = 
    new CPixAccumContribs[numValidPoints];
  // turn off smooth weight for the convergence
  // bool smoothWardWeight = TRecord::smoothWardWeight;
  // TRecord::smoothWardWeight = false;
  unsigned numRecordsBeforePass,passPixels,nContribs=0;
  // debug image saving
  // _initSaveDebugImages<TRecord>();
  timer.Stop();
  STATUS << "done in " << timer.UserTime() << "s" << std::endl;

  unsigned maxPixelContribs = 300;

  // --------------------------
  //  PIXEL TRAVERSAL
  // --------------------------
  
  while( !pixelList.empty() ) {
    _printPercentage(pixelList.size(),passPixels);
    unsigned idx = pixelList.front();
    pixelList.pop();
    // current hit point info
    const CHitInfoLite &hit = hitPoints[idx];
    // list of contributions for this pixel
    CPixAccumContribs &pixContribs = contribImage[hit.dataIdx];

    if(pixContribs.n < maxPixelContribs)  {
      // ----------
      // send a new ray and splat its contributions
      // ----------
      // generate a random direction
      contrib.prob =  hemisph_cos(raydir_canonical);
      // compute the direction to be traced
      CVector3D raydir = ToGlobalFrame(raydir_canonical, hit.U, hit.V, hit.N);   
      // initialize the ray
      contrib.ray.Init(hit.P,raydir,REFLECTED_RAY,hit.termination,hit.object,
        0, true);
      contrib.invalid = false;
      contrib.diffuseness = 1;

      // cast the ray an compute higher order light contribution
      float hitDist;
      if(_incRad->IncidentRadiance(outHitInfo,&contrib,0,hit.footprintArea))
        hitDist = outHitInfo.GetT();
      else
        hitDist = CLimits::Infinity;

      _splatIncidentRayIrreg(contrib.color.data(),hitDist,contribImage,
        splatter,pixelList);
      if(pixContribs.n < maxPixelContribs)
        pixelList.push(hit);
    }
  }

  // normalize & multiply by the kdImage
  for(unsigned idx=0; idx<numHitPoints; idx++) {
    const CHitInfoLite &hit = hitPoints[idx];
    if(hit.IsSet(CHitInfoLite::EE_IrrCaching)) {
      float *res = resultImage + 3*idx;
      CPixAccumContribs &pixContribs = contribImage[hit.dataIdx];
      const float *irrad = splatter.GetIrradImage() + 3*hit.dataIdx;
      CTriCol c = (hit.kd * pixContribs.Li) / w;
      c.CopyData(res);
    }
  }

  delete [] contribImage;
}
*/

__END_GOLEM_SOURCE
