// ===================================================================
// $Id: icalgo.cpp,v 1.4 2005/11/29 16:04:05 xkrivanj Exp $
//
// icalgo.cpp
//     Implementation file for the CICAlgo class.
//     A high level interface for irradiance caching.
//
// Initial coding by Jaroslav Krivanek, Oct 2003.

// GOLEM headers
#include "config.h"
#include "basmacr.h"
#include "icalgo.h"
#include "icache.h"
#include "rcache.h"
#include "diffsampl.h"
#include "surface.h"
#include "environ.h"

// standard headers
#include <iomanip>

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  CCachingAlgoBase::CCachingAlgoBase
// --------------------------------------------------------------------------
template<class TCache>
CCachingAlgoBase<TCache>::CCachingAlgoBase(const string &envPrefix,
                                           const CEnvironment& env,
                                           unsigned XXXdepthLimit, 
                                           CIncidentRadiance* incRad):
_depthLimit(XXXdepthLimit),
  _visualRecColor(CStandardValues::GetDefaultColorType()),
  _visualNewRecColor(CStandardValues::GetDefaultColorType()),
  _recVisTmp(CStandardValues::GetDefaultColorType()),
  _zeroColor(CStandardValues::GetDefaultColorType())
{
  _zeroColor.setz();

  _incRad     = incRad;

  // no record position visualization by default
  _visualizeRecPos    = false;
  _visualizeNewRecPos = false;
  _visualizeRecPosCurrent = false;
  _visualizeNewRecPosCurrent = false;
  _visualRecColor.Set(1);
  _visualNewRecColor.Set(1);
  _visualRecSize = 1;

  _approxErrorRender    = 0;
  _approxErrorPresample = 0;
  _approxErrorCurrent   = 0;
  _numRays              = 0;

  if(_depthLimit <1)
    return;

  env.GetBool(envPrefix+".use", _useCaching);
  if(!_useCaching)
    return;
  
  // --------------------------------------------------------------------
  //  NUMBER OF RAYS
  // --------------------------------------------------------------------
  string str;
  float numRaysMultiplier;
  _numRays = new int[_depthLimit];
  env.GetString(envPrefix+".numRays", str);
  ParseIntegerRow(str, _numRays, _depthLimit);
  
  env.GetFloat(envPrefix+".numRaysMultiplier",numRaysMultiplier);
  
  // apply the multiplicator
  unsigned i;
  for(i = 0; i < _depthLimit; i++)  {
    int n = 
(int)
(_numRays[i] * 
numRaysMultiplier + 
0.5);
    if(n == 0 && _numRays[i]>0)  // do not cut away levels of recursion
      n = 1;
    _numRays[i] = n; 
  }

  // ------------------------------
  //  RECORD VISUALIZATION PARAMS
  // ------------------------------
  float  rCol[3];
  string  visualRecColorStr;
  env.GetBool(envPrefix+".visualizeRecPos",_visualizeRecPos);
  env.GetFloat(envPrefix+".visualRecSize",_visualRecSize);
  env.GetString(envPrefix+".visualRecColor",visualRecColorStr);
  ParseFloatRow(visualRecColorStr, rCol, 3);
  _visualRecColor.SetData(rCol);
  env.GetBool(envPrefix+".visualizeNewRecPos",_visualizeNewRecPos);
  env.GetString(envPrefix+".visualNewRecColor",visualRecColorStr);
  ParseFloatRow(visualRecColorStr, rCol, 3);
  _visualNewRecColor.SetData(rCol);

  // --------------------------------------------------------------------
  //  APPROXIMATION ERROR
  // --------------------------------------------------------------------
  float approxErrorMultiplier;
  _approxErrorRender = new float[_depthLimit];
  assert(_approxErrorRender);
  _approxErrorPresample = new float[_depthLimit];
  assert(_approxErrorPresample);
  _approxErrorCurrent = _approxErrorRender;
  env.GetFloat(envPrefix+".approxErrorMultiplier",approxErrorMultiplier);
  env.GetString(envPrefix+".approxError",str);
  ParseFloatRow(str, _approxErrorRender, _depthLimit);
  env.GetFloat(envPrefix+".presampleMultiplier",_presampleMultiplier);

  if(_presampleMultiplier < CLimits::Small) {
    WARNING 
      << envPrefix+".presampleMultiplyer too small. Setting to 1."
      << std::endl;
    _presampleMultiplier = 1;
  }

  if (approxErrorMultiplier < CLimits::Small) {
    WARNING 
      << envPrefix+".approxErrorMultiplier is too small, settitng to 1."
      << std::endl;
    approxErrorMultiplier = 1.0;
  }

  for(i = 0; i < _depthLimit; i++) {
    if (_approxErrorRender[i] < CLimits::Small) {
      WARNING 
        << envPrefix+".approxError["<<i<<"] is too small, setting to 0.1" 
        <<  std::endl;
      _approxErrorRender[i] = 0.1;
    }
    _approxErrorRender[i] *= approxErrorMultiplier;
    
    if(_approxErrorRender[i]>1.0f) {
      WARNING 
        << "Clamping IrradianceCache.approxError[" << i << "] to 1." 
        << std::endl;
      _approxErrorRender[i]=1.0f;
    }

    _approxErrorPresample[i] = _approxErrorRender[i];
  }

  if(_depthLimit>0)
    _approxErrorPresample[0] *= _presampleMultiplier;

  // --------------------------------------------------------------------
  //  (IR)RADIANCE CACHE
  // --------------------------------------------------------------------
  env.GetFloat(envPrefix+".octreeScale",_octreeScale);
  env.GetFloat(envPrefix+".minDist",_minDist);
  env.GetFloat(envPrefix+".maxDist",_maxDist);
  env.GetBool(envPrefix+".useGradients", _useGradients);
  env.GetBool(envPrefix+".cacheLastResult",_cacheLastResult);
  env.GetFloat(envPrefix+".rotWeightMultiplier",_rotWeightMultiplier);
  env.GetFloat(envPrefix+".nearestRecordAMul",_nearestRecordAMul);
  env.GetFloat(envPrefix+".centerDisplaceDist",_centerDisplaceDist);
  env.GetBool(envPrefix+".clampInScreenSpace",_clampInScreenSpace);
  env.GetBool(envPrefix+".useMinHitDist",_useMinHitDist);
  env.GetBool(envPrefix+".useGeometricNormal",_useGeometricNormal);
  env.GetBool(envPrefix+".clampRadiusByGradients",_clampRadiusByGradients);
  env.GetBool(envPrefix+".clampByNeighbors",_clampByNeighbors);
  env.GetBool(envPrefix+".clampNeighbors",_clampNeighbors);
  env.GetBool(envPrefix+".jitterWeight",_jitterWeight);
  env.GetFloat(envPrefix+".minAllowedHitDist",_minAllowedHitDist);
  env.GetBool(envPrefix+".smoothWardWeight",
    TCache::record_type::smoothWardWeight);
 
  if(_octreeScale < CLimits::Small) {
    FATAL 
      << envPrefix+".octreeScale = " << _octreeScale << " is too small." 
      << std::endl;
    FATAL_EXIT;
  }

  if(_maxDist<=_minDist) {
    FATAL 
      << envPrefix+".maxDist must be greater than 'IrradianceCache.minDist'.\n"
      << std::endl;
    FATAL_EXIT;
  }

  if(_rotWeightMultiplier<0) {
    FATAL 
      << envPrefix+".rotWeightMultiplier = " << _rotWeightMultiplier 
      << " should be greater or equal to zero." << std::endl;
    FATAL_EXIT;
  }

  // make minDist/maxDist independent of the value of 'a'
  _minDist /= 2*_approxErrorRender[0];
  _maxDist /= 2*_approxErrorRender[0];

  // --------------------------------------------------------------------
  //  CACHING MODE
  // --------------------------------------------------------------------
  if(CCachingMode::ReadCachingMode(env,envPrefix,
    _presampleModePrim, _presampleModeHi, _renderModePrim, _renderModeHi)) 
    FATAL_EXIT; // error
}

// --------------------------------------------------------------------------
//  CCachingAlgoBase::CCachingAlgoBase
// --------------------------------------------------------------------------
template<class TCache>
CCachingAlgoBase<TCache>::~CCachingAlgoBase()
{
  delete [] _numRays;
  SAFE_DELETE_ARRAY(_approxErrorRender);
  SAFE_DELETE_ARRAY(_approxErrorPresample);
  _approxErrorCurrent = 0;
}

// --------------------------------------------------------------------------
//  CCachingAlgoBase::SetMode()
// --------------------------------------------------------------------------
template<class TCache>
void 
CCachingAlgoBase<TCache>::SetMode(CCachingMode::EMode m, int level)
{
  if(level<0)  {
    for( unsigned i=0; i < _depthLimit; i++) {
      SAFE_CALL(_cache[i], SetCachingMode(m));
    }
  }
  else {
    if( (unsigned)level < _depthLimit ) {
      SAFE_CALL(_cache[level], SetCachingMode(m));
    }
  }
}

// --------------------------------------------------------------------------
//  CCachingAlgoBase::SetPresampleMode()
// --------------------------------------------------------------------------
template<class TCache>
void 
CCachingAlgoBase<TCache>::SetPresampleMode()
{
  _approxErrorCurrent = _approxErrorPresample;
  SAFE_CALL(_cache[0], SetCachingMode(_presampleModePrim) );
  for(unsigned i=1; i<_depthLimit; i++) {
    SAFE_CALL(_cache[i], SetCachingMode(_presampleModeHi) );
  }

  SAFE_CALL(_cache[0], SetJitterWeight(false));
  // Record visualization is not done in ic interface but directly
  // by ic algo.
  // SAFE_CALL(_cache[0], SetVisualizeRecPos(false));
  // SAFE_CALL(_cache[0], SetVisualizeNewRecPos(false));
  _visualizeRecPosCurrent    = false;
  _visualizeNewRecPosCurrent = false;
}

// --------------------------------------------------------------------------
//  CCachingAlgoBase::SetRenderMode()
// --------------------------------------------------------------------------
template<class TCache>
void 
CCachingAlgoBase<TCache>::SetRenderMode()
{
  _approxErrorCurrent = _approxErrorRender;
  SAFE_CALL(_cache[0], SetCachingMode(_renderModePrim) );
  for(unsigned i=1; i<_depthLimit; i++) {
    SAFE_CALL(_cache[i], SetCachingMode(_renderModeHi) );
  }

  SAFE_CALL(_cache[0], SetJitterWeight(_jitterWeight));
  // Record visualization is not done in ic interface but directly
  // by ic algo.
  // SAFE_CALL(_cache[0], SetVisualizeRecPos(_visualizeRecPos));
  // SAFE_CALL(_cache[0], SetVisualizeNewRecPos(_visualizeNewRecPos));
  _visualizeRecPosCurrent    = _visualizeRecPos;
  _visualizeNewRecPosCurrent = _visualizeNewRecPos;
}



// --------------------------------------------------------------------------
//  CCachingAlgoBase::VisualizeRecords()
// --------------------------------------------------------------------------
template<class TCache>
void
CCachingAlgoBase<TCache>::VisualizeNearestRecords(
  const CVector3D& P, 
  const CVector3D& N,
  const unsigned depth, 
  const float size, 
  CValues &result) const
{
  // visualize record positions
  float w = _cache[depth]->VisualizeNearestRecords(P,N,size);
  if(w>0) {
    // maximum blending weight is one
    if(w>1) w = 1;
    // blend old color and record visualization color
    _recVisTmp = _visualRecColor * w;
    result *= (1-w);
    result += _recVisTmp;
    result += _recVisTmp;
  }
}

// --------------------------------------------------------------------------
//  CRCAlgo::GetNumRecords()
// --------------------------------------------------------------------------
template<class TCache>
unsigned long 
CCachingAlgoBase<TCache>::GetNumRecords(int level) const
{
  if(level>=0) {
    if( _cache[level] )
      return _cache[level]->GetNumRecords();
    else
      return 0;
  }

  unsigned long sum = 0;
  for(unsigned i=0; i<_depthLimit; i++) {
    if( _cache[i] )
      sum += _cache[i]->GetNumRecords();
  }

  return sum;
}

// --------------------------------------------------------------------------
//  CICAlgo::CICAlgo()
// --------------------------------------------------------------------------
CICAlgo::CICAlgo(const CEnvironment& env,
                 unsigned depthLimit, 
                 CIncidentRadiance* incRad,
                 const CAABox& bbox) :
CCachingAlgoBase<CICache>("IrradianceCache",env,depthLimit,incRad),
_albedoTmp(CStandardValues::GetDefaultWeightType())
{
  float infRayLength = Magnitude(bbox.Diagonal());
  _diffSampler          = 0;
  _cache                = 0;

  // don't init anything for depthLimit = 0, since we're not going to compute 
  // anything anyway
  if(_depthLimit <1)
    return;
  

  string str;
  CDiffuseSampler::EDiffuseSamplerType samplerType;
  
  // --------------------------------------------------------------------
  //  USE IRRADIANCE CACHE?
  // -------------------------------------------------------------------- 
  _diffSampler = new CDiffuseSampler* [_depthLimit];
  _cache    = new CICache* [_depthLimit];
  if(!_useCaching) {
    for(unsigned i=0; i < _depthLimit; i++) {
      _diffSampler [i] = 0;  
      _cache[i] = 0;
    }
    return;
  }

  // --------------------------------------------------------------------
  //  TYPE OF THE DIFFUSE SAMPLER
  // --------------------------------------------------------------------
  string samplerString;
  env.GetString("IrradianceCache.samplerType", samplerString);
  if(samplerString == "ward" || samplerString == "cos_ward")
    samplerType = CDiffuseSampler::EE_DiffuseSamplerCosWard;
  else if(samplerString == "uni_ward")
    samplerType = CDiffuseSampler::EE_DiffuseSamplerUniformWard;
  else if(samplerString == "uni_displace")
    samplerType = CDiffuseSampler::EE_DiffuseSamplerUniformDisplace;
  else if(samplerString == "uni_deriv")
    samplerType = CDiffuseSampler::EE_DiffuseSamplerUniformDeriv;
  else if(samplerString == "cos_displace")
    samplerType = CDiffuseSampler::EE_DiffuseSamplerCosDisplace;
  else if(samplerString == "cos_deriv")
    samplerType = CDiffuseSampler::EE_DiffuseSamplerCosDeriv;
  else if(samplerString == "plane_uniform")
    samplerType = CDiffuseSampler::EE_DiffuseSamplePlaneUniform;
  else {
    FATAL << "Unknown value for 'IrradianceCache.samplerType': '" 
      <<  samplerString << "'.\n";
    FATAL << "Possible values are:\n"
      << "\t'ward'\n"
      << "\t'cos_ward' (equivalent to 'ward')\n"
      << "\t'uni_displace'\n"
      << "\t'uni_deriv'\n"
      << "\t'cos_displace'\n"
      << "\t'cos_deriv'\n"
      << "\t'plane_uniform'."
      << std::endl;
    FATAL_EXIT;
  }
  
  // Create diffuse samplers & reflection irradiance caches
  
  for( unsigned i=0; i < _depthLimit; i++) {
    if(_numRays[i]>0) {
      _diffSampler [i] = 
        CreateDiffuseSampler(samplerType, i, _numRays[i], 
        infRayLength, _minAllowedHitDist, _incRad);
      assert(_diffSampler [i]);
      
      if( _useGradients && !_diffSampler[i]->CanComputeGradients() ) {
        FATAL 
          << "Diffuse Sampler '" << samplerString << "' cannot compute "
          << "gradients.\n";
        FATAL << "Either set 'IrradianceCache.useGradients' to OFF \n"; 
        FATAL 
          << "or use a different diffuse sampler "
          << "(IrradianceCache.samplerType)."  << std::endl;
        FATAL_EXIT;
      }
      
      _cache[i] = 
        new CICache(
        bbox,
        _octreeScale,
        _minDist,
        _maxDist, 
        _centerDisplaceDist,
        _rotWeightMultiplier,
        _nearestRecordAMul,
        _useGradients, 
        _cacheLastResult,
        _clampInScreenSpace,
        _useMinHitDist, 
        _useGeometricNormal,
        _clampRadiusByGradients, 
        _clampByNeighbors,
        _clampNeighbors,
        i,
        _diffSampler[i]);

      assert(_cache[i]);      

      _cache[i]->SetRecPosVisualParams(_visualRecSize,_visualRecColor);
      _cache[i]->SetNewRecPosVisualParams(_visualNewRecColor);
    }
    else {
      _diffSampler [i] = 0;
      _cache[i] = 0;
    }
  }
 
  // rendering mode is the default
  SetRenderMode();
}

// --------------------------------------------------------------------------
//  CICAlgo::~CICAlgo()
// --------------------------------------------------------------------------
CICAlgo::~CICAlgo()
{
  for( unsigned i=0; i < _depthLimit; i++) {
    SAFE_DELETE(_cache[i]);
    SAFE_DELETE(_diffSampler[i]);
  }
  SAFE_DELETE_ARRAY(_diffSampler);
  SAFE_DELETE_ARRAY(_cache);
}

// --------------------------------------------------------------------------
//  CICAlgo::DoReport()
// --------------------------------------------------------------------------
bool
CICAlgo::DoReport(ostream &app) 
{
  
  unsigned long numSamples       = 0;
  unsigned long numQueriesSuc    = 0;
  unsigned long numQueriesFail   = 0;
  unsigned long numWeightZero    = 0;
  unsigned long numWeightNonzero = 0;
  unsigned      numCaches        = 0;
  unsigned i;
  
  assert(_cache);
  
  for(i=0; i<_depthLimit; i++) {
    if(_cache[i]) {
      numCaches ++;
      
      numSamples       += _cache[i]->GetNumRecords();
      numQueriesSuc    += _cache[i]->GetNumQueriesSuc();
      numQueriesFail   += _cache[i]->GetNumQueriesFail();
      numWeightZero    += _cache[i]->GetNumWeightZero();
      numWeightNonzero += _cache[i]->GetNumWeightNonzero();
    }
  }
  
  app
    << "\n"
    << "+--------------------------------------------------------------------------+\n"
    << "|        Irradiance caching statistics                                     |\n"
    << "+--------------------------------------------------------------------------+\n"
    << "| # | Records |  Queries  | Q-succ | Weight Evals | W-succ | Dec A | MovRec|\n";
  
  for(i=0; i<_depthLimit; i++) {
    if(_cache[i]) {
      app << "|" << setw(2) << i;
      app << " |" << setw(8) << _cache[i]->GetNumRecords();
      app << " |" << setw(10) << _cache[i]->GetNumQueriesSuc()+
        _cache[i]->GetNumQueriesFail();
      app << " |" << setw(6) << setprecision(4); 
      if( _cache[i]->GetNumQueriesSuc()+
        _cache[i]->GetNumQueriesFail() > 0 )
        app << (100.0f*(float)_cache[i]->GetNumQueriesSuc() / 
        (float)(_cache[i]->GetNumQueriesSuc()+
        _cache[i]->GetNumQueriesFail())) << "%";
      else
        app << "    -  ";
      app << " |" << setw(13) << _cache[i]->GetNumWeightZero() +
        _cache[i]->GetNumWeightNonzero();
      app << " |" << setw(6) << setprecision(4);
      if( (_cache[i]->GetNumWeightZero() + 
        _cache[i]->GetNumWeightNonzero()) > 0)
        app 
        << 100.0f*(float)_cache[i]->GetNumWeightNonzero() / 
        (float)(_cache[i]->GetNumWeightZero() +
        _cache[i]->GetNumWeightNonzero())
        << "%";
      else
        app << "    -  ";
      app << " |" << setw(6); 
      app << _cache[i]->GetNumDecreaseA();
      app << " |" << setw(6); 
      app << _cache[i]->GetNumMoveInOctree();
      app << " |\n";

#ifdef LAST_SAMPLE_CACHING
      app 
        << "_last.nCachedSamples= " 
        << _cache[i]->GetLast().nCachedSamples << std::endl;
      app 
        << "_last.nFailed= " 
        << _cache[i]->GetLast().nFailed << std::endl;
      app 
        << "_last.nSucceed= " 
        << _cache[i]->GetLast().nSucceed << std::endl;
#endif

    }
    else {
      app << "| - |         |           |        |              |        |       |       |\n";
    }
  }
  app 
    << "+--------------------------------------------------------------------------+\n";
  app << "|TOT|" << setw(8) << numSamples;
  app << " |" << setw(10) << numQueriesSuc+numQueriesFail;
  app << " |" << setw(6) << setprecision(4);
  if(numQueriesSuc+numQueriesFail > 0)
    app 
    << 100.0f*(float)numQueriesSuc / (float)(numQueriesSuc+numQueriesFail)
    << "%";
  else
    app << "    -  ";
  app << " |" << setw(13) << numWeightZero + numWeightNonzero;
  app << " |" << setw(6) << setprecision(4);
  if ( numWeightZero + numWeightNonzero > 0)
    app 
    <<  100.0f*(float)numWeightNonzero / 
    (float)(numWeightZero + numWeightNonzero)
    << "%";
  else
    app << "    -  ";
  app << " |\n";
  app  
    << "+--------------------------------------------------------------------------+\n";
 
	 return false; // ok
}


// --------------------------------------------------------------------------
//  CICAlgo::IndirectDiffuseLighting()
// --------------------------------------------------------------------------
bool
CICAlgo::
IndirectDiffuseLighting(const CHitPointInfo    &inHitInfo,
                        const float           &footprintArea,
                        const CSurface*         surface,
                        const unsigned          depth,
                        const CValues          &directIllum,
                        CHitPointInfo          &outHitInfo,                        
                        CValues                &E)
{
  assert( depth < _depthLimit);
  assert( surface->GetSupportedTypes() & EE_REFLECTED_DIFFUSE );
 
  E.setz();
  if(! (( _diffSampler[depth] ) && ( inHitInfo.GetUDotProd() < 0 )) )
  { E.setz(); return false; }

  CICache::EQueryResult queryResult = 
    _cache[depth]->GetIrradiance(inHitInfo,footprintArea,
    outHitInfo,_approxErrorCurrent[depth],directIllum,E);

  if(queryResult == CICache::EE_Failed)
  { E.setz(); return false; }

  if(surface->GetAlbedo(_albedoTmp,inHitInfo,EE_REFLECTED_DIFFUSE)) {
    _albedoTmp *= M_1_PI;
    E *= _albedoTmp;
  }
  else
    E.setz();

  // record visualization
  if(depth==0) {
    // visualize records
    if(_visualizeRecPosCurrent) {
      VisualizeNearestRecords(inHitInfo.GetPoint(),inHitInfo.GetNormal(),
      depth, sqrt(footprintArea)*_visualRecSize, E);
    }
    // visualize newly added record
    if(_visualizeNewRecPosCurrent&&(queryResult==CICache::EE_NewRecord)) {
      E = _visualNewRecColor;
    }
  }

  // indirect diffuse lighting computed and set in result
  return true;
}

// --------------------------------------------------------------------------
//  CICAlgo::Irradiance()
// --------------------------------------------------------------------------
bool
CICAlgo::Irradiance(const CHitPointInfo    &inHitInfo,
                    const float            &footprintArea,
                    const unsigned          depth,
                    const CValues          &directIllum,
                    CHitPointInfo          &outHitInfo,
                    CValues                &E)
{
  assert( depth < _depthLimit);
  assert(_approxErrorCurrent);
  assert(_approxErrorCurrent==_approxErrorRender||_approxErrorCurrent==_approxErrorPresample);
  
  if(! ( (_diffSampler[depth]) && (inHitInfo.GetUDotProd()<0) ) )
  { E.setz(); return false; }

  CICache::EQueryResult queryResult = 
    _cache[depth]->GetIrradiance(inHitInfo,footprintArea,
    outHitInfo,_approxErrorCurrent[depth],directIllum,E);

  if(queryResult == CICache::EE_Failed)
  { E.setz(); return false; }

  // record visualization
  if(depth==0) {
    // visualize records
    if(_visualizeRecPosCurrent)
      VisualizeNearestRecords(inHitInfo.GetPoint(),inHitInfo.GetNormal(),
      depth, sqrt(footprintArea)*_visualRecSize, E);
    // visualize newly added record
    if(_visualizeNewRecPosCurrent&&queryResult==CICache::EE_NewRecord)
      E = _visualNewRecColor;
  }

  // indirect irradiance computed and set in E
  return true;
}

template class CCachingAlgoBase<CICache>;
template class CCachingAlgoBase<CRCache>;

__END_GOLEM_SOURCE
