// ===================================================================
// $Id: icalgo.h,v 1.4 2005/11/02 12:58:44 xkrivanj Exp $
//
// icalgo.h
//     Header file for the CICAlgo class.
//     A high level interface for irradiance caching.
//
// Class: CICAlgo
//
// Initial coding by Jaroslav Krivanek, Oct 2003.

#ifndef __ICALGO_H__
#define __ICALGO_H__

// GOLEM headers
#include "configh.h"
#include "cachmode.h"
#include "color.h"

__BEGIN_GOLEM_HEADER

class CDiffuseSampler;
class CICache;
class CIncidentRadiance;
class CAABox;
class CSurface;
class CContributionArray;
class CHitPointInfo;
class CVector3D;

template<class TCache>
class CCachingAlgoBase {

public:
  CCachingAlgoBase(
    const string &envPrefix,
    const CEnvironment& env,
    unsigned depthLimit, 
    CIncidentRadiance* incRad);

  ~CCachingAlgoBase();

  /// Sets caching mode of irrradiance caches.
  /**
   @param m      (in) Caching mode to set. 
                      See CICache::SetMode().
   @param level  (in) Recursion level (=depth) of the cache to set. If this 
                      is a negative number, all caches are set.
  */
  void SetMode(CCachingMode::EMode m, int level=-1);
  
  /// Sets caching mode to 'presampling mode' (according to environment)
  void SetPresampleMode();
  
  /// Sets caching mode to 'rendering mode' (according to environment)
  /** 
    This method is called at the end of the constructor, as the
    '_renderMode' is the default caching mode.
   */
  void SetRenderMode();

  /// Visualize nearest records.
  void VisualizeNearestRecords(
    const CVector3D& P, 
    const CVector3D& N,
    const unsigned depth, 
    const float size, 
    CValues &result) const;

  /// Returns the number of radiance cache records.
  /**
    @param level (in) 
      Recursion level (=depth) of the cache whose number of records will be returned.     
      If this is a negative number, sum over all depths is returned.
  */
  unsigned long  GetNumRecords(int level=-1) const;

  /// return the depth limit with which the caching was initialized
  unsigned GetDepthLimit() const { return _depthLimit; }

  float GetJitterWeight() const { return _jitterWeight; }

  float GetApproxErrorCurrent(int i) const { return _approxErrorCurrent[i];}

  float GetPresampleMultiplier() { return _presampleMultiplier; }

protected:
  /// Array of irradiance caches for reflected rays.
  TCache **_cache;
  /// number of rays per recursion level
  int *_numRays;
  /// pointer an interface implementing IncidentRadiance() method
  CIncidentRadiance *_incRad;
  /// Depth limit for the recursion = length of the allocated arrays.
  const unsigned _depthLimit;
  /// use caching?
  bool _useCaching;
  /// Visualize record positions?
  bool  _visualizeRecPos;
  /// Visualize newly added record positions?
  bool  _visualizeNewRecPos;
  /// Visualize record positions?
  /// false in presample mode. Equal to _visualizeRecPos in render mode.
  bool  _visualizeRecPosCurrent;
  /// Visualize newly added record positions? 
  /// false in presample mode. Equal to _visualizeRecPos in render mode.
  bool  _visualizeNewRecPosCurrent;
  /// use weight jittering
  bool _jitterWeight;
  /// minimum allowed hist distance in the diffuse sampler
  float _minAllowedHitDist;

  float _octreeScale;
  float _minDist;
  float _maxDist;
  float _centerDisplaceDist;
  float _rotWeightMultiplier;
  float _nearestRecordAMul;
  float _presampleMultiplier;
  /// use rotational and translational gradients
  bool  _useGradients;
  bool  _cacheLastResult;
  bool  _clampInScreenSpace;
  bool  _useMinHitDist;
  bool  _useGeometricNormal;
  bool  _clampRadiusByGradients;
  bool  _clampByNeighbors;
  bool  _clampNeighbors;

  /// size of the visualized record position in pixels
  float  _visualRecSize;
  /// color of the visualized record position
  CValues _visualRecColor;
  /// color of the visualized newly added record position
  CValues _visualNewRecColor;

  /// Approximation error for the rendering pass
  float *_approxErrorRender;
  /// Approximation error for the presampling pass
  float *_approxErrorPresample;
  /// Poiner to the current approximation error.
  /** 
   Points either to _approxErrorRender or _approxErrorPresample depending
   on whether we are in the rendering or presampling mode.
  */
  float *_approxErrorCurrent;
  /// black color
  CValues _zeroColor;

private:
  /// presample caching mode for primary rays (level = 0)
  CCachingMode::EMode _presampleModePrim;
  /// presample caching mode for higher order rays (level >= 1)
  CCachingMode::EMode _presampleModeHi;
  /// render caching mode for primary rays (level = 0)
  CCachingMode::EMode _renderModePrim;
  /// render caching mode for higher order rays (level >= 1)
  CCachingMode::EMode _renderModeHi;
  /// to avoid automatic CValues
  mutable CValues _recVisTmp;
};


/// Encapsulation of all that is related to irradiance caching.
/**
 This class should be used by every app that wants to use irradiance 
 caching. The constructor creates everything that's needed according to 
 the environment. Once an instance is created, it suffices to call 
 IndirectDiffuseLighting() to get the diffuse lighting at a given point.
*/
class CICAlgo : public CCachingAlgoBase<CICache> {

public:  // methods
  
  /// Create all that is needed for irradiance caching according to the 
  /// environment.
  /**
   @param depthLimit (in) maximal depth of recursion
   @param incRad     (in) implementation of the CIncidentRadiance interface
   @param bbox       (in) bounding box of the whole scene
  */
  CICAlgo(
    const CEnvironment& env,
    unsigned depthLimit, 
    CIncidentRadiance* incRad,
    const CAABox& bbox);
  
  /// Destructor.
  ~CICAlgo();
  
  /// Compute indirect diffuse lighting at a given point.
  /**
   This function calls the irradiance cache at a given recursion depth to see 
   if a cached sample can be used. If yes, use it. If not, sample the 
   hemisphere, store the sample into the cache and return it.
  
   @param inHitInfo  (in) 
     Query point. 
   @param surface    (in) 
     Surface of the object hit by 'inHitInfo'. Used for albedo evaluation.
   @param depth      (in) 
     Current recursion depth.
   @param outHitInfo (in,out) 
     This hitinfo is passed to the diffuse sampler in the case that the 
     hemisphere sampling is used.
   @param result     (out) 
     Outgoing radiance due to indirect diffuse lighting. This is equal
     to indirect irradiance times diffuse albedo divided by PI.

   @return true if result is nonzero; false is result is zero.
  */
  bool IndirectDiffuseLighting(
    const CHitPointInfo   & inHitInfo,
    const float          & footprintArea,
    const CSurface        * surface,
    const unsigned          depth,
    const CValues         & directIllum,
    CHitPointInfo         & outHitInfo,
    CValues               & E);

  bool IndirectDiffuseLighting(
    const CHitPointInfo   & inHitInfo,
    const float          & footprintArea,
    const CSurface        * surface,
    const unsigned          depth,
    CHitPointInfo         & outHitInfo,
    CValues               & E)
  { 
    return IndirectDiffuseLighting(inHitInfo,footprintArea,surface,depth,
      _zeroColor,outHitInfo,E);
  }

  /// Compute incident irradiance.
  /**
    The same as 'IndirectDiffuseLighting' but no multiplication
    by the surface BRDF is done (the returning value is 
    incident irradiance in [W/m^2].
   */
  bool Irradiance(
    const CHitPointInfo   & inHitInfo,
    const float          & footprintArea,
    const unsigned          depth,
    const CValues         & directIllum,
    CHitPointInfo         & outHitInfo,
    CValues               & result);
  
  /// Print a report.
  bool DoReport(ostream &app);

  /// Return pointer to irradiance cache interface
  CICache *GetICache(unsigned i) 
  {assert(i<_depthLimit); assert(_cache[i]); return _cache[i]; }

protected: // data 
  /// Array of diffuse samplers to be used for hemisphere sampling.
  CDiffuseSampler             **_diffSampler;
private:
  /// to avoid automatic CValues
  mutable CValues _albedoTmp;

  friend class CAdaptiveCaching;
};

__END_GOLEM_HEADER

#endif // __ICALGO_H__
