// ===================================================================
// $Id: icontrib.h,v 1.2 2005/11/29 16:04:04 xkrivanj Exp $
//
// icontrib.h
//   /// Array of irradiance contributions to a point.
//
// Initial coding by Jaroslav Krivanek (March 2005).

#ifndef __ICONTRIB_H__
#define __ICONTRIB_H__

// GOLEM headers
#include "configh.h"
#include "icrecord.h"
#include "rcrecord.h"
#include "growvec.h"
#include "tricol.h"

__BEGIN_GOLEM_HEADER

/// list of pixels to process
typedef unsigned TPixelIndex;

/// Contribution of an irradiance record to a point.
template<class TRecord>
class CRecordContrib 
{
  typedef unsigned short flags_type;
  typedef unsigned short clock_type;

public: // data (ordered according to the access frequency)
  /// Record clock at the time the contribution was created
  clock_type changeClock;
protected:
  flags_type _flags;
public:
  /// Contributing record
  TRecord  *rec;
  /// Contribution weight
  float  w;

protected:
  /// Pixel identification
  // TPixelIndex _pixelIndex;

  /// Irradiance, before multiplication by the weight (with gradients applied)
  CTriCol _E;

  enum {
    /// The 'invalid' bit is the highest bit in the pixel index
    EE_InvalidMask = 1 << (sizeof(flags_type)*8-1),
    /// The 'irradiance is valid' flag is the second highest bit
    EE_IrradValidMask = 1 << (sizeof(flags_type)*8-2),
    /// _processPixelsOveral alread performed for this pixel
    EE_ProcessPixelDone = 1 << (sizeof(flags_type)*8-3),
    /// _addRec has failed for this pixel
    EE_AddRecFailed = 1 << (sizeof(flags_type)*8-4),
  };

public: // methods

  /// void constructor
  CRecordContrib() {}

  /// copy constructor for CIrradianceRecordID
  CRecordContrib(const TRecord *r,float weight,unsigned pi)
    : rec((TRecord*)r),
    changeClock(r->GetChangeClock())
  {
    w = weight;
    SetPixelIdx(pi);
    ResetFlags();
    // SetValid();
    // SetEInvalid();
  }

  CRecordContrib(const TRecord *r,float weight)
    : rec((TRecord*)r),
    changeClock(r->GetChangeClock())
  {
    w = weight;
    ResetFlags();
    // SetValid();
    // SetEInvalid();
  }

  void ResetFlags() { _flags=0; }

public:

  /// Return the irradiance contribution
  const CTriCol& GetE() const {return _E; }
  template<class T> void SetE(const T* x) { _E.Set(x); SetEValid(); }
  void SetE(const CTriCol& x) { _E = x; SetEValid(); }

  /// Return the pixel index 
  //@{
  TPixelIndex GetPixelIdx() const {
    FATAL << "Pixel index no more the part of CRecordContrib." << std::endl;
    FATAL_EXIT;
    return 0;
    // return _pixelIndex & ~EE_FlagsMask;
  }
  friend TPixelIndex GetPixelIdx(const CRecordContrib &c)
  {return c.GetPixelIdx();}
  friend TPixelIndex GetPixelIdx(const CRecordContrib *c)
  {return c->GetPixelIdx();}
  //@}

  /// Set the pixel index
  //@{
  void SetPixelIdx(TPixelIndex i) { 
    FATAL << "Pixel index no more the part of CRecordContrib." << std::endl;
    FATAL_EXIT;
    // assert( (i&~EE_FlagsMask)==i );
    // _pixelIndex = (i & ~EE_FlagsMask) | (_pixelIndex & EE_FlagsMask);
  }
  void SetPixelIdx(CRecordContrib &c,TPixelIndex i)
  { c.SetPixelIdx(i); }
  void SetPixelIdx(CRecordContrib *c,TPixelIndex i)
  { c->SetPixelIdx(i); }
  //@}

  /// Is the contribution invalid?
  //@{
  bool IsInvalid() const { return _flags & EE_InvalidMask; }
  friend bool IsInvalid(const CRecordContrib &c) { return c.IsInvalid(); }
  friend bool IsInvalid(const CRecordContrib *c) { return c->IsInvalid(); }
  //@}
  /// Set the invalid flag to true
  //@{
  void SetInvalid() { _flags |= EE_InvalidMask; }
  friend void SetInvalid(CRecordContrib &c) {c.SetInvalid();}
  friend void SetInvalid(CRecordContrib *c) {c->SetInvalid();}
  //@}
  /// Set the invalid flag to false
  //@{
  void SetValid()   { _flags &= ~EE_InvalidMask; }
  friend void SetValid(CRecordContrib &c) {c.SetValid();}
  friend void SetValid(CRecordContrib *c) {c->SetValid();}
  //@}

  unsigned ProcessPixelDone() const { return _flags & EE_ProcessPixelDone; }
  unsigned AddRecFailed() const { return _flags & EE_AddRecFailed; }
  void SetProcessPixelDone() { _flags |= EE_ProcessPixelDone; }
  void UnsetProcessPixelDone() { _flags &= ~EE_ProcessPixelDone; }
  void SetAddRecFailed() { _flags |= EE_AddRecFailed; }

  /// Is the extrapolated irradiance valid?
  //@{
  bool IsEValid() const { return _flags & EE_IrradValidMask; }
  friend bool IsEValid(const CRecordContrib &c) { return c.IsEValid(); }
  friend bool IsEValid(const CRecordContrib *c) { return c->IsEValid(); }
  //@}
  /// Set the invalid flag to true
  void SetEValid() { _flags |= EE_IrradValidMask; }
  /// Set the invalid flag to false
  void SetEInvalid()   { _flags &= ~EE_IrradValidMask; }

  /// Return the change clock
  //@{
  const clock_type& ChangeClock() const {return changeClock;}
  clock_type& ChangeClock() {return changeClock;}
  friend const clock_type& ChangeClock(const CRecordContrib &c) 
  { return c.changeClock; }
  friend const clock_type& ChangeClock(const CRecordContrib *c) 
  { return c->changeClock; }
  friend clock_type& ChangeClock(CRecordContrib &c) 
  { return c.changeClock; }
  friend clock_type& ChangeClock(CRecordContrib *c) 
  { return c->changeClock; }
  //@}

  /// Following is for uniform access through pointer or reference.
  /// It is used in some template functions which might either accept
  /// pointer or reference to CRecordContrib.
  //@{
  TRecord*& Rec() {return rec;}
  TRecord*const& Rec() const {return rec;}
  friend TRecord*& Rec(CRecordContrib &c) {return c.rec;}
  friend TRecord*const& Rec(const CRecordContrib &c) {return c.rec;}
  friend TRecord*& Rec(CRecordContrib *c) {return c->rec;}
  friend TRecord*const& Rec(const CRecordContrib *c) {return c->rec;}

  friend void SetE(CRecordContrib &c,const CTriCol &E){return c.SetE(E);}
  friend void SetE(CRecordContrib *c,const CTriCol &E){return c->SetE(E);}
  friend const CTriCol& GetE(const CRecordContrib &c) {return c.GetE();}
  friend const CTriCol& GetE(const CRecordContrib *c) {return c->GetE();}

  float& W() {return w;}
  const float& W() const {return w;}
  friend float& W(CRecordContrib &c) {return c.w;}
  friend const float& W(const CRecordContrib &c) {return c.w;}
  friend float& W(CRecordContrib *c) {return c->w;}
  friend const float& W(const CRecordContrib *c) {return c->w;}
  //@}

  // friend class CRecordContribArray<Trecord>;
};

typedef CRecordContrib<CIrradianceRecord> CICContrib;
typedef CRecordContrib<CRadianceRecord> CRCContrib;

/// Array of irradiance contributions to a point.
template<class TRecord>
class CRecordContribArray : public growvector< CRecordContrib<TRecord> > {
public:
  typedef growvector<CRecordContrib<TRecord> > GVT;
  typedef typename GVT::const_iterator const_iterator;

  /// Implicit constructor.
  CRecordContribArray() {}

  /// Sums the weight and irradiance constributions
  inline void Sum(float &sumWeight, CTriCol &sumE) const;

  /// Compute the weighted average of irradiances.
  /** 
     @return false if the array is empty. In such case 'E' is left unchanged.
   */
  inline bool WeightedAverage(float *E) const;
};

typedef CRecordContribArray<CIrradianceRecord> CICContribArray;
typedef CRecordContribArray<CRadianceRecord>   CRCContribArray;

// --------------------------------------------------------------------------
//  CRecordContribArray::Sum()
// --------------------------------------------------------------------------
template<class TRecord>
inline void 
CRecordContribArray<TRecord>::Sum(float &sumWeight, CTriCol &sumE) const
{
  // const_iterator iend = end();
  for(const_iterator i=begin(); i!=end(); i++) {
    sumWeight += i->w;
    sumE += i->w * i->GetE();
  }
}

// --------------------------------------------------------------------------
//  CRecordContribArray::WeightedAverage()
// --------------------------------------------------------------------------
template<class TRecord>
inline bool
CRecordContribArray<TRecord>::WeightedAverage(float *E) const
{
  switch(size()) {
  case 0: 
    return false;
  case 1:
    front().GetE().CopyData(E);
    return true;
  default:
    CTriCol sumE(0.0);
    float  sumW(0.0);
    Sum(sumW,sumE);
    sumE /= sumW;
    sumE.CopyData(E);
    return true;
  }
}

// --------------------------------------------------------------------------
//  SumContributions()
// --------------------------------------------------------------------------
template<class T>
inline void 
SumContributions(const T& clist, float &sumWeight, CTriCol &sumE)
{
  // const_iterator iend = end();
  for(typename T::const_iterator it=clist.begin(); it!=clist.end(); it++) {
    sumWeight += W(*it);
    sumE += W(*it) * GetE(*it);
  }
}

// --------------------------------------------------------------------------
//  SumValidContributions()
// --------------------------------------------------------------------------
template<class T>
inline void 
SumValidContributions(const T& clist, float &sumWeight, CTriCol &sumE)
{
  // const_iterator iend = end();
  for(typename T::const_iterator it=clist.begin(); it!=clist.end(); it++) {
    if(!IsInvalid(*it)) {
      sumWeight += W(*it);
      sumE += W(*it) * GetE(*it);
    }
  }
}

// --------------------------------------------------------------------------
//  ExtrapolateE()
// --------------------------------------------------------------------------
template<class ItT>
inline void 
ExtrapolateE(ItT first,ItT last,const CVector3D &P,const CVector3D &N)
{
  CTriCol E;
  for( ; first != last ; ++first) {
    if(!IsEValid(*first)) {
      Rec(*first)->ExtrapolateIrradiance(P,N,E.data());
      SetE(*first,E);
    }
  }
}

__END_GOLEM_HEADER

#endif // __ICONTRIB_H__
