// ===================================================================
// $Id: icrecord.h,v 1.1 2005/11/02 12:58:48 xkrivanj Exp $
//
// icrecord.h
//   Irradiance cache record definition
//   
// REPLACEMENT_STRING
//
// Initial coding by Jaroslav Krivanek (March, 2005).

#ifndef __ICRECORD_H__
#define __ICRECORD_H__

// GOLEM headers
#include "configh.h"
#include "color.h"
#include "vector3.h"

#include "rnd.h"
#include "basmacr.h"

// standard headers
#include <list>

__BEGIN_GOLEM_HEADER

// forward declarations
class CIrradianceRecord;
template<class T> class CRCOctreeNode;

template<class T>
class CRecordList : public std::list<T*> 
{
};

/// Typedef for list of irradiance records.
typedef CRecordList<CIrradianceRecord> CIrradianceList;

/// use Greg Ward's behind test from Radiance
#define WARD_RADIANCE_BEHIND_TEST
/// use my behind test based on surface curvature
// #define JK_CURVATURE_BEHIND_TEST

/// The entry of the irradiance cache (irradiance record).
template<class T>
class CICRCRecord
{
protected:

   /// Constructor of the cache entry without gradients.
   CICRCRecord(
     const CVector3D &P,
     const CVector3D &N,
     float a,
     float R,
     float unclampedR,
     float Rminus,
     float Rplus,
     float rotMult,
     float stdDev,
     void   *data):
     _P(P),
     _N(N),
     _a(a),
     _R(R),
     _unclampedR(unclampedR),
     _Rminus(Rminus),
     _Rplus(Rplus),
     _sqrInvR(1.0f/(R*R)),
     _sqrRotMult(rotMult*rotMult),
     _stdDev(stdDev),
     _changeClock(0),
     _node(0),
     _data(data),
     _flags(0)
   { 
   }
   
   CICRCRecord():
   _changeClock(0),_data(0), _node(0), _flags(0)
   {}
   
   ~CICRCRecord() {}

 public:

   enum EFlags {
     /// has 'R' or 'a' changed?
     EE_Changed = 1
   };

   /// mamimum allowed dot product between the normal at a query point and
   /// the accepted record's normal
   static float maxAllowedDot;

   const float&    GetStdDev()           const { return _stdDev; }
   const CVector3D& GetPoint()            const { return _P; }
   const CVector3D& GetP()                const { return _P; }
   const CVector3D& GetNormal()           const { return _N; }
   const CVector3D& GetN()                const { return _N; }
   float GetSqrInvR()              const { return _sqrInvR; }
   float GetR()                    const { return _R; }
   float GetUnclampedR()           const { return _unclampedR; }
   float GetRMinus()               const { return _Rminus; }
   float GetRPlus()                const { return _Rplus; }
   float GetSqrRotMult()           const { return _sqrRotMult; }
   float GetValidDomain(const float &a)  const { return _R*a; }
   float GetValidDomain()                 const { return _R*_a; }
   float GetA()                           const { return _a; } 
   unsigned GetChangeClock()              const { return _changeClock; }
   void* GetData() const { return _data; }
   void  SetData(void *data) const { _data=data; }
   void  SetStdDev(float stdDev) { _stdDev=stdDev; }
   const CRCOctreeNode<T>* GetTreeNode() const { return _node; }
   CRCOctreeNode<T>* GetTreeNode() { return _node; }
   typename CRecordList<T>::iterator GetTreeIterator() {return _iter; }

   /// Flag setting and querying
   //@{
   bool Changed() const { return _flags & EE_Changed; }
   void UnsetChanged() { _flags &= ~EE_Changed; }
   void SetChanged() { _flags |= EE_Changed; }
   //@}

   void SetUnclampedR(const float uR) { _unclampedR=uR;}
   void SetR(const float R) {_R=R; _sqrInvR=1.0f/(R*R); _incChangeClock(); }
   void SetRClamp(const float R) {SetR(std::max(std::min(R,_Rplus),_Rminus));}
   void SetA(const float a) {_a=a; _incChangeClock(); }
   void SetTreeLoc(CRCOctreeNode<T> *node, typename CRecordList<T>::iterator &iter)
   { _node = node; _iter = iter; }


   /// Compute weight of this record with respect to a given point
   /**
    The new 'behind' test looks like this:
    
     h = dot(P-Pi, Ni)
     h_max^2 = (0.5+eps) * ||P-Pi||^2 * (1-dot(N,Ni)) + eps;
     if(h^2 > h_max^2) return false;

    The rationale behind the test is the following:

    If a record ri={Pi,Ni,...} is accepted at position {P,N},
    we assume that the two point Pi and P lie on a smooth
    curved surface of a constant curvature. If they did not lie
    on a smooth surface, we want to reject the record ri - 
    and this is what the behind test does.

    The radius of the assumed curved surface is given by the formula:
              |P-Pi||/2    sqrt(2)        ||P-Pi||
        r = ------------ = ------- . ------------------,       (1)
            sin(alpha/2)      2       sqrt(1-dot(N,Ni))
    where alpha is the angle between the two normals N and Ni.

    The basic idea of the new behind test is to limit maximum distance
    of point P from the tangnet plane at {Pi,Ni} so that it still lies on
    the curved surface with the radius given in Equation (1).
    We assume that the curved surface has constant curvature in the considered
    region and thus can be replaced by a sphere of radius $r$.

    The distance of point $P$ lying on the sphere of radius $r$ from the
    tangent plane at {Pi,Ni} is given by 

       h_max = r * ( 1-cos(alpha) ) = r * (1-dot(N,Ni)),       (2)

    and this is also the maximum distance that we will allow in our 
    'behind' test.

    Combining Equations (1) and (2) we get 
      h_max^2 = 0.5 * ||P-Pi||^2 * (1-dot(N,Ni)).

    To avoid numerical instabilities, we add the additive and multiplicative
    epsilons:
     h_max^2 = (0.5+eps) * ||P-Pi||^2 * (1-dot(N,Ni)) + eps;

   The actual distance of point 'P' from the tangent plane at {Pi,Ni}
   is given by 
     h = dot(P-Pi, Ni)

   And the behind test is
     if(h^2 > h_max^2) {
       // reject record
       return false;
     }
   */
   template<bool jitter>
   inline bool ComputeWeightTemplate(
     const CVector3D  &P,
     const CVector3D  &N,
     float           a,
     float           &w) const;

   template<bool jitter>
   inline bool ComputeWeightTemplate(
     const CVector3D  &P,
     const CVector3D  &N,
     float           &w) const
   {return ComputeWeightTemplate<jitter>(P,N,_a,w);}

   /// specialized version  - the distance test is known to have passed
   template<bool jitter>
   inline bool ComputeWeightTemplateDistPassed(
   const CVector3D &P,
   const CVector3D &N,
   float           sqrDistance,
   float          &weight) const;

   inline bool ComputeWeight(
     const CVector3D  &P,
     const CVector3D  &N,
     float            a,
     float           &w) const
   { return ComputeWeightTemplate<false>(P,N,a,w); }

   inline bool ComputeWeight(const CVector3D &P,const CVector3D &N,float &w)const
   { return ComputeWeight(P,N,_a,w); }

   inline float FindJustExclusionA(
     const CVector3D       &point,
     const CVector3D       &normal) const;

   /// Is this behind r2 ?  Returns true=behind, false=not behind.
   inline bool Behind(const CICRCRecord &r2) const;

 public: //data

   /// use linear weight falloff for Ward's weight instead of 1/w
   static bool smoothWardWeight;

   /// amout of jittering
   const static float jitterAmount;

 protected: // data
   // sorted according to typical access pattern
   // 1. check clock change 2. re-evaluate weight 3. evaluate contribution

   /// clock when 'R' or 'a' was last changed for this record
   unsigned    _changeClock;
   /// position
   CVector3D   _P;
   /// approximation error
   float      _a;
   /// 1/(R^2)
   float      _sqrInvR;
   /// surface normal at P
   CVector3D   _N;
   /// squared multiplier for the rotation error in the weight evaluation
   float      _sqrRotMult;
   /// estimate of the standard deviation of E
   float      _stdDev;
   /// radius clamped betweeen _Rminus and _Rplus
   float      _R;
   /// unclamped radius
   float      _unclampedR;
   /// clamping values at the record position
   float      _Rminus,_Rplus;
   /// Octree node it resides in
   CRCOctreeNode<T>  *_node;
   /// Iterator in the node's record list
   typename CRecordList<T>::iterator _iter;
   /// Record flags
   unsigned _flags;
   /// additional data (application dependent)
   mutable void *_data;

protected: // methods
  void _incChangeClock() {_changeClock++; SetChanged(); }
};

// --------------------------------------------------------------------------
//  class CIrradianceRecord 
// --------------------------------------------------------------------------
class CIrradianceRecord : public CICRCRecord<CIrradianceRecord> {

public:
   /// Constructor of the cache entry.
   CIrradianceRecord(
     const CVector3D &P,
     const CVector3D &N,
     const CValues   &E,
     float a,
     float R,
     float unclampedR,
     float Rminus,
     float Rplus,
     float rotMult,
     float stdDev,
     CVector3D *rotGrad,
     CVector3D *transGrad,
     void      *data=0):
   CICRCRecord<CIrradianceRecord>(P,N,a,R,unclampedR,Rminus,Rplus,rotMult,stdDev,data),
     _E(E),
     _rotGrad(rotGrad), 
     _transGrad(transGrad)
   { 
   }
   
   CIrradianceRecord():
   _rotGrad(0),_transGrad(0)     
   {}

   ~CIrradianceRecord() {delete [] _rotGrad; delete [] _transGrad;}

   const CValues&   GetIrradiance()       const { return _E; }
   const CValues&   GetE()                const { return _E; }
   bool   GradientsUsed()           const { return _rotGrad!=0; }
   const CVector3D* GetRotGrad()           const { return _rotGrad;   }
   const CVector3D* GetTransGrad()         const { return _transGrad; }
   const CVector3D& GetRotGrad(int i)      const { return _rotGrad[i];   }
   const CVector3D& GetTransGrad(int i)    const { return _transGrad[i]; }

   /// Get extrapolated irradiance at a given position.
   void ExtrapolateIrradiance(
     const CVector3D& P,
     const CVector3D &N,
     float *destE) const;

 protected: // data
   /// irradiance at P(N)
   CValues     _E;
   /// translation gradient
   CVector3D   *_transGrad;
   /// rotation gradient
   CVector3D   *_rotGrad;
};

// --------------------------------------------------------------------------
//  CIrradianceRecord::ExtrapolateIrradiance()
// --------------------------------------------------------------------------
inline void 
CIrradianceRecord::ExtrapolateIrradiance(const CVector3D& P,
                                         const CVector3D &N,
                                         float *destE) const
{
  const float *srcE = GetE().data();
  const int channels = GetE().GetCount();

  // apply gradients if available
  if ( GradientsUsed() ) {
    CVector3D diff  = P - GetPoint();
    CVector3D nnorm = CrossProd(N, GetNormal());
    // accumulate with gradients
    for(int i=0; i<channels; i++) {
      destE[i] = srcE[i] +
        DotProd(diff,  GetTransGrad(i)) + 
        DotProd(nnorm, GetRotGrad(i));      
    }
  }
  else {
    // accumulate the value without gradients
    for(int i=0; i<channels; i++)
      destE[i] = srcE[i];
  }
}

// --------------------------------------------------------------------------
//  CICRCRecord::FindJustExclusionA()
// --------------------------------------------------------------------------
template<class T>
inline float
CICRCRecord<T>::FindJustExclusionA(const CVector3D       &P,
                                   const CVector3D       &N) const
{
  float sqrDistance    =  SqrMagnitude(P-GetPoint());
  float e1             =  sqrDistance * GetSqrInvR();

  return  sqrt(e1) - 1e-6;
}


// --------------------------------------------------------------------------
//  CICRCRecord::BehindTest()
// --------------------------------------------------------------------------
template<class T>
inline bool
CICRCRecord<T>::Behind(const CICRCRecord &r2) const
{
#ifdef WARD_RADIANCE_BEHIND_TEST 
  {
  CVector3D diff = GetP() - r2.GetP();
  float sqrDistance = SqrMagnitude(diff);
  if(sqrDistance > CLimits::Small) {
    float di = DotProd(diff, (_N + r2._N) );
    if (di*0.5 < -_a*_Rminus-0.001)
    {/*static int WW=0;WW++;if(WW%100==0)PRINTVAR(WW);*/ return true;} // behind
  }
  }
#endif

#ifdef JK_CURVATURE_BEHIND_TEST
  {
  // reject backfacing
  float dot = DotProd(_N, r2._N);
  if(dot<=maxAllowedDot) return true; // behind
  float eni2 = wardWeightRotationMult2 * (1.0f-dot);
  CVector3D diff = _P - r2._P;
  float sqrDistance = SqrMagnitude(diff);

  // one direction
  float h = DotProd(diff, r2._N);
  float maxh2 = 0.52*sqrDistance*eni2 + 0.02;
  if(h*h > maxh2)
  { static int MM=0;MM++;if(MM%100==0)PRINTVAR(MM); return true;} // behind
  }
#endif

  return false; // not behind
}

// --------------------------------------------------------------------------
//  CICRCRecord::ComputeWeightTemplate()
// --------------------------------------------------------------------------
template<class T>
template<bool jitter>
inline bool
CICRCRecord<T>::ComputeWeightTemplate(const CVector3D &P,
                                         const CVector3D &N,
                                         float           a,
                                         float          &weight) const
{
  if(jitter)
    a *= jitterAmount;

  CVector3D  diff = P - GetPoint();

  // -----------------------------
  //  reject based on distance ?
  // -----------------------------
  float sqrDistance  =  SqrMagnitude(diff);
  float epi2         =  sqrDistance * GetSqrInvR();
  float a2 = a*a;
  if(epi2 > a2) return false;

  // ------------------------------------------
  //  reject based on the normal divergence? 
  // ------------------------------------------
  // orientation error
  float dot = DotProd(N , GetNormal());
  // reject backfacing
  if(dot<=maxAllowedDot) return false;
  float eni2 = GetSqrRotMult() * (1.0f-dot);
  if(epi2 + eni2 > a2) return false;

  // -------------------------------------
  //  reject based on the 'behind' test?
  // -------------------------------------
#ifdef WARD_RADIANCE_BEHIND_TEST
  // ward's test (from Radiance code)
  if(sqrDistance > CLimits::Small) {
    float di = DotProd(diff, (N + GetNormal()) );
    if (di*0.5 < -a*_Rminus-0.001)
    { /*static int W=0;W++;if(W%100==0)PRINTVAR(W);*/ return false;}
  }
#endif
#ifdef JK_CURVATURE_BEHIND_TEST
  // my new test
  float h = DotProd(diff, GetNormal());
  float maxh2 = 0.52*sqrDistance*eni2 + 0.02;
  if(h*h > maxh2) 
  { /*static int M=0;M++;if(M%100==0)PRINTVAR(M);*/ return false;}
#endif

  // compute weight
  float iw = sqrtf(epi2);
  if(eni2 > 0) iw += sqrtf(eni2);

  if ( iw <= CLimits::Small ) {
    weight = 1.0f / CLimits::Small;
    return true;
  }

  // test the final weight against a 
  if(jitter) {
    //PRINTVAR(jitter);
    // with jittering
    float effectiveA = a*RND::randabf(1.0/jitterAmount,1);
    if (effectiveA > iw) {
      weight = smoothWardWeight ? (1.0f - iw/effectiveA) : (1.0f / iw);
      return true;
    }
    else
      return false;
  }
  else {
    // without jittering 
    if (a > iw) {
      weight = smoothWardWeight ? (1.0f - iw/a) : (1.0f / iw);
      return true;
    }
    else
      return false;
  }
}


// --------------------------------------------------------------------------
//  CICRCRecord::ComputeWeightTemplateDistPassed()
// --------------------------------------------------------------------------
template<class T>
template<bool jitter>
inline bool
CICRCRecord<T>::ComputeWeightTemplateDistPassed(const CVector3D &P,
                                                   const CVector3D &N,
                                                   float           sqrDistance,
                                                   float          &weight) const
{
  float a = GetA();
  if(jitter)
    a *= jitterAmount;

  CVector3D  diff = P - GetPoint();

  // -----------------------------
  //  reject based on distance ?
  // -----------------------------
  float epi2         =  sqrDistance * GetSqrInvR();
  float a2 = a*a;

  // ------------------------------------------
  //  reject based on the normal divergence? 
  // ------------------------------------------
  // orientation error
  float dot = DotProd(N , GetNormal());
  // reject backfacing
  if(dot<=maxAllowedDot) return false;
  float eni2 = GetSqrRotMult() * (1.0f-dot);
  if(epi2 + eni2 > a2) return false;

  // -------------------------------------
  //  reject based on the 'behind' test?
  // -------------------------------------
#ifdef WARD_RADIANCE_BEHIND_TEST
  // ward's test (from Radiance code)
  if(sqrDistance > CLimits::Small) {
    float di = DotProd(diff, (N + GetNormal()) );
    if (di*0.5 < -a*_Rminus-0.001)
    { /*static int W=0;W++;if(W%100==0)PRINTVAR(W);*/ return false;}
  }
#endif
#ifdef JK_CURVATURE_BEHIND_TEST
  // my new test
  float h = DotProd(diff, GetNormal());
  float maxh2 = 0.52*sqrDistance*eni2 + 0.02;
  if(h*h > maxh2) 
  { /*static int M=0;M++;if(M%100==0)PRINTVAR(M);*/ return false;}
#endif

  // compute weight
  float iw = sqrtf(epi2);
  if(eni2 > 0) iw += sqrtf(eni2);

  if ( iw <= CLimits::Small ) {
    weight = 1.0f / CLimits::Small;
    return true;
  }

  // test the final weight against a 
  if(jitter) {
    //PRINTVAR(jitter);
    // with jittering
    float effectiveA = a*RND::randabf(1.0/jitterAmount,1);
    if (effectiveA > iw) {
      weight = smoothWardWeight ? (1.0f - iw/effectiveA) : (1.0f / iw);
      return true;
    }
    else
      return false;
  }
  else {
    // without jittering 
    if (a > iw) {
      weight = smoothWardWeight ? (1.0f - iw/a) : (1.0f / iw);
      return true;
    }
    else
      return false;
  }
}


__END_GOLEM_HEADER

#endif // __ICRECORD_H__
