// ===================================================================
// $Id: icsplat.h,v 1.1 2005/11/02 12:58:48 xkrivanj Exp $
//
// icsplat.h
//   Irradiance cache splatter (predecesor for reular/irregular splatter).
//   
// REPLACEMENT_STRING
//
// Initial coding by Jaroslav Krivanek (March, 2005).
// Based on Pascal Gautron's splatting code.

#ifndef __ICSPLAT_H__
#define __ICSPLAT_H__

// GOLEM headers
#include "configh.h"
#include "fixvec.h"
#include "hinflite.h"
#include "icontrib.h"

// standard headers
#include <list>

__BEGIN_GOLEM_HEADER

// forward declarations
class CIrradianceRecord;

/// List that does not store anything.
template<class T>
class CVoidList{
public:
  void push_back(const T&) {}
};

/// List of contributions for each record.
typedef std::list<CICContrib> CRecContribList;

/// List of contributions for each pixel.
// typedef fixvector<CICContrib,6> CPixICContribList;
template<class TRecord>
class CPixContribList : public fixvector<CRecordContrib<TRecord>,6> {
public:
  // pixel flags are stored in the first contribution array
  typedef fixvector<CRecordContrib<TRecord>,6> super;
  unsigned ProcessPixelDone() const { return front().ProcessPixelDone(); }
  unsigned AddRecFailed() const { return front().AddRecFailed(); }
  void SetProcessPixelDone() { front().SetProcessPixelDone(); }
  void UnsetProcessPixelDone() { front().UnsetProcessPixelDone(); }
  void SetAddRecFailed() { front().AddRecFailed(); }
};
inline void fixvec_not_fit(CICContrib& c) {c.SetInvalid();}
inline void fixvec_not_fit(CRCContrib& c) {c.SetInvalid();}
typedef CPixContribList<CIrradianceRecord> CPixICContribList;
typedef CPixContribList<CRadianceRecord> CPixRCContribList;

/// Pointer to a contribution stored in the pixel list.
typedef CICContrib* TPixICContribPtr;

/// Pointer to a contribution stored in the record list.
typedef CICContrib* TRecICContribPtr;

/// List of contribution pointers for each record.
typedef std::list<TRecICContribPtr> CRecICContribPtrList;

/// List of contribution pointers for each pixel.
typedef fixvector<TRecICContribPtr,6> CPixICContribPtrList;
inline void fixvec_not_fit(CICContrib* c) {c->SetInvalid();}


/// number of contributions for one pixel
typedef unsigned TPixContribCount;

/// Irradiance cache splatter (predecesor for reular/irregular splatter).
class CICSplatter {

public: // methods

  /// canvas bits names
  enum ECanvasType {
    EE_Weight      = 0x00000001,
    EE_Irradiance  = 0x00000002,
    EE_ContribPtr  = 0x00000004,
    EE_Contrib     = 0x00000008,
    EE_ContribCnt  = 0x00000010
  };

  /// constructor
  CICSplatter(
    const CHitInfoLite* hitImg,
    unsigned numHitPoints,
    unsigned numPixels,
    unsigned handleFlag) :
  _recPosCol(0.0)
  {
    assert(hitImg);
    assert(numHitPoints>0);

    _hitImg = hitImg;
    _numHitPoints = numHitPoints;
    _numPixels    = numPixels;
    _handleFlag = handleFlag;

    SetWeightImage(0);
    SetIrradImage(0);
    SetContribPtrImage(0);
    SetContribCountImage(0);
  }

  /// should the given hit info be handled
  unsigned IsHandled(const CHitInfoLite* h) const {return h->flags & _handleFlag;}
  /// should the given hit info be handled
  unsigned IsHandled(const CHitInfoLite& h) const {return h.flags & _handleFlag;}

  /// Set the canvas for accumulating weights.
  void SetWeightImage(float *wi) {_weightImg = wi;}
  /// return the image for accumulating weights
  float* GetWeightImage() { return _weightImg; }
  /// return the image for accumulating weights
  const float* GetWeightImage() const { return _weightImg; }

  /// Set the canvas for accumulating irradiances.
  void SetIrradImage(float *ii)  {_irradImg  = ii;}
  /// return the image for accumulating irradiances
  float* GetIrradImage() { return _irradImg; }
  /// return the image for accumulating irradiances
  const float* GetIrradImage() const { return _irradImg; }

  /// Set the canvas for adding record contribution pointers.
  void SetContribPtrImage(CPixICContribPtrList *ci) {_contribPtrImg = ci;}
  /// Return the canvas for adding record contribution pointers.
  CPixICContribPtrList *GetContribPtrImage() { return _contribPtrImg;}
  /// Return the const canvas for adding record contributions pointers.
  const CPixICContribPtrList *GetContribPtrImage() const { return _contribPtrImg;}

  /// Set the canvas for adding record contributions.
  void SetContribImage(CPixICContribList *ci) {_contribImg = ci;}
  /// Return the canvas for adding record contributions.
  CPixICContribList *GetContribImage() { return _contribImg;}
  /// Return the const canvas for adding record contributionss
  const CPixICContribList *GetContribImage() const { return _contribImg;}

  /// Set the canvas for counting contributions.
  void SetContribCountImage(TPixContribCount *contribCountImg)
  { _contribCountImg = contribCountImg; }
  /// Return the mutable canvas for counting contributions.
  TPixContribCount *GetContribCountImage() { return _contribCountImg; }
  /// Return the const canvas for counting contributions.
  const TPixContribCount *GetContribCountImage()const{return _contribCountImg;}

  /// return the hit info map
  const CHitInfoLite* GetHitInfoImage() const { return _hitImg; }

  /// Return the number of hit points in _hitInfoImage.
  unsigned GetNumHitPoints() const { return _numHitPoints; }

  /// Return the number of pixels in splatting canvases.
  unsigned GetNumPixels() const { return _numPixels; }

  /// clear the specified canvases
  void Clear(unsigned whichCanvas);

  /// clear the irradiance canvas with the given color
  void ClearIrradiance(const float *c);

  /// Divide the accumulated irradiances by weight.
  void NormalizeBuffer();

  void SetRecPosColor(const float *c) { _recPosCol.Set(c); }

  /// return the handle flag depending on the template argument
  template<class TRecord>
    static inline unsigned GetHandleFlag();

protected: // data
  /// Only hit infos having this flag set are handled.
  unsigned _handleFlag;

  /// Number of points in hitImg
  unsigned _numHitPoints;

  /// Number of pixels in splatting canvases (can be lower if
  /// than _numHitPoint in case some hit point are invalid.
  unsigned _numPixels;

  /// image of hit points (position, normal, object, ... etc.)
  const CHitInfoLite *_hitImg;
  /// destination image for accumulating weights
  float *_weightImg;
  /// destination image for accumulating irradiances
  float *_irradImg;
  /// Destination image for adding contribution pointers.
  CPixICContribPtrList *_contribPtrImg;
  /// Destination image for adding contributions.
  CPixICContribList *_contribImg;
  /// canvas for counting the number of contributions
  TPixContribCount *_contribCountImg;
  /// color for record position
  CTriCol _recPosCol;
};

// --------------------------------------------------------------------------
//  CICSplatter::GetHandleFlag()
// --------------------------------------------------------------------------
template<class TRecord>
inline unsigned
CICSplatter::GetHandleFlag()
{ 
  FATAL << "Only specialized version of GetHandleFlag() can be called.\n";
  FATAL_EXIT;
  return 0;
}

template<>
inline unsigned
CICSplatter::GetHandleFlag<CIrradianceRecord>()
{ 
  return CHitInfoLite::EE_IrrCaching;
}

template<>
inline unsigned
CICSplatter::GetHandleFlag<CRadianceRecord>()
{ 
  return CHitInfoLite::EE_RadCaching;
}



__END_GOLEM_HEADER

#endif // __ICSPLAT_H__
