// ===================================================================
// $Id: icsplati.cpp,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// icache.cpp
//     Implementation of the irradiance cache.
//
// Class: CRCOctree
//
// REPLACEMENT_STRING
//
// Initial coding by Jaroslav Krivanek (August 2005).

// GOLEM headers
#include "config.h"
#include "icsplati.h"

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  CICSplatterIrregular::CICSplatterIrregular()
// --------------------------------------------------------------------------
CICSplatterIrregular::CICSplatterIrregular(const CHitInfoLite *hitPoints,
                                           unsigned numHitPoints,
                                           unsigned numPixels,
                                           unsigned handleFlag):
CICSplatter(hitPoints,numHitPoints,numPixels,handleFlag)
{
  assert(numHitPoints>0);
  _kdTree.Reserve(numPixels);
  for(unsigned i=0; i<numHitPoints;i++) {
    if(IsHandled(hitPoints+i))
      _kdTree.AddItem(hitPoints+i);
  }
  unsigned numTreePoints = _kdTree.GetNumPoints();
  if(numTreePoints != numPixels) {
    FATAL 
      << "CICSplatterIrregular::CICSplatterIrregular(): "
      << "Inconsistency in data!\n"
      << "The number of hit points passing IsHandled() test ("
      << numTreePoints
      << ") is different from the number of pixels passed to the "
      << "constructor (" << numPixels << ")." << std::endl;
    FATAL_EXIT;
  }

  STATUS << "building up kdtree...";
  _kdTree.BuildUp();
  STATUS << "tree built";

  treeTimer.Reset();
}

// --------------------------------------------------------------------------
//  CICSplatterIrregular::~CICSplatterIrregular()
// --------------------------------------------------------------------------
CICSplatterIrregular::~CICSplatterIrregular()
{
}

// --------------------------------------------------------------------------
//  CICSplatterIrregular::Splat()
// --------------------------------------------------------------------------
template<bool jitter>
unsigned
CICSplatterIrregular::Splat(const CIrradianceRecord *rec)
{
  unsigned numPoints = 0;
  float weight;
  const float* E = rec->GetE().data();
  // enlarge radius if jittering is used
  float r = rec->GetValidDomain();
  if(jitter)
    r *= CIrradianceRecord::jitterAmount;
  // find shading points within the record's influence radius
  const CNearestPointArray &nearestPoints = FindInSphere(rec->GetP(), r);
  // test weight for those points
  CNearestPointArray::const_iterator it = nearestPoints.begin();
  for( ; it != nearestPoints.end(); ++it ) {
    const CHitInfoLite &hit = *(it->pt);
    if(rec->template ComputeWeightTemplateDistPassed<jitter>(hit.P,hit.N,it->d2,weight)){
      // record accepted 
      numPoints++;
      unsigned idx = hit.dataIdx;
      // accumulate weight
      _weightImg[idx] += weight;
      // accumulate irradiance
      float *irrad = _irradImg+3*idx;
      if ( rec->GradientsUsed() ) {
        // accumulate with trans & rot gradients
        CVector3D diff  = hit.P - rec->GetPoint();
        CVector3D nnorm = CrossProd(hit.N, rec->GetNormal());
        for(int i=0; i<3; i++) {
          float Eprim = E[i] +
            DotProd(diff,  rec->GetTransGrad(i)) + 
            DotProd(nnorm, rec->GetRotGrad(i));
          irrad[i] += Eprim*weight;
        }
      }
      else 
      {
        // accumulate the value without gradients
        for(int i=0; i<3; i++)
          irrad[i] += E[i]*weight;
      }
    }
  }
  return numPoints;
}

template unsigned CICSplatterIrregular::Splat<true>(const CIrradianceRecord*);
template unsigned CICSplatterIrregular::Splat<false>(const CIrradianceRecord*);

// --------------------------------------------------------------------------
//   CICSplatterIrregular::Splat()
// --------------------------------------------------------------------------
template<bool jitter>
unsigned
CICSplatterIrregular::Splat(const CIrradianceList &recList)
{
  int numPixels = 0;
  CIrradianceList::const_iterator it = recList.begin();
  for( ; it!=recList.end(); it++) {
    numPixels += Splat<jitter>(*it);
  }
  return numPixels;
}
template unsigned CICSplatterIrregular::Splat<true>(const CIrradianceList&);
template unsigned CICSplatterIrregular::Splat<false>(const CIrradianceList&);

// --------------------------------------------------------------------------
//  CICSplatterIrregular::FindInSphereLinear()
// --------------------------------------------------------------------------
void
CICSplatterIrregular::
FindInSphereLinear(CICSplatterIrregular::CNearestPointStruct &nearestPoints)
{
  // linear search
  float r2 = nearestPoints.GetR2();
  const CVector3D &P = nearestPoints.GetP();
  for(unsigned i=0; i<_numHitPoints; i++) {
    if(IsHandled(GetHitPoints()[i])) {
      float d2 = SqrDistance(P,GetHitPoints()[i].P);
      if(d2<=r2)
        nearestPoints.list.push_back(CNearestHitInfoLite(&GetHitPoints()[i],d2));
    }
  }
}

__END_GOLEM_SOURCE
