// ===================================================================
// $Id: icsplatr.cpp,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// icsplat.cpp
//   Irradiance cache splatter
//   
//
// Initial coding by Jaroslav Krivanek (March, 2005).

// GOLEM headers
#include "config.h"
#include "icsplatr.h"

// SPHAR headers
#include "shvec.h"

// standard headers
#include <cassert>

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//   CICSplatterRegular::CICSplatterRegular()
// --------------------------------------------------------------------------
CICSplatterRegular::CICSplatterRegular(const CMatrix4 &projMatrix, 
                         const CMatrix4 &modelviewMatrix,
                         const int w, 
                         const int h, 
                         const CHitInfoLite *hitImg,
                         unsigned handleFlag):
CICSplatter(hitImg,w*h,w*h,handleFlag)
{
  assert(w>0);
  assert(h>0);
  assert(hitImg);
  assert(handleFlag);

  // save the matrices
  _projMatrix      = projMatrix;
  _modelviewMatrix = modelviewMatrix;

  // compute the viewer position
  CVector3D at, up;
  GluUnLookAt(modelviewMatrix, _viewPos, at, up);

  // precompute projection*viewport matrix
  CMatrix4 tr=TranslationMatrix(CVector3D(1,1,0));
  CMatrix4 sc=ScaleMatrix(0.5f*w, 0.5f*h, 1);
  CMatrix4 vportMatrix = tr*sc;
  _projViewportMatrix = projMatrix*vportMatrix;
 
  _w = w;
  _h = h;

  SetDrawCenter(false);
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::CheckValidBBox()
// --------------------------------------------------------------------------
bool
CICSplatterRegular::CheckValidBBox(const CIrradianceRecord *irr, 
                            const float a, 
                            const CICSplatterRegular::TBBox& r)
{
  bool errFound = false;

  if(r.xmin<0 || r.ymin<0 || r.xmax>=_w || r.ymax>=_h) {
    FATAL << "BBox is off-screen. " << std::endl;
    errFound = true;
  }

  // test if there is a valid contribution outside the bounding rectangle
  for(int y=0; y < _h; y++)
  {
    int index = _w * y; 

    const CHitInfoLite *hitptr = _hitImg    +   index;
    float *weightptr           = _weightImg +   index;
    float *irradptr            = _irradImg  + 3*index;

    for(int x=0; x < _w; x++, hitptr++, weightptr++, irradptr+=3)
    {
      float weight;
      if(IsHandled(hitptr) && irr->ComputeWeight(hitptr->P,hitptr->N,a,weight)) {
        // valid contribution - we have to be inside the passed bounding rect
        if(!r.Contains(x,y)) {
          FATAL << "Invalid bbox for pixel " << x << ", " << y << std::endl;
          errFound = true;
        }
      }
    }
  }
  return errFound;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::AccumulateContribPtrs()
// --------------------------------------------------------------------------
void
CICSplatterRegular::AccumulateContribPtrs(const CICSplatterRegular::TBBox& r)
{
  for(int y=r.ymin; y <= r.ymax; y++)
  {
    int idx = Idx(r.xmin,y);
    CPixICContribPtrList *contribptrptr = _contribPtrImg +   idx;
    float *weightptr                  = _weightImg     +   idx;
    float *irradptr                   = _irradImg      + 3*idx;
    for(int x=r.xmin; x <= r.xmax; x++,contribptrptr++,weightptr++,irradptr+=3)
    {
      CPixICContribPtrList::const_iterator citer = contribptrptr->begin();
      CPixICContribPtrList::const_iterator citer_end = contribptrptr->end();
      for( ; citer != citer_end; citer++) {
        TPixICContribPtr item = *citer;
        weightptr[0] += item->W();
        for(int ch=0; ch<3; ch++)
          irradptr[ch] += item->W() * item->GetE()[ch];
      }
    }
  }
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::RasterizeContrib()
// --------------------------------------------------------------------------
template<class T>
int 
CICSplatterRegular::RasterizeContrib(const CIrradianceRecord *rec, 
                              const CICSplatterRegular::TBBox& r,
                              T &addList)
{
  assert(r.xmin>=0); assert(r.ymin>=0); assert(r.xmax<_w); assert(r.ymax<_h);

  int numPixels=0;
  for(int y=r.ymin; y <= r.ymax; y++) {
    int idx = Idx(r.xmin,y);
    const CHitInfoLite *hitptr = _hitImg + idx;
    CPixICContribPtrList *contribptrptr = _contribPtrImg + idx;
    for(int x=r.xmin; x <= r.xmax; x++, hitptr++, contribptrptr++, idx++) {
      float weight;
      if(IsHandled(hitptr) && rec->ComputeWeight(hitptr->P,hitptr->N,weight)){
        // add the contribution to the additional list
        addList.push_back(CICContrib(rec,weight,idx));
        // add the contribution to the pixel list
        contribptrptr->push_back(&addList.back());
        // increase counter
        numPixels++;
      }
    }
  }
  return numPixels;
}

template int 
CICSplatterRegular::RasterizeContrib<CRecContribList>
(const CIrradianceRecord *rec, const CICSplatterRegular::TBBox& r, 
 CRecContribList &addList);

// --------------------------------------------------------------------------
//   CICSplatterRegular::SplatContrib()
// --------------------------------------------------------------------------
template<class T>
int 
CICSplatterRegular::SplatContrib(const CIrradianceRecord* rec, T &addList)
{
  if(CheckBackFacing(rec)) return 0;
  // number of affected pixels
  int numPixels = 0;
  // Compute the record position in camera coordinates
  CVector3D projPos = Modelview(rec->GetPoint());
  // compute the screen bounding box
  CICSplatterRegular::TBBox bbox;
  if( ComputeBBox(rec,rec->GetA(),projPos,bbox) ) {
    // rasterize the record
    numPixels += RasterizeContrib<T>(rec, bbox, addList);
  }
  return numPixels;
}

template int CICSplatterRegular::SplatContrib<CRecContribList>
(const CIrradianceRecord *rec, CRecContribList &addList);

// --------------------------------------------------------------------------
//   CICSplatterRegular::Splat()
// --------------------------------------------------------------------------
int
CICSplatterRegular::Splat(const CIrradianceRecord *irr, const float *aMap)
{
  if(CheckBackFacing(irr)) return 0;

  // number of affected pixels
  int numPixels = 0;

  // Compute the record position in camera coordinates
  CVector3D projPos = Modelview(irr->GetPoint());

  // Find the record center on the screen
  float ctrf[2];
  MultCol32(_projViewportMatrix, projPos, ctrf);
  int cx=(int)ctrf[0], cy=(int)ctrf[1];

  /*
  if( !IsOnScreen(cx,cy) ) {
    FATAL 
      << "Record center off screen. " 
      << "Float pos=(" << ctrf[0] << ", " << ctrf[1] << "), "
      << "Int pos=(" << cx << ", " << cy << "), " 
      << std::endl;
    // FATAL_EXIT;
  }
  */

  /// get the 'a' value from the closest pixel on the screen
  if(cx<0)   cx=0;
  if(cx>=_w) cx=_w-1;
  if(cy<0)   cx=0;
  if(cy>=_h) cx=_h-1;

  // get the weight from the record center - use it as the initial value for
  // the influence radius
  float a = aMap[Idx(cx,cy)];

  // compute the screen bounding box
  CICSplatterRegular::TBBox bbox;
  if( ComputeBBox(irr,a,projPos,bbox) ) {
    // rasterize the record
    numPixels += Rasterize(irr, aMap, bbox);
    // center as a white dot
    // PutPixel(cx,cy);
  }

  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::RasterizeA()
// --------------------------------------------------------------------------
template<class TRecord>
int 
CICSplatterRegular::RasterizeA(const TRecord *rec, 
                               const float a, 
                               const CICSplatterRegular::TBBox& r)
{
  assert(r.xmin>=0);
  assert(r.ymin>=0);
  assert(r.xmax<_w);
  assert(r.ymax<_h);

  int numPixels=0;
  float weight;

  for(int y=r.ymin; y <= r.ymax; y++)
  {
    int idx = Idx(r.xmin,y);

    const CHitInfoLite *hitptr = _hitImg    +   idx;
    float *weightptr           = _weightImg +   idx;
    float *recadptr            = _irradImg  + 3*idx;

    for(int x=r.xmin; x <= r.xmax; x++, hitptr++, weightptr++, recadptr+=3)
    {
      if(hitptr->IsSet(GetHandleFlag<TRecord>()) && 
        rec->ComputeWeight(hitptr->P,hitptr->N,a,weight) ) {
        // record accepted
        numPixels ++;
        weightptr[0] += weight;
        for(int ch=0; ch<3; ch++) recadptr[ch] += a*weight;
      }
    }
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::RasterizeA()
// --------------------------------------------------------------------------
template<class TRecord>
int 
CICSplatterRegular::RasterizeStdDev(const TRecord *rec, 
                                    const float a, 
                                    const CICSplatterRegular::TBBox& r)
{
  assert(r.xmin>=0);assert(r.ymin>=0);assert(r.xmax<_w);assert(r.ymax<_h);

  int numPixels=0;
  float weight;

  for(int y=r.ymin; y <= r.ymax; y++) {
    int idx = Idx(r.xmin,y);
    const CHitInfoLite *hitptr = _hitImg    +   idx;
    float *weightptr           = _weightImg +   idx;
    float *recadptr            = _irradImg  + 3*idx;
    for(int x=r.xmin; x <= r.xmax; x++, hitptr++, weightptr++, recadptr+=3){
      if(hitptr->IsSet(GetHandleFlag<TRecord>()) && 
        rec->ComputeWeight(hitptr->P,hitptr->N,a,weight) ) {
        numPixels ++;
        weightptr[0] += weight;
        for(unsigned ch=0; ch<3; ch++)
          recadptr[ch] += weight*rec->GetStdDev();
      }
    }
  }
  return numPixels;
}


// --------------------------------------------------------------------------
//   CICSplatterRegular::SplatRecordPos()
// --------------------------------------------------------------------------
template<class TRecord>
void
CICSplatterRegular::SplatRecordPos(const TRecord *rec)
{
  if(CheckBackFacing(rec)) return;
  // Compute the record position in camera coordinates
  CVector3D projPos = Modelview(rec->GetPoint());
  // center as a white dot
  float c[2];
  MultCol32(_projViewportMatrix, projPos, c);
  int x=(int)c[0];
  int y=(int)c[1];
  PutPixel(x,y,_recPosCol.data());
  PutPixel(x+1,y,_recPosCol.data());
  PutPixel(x+1,y+1,_recPosCol.data());
  PutPixel(x,y+1,_recPosCol.data());
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::SplatA()
// --------------------------------------------------------------------------
template<class TRecord>
int
CICSplatterRegular::SplatA(const TRecord *rec)
{
  if(CheckBackFacing(rec)) return 0;

  // number of affected pixels
  int numPixels = 0;

  // Compute the record position in camera coordinates
  CVector3D projPos = Modelview(rec->GetPoint());

  // compute the screen bounding box
  CICSplatterRegular::TBBox bbox;
  if( ComputeBBox(rec,rec->GetA(),projPos,bbox) ) {

    // DEBUG << "x=" << bbox.xmin << ", " << bbox.xmax << std::endl;
    // DEBUG << "y=" << bbox.ymin << ", " << bbox.ymax << std::endl;

    // rasterize the record
    numPixels += RasterizeA(rec, rec->GetA(), bbox);

    // center as a white dot
#if 0
     float c[2];
     MultCol32(_projViewportMatrix, projPos, c);
     PutPixel(c[0],c[1], _recPosCol.data());
     PutWeight(c[0],c[1], 1);
#endif
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::SplatStdDev()
// --------------------------------------------------------------------------
template<class TRecord>
int
CICSplatterRegular::SplatStdDev(const TRecord *rec)
{
  if(CheckBackFacing(rec)) return 0;
  int numPixels = 0;
  CVector3D projPos = Modelview(rec->GetPoint());
  CICSplatterRegular::TBBox bbox;
  if( ComputeBBox(rec,rec->GetA(),projPos,bbox) ) {
    numPixels += RasterizeStdDev(rec, rec->GetA(), bbox);
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::SplatRecordPos()
// --------------------------------------------------------------------------
template<class TRecord>
void 
CICSplatterRegular::SplatRecordPos(const CRecordList<TRecord> &recList)
{
  typename CRecordList<TRecord>::const_iterator it;
  for(it = recList.begin(); it!=recList.end(); it++)
    SplatRecordPos(*it);
}

template void CICSplatterRegular::SplatRecordPos(
  const CRecordList<CIrradianceRecord> &recList);

template void CICSplatterRegular::SplatRecordPos(
  const CRecordList<CRadianceRecord> &recList);

// --------------------------------------------------------------------------
//   CICSplatterRegular::Splat()
// --------------------------------------------------------------------------
int 
CICSplatterRegular::Splat(const CIrradianceList &recList, const float *aMap)
{
  int numPixels = 0;
  CIrradianceList::const_iterator it;
  for(it = recList.begin(); it!=recList.end(); it++) {
    numPixels += Splat(*it, aMap);
  }
  return numPixels;
}


// --------------------------------------------------------------------------
//   CICSplatterRegular::SplatA()
// --------------------------------------------------------------------------
template<class TRecord>
int
CICSplatterRegular::SplatA(const CRecordList<TRecord> &recList)
{
  int numPixels = 0;
  typename CRecordList<TRecord>::const_iterator it;
  for(it = recList.begin(); it!=recList.end(); it++) {
    numPixels += SplatA(*it);
  }
  return numPixels;
}

template int CICSplatterRegular::SplatA(const CIrradianceList &recList);
template int CICSplatterRegular::SplatA(const CRadianceList &recList);

// --------------------------------------------------------------------------
//   CICSplatterRegular::SplatStdDev()
// --------------------------------------------------------------------------
template<class TRecord>
int
CICSplatterRegular::SplatStdDev(const CRecordList<TRecord> &recList)
{
  int numPixels = 0;
  typename CRecordList<TRecord>::const_iterator it;
  for(it = recList.begin(); it!=recList.end(); it++) {
    numPixels += SplatStdDev(*it);
  }
  return numPixels;
}

template int CICSplatterRegular::SplatStdDev(const CIrradianceList &recList);
template int CICSplatterRegular::SplatStdDev(const CRadianceList &recList);

// --------------------------------------------------------------------------
//   CICSplatterRegular::NormalizeContribs()
// --------------------------------------------------------------------------
int
CICSplatterRegular::NormalizeContribs(const CPixICContribPtrList* contribPtrImg,
                               float *destImg) const
{
  assert(destImg);
  assert(_contribImg);
  float *destptr  = destImg;
  const CPixICContribPtrList *contribptrptr  = contribPtrImg;
  int numPixels = 0;

  for(int i=0; i<_w*_h; i++, destptr+=3, contribptrptr++) {
    // const CICSplatterRegular::CPixICContribPtrList &pixContribs = *contribptr;
    switch(contribptrptr->size()) {
      case 0:
        break;
      case 1:
        contribptrptr->front()->GetE().CopyData(destptr);
        numPixels ++;
        break;
      default: 
        {
          // sum & normalize
          CPixICContribPtrList::const_iterator i=contribptrptr->begin();
          CPixICContribPtrList::const_iterator iend=contribptrptr->end();
          CTriCol sumE(0.0);
          float  sumW=0;
          for(  ; i!=iend; i++) {
            sumW += (*i)->W();
            sumE += (*i)->W() * (*i)->GetE();
          }
          sumE /= sumW;
          sumE.CopyData(destptr);
        }
        numPixels ++;
        break;
    }
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::NormalizeContribs()
// --------------------------------------------------------------------------
int
CICSplatterRegular::NormalizeContribs(const CPixICContribList* contribImg,
                               float *destImg) const
{
  assert(destImg);
  assert(_contribImg);
  float *destptr  = destImg;
  const CPixICContribList *contribptr = contribImg;
  int numPixels = 0;

  for(int i=0; i<_w*_h; i++, destptr+=3, contribptr++) {
    // const CICSplatterRegular::CPixICContribPtrList &pixContribs = *contribptr;
    switch(contribptr->size()) {
      case 0:
        break;
      case 1:
        if(!contribptr->front().IsInvalid()) {
          contribptr->front().GetE().CopyData(destptr);
          numPixels ++;
        }
        break;
      default: 
        {
          // sum & normalize
          CPixICContribList::const_iterator i=contribptr->begin();
          CPixICContribList::const_iterator iend=contribptr->end();
          CTriCol sumE(0.0);
          float  sumW=0;
          for(  ; i!=iend; i++) {
            if(!(*i).IsInvalid()) {
              sumW += (*i).W();
              sumE += (*i).W() * (*i).GetE();
            }
          }
          if(sumW>0) {
            sumE /= sumW;
            sumE.CopyData(destptr);
            numPixels ++;
          }
        }
        break;
    }
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatter::Clear()
// --------------------------------------------------------------------------
void
CICSplatter::Clear(unsigned whichCanvas)
{
  unsigned nPixels = GetNumPixels();

  if( (whichCanvas&EE_Weight) && _weightImg ) {
    shv<float>::setz(nPixels,_weightImg);
  }

  if( (whichCanvas&EE_Irradiance) && _irradImg) {
    shv<float>::setz(nPixels*3,_irradImg);
  }

  if( (whichCanvas&EE_ContribPtr) && _contribPtrImg) {
    for(unsigned i=0; i<nPixels; i++)
      _contribPtrImg[i].clear();
  }

  if( (whichCanvas&EE_Contrib) && _contribImg) {
    for(unsigned i=0; i<nPixels; i++)
      _contribImg[i].clear();
  }

  if( (whichCanvas&EE_ContribCnt) && _contribCountImg) {
    shv<TPixContribCount>::setz(nPixels,_contribCountImg);
  }
}

// --------------------------------------------------------------------------
//   CICSplatter::ClearIrradiance()
// --------------------------------------------------------------------------
void
CICSplatter::ClearIrradiance(const float *c)
{
  unsigned nPixels = GetNumPixels();
  float *irr = _irradImg;
  for(unsigned i=0; i<nPixels; i++) {
    *irr++ = c[0];
    *irr++ = c[1];
    *irr++ = c[2];
  }
}

// --------------------------------------------------------------------------
//   CICSplatter::NormalizeBuffer()
// --------------------------------------------------------------------------
void
CICSplatter::NormalizeBuffer()
{
  float *irradptr   = _irradImg;
  const float *weightptr  = _weightImg;
  unsigned nPixels = GetNumPixels();

  for(unsigned i=0; i<nPixels; i++, irradptr+=3, weightptr++) {
    if(weightptr[0] > 0) {
      float iw=1.0f/weightptr[0];
      for(int ch=0; ch<3; ch++)
        irradptr[ch] *= iw;
    }
    else {
      // irradptr[1] = 100;
    }
  }
}

__END_GOLEM_SOURCE
