// ===================================================================
// $Id: icsplatr.h,v 1.2 2005/11/29 16:04:04 xkrivanj Exp $
//
// icsplatr.h
//   Irradiance cache splatter for regular images.
//   
// REPLACEMENT_STRING
//
// Initial coding by Jaroslav Krivanek (March, 2005).
// Based on Pascal Gautron's splatting code.

#ifndef __ICSPLATR_H__
#define __ICSPLATR_H__

// GOLEM headers
#include "configh.h"
#include "icsplat.h"
#include "matrix4.h"
#include "rect.h"


__BEGIN_GOLEM_HEADER

// forward declarations

class CICSplatterRegular : public CICSplatter {

public:

  typedef CRect<int> TBBox;

protected:

  /// image width
  int _w;
  /// image height
  int _h;

  /// modelview matrix stored in the OpenGL compatible format
  CMatrix4 _modelviewMatrix;
  /// projection matrix stored in the OpenGL compatible format
  CMatrix4 _projMatrix;
  /// projection * viewport matrix
  CMatrix4 _projViewportMatrix;
  /// view position
  CVector3D _viewPos;
  /// whether or not to draw splatted records's center as a white dot
  bool _drawCenter;

public:

  /// constructor with no canvases specified
  CICSplatterRegular(
    const CMatrix4 &projMatrix, 
    const CMatrix4 &modelviewMatrix,
    const int w, 
    const int h, 
    const CHitInfoLite *hitImg,
    unsigned handleFlag);

public:
  /// Whether or not to draw splatted records's center as a white dot.
  void SetDrawCenter(bool dc) { _drawCenter = dc; }

  /// Return true if a pixel is on-screen.
  inline bool IsOnScreen(int cx, int cy) const
  { return (cx>=0 && cx<_w && cy>=0 && cy<_h); }

  inline int Idx(int cx, int cy) const
  { return _w*cy+cx; }

  /// return the image width
  inline int GetWidth() const  { return _w; }
  /// return the image height
  inline int GetHeight() const { return _h; }

  /// Set the irradiance buffer at a given position to white. Return the 
  /// number of affected pixels (0 or 1).
  inline int PutPixel(int cx, int cy,const float col[]);
  inline int PutPixel(int cx, int cy) 
  {static const float col[]={1000,1000,1000}; return PutPixel(cx,cy,col);}
  inline int PutWeight(int cx, int cy,const float w);

  /// Draw a dot at a position for a point given in camera space.
  /// Used for drawing record center position.
  inline void DrawCenter(const CVector3D& projPos);

  /// Accumulate contributions from the contribution ptr buffer into the weight
  /// buffer and irradiance buffer.
  //@{
  void AccumulateContribPtrs(const TBBox& r);
  void AccumulateContribPtrs() {AccumulateContribPtrs(TBBox(0,0,_w-1,_h-1));}
  //@}

  /// Splat an irradiance record. Return number of affected pixels 
  template<bool jitter>
  int Splat(const CIrradianceRecord *irr, const float a);

  /// Splat an irradiance record using the value of 'a' stored in the record. 
  /// Return number of affected pixels 
  template<bool jitter>
  inline int Splat(const CIrradianceRecord *irr)
  { return Splat<jitter>(irr, irr->GetA()); }

  template<bool jitter>
    inline int Splat(const CRadianceRecord *rec) {
      FATAL << "RC splatting not implemented" << std::endl;
      FATAL_EXIT;
    }

  /// Splat an irradiance record. Return number of affected pixels 
  int Splat(const CIrradianceRecord *irr, const float *aMap);

  /// Splat the value of 'a' for the given record.
  template<class TRecord>
    int SplatA(const TRecord *rec);

  /// Splat the value of 'a' for the given record.
  template<class TRecord>
    int SplatStdDev(const TRecord *rec);

  /// Splat record contributions and add the contributions to 'addList'.
  template<class T>
  int SplatContrib(const CIrradianceRecord *rec, T &addList);

  /// Splat a list of irradiance records given a single value of 'a'
  template<bool jitter,class TRecord>
  int Splat(const CRecordList<TRecord> &recList, const float a);

  /// Splat a list of irradiance records given a map of 'a' for each pixel
  int Splat(const CIrradianceList &recList, const float *aMap);

  /// Splat a list of irradiance records useing the value of 'a' stored with
  // in the records
  template<bool jitter,class TRecord>
  int Splat(const CRecordList<TRecord> &recList);

  /// Splat the value of 'a' for each records in the list.
  template<class TRecord>
  int SplatA(const CRecordList<TRecord> &recList);

  /// Splat the standard deaviation for each records in the list.
  template<class TRecord>
  int SplatStdDev(const CRecordList<TRecord> &recList);

  /// Splat record position as a white dot
  //@{
  template<class TRecord>
  void SplatRecordPos(const TRecord *rec);
  template<class TRecord>
  void SplatRecordPos(const CRecordList<TRecord> &recList);
  //@}

  /// Rasterize an irradiance record given its screen bounding box.
  /**
    No checking is done if the bbox is off-screen. It is the responsability
    of the caller to assure the bounding box is all in a valid screen area.
    @return number of affectdd pixels
   */
  template<bool jitter>
  inline int Rasterize(
    const CIrradianceRecord *irr, 
    const float a, 
    const TBBox& r);

  inline int Rasterize(
    const CIrradianceRecord *irr, 
    const float *aMap, 
    const TBBox& r);

  /// Put the value of 'a' instead of irradiance
  template<class TRecord>
  int RasterizeA(
    const TRecord *rec,
    const float a, 
    const TBBox& r);

  /// Put the value of stdDev instead of irradiance
  template<class TRecord>
  int RasterizeStdDev(
    const TRecord *rec,
    const float a, 
    const TBBox& r);

  template<class T>
    int RasterizeContrib(
    const CIrradianceRecord *rec, 
    const TBBox& r,
    T &addList);

  /// Check whether there is a valid contribution outside the record's 
  /// bounding box. Return true if problem detected.
  bool CheckValidBBox(
    const CIrradianceRecord *irr, 
    const float a, 
    const TBBox& r);

  /// Return true if the record normal is back-facing for current view.
  template<class TRecord>
    inline bool CheckBackFacing(const TRecord *irr);

  /// Compute the bounding box for a record. Return false if bbox is off-screen.
  template<class TRecord>
    inline bool ComputeBBox(
    const TRecord *rec, 
    const float a,
    const CVector3D& projPos, 
    TBBox &bbox);

  template<class TRecord>
    inline bool ComputeBBox(
    const TRecord *rec, 
    const float a, 
    TBBox &bbox);

  /// Sum up contributions and write the resulting irradiance into the provided
  /// image. Return number of affected pixels.
  //@{
  int NormalizeContribs(
    const CPixICContribPtrList* contribImg,
    float *destImg) const;
  int NormalizeContribs(
    const CPixICContribList* contribImg,
    float *destImg) const;
  //@}

  /// Multiply modelview matrix by (v|1). Assumes that lower row of M is 
  /// (0,0,0,1) (no division is neded in this case).
  inline CVector3D Modelview(const float *v);
};

// --------------------------------------------------------------------------
//   Modelview()
// --------------------------------------------------------------------------
// Multiply M*(v|1). Assumes that lower row of M is (0,0,0,1) (no division is neded in this case).
inline CVector3D 
Modelview(const CMatrix4& M, const float* v)
{
  return CVector3D(
    v[0]*M.x[0][0]+v[1]*M.x[1][0]+v[2]*M.x[2][0] + M.x[3][0],
    v[0]*M.x[0][1]+v[1]*M.x[1][1]+v[2]*M.x[2][1] + M.x[3][1],
    v[0]*M.x[0][2]+v[1]*M.x[1][2]+v[2]*M.x[2][2] + M.x[3][2]);
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::Modelview()
// --------------------------------------------------------------------------
inline CVector3D
CICSplatterRegular::Modelview(const float *v)
{
  return ::Modelview(_modelviewMatrix,v);
}

// --------------------------------------------------------------------------
//   MultCol32()
// --------------------------------------------------------------------------
inline void 
MultCol32(const float m[16], const float v[3], float dest[2]) 
{
  float t0 = v[0]*m[0*4+0] + v[1]*m[1*4+0] + v[2]*m[2*4+0] + m[3*4+0];
  float t1 = v[0]*m[0*4+1] + v[1]*m[1*4+1] + v[2]*m[2*4+1] + m[3*4+1];
  float id = 1.0/(v[0]*m[0*4+3] + v[1]*m[1*4+3] + v[2]*m[2*4+3] + m[3*4+3]);
  dest[0] = t0*id;
  dest[1] = t1*id;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::ComputeBBox()
// --------------------------------------------------------------------------
template<class TRecord>
inline bool
CICSplatterRegular::ComputeBBox(const TRecord *rec, 
                                const float a, 
                                TBBox &bbox)
{
  return ComputeBBox(rec,a,Modelview(rec->GetPoint()),bbox);
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::Rasterize()
// --------------------------------------------------------------------------
inline int 
CICSplatterRegular::Rasterize(const CIrradianceRecord *irr, 
                       const float *aMap, 
                       const TBBox& r)
{
  assert(r.xmin>=0);
  assert(r.ymin>=0);
  assert(r.xmax<_w);
  assert(r.ymax<_h);

  int numPixels=0;
  float weight;
  const float *E = irr->GetIrradiance().data();

  // Within the record's domain, the value of 'a' read from 
  // aMap will not be allowed to exceed the value of 'a' stored in 
  // the record.
  float irra = irr->GetA();

  for(int y=r.ymin; y <= r.ymax; y++)
  {
    int idx = Idx(r.xmin,y);

    const CHitInfoLite *hitptr = _hitImg    +   idx;
    float *weightptr           = _weightImg +   idx;
    float *irradptr            = _irradImg  + 3*idx;
    const float *aptr          =  aMap      +   idx;

    for(int x=r.xmin; x<=r.xmax; x++,hitptr++,weightptr++,aptr++,irradptr+=3)
    {
      if(IsHandled(hitptr) && 
        irr->ComputeWeight(hitptr->P,hitptr->N,Min(*aptr,irra),weight)) 
      {
        numPixels++;
        // record accepted 
        weightptr[0] += weight;
        // apply gradients if available
        if ( irr->GradientsUsed() ) {
          CVector3D diff  = hitptr->P - irr->GetPoint();
          CVector3D nnorm = CrossProd(hitptr->N, irr->GetNormal());
          // accumulate with trans & rot gradients
          for(int i=0; i<3; i++) {
            float Eprim = E[i] +
              DotProd(diff,  irr->GetTransGrad(i)) + 
              DotProd(nnorm, irr->GetRotGrad(i));
            irradptr[i] += Eprim*weight;
          }
        }
        else 
        {
          // accumulate the value without gradients
          for(int i=0; i<3; i++)
            irradptr[i] += E[i]*weight;
        }
      }
    }
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::Rasterize()
// --------------------------------------------------------------------------
template<bool jitter>
inline int 
CICSplatterRegular::Rasterize(const CIrradianceRecord *irr, 
                       const float a, 
                       const TBBox& r)
{
  assert(r.xmin>=0);
  assert(r.ymin>=0);
  assert(r.xmax<_w);
  assert(r.ymax<_h);

  int numPixels=0;
  float weight;
  const float *E = irr->GetIrradiance().data();

  for(int y=r.ymin; y <= r.ymax; y++)
  {
    int idx = Idx(r.xmin,y);

    const CHitInfoLite *hitptr = _hitImg    +   idx;
    float *weightptr           = _weightImg +   idx;
    float *irradptr            = _irradImg  + 3*idx;

    for(int x=r.xmin; x <= r.xmax; x++, hitptr++, weightptr++, irradptr+=3)
    {
      if(IsHandled(hitptr) && 
        irr->template ComputeWeightTemplate<jitter>(hitptr->P,hitptr->N,a,weight)) {
        numPixels++;
        // record accepted 
        weightptr[0] += weight;
        // apply gradients if available
        if ( irr->GradientsUsed() ) {
          CVector3D diff  = hitptr->P - irr->GetPoint();
          CVector3D nnorm = CrossProd(hitptr->N, irr->GetNormal());
          // accumulate with trans & rot gradients
          for(int i=0; i<3; i++) {
            float Eprim = E[i] +
              DotProd(diff,  irr->GetTransGrad(i)) + 
              DotProd(nnorm, irr->GetRotGrad(i));
            irradptr[i] += Eprim*weight;
          }
        }
        else 
        {
          // accumulate the value without gradients
          for(int i=0; i<3; i++)
            irradptr[i] += E[i]*weight;
        }
      }
    }
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::CheckBackFacing()
// --------------------------------------------------------------------------
template<class TRecord>
inline bool
CICSplatterRegular::CheckBackFacing(const TRecord *rec)
{
  // If the record is on a back facing polygon, no need to render it
#if 0
  if(DotProd(_viewPos-irr->GetPoint(), rec->GetNormal()) < 0.0) { 
    DEBUG << "backfacing record in splatter" << std::endl;
    return true;
  }
#endif
  return false;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::ComputeBBox()
// --------------------------------------------------------------------------
template<class TRecord>
inline bool
CICSplatterRegular::ComputeBBox(const TRecord *rec, 
                         const float a, 
                         const CVector3D& projPos, 
                         TBBox &bbox)
{
  // Code assumes modelview matrix is only rotation-translation (no skew)

  // Bounding box corners in camera space 
  float radius = a * rec->GetR(); 
  CVector3D toprightcam   (projPos.x+radius, projPos.y+radius, projPos.z);
  CVector3D bottomleftcam (projPos.x-radius, projPos.y-radius, projPos.z);

  // Project the corners using projection matrix
  float topright[2],bottomleft[2];
  MultCol32(_projViewportMatrix, toprightcam, topright);
  MultCol32(_projViewportMatrix, bottomleftcam, bottomleft);

  if (topright[0] < bottomleft[0]) Swap(topright[0], bottomleft[0]);
  if (topright[1] < bottomleft[1]) Swap(topright[1], bottomleft[1]);

  // If the top-right corner of the bounding box is lower than the bottom left of 
  // the image, the record will not be used

  // BBox corners in screen space
  bbox.xmax = Min( (int)(topright[0]+0.9999999f), _w-1);
  bbox.ymax = Min( (int)(topright[1]+0.9999999f), _h-1);
  bbox.xmin = Max( (int)(bottomleft[0]), 0);
  bbox.ymin = Max( (int)(bottomleft[1]), 0);

  // reject if off-screen
  if (bbox.xmax < 0 || bbox.ymax < 0 || bbox.xmin>=_w || bbox.ymin>=_h)
    return false;

  return true;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::PutPixel()
// --------------------------------------------------------------------------
inline int
CICSplatterRegular::PutPixel(int cx, int cy, const float col[])
{
  if(IsOnScreen(cx,cy)) {
    int idx = Idx(cx,cy);
    // _weightImg[idx] = 1;
    _irradImg[3*idx+0] = col[0];
    _irradImg[3*idx+1] = col[1];
    _irradImg[3*idx+2] = col[2];
    return 1;
  }
  else
    return 0;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::PutWeight()
// --------------------------------------------------------------------------
inline int
CICSplatterRegular::PutWeight(int cx, int cy, const float w)
{
  if(IsOnScreen(cx,cy)) {
    int idx = Idx(cx,cy);
    _weightImg[idx] = w;
    return 1;
  }
  else
    return 0;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::DrawCenter()
// --------------------------------------------------------------------------
inline void 
CICSplatterRegular::DrawCenter(const CVector3D& projPos)
{
  float c[2];
  MultCol32(_projViewportMatrix, projPos, c);
  PutPixel((int)c[0],(int)c[1]);
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::Splat()
// --------------------------------------------------------------------------
template<bool jitter>
int
CICSplatterRegular::Splat(const CIrradianceRecord *irr, const float a)
{
  if(CheckBackFacing(irr)) return 0;
  // number of affected pixels
  int numPixels = 0;
  // Compute the record position in camera coordinates
  CVector3D projPos = Modelview(irr->GetPoint());
  // compute the screen bounding box
  CICSplatterRegular::TBBox bbox;
  if( ComputeBBox(irr,a,projPos,bbox) ) {
    // DEBUG << "x=" << bbox.xmin << ", " << bbox.xmax << std::endl;
    // DEBUG << "y=" << bbox.ymin << ", " << bbox.ymax << std::endl;
    // rasterize the record
    numPixels += Rasterize<jitter>(irr, a, bbox);
    // center as a white dot
    if(_drawCenter) DrawCenter(projPos);
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::Splat()
// --------------------------------------------------------------------------
template<bool jitter,class TRecord>
int 
CICSplatterRegular::Splat(const CRecordList<TRecord> &recList, const float a)
{
  int numPixels=0;
  CIrradianceList::const_iterator it;
  for(it = recList.begin(); it!=recList.end(); it++) {
    numPixels += Splat<jitter>(*it, a);
  }
  return numPixels;
}

// --------------------------------------------------------------------------
//   CICSplatterRegular::Splat()
// --------------------------------------------------------------------------
template<bool jitter,class TRecord>
int 
CICSplatterRegular::Splat(const CRecordList<TRecord> &recList)
{
  int numPixels=0;
  typename CRecordList<TRecord>::const_iterator it;
  for(it = recList.begin(); it!=recList.end(); it++) {
    numPixels += Splat<jitter>(*it);
  }
  return numPixels;
}


__END_GOLEM_HEADER

#endif // __ICSPLATR_H__
