// ==================================================================
// $Id: radsamp2.cpp,v 1.5 2005/11/29 16:04:05 xkrivanj Exp $
//
// diffsampl.cpp
//    Support routines for sampling hemispheres.
//    Used for radiance caching.
//
// Initial coding by Jaroslav Krivanek, Descember 2003

// GOLEM headers
#include "config.h"
#include "contrib.h"
#include "incrad.h"
#include "radsampl.h"
#include "rnd.h"
#include "surface.h"

// SPHAR headers
#include "shvector.h"
#include "shbasis.h"

__BEGIN_GOLEM_SOURCE

/////////////////////////////////////////////////////////////////////////////
//
//   CRadianceSamplerDeriv
//
/////////////////////////////////////////////////////////////////////////////

bool
CRadianceSamplerDeriv::ComputeGradients(const CVector3D &UU,
                                        const CVector3D &VV,
                                        const CVector3D &NN,
                                        const CCoefVec<float>& result,
                                        CCoefVec<float>& dx,
                                        CCoefVec<float>& dy,
                                        int desiredOrder) const
{
  int sqrDesiredOrder = desiredOrder*desiredOrder;
  int vecLen = PAD_FOUR(sqrDesiredOrder);
  int channels = dx.GetNumChannels();
  ALIGN16 float dx_gray [MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
  ALIGN16 float dy_gray [MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
  for(int i=sqrDesiredOrder; i<vecLen; i++) {dx_gray[i]=dy_gray[i]=0;}
  float cos_xi;

  dx.setz();
  dy.setz();

  for (int k = 0; k < _N; k++) {
    for (int j = 0; j < _M; j++) {

      int idx = Idx(j,k);
      // skip rays going to infinity
      if(_r[idx]==CLimits::Infinity || (cos_xi=-_dot[idx])<0.05)
        continue;
      const CDirData &dd = _precomputedDirData[idx];
      // ray direction in the middle of strata
      const CVector3D raydir_loc = dd.v;
      // transform normal at the hit point to local frame
      CVector3D nl = ToLocalFrame(_normal[idx],UU,VV,NN);
      float r = _r[idx], ir=1.0/r;
      // derivatives of Omega w.r.t x and y
      float dOmega_dx = ( nl.x + 3.0f*raydir_loc.x*cos_xi ) / (r*cos_xi);
      float dOmega_dy = ( nl.y + 3.0f*raydir_loc.y*cos_xi ) / (r*cos_xi);
      const float* Li = _Li[idx].data();
      const float *H = dd.H;
#if 0  // SIMD
      shvf::lerpvv4(sqrDesiredOrder,dx_gray,dd.H,dOmega_dx,dd.dHdx,ir);
      shvf::lerpvv4(sqrDesiredOrder,dy_gray,dd.H,dOmega_dy,dd.dHdy,ir);
#else  // original code
      const float *dHdx = dd.dHdx;
      const float *dHdy = dd.dHdy;
      for(int i=0; i<sqrDesiredOrder; i++) {
        dx_gray[i] = H[i] * dOmega_dx + dHdx[i] * ir;
        dy_gray[i] = H[i] * dOmega_dy + dHdy[i] * ir;
      }
#endif
      // Li is the incoming radiance along the rays
      if(_multichannelGrad) {
        for(int ch=0; ch < channels; ch++) {
          if(Li[ch]<=0) { continue; }
          //  multiply + add
          shvf::madvsip4(vecLen,dx.GetChannel(ch),dx_gray,Li[ch]);
          shvf::madvsip4(vecLen,dy.GetChannel(ch),dy_gray,Li[ch]);
        }
      }
      else {
        // gradient only for luminance
        float lum = _Li[idx].Luminance();
        if(lum>0) {
          shvf::madvsip4(vecLen,dx.GetChannel(0),dx_gray,lum);
          shvf::madvsip4(vecLen,dy.GetChannel(0),dy_gray,lum);
        }
      }

    }
  }

  // normalize
  dx.mul(_invNumCells*M_PI*2.0);
  dy.mul(_invNumCells*M_PI*2.0);

/*
  for(int ch=0; ch<channels; ch++) {
    for(int i=0; i<vecLen; i++) {
      if(!_finite(dx.GetCoeff(ch,i)) ||  _isnan(dx.GetCoeff(ch,i))) {
        PRINTVAR(ch);
        PRINTVAR(i);
        PRINTVAR(dx.GetCoeff(ch,i));
      }
      if(!_finite(dy.GetCoeff(ch,i)) ||  _isnan(dy.GetCoeff(ch,i))) {
        PRINTVAR(ch);
        PRINTVAR(i);
        PRINTVAR(dy.GetCoeff(ch,i));
      }
    }
  }
*/

  return true; // ok
}

/////////////////////////////////////////////////////////////////////////////
//
//   CRadianceSamplerWard
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CRadianceSamplerWard::CRadianceSamplerWard()
// --------------------------------------------------------------------------
CRadianceSamplerWard::
CRadianceSamplerWard(int                depth, 
                     int                numCells, 
                     int                raysPerCell,
                     int                maxOrder,
                     float              infRayLength, 
                     float              minAllowedHitDist,
                     CIncidentRadiance* incRad, 
                     const CSphBasis<float,float>   *sphBasis,
                     bool               localFrame,
                     bool               multichannelGrad,
                     const void*        data): 
CRadianceSampler(depth, numCells, raysPerCell, maxOrder, infRayLength,
                 minAllowedHitDist,incRad, sphBasis, localFrame, 
                 multichannelGrad, data)
{
  int channels = CStandardValues::GetDefaultWeightCount();

  M_PI_mul_2_div_N  =  M_PI*2.0 / _N; 

  v_factor          = new float[_M];
  u_factor          = new float[_M + 1];
  u_vect_canonical  = new CVector3D[_N];
  v_vect_canonical  = new CVector3D[_N];

  // -----------------------
  //  precompute phi arrays
  // -----------------------
  for (int k = 0; k < _N; k++) {
    float phi = 2.0 * M_PI * ((float)k + 0.5) *_invN;
    float cos_phi_k = cos(phi);
    float sin_phi_k = sin(phi);
    
    u_vect_canonical[k] = CVector3D(cos_phi_k, sin_phi_k, 0);
    
    float phiBase = phi + 0.5*M_PI;
    if (phiBase >= 2.0*M_PI) {
      phiBase -= 2.0*M_PI;
    }
    float wv = 2.0*M_PI*float(k)/float(_N) + 0.5*M_PI;
    if (wv >= 2.0 * M_PI) { wv -= 2.0*M_PI; }
    v_vect_canonical[k] = CVector3D(cos(wv), sin(wv), 0);
  }

  // -------------------------
  //  precompute theta arrays
  // -------------------------
  for(int j = 0; j <= _M; j++) {
    float cos_theta_grid = 1.0 - (float)j*_invM;
    float sin_theta_grid = sqrtf(1.0f - sqr(cos_theta_grid));
    u_factor[j]          = M_PI_mul_2_div_N * sin_theta_grid * cos_theta_grid;
  }

  for (int j = 0; j < _M; j++) {
    float cos_theta = 1.0 - ((float)j+0.5)*_invM;
    float sin_theta = sqrtf(1.0f - sqr(cos_theta));
    v_factor[j] = 1.0f/(_M*sin_theta);
  }
}

// --------------------------------------------------------------------------
//  CRadianceSamplerWard::~CRadianceSamplerWard()
// --------------------------------------------------------------------------
CRadianceSamplerWard::~CRadianceSamplerWard() 
{
  // precomputed data
  delete[] u_vect_canonical;
  delete[] v_vect_canonical;

  delete[] u_factor;
  delete[] v_factor;
}

// --------------------------------------------------------------------------
//  CRadianceSamplerWard::ComputeGradients()
// --------------------------------------------------------------------------
bool
CRadianceSamplerWard::ComputeGradients(const CVector3D &UU,
                                       const CVector3D &VV,
                                       const CVector3D &NN,
                                       const CCoefVec<float>& result,
                                       CCoefVec<float>& dx,
                                       CCoefVec<float>& dy,
                                       int desiredOrder) const
{
  // number of color channels
  int channels = dx.GetNumChannels();
  int ch;

  int sqrDesiredOrder = desiredOrder * desiredOrder;
  int vecLen = PAD_FOUR(desiredOrder * desiredOrder);

  // reset the gradient
  dx.setz(); dy.setz();

  CCoefVec<float> _gradU(channels,desiredOrder);
  CCoefVec<float> _gradV(channels,desiredOrder);

  // phi intervals
  for (int k = 0; k < _N ; k++)
  {
    _gradU.setz();
    _gradV.setz();

    // theta intervals
    for(int j = 0; j<_M ; j++) {
      // indices for this cell and neigboring cells
      int ijk  = Idx(j,k);
      // skip this cell - it was under the horizon
      if(
        _surfaceType[ijk]==EE_REFLECTED_GLOSSY || 
        _r[ijk]==-1) continue;
      int ij1k = Idx(j-1,k);
      int ijk1 = k>0 ? Idx(j,k-1) : Idx(j,_N-1);
      // basis function evaluated for this cell
      const float *Hjk  = _precomputedDirData[ijk].H;
      // incoming radiance for thi and neigboring cells
      const float *Ljk  = _Li[ijk].data();
      const float *Ljk1 = _Li[ijk1].data();

      //  u - translational gradient
      if (j>0) {
        // skip this cell - it was under the horizon
        if(_r[ij1k]==-1) continue;
        const float *Lj1k = _Li[ij1k].data();
        float m = u_factor[j] / Min(_r[ijk], _r[ij1k]);
        for(ch=0; ch<channels;ch++)
          shvf::madvsip(vecLen,_gradU.GetChannel(ch),Hjk,(Ljk[ch]-Lj1k[ch])*m);
      }
      // v - translational gradient
      // skip this cell - it was under the horizon
      if(_r[ijk1]==-1) continue;
      float m = v_factor[j] / Min(_r[ijk], _r[ijk1]);
      for(ch=0; ch<channels;ch++)
        shvf::madvsip(vecLen,_gradV.GetChannel(ch),Hjk,(Ljk[ch]-Ljk1[ch])*m);
    }

    // accumulate rotational gradients
    for(ch=0; ch<channels; ch++) {
      float *dxch = dx.GetChannel(ch);
      float *dych = dy.GetChannel(ch);
      const float *tuch = _gradU.GetChannel(ch);
      const float *tvch = _gradV.GetChannel(ch);
      // x-derivative
      shvf::madvsip(vecLen, dxch, tuch, u_vect_canonical[k].x );
      shvf::madvsip(vecLen, dxch, tvch, v_vect_canonical[k].x );
      // y-derivative
      shvf::madvsip(vecLen, dych, tuch, u_vect_canonical[k].y );
      shvf::madvsip(vecLen, dych, tvch, v_vect_canonical[k].y );
    }
  } // for k

  // normalization
  // float m = invRayNum * 2.0f*M_PI * gradMul;
  // dx->mul(m); dy->mul(m);

  return true; // ok
}


// --------------------------------------------------------------------------
//  CRadianceSamplerWard::_fillMissingDataAndFilter()
// --------------------------------------------------------------------------
void 
CRadianceSamplerWard::_fillMissingDataAndFilter(CValues *_Li, float *_r,
                                                CValues *LLi, float *rr) const
{
#if 0
  // theta intervals
  for(int j = 0; j<_M ; j++) {
    // phi intervals
    for (int k = 0; k<_N ; k++)
    {
      int ijk  = idx(j,k);

      // this cell is ok
      // if(_r[ijk]!=-1) continue;

      // this cell has to be reconstructed

      // number of contributing cells
      int     n=0;
      // summed weight
      float   w=0;
      // summed distances
      float   r=0;
      // average luminance, average squared luminance
      float lumAvg=0, sqrLumAvg=0;
      // summed incoming radiances
      CValues L(_Li[0]);
      L.setz();
      // surrounding cells
      int jprev = idx(j-1,k);
      if(j>0 && _r[jprev]!=-1) {
        L += _Li[jprev];
        r += _r[jprev];
        w += 1;
        float lum  = _Li[jprev].MeanValue();
        lumAvg    += lum;
        sqrLumAvg += (lum*lum);
        n++;
      }
      int jnext = idx(j+1,k);
      if(j<M-1 && _r[jnext]!=-1) {
        L += _Li[jnext];
        r +=  _r[jnext];
        w += 1;
        float lum  = _Li[jnext].MeanValue();
        lumAvg    += lum;
        sqrLumAvg += (lum*lum);
        n++;
      }
      int kprev = k>0 ? idx(j,k-1) : idx(j,N-1);
      if(_r[kprev]!=-1) {
        L += _Li[kprev];
        r +=  _r[kprev];
        w += 1;
        float lum  = _Li[kprev].MeanValue();
        lumAvg    += lum;
        sqrLumAvg += (lum*lum);
        n++;
      }
      int knext = k<N-1 ? idx(j,k+1) : idx(j,0);
      if(_r[knext]!=-1) {
        L += _Li[knext];
        r +=  _r[knext];
        w += 1;
        float lum  = _Li[knext].MeanValue();
        lumAvg    += lum;
        sqrLumAvg += (lum*lum);
        n++;
      }

      // this cell
      if(_r[ijk]!=-1) {
        // this cell has a valid contribution
        if(n>1) {
          // check if this is an outlier
          float lum  = _Li[ijk].MeanValue();
          /*
          lumAvg    += lum;
          sqrLumAvg += (lum*lum);
          n++;
          */
          // compute variance
          float in = 1.0f/(float)n;
          lumAvg    *= in;
          sqrLumAvg *= in;
          float variance = sqrLumAvg - sqr(lumAvg);

          if( variance>1e-2 && fabsf(lum-lumAvg) > variance*10 ) {
            // PRINTVAR(variance);
            // DEBUG << "outlier" << std::endl;
            goto normalize;
          }
        }

        // it was not an outlier, add it to the average
        const float thisw = 1.5;
        L.Madd(_Li[ijk], thisw);
        r += thisw * _r[ijk];
        w += thisw;
      }

      // normalize
normalize:
      if(w != 0) {
        float iw = 1.0f/w;
        LLi[ijk] = L*iw;
        rr[ijk]  = r*iw;
      }
      else {
        DEBUG 
          << "Could not reconstruct incoming radiance for "
          << "(" << j << "," << k << ")" << std::endl;
        LLi[ijk].setz();
        rr[ijk] = CLimits::Infinity;
      }

      if(_r[ijk]==-1) {
        _Li[ijk] = LLi[ijk];
        _r [ijk] = rr [ijk];
      }
    }
  }
#endif
}


__END_GOLEM_SOURCE
