// ==================================================================
// $Id: radsampl.cpp,v 1.6 2005/11/02 12:59:02 xkrivanj Exp $
//
// diffsampl.cpp
//    Support routines for sampling hemispheres.
//    Used for radiance caching.
//
// Initial coding by Jaroslav Krivanek, Descember 2003

// GOLEM headers
#include "config.h"
#include "contrib.h"
#include "incrad.h"
#include "radsampl.h"
#include "rnd.h"
#include "surface.h"

#include "appear.h"
#include "object3d.h"

// SPHAR headers
#include "shvector.h"
#include "shbasis.h"
#include "shrot.h"

// #include <cfloat>

__BEGIN_GOLEM_SOURCE

/////////////////////////////////////////////////////////////////////////////
//
//   CRadianceSampler
//
/////////////////////////////////////////////////////////////////////////////

// bool _weird(double f) { return !_finite(f) || _isnan(f); }

// --------------------------------------------------------------------------
//  CRadianceSampler::CRadianceSampler()
// --------------------------------------------------------------------------
CRadianceSampler::
CRadianceSampler(int                depth,
                 int                numCells,
                 int                raysPerCell,
                 int                maxOrder,
                 float             infRayLength,
                 float             minAllowedHitDist,
                 CIncidentRadiance* incRad,
                 const CSphBasis<float,float>  *sphBasis,
                 bool               localFrame,
                 bool               multichannelGrad,
                 const void*        data)
{
  (void)data;
  assert(depth>=0);
  assert(maxOrder>0);
  assert(incRad!=0);
  assert(sphBasis!=0);

  zxzxzRotObject.LiftOrder(maxOrder);

  ResetTimers();

  if(numCells<1) numCells = 1;
  
  _depth    = depth;
  _numCells = numCells;
  _raysPerCell = raysPerCell;
  _invRaysPerCell = 1.0/raysPerCell;
  _incRad   = incRad;
  _sphBasis = sphBasis;
  _maxOrder          = maxOrder;
  _infRayLength      = infRayLength;
  int maxVecLen      = PAD_FOUR(_maxOrder*_maxOrder);
  _localFrame        = localFrame;
  _multichannelGrad  = multichannelGrad;
  _minAllowedHitDist = minAllowedHitDist;
  
  _r     = 0;
  _dir   = 0;
  _Li    = 0;
  _normal= 0;
  _dot   = 0;
  _precomputedDirData = 0;
  _surfaceType = 0;

  // compute the number of divisions
  if (numCells==1) {
    _M = _N  = 1;
    _invM = _invN = 1;
  }
  else {
    _M = (int)(sqrt((double)numCells / 2.0));
    if (_M < 1) _M=1;
    _N = (int)((float)numCells/(float)_M);
    _invN = 1.0f / _N;
    _invM = 1.0f / _M;
  }
  _invNumCells = 1.0f/(_M*_N);
  
  _r      = new float   [_M * _N];
  _dir    = new CVector3D[_M * _N];
  _Li     = new CValues  [_M * _N];
  _normal = new CVector3D[_M * _N];
  _dot    = new float   [_M * _N];
  _surfaceType = new unsigned[_M * _N];

  for(int i=0; i<_M*_N; i++) {
    _Li[i].SetRepresentation(CStandardValues::GetDefaultColorType());
  }

  // Allocate direction data
  int vecLen = PAD_FOUR(maxOrder*maxOrder);

  _precomputedDirData = new CDirData[_N*_M];
  assert(_precomputedDirData);

  float *dHdt = NEW_ALIGN16(float,vecLen);
  float *dHdp = NEW_ALIGN16(float,vecLen);
  assert(dHdt);
  assert(dHdp);
  shvf::setz4(vecLen,dHdt);
  shvf::setz4(vecLen,dHdp);

  for (int k = 0; k < _N; k++) {
    for (int j = 0; j < _M; j++) {
      int i = Idx(j,k);
      CDirData &dd = _precomputedDirData[i];
      // generate direction in the middle of the statum
      float cos_theta = 1.0f - ((float)j + 0.5f) * _invM;
      float sin_theta = sqrtf(1.0f - cos_theta * cos_theta);
      float phi       = 2.0 * M_PI * ((float)k + 0.5f) * _invN;
      float cos_phi   = cosf(phi);
      float sin_phi   = sinf(phi);
      dd.v = CVector3D(cos_phi*sin_theta,sin_phi*sin_theta,cos_theta);
      // evaluate the basis for the direction
      dd.H    = NEW_ALIGN16(float,vecLen);
      assert(dd.H);
      shvf::setz4(vecLen,dd.H);
      dd.dHdx = NEW_ALIGN16(float,vecLen);
      dd.dHdy = NEW_ALIGN16(float,vecLen);
      assert(dd.dHdx);
      assert(dd.dHdy);
      sphBasis->GetAnalytic()->EvalDerivCartV(maxOrder,dd.v,dd.H,dHdt,dHdp);
      // compute derivatives w.r.t x and y
      const float r=1;
      const float ir=1;
      float dTheta_dx = -cos_theta * cos_phi * ir;
      float dTheta_dy = -cos_theta * sin_phi * ir;
      float dPhi_dx   =  sin_phi / (r * sin_theta);
      float dPhi_dy   = -cos_phi / (r * sin_theta);
      for(int i=0; i<vecLen; i++) {
        dd.dHdx[i] = dHdt[i] * dTheta_dx + dHdp[i] * dPhi_dx;
        dd.dHdy[i] = dHdt[i] * dTheta_dy + dHdp[i] * dPhi_dy;
      }
    }
  }
  FREE_ALIGN16(dHdt);
  FREE_ALIGN16(dHdp);
}

// --------------------------------------------------------------------------
//  CRadianceSampler::~CRadianceSampler()
// --------------------------------------------------------------------------
CRadianceSampler::~CRadianceSampler() 
{
  delete [] _Li;
  delete [] _r;
  delete [] _dir;
  delete [] _normal;
  delete [] _dot;
  delete [] _surfaceType;

  if(_precomputedDirData) {
    for(int i=0; i<_M*_N; i++) {
      CDirData &dd = _precomputedDirData[i];
      FREE_ALIGN16(dd.H);
      FREE_ALIGN16(dd.dHdx);
      FREE_ALIGN16(dd.dHdy);
    }
    delete _precomputedDirData;
  }
}

// --------------------------------------------------------------------------
//  GetSurfaceType()
// --------------------------------------------------------------------------
inline static unsigned 
GetSurfaceType(const CHitPointInfo &hitInfo)
{
  CAppearance *apr=(CAppearance*)hitInfo.GetObject()->GetAttr(CHitPointInfo::GetAppearanceIndex());
  CSurface    *surf = apr->GetSurface();
  return surf->GetSupportedTypes();
}

// --------------------------------------------------------------------------
//  CRadianceSampler::SampleHemisphereOneRPC()
// --------------------------------------------------------------------------
bool
CRadianceSampler::SampleHemisphereOneRPC(const CVector3D& pt,
                                          const CVector3D& UU,
                                          const CVector3D& VV,
                                          const CVector3D& NN,
                                          const CVector3D& uN,
                                          const CObject3D* object,
                                          const void*      termination,
                                          const float     &footprintArea,
                                          CHitPointInfo&   outHitInfo,
                                          int              desiredOrder,
                                          CCoefVec<float>& result,
                                          float&         outHarmMeanDistance,
                                          float&         outMinHitDist,
                                          float&         outStdDev,
                                          const void*      data) const
{
  /// Sample hemisphere and save 1) incoming radiance, 2) hit distance

  assert(desiredOrder<=_maxOrder);
  // number of rays hitting a surface
  int numHits = 0;
  // sum of inverse distances to the intersected objects 
  float distancesInv = 0;
  // minimum hit distance
  float minHitDist = CLimits::Infinity;
  // contribution item used to gather IncidentRadiance()
  CContributionItem contrib;
  // set the color & weight type
  contrib.color.SetRepresentation (CStandardValues::GetDefaultColorType() );
  // contrib.weigth.SetRepresentation(CStandardValues::GetDefaultWeightType() );
  // set the interaction type
  contrib.interaction = EE_REFLECTED_GLOSSY;
  // we'll make the contribution go through a diffuse relection, thus:
  /// @@@ tohke nastavit podle plochy, na ktere se sampluje
  contrib.diffuseness = 1;
  // and this is a reflected ray
  contrib.flags = CContributionItem::EE_ReflectedRay;
  // uniform probability density
  contrib.prob = 0.5f*M_1_PI;
  // total radiance for odd and even rays
  float L[2] = {0,0};
  // total number of odd and even rays
  int numRays[2] = {0,0};
  // CVector3D unormal = (DotProd(NN,uN)>=0)?uN:-uN;

  for (int k = 0; k < _N; k++) {
    for (int j = 0; j < _M; j++) {
      int idx = Idx(j,k);
      // reject rays close to horizon
      /*
      if(j>=_M-1) {
        _r[idx] = -1; 
        continue; 
      }
      */
      // sample the hemisphere uniformly in local coordinates 
      // (avoiding rays very near to the hirizon)
      float rnd_theta = (j == _M-1) ? 0.25 : RND::rand01f();
      float cos_theta = 
        1.0f - ((float)j + rnd_theta ) * _invM;

      float sin_theta = sqrtf(1.0f - cos_theta * cos_theta);
      float phi       = 2.0 * M_PI * ( (float)k + RND::rand01f() ) * _invN;
      const CVector3D raydir_loc =  
        CVector3D(cosf(phi)*sin_theta, sinf(phi)*sin_theta, cos_theta);
      // transform from local coordinates
      CVector3D raydir_glob = ToGlobalFrame(raydir_loc, UU,VV,NN);
      // reject rays entering the surface
      /* makes problems on the silhouettes of the disney hall
      if(DotProd(raydir_glob,uN)<=0 ){
        _r[idx] = -1; 
        continue; 
      }
      */
      // save for gradient computation
      _dir[idx] = raydir_glob;
      // initialize the ray
      contrib.ray.Init(pt,raydir_glob,REFLECTED_RAY,termination,object,0,true);
      // unset the i'invalid' flag, if it was set from previous iteration
      contrib.invalid = false;
      // cast the ray an compute higher order light contribution
      if(_incRad->IncidentRadiance(outHitInfo,&contrib,_depth,footprintArea)){
        // the ray hit something
        numHits++; 
        // save the hit distance
        float r = _r[idx] = outHitInfo.GetT();
        // update minimum distance
        if(r < minHitDist) minHitDist = r;
        // update harmonic mean distance
        distancesInv += 1.0f/r;
        // check if the intersection was too close
        CHECK_HIT_DIST_RC
        // save the hit normal and dot product for gradient computation
        _normal[idx] = outHitInfo.GetNormal();
        _dot[idx]    = outHitInfo.GetDotProd();
        _surfaceType[idx] = GetSurfaceType(outHitInfo);
      }
      else {
        _surfaceType[idx] = 0;
        // the ray did not hit anything
        _r[idx] = CLimits::Infinity;
      }

      // increment the number of even/odd rays
      numRays[(j+k)&1]++;
      // valid / invalid contribution?
      if( !contrib.invalid ) {
        // there was a valid contribution - save the incoming radiance
        _Li[idx] = contrib.color;
        // increment even/odd radiance sum
        L[(j+k)&1] += contrib.color.Luminance();
      }
      else {
        // not a valid contribution - set incoming radiance to zero
        _Li[idx].setz();
      }
    }
  }

  // compute standard deviation
  // estimate result variance
  if(numRays[0]>0 && numRays[1]>0) {
    L[0] /= numRays[0];
    L[1] /= numRays[1];
    outStdDev = fabs(L[0]-L[1]);
    // this should actually be multiplied by 0.5 to make it correct
  }
  else
    outStdDev = 0;

  // check if the intersection was too close
  CHECK_HIT_DIST_RC_FINAL;

  // compute harmonic mean distance
  if(numHits == 0)
    outMinHitDist = outHarmMeanDistance = _infRayLength;
  else {
    outHarmMeanDistance = (float)numHits / distancesInv;
    outMinHitDist = minHitDist;
  }

  return true; // OK
}


// --------------------------------------------------------------------------
//  CRadianceSampler::SampleHemisphereManyRPC()
// --------------------------------------------------------------------------
bool
CRadianceSampler::SampleHemisphereManyRPC(const CVector3D& pt,
                                          const CVector3D& UU,
                                          const CVector3D& VV,
                                          const CVector3D& NN,
                                          const CVector3D& uN,
                                          const CObject3D* object,
                                          const void*      termination,
                                          const float     &footprintArea,
                                          CHitPointInfo&   outHitInfo,
                                          int              desiredOrder,
                                          CCoefVec<float>& result,
                                          float&         outHarmMeanDistance,
                                          float&         outMinHitDist,
                                          float&         outStdDev,
                                          const void*      data) const
{
  /// Sample hemisphere and save 1) incoming radiance, 2) hit distance

  assert(desiredOrder<=_maxOrder);
  // number of rays hitting a surface
  int numHits = 0;
  // sum of inverse distances to the intersected objects 
  float distancesInv = 0;
  // minimum hit distance
  float minHitDist = CLimits::Infinity;
  // contribution item used to gather IncidentRadiance()
  CContributionItem contrib;
  // set the color & weight type
  contrib.color.SetRepresentation (CStandardValues::GetDefaultColorType() );
  // contrib.weigth.SetRepresentation(CStandardValues::GetDefaultWeightType() );
  // set the interaction type
  contrib.interaction = EE_REFLECTED_GLOSSY;
  // we'll make the contribution go through a diffuse relection, thus:
  /// @@@ tohke nastavit podle plochy, na ktere se sampluje
  contrib.diffuseness = 1;
  // and this is a reflected ray
  contrib.flags = CContributionItem::EE_ReflectedRay;
  // uniform probability density
  contrib.prob = 0.5f*M_1_PI;
  // total radiance for odd and even rays
  float L[2] = {0,0};
  // total number of odd and even rays
  int numRays[2] = {0,0};

  for (int k = 0; k < _N; k++) {
    for (int j = 0; j < _M; j++) {
      int idx = Idx(j,k);

      _Li[idx].setz();
      // _r[idx] = CLimits::Infinity;
      _r[idx] = 0;
      _dir[idx].SetValue(0);
      _normal[idx].SetValue(0);
      _dot[idx]=0;

      for(int rayIdx = 0; rayIdx<_raysPerCell; rayIdx++) {
        // sample the hemisphere uniformly in local coordinates
        float cos_theta = 1.0f - ((float)j + RND::rand01f() ) * _invM;
        float sin_theta = sqrtf(1.0f - cos_theta * cos_theta);
        float phi       = 2.0 * M_PI * ( (float)k + RND::rand01f() ) * _invN;
        const CVector3D raydir_loc =  
          CVector3D(cosf(phi)*sin_theta, sinf(phi)*sin_theta, cos_theta);
        // transform from local coordinates
        CVector3D raydir_glob = ToGlobalFrame(raydir_loc, UU,VV,NN);
        // reject rays entering the surface
        /*
        if(DotProd(raydir_glob,uN)<=0){
        _r[idx] = -1; 
        continue; 
        }
        */
        // save for gradient computation
        _dir[idx] += raydir_glob;
        // initialize the ray
        contrib.ray.Init(pt,raydir_glob,REFLECTED_RAY,termination,object,0,true);
        // unset the i'invalid' flag, if it was set from previous iteration
        contrib.invalid = false;
        // cast the ray an compute higher order light contribution
        if(_incRad->IncidentRadiance(outHitInfo,&contrib,_depth,footprintArea)){
          // the ray hit something
          numHits++; 
          // save the hit distance
          float r = outHitInfo.GetT();
          // if(r<_r[idx]) _r[idx]=r;
          _r[idx] += r;

          // update minimum distance
          if(r < minHitDist) minHitDist = r;
          // update harmonic mean distance
          distancesInv += 1.0f/r;
          // check if the intersection was too close
          CHECK_HIT_DIST_RC
          // save the hit normal and dot product for gradient computation
          _normal[idx] += outHitInfo.GetNormal();
          _dot[idx]    += outHitInfo.GetDotProd();
        }
        else {
          // the ray did not hit anything
          _r[idx] += CLimits::Infinity;
        }

        // increment the number of even/odd rays
        numRays[(j+k)&1]++;
        // valid / invalid contribution?
        if( !contrib.invalid ) {
          // there was a valid contribution - save the incoming radiance
          _Li[idx] += contrib.color;
          // increment even/odd radiance sum
          L[(j+k)&1] += contrib.color.Luminance();
        }
      } // for rays in cell
      _Li[idx]     *= _invRaysPerCell;
      _r[idx]      *= _invRaysPerCell;
      _dir[idx]    *= _invRaysPerCell;
      _normal[idx] *= _invRaysPerCell;
      _dot[idx]    *= _invRaysPerCell;
    }
  }

  // compute standard deviation
  // estimate result variance
  L[0] /= numRays[0];
  L[1] /= numRays[1];
  outStdDev = fabs(L[0]-L[1])/(L[0]+L[1]+1e-6);

/*
  PRINTVAR(outStdDev);
  PRINTVAR(numRays[0]);
  PRINTVAR(numRays[1]);
  PRINTVAR(L[0]);
  PRINTVAR(L[1]);
  if(_isnan(outStdDev))
    FATAL_EXIT;
*/

  // check if the intersection was too close
  CHECK_HIT_DIST_RC_FINAL;

  // compute harmonic mean distance
  if(numHits == 0)
    outMinHitDist = outHarmMeanDistance = _infRayLength;
  else {
    outHarmMeanDistance = (float)numHits / distancesInv;
    outMinHitDist = minHitDist;
  }

  return true; // OK
}

// --------------------------------------------------------------------------
//  CRadianceSampler::Project()
// --------------------------------------------------------------------------
bool
CRadianceSampler::Project(CCoefVec<float>& result,int desiredOrder) const
{
  int channels = result.GetNumChannels();
  int vecLen = PAD_FOUR(desiredOrder * desiredOrder);
  result.setz();
  // sum over all incoming radiance contributions
  for (int k = 0; k < _N; k++) {
    for (int j = 0; j < _M; j++) {
      int idx = Idx(j,k);
      // skip this cell - it was under the horizon
      if(_r[idx]==-1) continue;
      // incoming radiance 
      const float *Li = _Li[idx].data();
      // precomputed data for this direction
      const CDirData &dd = _precomputedDirData[idx];
      // update the projection sum
      for(int ch=0; ch < channels; ch++) {
        if(Li[ch]<=0) { continue; }
        shvf::madvsip4(vecLen,result.GetChannel(ch),dd.H,Li[ch]);
      }
    }
  }
  // normalization: probability / NM
  result.mul(_invNumCells*M_PI*2.0);
  return true; // ok
}

// --------------------------------------------------------------------------
//  CRadianceSampler::SampleHemisphereSaveAll()
// --------------------------------------------------------------------------
bool CRadianceSampler::RotateToGlobalFrame(const CVector3D &UU,
                                           const CVector3D &VV,
                                           const CVector3D &NN,
                                           CCoefVec<float> &result,
                                           CCoefVec<float> *dx,
                                           CCoefVec<float> *dy,
                                           int desiredOrder) const
{
  int channels = result.GetNumChannels();

#if 1  // ZXZXZ rotation

  zxzxzRotObject.ToGlobalFrame<float>(UU,VV,NN,desiredOrder);
  // rotate & compute the dot product
  for(int ch=0; ch<channels; ch++) {
    zxzxzRotObject.TransformVector(result.GetChannel(ch),
      result.GetChannel(ch),desiredOrder,rotGenerator);
  }
  if(dx && dy) {
    for(int ch=0; ch<channels; ch++) {
      zxzxzRotObject.TransformVector(dx->GetChannel(ch),dx->GetChannel(ch),
        desiredOrder,rotGenerator);
      zxzxzRotObject.TransformVector(dy->GetChannel(ch),dy->GetChannel(ch),
        desiredOrder,rotGenerator);
    }
  }

#else // Ivanic & Rudenberg [testing code - not optimized]
  float ALIGN16 tmpvec[SQR_MAX_FIXED_SH_ORDER];
  int sqrDesiredOrder = sqr(desiredOrder);

  CSHRotMatrix M(desiredOrder);
  rotGenerator.ToGlobalFrame(M,desiredOrder,UU,VV,NN);
  // rotate & compute the dot product
  for(int ch=0; ch<channels; ch++) {		
    M.TransformVector(tmpvec,result.GetChannel(ch),desiredOrder);
    shvf::movvv(sqrDesiredOrder,result.GetChannel(ch),tmpvec);
  }
  if(dx && dy) {
    for(int ch=0; ch<channels; ch++) {
      M.TransformVector(tmpvec,dx->GetChannel(ch),desiredOrder);
      shvf::movvv(sqrDesiredOrder,dx->GetChannel(ch),tmpvec);
      M.TransformVector(tmpvec,dy->GetChannel(ch),desiredOrder);
      shvf::movvv(sqrDesiredOrder,dy->GetChannel(ch),tmpvec);
    }
  }
#endif

  return true; //ok
}

// --------------------------------------------------------------------------
//  CRadianceSampler::SampleHemisphereSaveAll()
// --------------------------------------------------------------------------
bool CRadianceSampler::ComputeRadiance(const CVector3D &pt,
                                       const CVector3D& UU,
                                       const CVector3D& VV,
                                       const CVector3D& NN,
                                       const CVector3D& uN,
                                       const CObject3D* object,
                                       const void* termination,
                                       const float     &footprintArea,
                                       CHitPointInfo& outHitInfo,
                                       int desiredOrder,
                                       CCoefVec<float>& result,
                                       CCoefVec<float>* dx,
                                       CCoefVec<float>* dy,
                                       float&         outHarmMeanDistance,
                                       float&         outMinHitDist,
                                       float&         outStdDev,
                                       const void*     data) const
{
  bool res;

  _sampleTimer.Start();
  if(_raysPerCell<=1)
    res = SampleHemisphereOneRPC(pt,UU,VV,NN,uN,object,termination,
    footprintArea,outHitInfo,desiredOrder,result,outHarmMeanDistance,
    outMinHitDist,outStdDev,data);
  else
    res = SampleHemisphereManyRPC(pt,UU,VV,NN,uN,object,termination,
    footprintArea,outHitInfo,desiredOrder,result,outHarmMeanDistance,
    outMinHitDist,outStdDev,data);
  _sampleTimer.Stop();
  if(!res) return false; // error

  // project to spherical harmonics
  _projectTimer.Start();
  res = Project(result,desiredOrder);
  _projectTimer.Stop();
  if(!res) return false; // error

  if(dx && dy) {
    if(outHarmMeanDistance == _infRayLength) {
      // no hits - avoid gradient computation
      dx->setz(); dy->setz();
      dx = dy = 0;
    }
    else {
      _gradTimer.Start();
      res = ComputeGradients(UU,VV,NN,result,*dx,*dy,desiredOrder);
      _gradTimer.Stop();
      if(!res) return false; // error
    }
  }

  if(!_localFrame) {
    if(!RotateToGlobalFrame(UU,VV,NN,result,dx,dy,desiredOrder))
      return false; // error
  }

  return true; // ok
}

/////////////////////////////////////////////////////////////////////////////
//
//   CRadianceSamplerDisplace
//
/////////////////////////////////////////////////////////////////////////////

bool
CRadianceSamplerDisplace::ComputeGradients(const CVector3D &UU,
                                           const CVector3D &VV,
                                           const CVector3D &NN,
                                           const CCoefVec<float>& result,
                                           CCoefVec<float>& dx,
                                           CCoefVec<float>& dy,
                                           int desiredOrder) const
{ 
  FATAL << "Displace nefunguje a uz me to nebavi debugovat.\n";
  FATAL << "Stejne je to jenom relikt." << std::endl;
  FATAL_EXIT;


  float displacement = 0.001;
  
  // just for debugging; gradient use (H)SH order as specified here 
  //   (must be <= desiredOrder)
  int gradientOrder    = desiredOrder;
  int sqrGradientOrder = sqr(gradientOrder);
  int gradVecLen       = PAD_FOUR(sqrGradientOrder);
  const float *H;
  int channels         = dx.GetNumChannels();
  int ch;
  
  // displacement in the tangent plane transformed to global frame
  CVector3D xdisplaced = displacement*UU;
  CVector3D ydisplaced = displacement*VV;
  
  CVector3D  xrayDir, yrayDir;
  float     xmul, ymul, xmag, ymag;
  float     xdistratio, ydistratio, xdotratio, ydotratio;

  // reset 
  dx.setz();
  dy.setz();
  
  for (int k = 0; k < _N; k++) {
    for (int j = 0; j < _M; j++) {
      int i = Idx(j,k);
      // ray direction in the shifted hemisphere
      if( _r[i] < CLimits::Infinity && _dot[i] < -0.05) {
        xrayDir = (_r[i] * _dir[i] - xdisplaced);
        yrayDir = (_r[i] * _dir[i] - ydisplaced);
        xmag = Magnitude(xrayDir);
        ymag = Magnitude(yrayDir);
        if( xmag < CLimits::Small || ymag < CLimits::Small )
          continue;       
        xrayDir /= xmag;
        yrayDir /= ymag;
        /// calculate the change of the solid angle
        /// mul = cos(theta_) * r_old^2 / r_new^2
        xdistratio = sqr(_r[i] / xmag);
        ydistratio = sqr(_r[i] / ymag);
        xdotratio = DotProd(xrayDir,_normal[i]) / _dot[i];
        ydotratio = DotProd(yrayDir,_normal[i]) / _dot[i];
        xmul = xdotratio > 0 ? xdotratio * xdistratio : 0;
        ymul = ydotratio > 0 ? ydotratio * ydistratio : 0;
      }
      else { 
        // ray goes to infinity
        xrayDir = yrayDir = _dir[i];
        xmul    = ymul    = 1;
      }

      // replace xrayDir and yrayDir by their local counterparts for SH 
      // evaluation
      xrayDir = ToLocalFrame(xrayDir, UU,VV,NN);
      yrayDir = ToLocalFrame(yrayDir, UU,VV,NN);

      if(_multichannelGrad) {
        // gradient per color channel
        const float *Li = _Li[i].data();
        if(xmul>0) {
          // evaluate SH for x displacement and update the integral
          _sphBasis->EvalBasisCartVInt(gradientOrder,xrayDir,&H);
          for(ch=0; ch < channels; ch++) {
            if(Li[ch]>0)
              shvf::madvsip4(gradVecLen,dx.GetChannel(ch),H,Li[ch]*xmul);
          }
        }
        if(ymul>0) {
          // evaluate SH for y displacement and update the integral
          _sphBasis->EvalBasisCartVInt(gradientOrder,(const float*)&yrayDir,&H);
          for(ch=0; ch < channels; ch++) {
            if(Li[ch]>0) 
              shvf::madvsip4(gradVecLen,dy.GetChannel(ch),H,Li[ch]*ymul);
          }
        }
      }
      else {
        // one gradient for luminance
        const float lum = _Li[i].Luminance();
        if(xmul>0) {
          _sphBasis->EvalBasisCartVInt(gradientOrder,(const float*)&xrayDir,&H);
          shvf::madvsip4(gradVecLen,dx.GetChannel(0),H,lum*xmul);
        }
        if(ymul>0) {
          _sphBasis->EvalBasisCartVInt(gradientOrder,(const float*)&yrayDir,&H);
          shvf::madvsip4(gradVecLen,dy.GetChannel(0),H,lum*ymul);
        }
      }
    }
  }

  // at this point we have the projection coefficients of Li in the 
  // derivative array - let's convert them to differences.
  float m = 1.0f / displacement;
  float normVal = _invNumCells*M_PI*2.0;
  if(_multichannelGrad) {
    for(ch=0; ch < channels; ch++) {
      float* xch = dx.GetChannel(ch);
      float* ych = dy.GetChannel(ch);
      const float* rch = result.GetChannel(ch);
      for(int i=0; i<sqrGradientOrder; i++) {
        //PRINTVAR(rch[i]);
        //PRINTVAR(_invNumRays*M_PI*2.0*xch[i]);
        xch[i] = (normVal*xch[i] - rch[i]) * m;
        ych[i] = (normVal*ych[i] - rch[i]) * m;
      }
    }
  }
  else {
    assert(channels==3);
    float* xch = dx.GetChannel(0);
    float* ych = dy.GetChannel(0);
    const float* rch[3];
    for(int i=0;i<3;i++) rch[i] = result.GetChannel(i);   
    for(int i=0; i<sqrGradientOrder; i++) {
      float rlum = rch[0][i]*0.299f + rch[1][i]*0.587f + rch[2][i]*0.114f;
      xch[i] = (_invNumCells*xch[i] - rlum) * m;
      ych[i] = (_invNumCells*ych[i] - rlum) * m;
    }
  }

  // dx.setz();
  // dy.setz();
/*
  for(int ch=0; ch<channels; ch++) {
    for(int i=0; i<gradVecLen; i++) {
      if(!_finite(dx.GetCoeff(ch,i)) ||  _isnan(dx.GetCoeff(ch,i))) {
        PRINTVAR(ch);
        PRINTVAR(i);
        PRINTVAR(dx.GetCoeff(ch,i));
      }
      if(!_finite(dy.GetCoeff(ch,i)) ||  _isnan(dy.GetCoeff(ch,i))) {
        PRINTVAR(ch);
        PRINTVAR(i);
        PRINTVAR(dy.GetCoeff(ch,i));
      }
    }
  }
*/


  return true;
}

/*
inline void gradw4(int n,
                  float *dest1,float *dest2, 
                  const float *src1, const float *src2, const float *src3,
                  float w11, float w12, float w13,
                  float w21, float w22, float w23)
#ifdef SSE_INTRINSIC
{
  assert( (n&3) == 0 );
  n /= 4;

  __m128* p_dest1 = (__m128*) dest1;
  __m128* p_dest2 = (__m128*) dest2;

  __m128* p_src1 = (__m128*) src1;
  __m128* p_src2 = (__m128*) src2;
  __m128* p_src3 = (__m128*) src3;

  __m128  m_w11   = _mm_set_ps1(w11);
  __m128  m_w12   = _mm_set_ps1(w12);
  __m128  m_w13   = _mm_set_ps1(w13);
  __m128  m_w21   = _mm_set_ps1(w21);
  __m128  m_w22   = _mm_set_ps1(w22);
  __m128  m_w23   = _mm_set_ps1(w23);

  __m128  m11,m12,m13,m21,m22,m23;

  for ( int i = 0; i < n; i++ ) 
  {
    m11 = _mm_mul_ps(p_src1[i],m_w11);
    m21 = _mm_mul_ps(p_src1[i],m_w21);
    m12 = _mm_mul_ps(p_src2[i],m_w12);
    m22 = _mm_mul_ps(p_src2[i],m_w22);
    m13 = _mm_mul_ps(p_src3[i],m_w13);
    m23 = _mm_mul_ps(p_src3[i],m_w23);

    p_dest1[i] = _mm_add_ps(_mm_add_ps(m11,m12),m13);
    p_dest2[i] = _mm_add_ps(_mm_add_ps(m21,m22),m23);
  }
}
#else
{
  assert( (n&3) == 0 ); 
  gradw(n,dest1,dest2,src1,src2,src3,w11,w12,w13,w21,w22,w23);
}
#endif

bool 
CRadianceSamplerDeriv::ComputeRadiance(const CVector3D &pt,
                                       const CVector3D& U,
                                       const CVector3D& V,
                                       const CVector3D& N,
                                       const CVector3D& uN,
                                       const CObject3D* object,
                                       const void* termination,
                                       const float     &footprintArea,
                                       CHitPointInfo& outHitInfo,
                                       int desiredOrder,
                                       CCoefVec<float>& result,
                                       CCoefVec<float>* dx,
                                       CCoefVec<float>* dy,
                                       float&         outHarmMeanDistance,
                                       float&         outMinHitDist,
                                       float&         outStdDev,
                                       const void*          data) const
{
  // gradients don't work in global frame with this sampler
  assert(!dx || dx && _localFrame); 
  assert(desiredOrder<=_maxOrder);

  int sqrDesiredOrder = desiredOrder * desiredOrder;
  int vecLen = PAD_FOUR(desiredOrder * desiredOrder);

  // number of rays hitting any surface
  int numHits = 0;
  // sum of inverse distances to the intersected objects 
  float distancesInv = 0;
  // minimum hit distance
  float minHitDist = CLimits::Infinity;
  // reset the result
  result.setz();
  // contribution item used to gather IncidentRadiance()
  CContributionItem contrib;
  // set the color & weight type
  contrib.color.SetRepresentation (CStandardValues::GetDefaultColorType() );
  contrib.weigth.SetRepresentation(CStandardValues::GetDefaultWeightType() );
  // set the interaction type
  contrib.interaction = EE_REFLECTED_GLOSSY;
  // we'll make the contribution go through a diffuse relection, thus:
  /// @@@ tohke nastavit podle plochy, na ktere se sampluje
  contrib.diffuseness = 1;
  // and this is a reflected ray
  contrib.flags = CContributionItem::EE_ReflectedRay;
  // outgoing ray direction in local coords (Z = up = surface normal)
  CVector3D raydir_loc;
  // number of color channels
  int channels = result.GetNumChannels();
  int i=0, ch;
  
  if(dx) { assert(dy); dx->setz(); dy->setz(); }
 
  ALIGN16 float dx_gray [MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
  ALIGN16 float dy_gray [MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
  for(i=sqrDesiredOrder; i<vecLen; i++) {dx_gray[i]=dy_gray[i]=0;}
  
  bool inf;
  float r, ir;
  float r1 = 0.0, r2 = 0.0;
  float cos_xi;

  const float *H, *dH_dt, *dH_dp;

  i = 0;
  for(int ir1 = phiInt; ir1-- ; r1 += r1w, r2 = 0 ) {
	   for(int ir2 = thetaInt ; ir2-- ; r2 += r2w, i++ ) {
       // generate random vector
       CVector2D v(r1 + RND::rand0bf(r1w),r2 + RND::rand0bf(r2w));
       // sample the hemisphere uniformly in local coordinates
       float cos_theta = v.xx;
       float sin_theta = sqrtf(1.0f-cos_theta*cos_theta);
       float phi       = 2.0f * M_PI * v.yy;
       float cos_phi   = cosf(phi);
       float sin_phi   = sinf(phi);
       raydir_loc =  
         CVector3D(  cos_phi*sin_theta, sin_phi*sin_theta, cos_theta);
       contrib.prob = 0.5f*M_1_PI;
       // transform from local coordinates
       CVector3D raydir_glob = 
         raydir_loc.x*U + raydir_loc.y*V + raydir_loc.z*N;   
       // initialize the ray
       contrib.ray.Init(pt, raydir_glob, REFLECTED_RAY, termination, object, 0,
         true);
       // unset,if it was set from previous iteration
       contrib.invalid = false;
       // cast the ray an compute higher order light contribution
       if( _incRad->IncidentRadiance( outHitInfo, &contrib, _depth )) {
         // save the hit distance
         r = outHitInfo.GetT(); ir = 1.0f/r;
         // update minimum distance
         if(r < minHitDist) minHitDist = r;
         // update harmonic mean distance
         distancesInv += ir;
         // if ray goes to infty, there is no contrib to distancesInv 
         // (1/infty)
         numHits++; 
         inf = false;
         CHECK_HIT_DIST_RC;
       }
       else {
         r = CLimits::Infinity;
         inf = true;
       }
       
       if( ! contrib.invalid ) {
         contrib.color *= (1.0 / contrib.prob);
         const float* Li = contrib.color.data();
         
        if(!dx || inf || (cos_xi=-outHitInfo.GetDotProd()) < 0.05) {
           /////////////////
           // NO GRADIENTS
           /////////////////

           // evaluate sh coefficients
          // DEBUG << "111" << std::endl;
           _sphBasis->EvalBasisCartVInt(desiredOrder, 
             _localFrame ? raydir_loc : raydir_glob, &H);

          // DEBUG << "222" << std::endl;
           // integrate
           for(ch=0; ch < channels; ch++)
             shvf::madvsip4(vecLen,result.GetChannel(ch),H,Li[ch]);
          // DEBUG << "333" << std::endl;
         }
         else {
           ////////////////////////////////////
           //  GRADIENTS
           ////////////////////////////////////

           // evaluate (H)SH coefficients + derivatives
           _sphBasis->EvalDerivCartVInt(desiredOrder, 
             _localFrame ? raydir_loc : raydir_glob,
             &H, &dH_dt, &dH_dp);

           // transform normal at the hit point to local frame
           CVector3D nl = ToLocalFrame(outHitInfo.GetNormal(),U,V,N);
           
           // derivatives of Omega w.r.t x and y
           float dOmega_dx = ( nl.x + 3.0f*raydir_loc.x*cos_xi ) / (r*cos_xi);
           float dOmega_dy = ( nl.y + 3.0f*raydir_loc.y*cos_xi ) / (r*cos_xi);

           // derivatives needed to evaluate dH/dx 
           // using  dH/dx = dH/dtheta*dtheta/dx + dH/dphi*dphi/dx
           float dTheta_dx = -cos_theta * cos_phi * ir;
           float dTheta_dy = -cos_theta * sin_phi * ir;
           float dPhi_dx   =  sin_phi / (r * sin_theta);
           float dPhi_dy   = -cos_phi / (r * sin_theta);

#if 1     // SIMD
           gradw4(vecLen,
                 dx_gray,dy_gray, 
                 H, dH_dt, dH_dp,
                 dOmega_dx, dTheta_dx, dPhi_dx,
                 dOmega_dy, dTheta_dy, dPhi_dy);
#else     // original code
           for(int i=0; i<sqrDesiredOrder; i++) 
           {
             dx_gray[i] = H[i]     * dOmega_dx + 
                          dH_dt[i] * dTheta_dx + 
                          dH_dp[i] * dPhi_dx;
             dy_gray[i] = H[i]     * dOmega_dy +
                          dH_dt[i] * dTheta_dy + 
                          dH_dp[i] * dPhi_dy;
           }
#endif
           
           // Li is the incoming radiance along the rays
           if( _multichannelGrad )
           {
             // gradient for all channels
             for(ch=0; ch < channels; ch++) {
               if(Li[ch]<=0) { continue; }
               // multiply vector 'H' by scalar Li[ch] and add to 'result'
               shvf::madvsip4(vecLen,result.GetChannel(ch),H,      Li[ch]);
               //  multiply + add
               shvf::madvsip4(vecLen,dx->GetChannel(ch),   dx_gray,Li[ch]);
               shvf::madvsip4(vecLen,dy->GetChannel(ch),   dy_gray,Li[ch]);
             }
           }
           else 
           {
             // gradient only for luminance
             for(ch=0; ch < channels; ch++) {
               if(Li[ch]<=0) { continue; }
               shvf::madvsip4(vecLen,result.GetChannel(ch),H,Li[ch]);
             }

             // float lum = contrib.color.MeanValue();
             float lum = contrib.color.Luminance();
             shvf::madvsip4(vecLen,dx->GetChannel(0),dx_gray,lum);
             shvf::madvsip4(vecLen,dy->GetChannel(0),dy_gray,lum);
           }
         }
       }
       else {
       }
     }
   }
   
   result.mul(invRayNum);
   
   if(numHits == 0) {
     outMinHitDist = outHarmMeanDistance = _infRayLength;
     // if all rays go to infinity, the translational gradient is zero
     if(dx) {
       dx->setz();
       dy->setz();
     }
     return true;
   }

   CHECK_HIT_DIST_RC_FINAL;

   // harmonic mean distance is the inverse of mean inverses 
   outHarmMeanDistance = (float)numHits / distancesInv;
   outMinHitDist = minHitDist;
   
   if(dx) {
     dx->mul(invRayNum);
     dy->mul(invRayNum);
   }
   
   return true;
}
*/

__END_GOLEM_SOURCE
