// ===================================================================
// $Id: radsampl.h,v 1.7 2005/11/02 12:58:46 xkrivanj Exp $
//
// radsampl.h
//     Header file for the CRadianceSampler class - sampling of the 
//     hemisphere and storing the incident radiance field as a 
//     (hemi)spherical function with (hemi) spherical harmonics.
//
// Class: CRadianceSampler
//
// Initial coding by Jaroslav Krivanek, December 2003.

#ifndef __RADSAMPL_H__
#define __RADSAMPL_H__

// GOLEM headers
#include "configh.h"
#include "hminfo.h"
#include "timer.h"

__BEGIN_GOLEM_HEADER

// forward ceclarations
class CIncidentRadiance;
template<class TV,class TA> class CSphBasis;
template<class TV> class CCoefVec;

#if 0
#define CHECK_HIT_DIST_RC
#else
#define CHECK_HIT_DIST_RC \
  if(minHitDist < _minAllowedHitDist) {  \
    /*DEBUG << "RC:(minHitDist=" << minHitDist << ") < (_minAllowedHitDist=" << _minAllowedHitDist << ")" << std::endl; */  \
    /* seems that we were dangerously close to something */ \
    outHarmMeanDistance = -minHitDist;  \
    return false; \
  }
#endif
#define CHECK_HIT_DIST_RC_FINAL CHECK_HIT_DIST_RC

/// An object for computing radiance field at a point.
/**
  This class exists mainly as an interface for the (pure) virtual method 
  ComputeRadiance() and also as a storage of some precomputed data
  of a particular implementation of the radiance sampler.
  This can (and actually is) be used e.g. for radiance caching (not 
  IRrradiance caching) to calculate the samples if nothing is found in 
  the cache. To be able to sample the hemisphere and still be independent on the 
  particular rendering algorithm, the class is initialized with a pointer to the
  CIncidentRadiance interface.
*/
class CRadianceSampler
{

protected: // data

	/// recursion depth of this diffuse sampler
	int _depth;
	/// number of rays to sample the diffuse hemisphere
	int _numCells;
  /// number of rays per cell
  int _raysPerCell;
  /// 1/raysPerCell
  float _invRaysPerCell;
	/// order of the (hemi)spherical harmonics
	int _maxOrder;
	/// what length should be assigned to a ray that does not hit anything
	float _infRayLength;
  /// minimum allowed hit distance during hemisphere sampling
  float _minAllowedHitDist;
	/// pointer an interface implementing IncidentRadiance() method
	CIncidentRadiance *_incRad;
	/// pointer to a (hemi)spherical basis interface
	const CSphBasis<float,float> *_sphBasis;
	/// Store radiance in local or global coordinates.
	bool _localFrame;
  /// Compute gradient for all color channel or only for luminance?
  bool _multichannelGrad;

  /// Number of divisions for theta.
  int   _M;
  /// Number of divisions for phi.
  int   _N;
  /// invM = 1/M;
  float _invM;
  /// invN = 1/N;
  float _invN;
  /// 1/(N*M)
  float _invNumCells;

  /// index to the 2d-array
  inline int Idx(int j, int k) const { return j+k*_M; }

	/// array of incoming radiances
	CValues *_Li;
	/// array of sample distances
	float *_r;
	/// array of sample directions (in global frame)
	CVector3D *_dir;
	/// array of normals
	CVector3D *_normal;
	/// array of  dot products
	float *_dot;
  unsigned *_surfaceType;

  class  CDirData {
  public: // data
    /// direction in cartesian coordinates
    CVector3D v;
    /// vector of (H)SH coefficients for the direction
    float *H;
    /// derivatative w.r.t. to x and y displacement
    float *dHdx, *dHdy;
  };

  CDirData *_precomputedDirData;

  /// Hemisphere sampling timer.
  mutable CTimer _sampleTimer;
  /// Basis projection timer.
  mutable CTimer _projectTimer;
  /// gradient computation timer/
  mutable CTimer _gradTimer;

public: // types

	enum EType {
		/// Gradient computation by displacement
		ERadianceSamplerDisplace = 2,
		/// Gradient computation by derivatives
		ERadianceSamplerDeriv = 3,
		/// Gradients by Ward
		ERadianceSamplerWard = 5
	};

public: // methods

	/// The one and only constructor.
	/**
	  Initialize the object.

    @param depth   (in) 
      Recursion depth for which this is going to be used.
	  @param numRays (in) 
     Number of rays to be shot.
    @param order   (in) 
     Order of the (hemi)spherical harmonics.
	  @param infRayLength(in) 
     Harmonic mean distance to be set if all rays go to infinity.
    @param minAllowedHitDist (in)
     Minimum allowed hit distance during hemisphere sampling.
     If any intersection is closer than this value during hemisphere sampling
     (ComputeRadiance() method), error is indicated and the record will not be 
     stored in the cache. Such a record is very close to something and chances 
     are that it will have some ray leaking and errorneous results.
	  @param sphBasis (in) 
     Const pointer to an interface for evaluation of (hemi)spherical functions.
	  @param incRad  (in) 
     Pointer to a particular implementation of the CIncidentRadiance interface. 
		 incRad->IncidentRadiance() is called to get the incoming radiance from a 
     given direction.
	  @param localFrame (in) 
     Should the incoming radiance be stored in the local coordinates (with 
     normal=z-axis) or in the global coordinates?
    @pram multichannelGrad (in) 
       Should the translational gradient be computed for all color channels 
       or only for luminance?
    @param data    (in) 
      Additional per-instance data.
	*/
	CRadianceSampler (
    int           depth, 
    int           numRays,
    int           numRaysPerCell,
    int           order,
    float         infRayLength, 
    float         minAllowedHitDist,
    CIncidentRadiance* incRad,
    const CSphBasis<float,float>       *sphBasis,
    bool               localFrame,
    bool               multichannelGrad,
    const void*        data);

	/// Virtual destructor.
	virtual ~CRadianceSampler();

	/// Compute the radiance at a given point and given normal.
	/**
	  @param outHitInfo  (tmp) This MUST be used to shoot the rays for 
	                                hemisphere sampling.
	  @param result           (out) Set the result here. Undefined on fail.
    @param harmMeanDistance (out) Harmonic mean of the distances of the rays.

    @return 'true' if the irradiance has been succesfully computed and the 
             result stored in 'result'. Otherwise, 'false' is returned and 
            'result' is undefined.
	 */
  virtual bool ComputeRadiance(
    const CVector3D &pt,
    const CVector3D& U,
    const CVector3D& V,
    const CVector3D& N,
    const CVector3D& uN,
    const CObject3D* object,
    const void* termination,
    const float     &footprintArea,
    CHitPointInfo& outHitInfo,
    int desiredOrder,
    CCoefVec<float>& result,
    CCoefVec<float>* xderivative,
    CCoefVec<float>* yderivative,
    float&         outHarmMeanDistance,
    float&         outMinHitDist,
    float&         outStdDev,
    const void*          data) const;

  /// Sample the hemisphere, save hit information for each cell in arrays.
  /// Returns true on success.
  /**
     This version casts one ray per cell (RPC).
   */
  bool SampleHemisphereOneRPC(
    const CVector3D& pt,
    const CVector3D& UU,
    const CVector3D& VV,
    const CVector3D& NN,
    const CVector3D& uN,
    const CObject3D* object,
    const void*      termination,
    const float     &footprintArea,
    CHitPointInfo&   outHitInfo,
    int              desiredOrder,
    CCoefVec<float>& result,
    float&         outHarmMeanDistance,
    float&         outMinHitDist,
    float&         outStdDev,
    const void*      data) const;

  bool SampleHemisphereManyRPC(
    const CVector3D& pt,
    const CVector3D& UU,
    const CVector3D& VV,
    const CVector3D& NN,
    const CVector3D& uN,
    const CObject3D* object,
    const void*      termination,
    const float     &footprintArea,
    CHitPointInfo&   outHitInfo,
    int              desiredOrder,
    CCoefVec<float>& result,
    float&         outHarmMeanDistance,
    float&         outMinHitDist,
    float&         outStdDev,
    const void*      data) const;

  /// Project sampled irradiances to the basis.
  bool Project(CCoefVec<float>& result,int desiredOrder) const;

  /// Compute gradients in local coordinate frame. Return false on error.
  virtual bool ComputeGradients(
    const CVector3D &U,
    const CVector3D &V,
    const CVector3D &N,
    const CCoefVec<float>& result,
    CCoefVec<float> &dx,
    CCoefVec<float> &dy,
    int desiredOrder) const = 0;

  /// Rotate vectors to global frame
  bool RotateToGlobalFrame(
    const CVector3D &U,
    const CVector3D &V,
    const CVector3D &N,
    CCoefVec<float> &result,
    CCoefVec<float> *dx,
    CCoefVec<float> *dy,
    int desiredOrder) const;

	/// Returns the number of rays to be shot.
	int GetNumCells() const { return _numCells; }

	/// Return the (hemi)spheriacal harmonics order.
	int GetMaxOrder()   const  { return _maxOrder; }

  float GetSampleTime()  const {return _sampleTimer.UserTime(); }
  float GetProjectTime() const {return _projectTimer.UserTime(); }
  float GetGradTime()    const {return _gradTimer.UserTime(); }

  void ResetTimers() {
    _sampleTimer.Reset();
    _projectTimer.Reset();
    _gradTimer.Reset();
  }
};

/// Gradients computed by a finite hemisphere displacement.
class CRadianceSamplerDisplace : public CRadianceSampler
{
public: // methods

  CRadianceSamplerDisplace(
    int                depth, 
    int                numCells, 
    int                raysPerCell,
    int                order,
    float             infRayLength, 
    float             minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const CSphBasis<float,float>   *sphBasis,
    bool               localFrame,
    bool               multichannelGrad,
    const void*        data=0):
  CRadianceSampler(depth,numCells,raysPerCell,order,infRayLength, 
    minAllowedHitDist,incRad,sphBasis,localFrame,multichannelGrad,data) {}
  
  virtual ~CRadianceSamplerDisplace() {}

  virtual bool ComputeGradients(
    const CVector3D &U,
    const CVector3D &V,
    const CVector3D &N,
    const CCoefVec<float>& result,
    CCoefVec<float> &dx,
    CCoefVec<float> &dy,
    int desiredOrder) const;
};

/// Gradients computed by computing derivative of SH bases and incoming
/// radiance field.
class CRadianceSamplerDeriv : public CRadianceSampler
{
public: // methods

  CRadianceSamplerDeriv(
    int                depth, 
    int                numCells, 
    int                raysPerCell,
    int                maxOrder,
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const CSphBasis<float,float>   *sphBasis,
    bool               localFrame,
    bool               multichannelGrad,
    const void*        data) :
  CRadianceSampler(depth,numCells,raysPerCell,maxOrder,infRayLength, 
    minAllowedHitDist,incRad,sphBasis,localFrame,multichannelGrad,data) {}

  virtual ~CRadianceSamplerDeriv() {}

  virtual bool ComputeGradients(
    const CVector3D &U,
    const CVector3D &V,
    const CVector3D &N,
    const CCoefVec<float>& result,
    CCoefVec<float> &dx,
    CCoefVec<float> &dy,
    int desiredOrder) const;
};


/// Like CRadianceSamplerDerivDir, but gradient computation besed on Ward
class CRadianceSamplerWard : public CRadianceSampler
{
protected: // data

  float *    v_factor;
  float *    u_factor;
  
  CVector3D* u_vect_canonical;
  CVector3D* v_vect_canonical;

  // 2*PI / N
  float M_PI_mul_2_div_N;

protected: // methods

  /// completely screwed up experimentation, better not call
  void _fillMissingDataAndFilter(
    CValues *_Li, 
    float   *_r,
    CValues *LLi, 
    float   *rr) const;

public: // methods

  CRadianceSamplerWard(
    int                depth, 
    int                numCells, 
    int                raysPerCell,
    int                order,
    float              infRayLength, 
    float              minAllowedHitDist,
    CIncidentRadiance* incRad, 
    const CSphBasis<float,float>   *sphBasis,
    bool               localFrame,
    bool               multichannelGrad,
    const void*        data);

	virtual ~CRadianceSamplerWard();

  virtual bool ComputeGradients(
    const CVector3D &U,
    const CVector3D &V,
    const CVector3D &N,
    const CCoefVec<float>& result,
    CCoefVec<float> &dx,
    CCoefVec<float> &dy,
    int desiredOrder) const;
};

__END_GOLEM_HEADER

#endif // __RADSAMPL_H__
