// ===================================================================
// $Id: rcache.cpp,v 1.8 2005/11/29 16:04:05 xkrivanj Exp $
//
// rcache.cpp
//     Implementation of the CRCache class.
//     Encapsulation of the radiance caching and (hemi)spherical harmonic hemisphere sampling.
//
// Initial coding by Jaroslav Krivanek, December 2003.

// GOLEM headers
#include "config.h"
#include "basmacr.h"
#include "rcache.h"
#include "radsampl.h"
#include "surface.h"

#include "pharr.h"  // only because of the Clamp() function

// SPHAR headers
#include "shrot.h"
#include "shbrdf.h"

// #define PLAY_WITH_MATLAB

#ifdef PLAY_WITH_MATLAB
#include "C:\\matlabR12\\extern\\include\\matlab.h"
#include "matrix3.h"
#endif // PLAY_WITH_MATLAB

// D3DX headers (for SH-rotation)
//#define D3DXROTATION
#ifdef D3DXROTATION
#include "d3dx9math.h"
#endif

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  DumpErrAnalysisData()
// --------------------------------------------------------------------------
bool
DumpErrAnalysisData(vector<CRotErrAnalysisData> &data,const char *fname)
{
  assert(fname);
  FILE* f=fopen(fname,"wt");
  if(!f) {
    WARNING << "Could not open '" << fname << "' for writing." << std::endl;
    return true;
  }
  vector<CRotErrAnalysisData>::iterator i;
  for(i=data.begin();i!=data.end();i++) {
    fprintf(f,"%.4f  ",i->degrees);
    for(int k=0; k<4; k++)
      fprintf(f,"%.4f %.4f ",i->lighterr[k],i->brdferr[k]);
    for(int k=0; k<4; k++)
      fprintf(f,"%.4f %.4f ",i->doterrlight[k],i->doterrbrdf[k]);
    fprintf(f,"\n");
  }
  fclose(f);

  return false; // no error
}


// --------------------------------------------------------------------------
//  MeasureRotError()
// --------------------------------------------------------------------------
static void
MeasureRotError(const CSHRotMatrix &Rapprox,
                const CSHRotMatrix &Rok,
                const float* lightCoeffs,
                const float* brdfCoeffs,
                int   order,
                CRotErrAnalysisData &res, int erridx)
{
  assert(Rapprox.GetOrder()>=order);
  assert(Rok.GetOrder()>=order);

  float lightRotApprox[SQR_MAX_FIXED_SH_ORDER];
  float lightRotOK    [SQR_MAX_FIXED_SH_ORDER];
  float brdfRotApprox [SQR_MAX_FIXED_SH_ORDER];
  float brdfRotOK     [SQR_MAX_FIXED_SH_ORDER];

  // 1) L2 error of rotated light
  // 2) L2 error of rotated brdf
  // 3) dot prod err (BRDF rot is approximated)
  // 4) dot prod err (lighting rot is approximated)

  // 1) L2 error of rotated light
  Rapprox.TransformVector(lightRotApprox, lightCoeffs, order);
  Rok.TransformVector(lightRotOK, lightCoeffs, order);
  res.lighterr[erridx] = shvf::sqrdist(order*order, lightRotApprox, lightRotOK);

  // 2) L2 error of rotated light
  Rapprox.TransformVector(brdfRotApprox, brdfCoeffs, order);
  Rok.TransformVector(brdfRotOK, brdfCoeffs, order);
  res.brdferr[erridx] = shvf::sqrdist(order*order, brdfRotApprox, brdfRotOK);

  float dotCorrect = shvf::dot(order*order, lightCoeffs, brdfCoeffs);

  // 3) dot prod err (BRDF rot is approximated)
  float dotbrdf = shvf::dot(order*order, lightRotOK, brdfRotApprox);
  // res.doterrbrdf[erridx] = fabs(dotCorrect - dotbrdf);
  res.doterrbrdf[erridx] = fabs( (dotbrdf-dotCorrect)/dotCorrect );

  // 4) dot prod err (BRDF rot is approximated)
  float dotlight = shvf::dot(order*order, lightRotApprox, brdfRotOK);
  // res.doterrlight[erridx] = fabs(dotCorrect - dotlight);
  res.doterrlight[erridx] = fabs( (dotlight-dotCorrect)/dotCorrect );
}

// --------------------------------------------------------------------------
//  RotErrAnalysis()
// --------------------------------------------------------------------------
void 
RotErrAnalysis(const CCoefVec<float>& coeffs,
               const CCoefVec<float>& shbrdf,
               vector<CRotErrAnalysisData> &averages,
               vector<CRotErrAnalysisData> &maxima)
{
  assert(coeffs.GetNumChannels()==3);
  int maxOrder = min(coeffs.GetOrder(), shbrdf.GetOrder());
  int vecLen = min(coeffs.GetVecLen(), shbrdf.GetVecLen());

  // compute luminance for the incoming light and BRDF coefficient vectors
  float lightCoeffs[SQR_MAX_FIXED_SH_ORDER];
  float brdfCoeffs[SQR_MAX_FIXED_SH_ORDER];
  shvf::setz(vecLen,lightCoeffs);
  shvf::setz(vecLen,brdfCoeffs);
  float lumweights[] = {0.299f,0.587f,0.114f};
  for(int ch=0; ch<coeffs.GetNumChannels(); ch++) {
    shvf::madvsip(vecLen, lightCoeffs, coeffs.GetChannel(ch), lumweights[ch]);
    shvf::madvsip(vecLen, brdfCoeffs, shbrdf.GetChannel(ch), lumweights[ch]);
  }

  // rotation of coeffs, rotation of the brdf
  // error of dot product, if a rotation of D degrees is replaced by an 
  // approximation

  // identity SH rot matrix
  CSHRotMatrix I(maxOrder);
  // first derivative
  CSHRotMatrix d10Ry(maxOrder);
  // diagonal of the second derivative
  CSHRotMatrix d15Ry(maxOrder);
  // full second derivative
  CSHRotMatrix d20Ry(maxOrder);
  // build the zeroth (identity), first and second derivative in zero
  CSHRotMatrix *dRarray[] = {&I,&d10Ry,&d20Ry};
  rotGenerator.BuildMatrix_dY(dRarray,2,maxOrder,cos(0.0),sin(0.0));
  // copy the full second derivative to one and half-th derivative
  d15Ry.assign(d20Ry);
  // multiply element-by-element by the identity, to keep the diagonal only
  d15Ry.mul(I);
  // remove the diagonal elements from d20Ry
  d20Ry.sub(d15Ry);
  // tmp matrix for computation
  CSHRotMatrix Rtmp1(maxOrder), Rtmp2(maxOrder);
  // correct y rotation matrix
  CSHRotMatrix Ryok(maxOrder);

  vector<CRotErrAnalysisData> errdatavec;

  for(int order=maxOrder; order<=maxOrder; order++) {
    for(int degrees=0; degrees<=90; degrees += 2) {
      double angle = (double)degrees*M_PI/180;
      // mlfPrintf("W[l=%d,a=%.2f]\n",l,angle/M_PI*180); 

      errdatavec.push_back(CRotErrAnalysisData());
      CRotErrAnalysisData &errdata = errdatavec.back();

      errdata.degrees = degrees;

      // build the correct rotation matrix
      rotGenerator.BuildMatrixY(Ryok,order,cos(angle),sin(angle));

      // I vs. R
      MeasureRotError(I, Ryok, lightCoeffs, brdfCoeffs, order, errdata,
        CRotErrAnalysisData::EE_Ignore);

      // dR vs. R
      Rtmp1  = d10Ry;
      Rtmp1 *= angle;
      Rtmp1 += I;
      MeasureRotError(Rtmp1, Ryok, lightCoeffs, brdfCoeffs, order, errdata,
        CRotErrAnalysisData::EE_d10R);

      // d15R vs. R
      Rtmp2  = d15Ry;
      Rtmp2 *= 0.5*angle*angle;
      Rtmp1 += Rtmp2;
      MeasureRotError(Rtmp1, Ryok, lightCoeffs, brdfCoeffs, order, errdata,
        CRotErrAnalysisData::EE_d15R);

      // d20R vs. R
      Rtmp2  = d20Ry;
      Rtmp2 *= 0.5*angle*angle;
      Rtmp1 += Rtmp2;
      MeasureRotError(Rtmp1, Ryok, lightCoeffs, brdfCoeffs, order, errdata,
        CRotErrAnalysisData::EE_d20R);
    }
  }

  if(averages.size()==0)
    averages.resize(errdatavec.size());
  if(maxima.size()==0)
    maxima.resize(errdatavec.size());

  assert(averages.size()==errdatavec.size());
  assert(maxima.size()==errdatavec.size());

  for(size_t i=0; i<errdatavec.size(); i++) {
    averages[i] += errdatavec[i];
    maxima[i].Maximize(errdatavec[i]);
  }
}

// --------------------------------------------------------------------------
//  CRCache::~CRCache() 
// --------------------------------------------------------------------------
CRCache::~CRCache () 
{ 
  // dump the error analysis averages
  if(_errAverages.size()>0) {
    for(unsigned i=0; i<_errAverages.size(); i++) {
      _errAverages[i].mul(1.0/_errAverages[i].n);
    }
 
    DumpErrAnalysisData(_errAverages,"erranalysis-avg.dat");
  }

  // dump the error analysis maxima
  if(_errMaxima.size()>0) {
    DumpErrAnalysisData(_errMaxima,"erranalysis-max.dat");
  }

  delete _tmpVec;
}

/// for debugging
static float phong(const CVector3D &inDir, const CVector3D& N, const CVector3D &outDir, float n)
{
	if( DotProd(N, inDir) < 0)
		FATAL_EXIT;
	CVector3D R = N * (2.0 * DotProd(N, inDir)) - inDir;
	float cosR_L = DotProd(R, outDir);
	if (cosR_L <= 0.0) return 0;
	return  powf(cosR_L, n) * (n + 2.0f) * (0.5f * M_1_PI) * DotProd(N,outDir);
}



// compute the covariance matrix for a given coefficient vector
void CovarianceMatrix(float C[3][3],float **v, int n)
{
  int channels = 3;
  int i,j;

  // zero the covariance matrix
  for(i=0; i<channels; i++)
    for(j=0; j<channels; j++)
      C[i][j]=0;

  // compute the covariance matrix for the gradient
  for(int k=0; k<n; k++)
    for(i=0; i<channels; i++)
      for(j=0; j<channels; j++)
        C[i][j] += v[i][k] * v[j][k];
}


#ifdef PLAY_WITH_MATLAB
/*
  datain format:
     r[0] g[0] b[0] r[1] g[1] b[1] ... r[n] g[n] b[n]

  dataout format (new base is names (x,y,z))
     x[0] y[0] z[0] x[1] y[1] z[1] ... x[n] y[n] z[n]

  weightout format:
     w_r[x] w_r[y] w_r[z]  w_g[x] w_g[y] w_g[z]  w_b[x] w_b[y] w_b[z]
*/
void PCA3Vector(float *dataout, float *weightout, float *datain, int n)
{
  const int channels = 3;
  int i,j;

  mlfEnterNewContext(0, 0);
  mxArray *A = NULL, *U = NULL, *S = NULL, *V = NULL, *W = NULL;

  // create the Matlab matrix for input
  double *data = new double[n*channels];
  for(i=0; i<n*channels; i++) data[i]=datain[i];
  mlfAssign(&A, mlfDoubleMatrix(channels,n,data,NULL));
  delete [] data;
  // mlfPrintf("A:\n"); mlfPrintMatrix(A);

  mlfAssign(&U, mlfSvd(&S, &V, A, NULL));
  // mlfPrintf("U = \n"); mlfPrintMatrix(U); 
  // mlfPrintf("S = \n"); mlfPrintMatrix(S);
  // mlfPrintf("V = \n"); mlfPrintMatrix(V);

  mlfAssign(&W, mlfMtimes(U,S));
  // mlfPrintf("W (pca weights):\n"); mlfPrintMatrix(W);

  // mlfAssign(&AA, mlfMtimes(W,mlfTranspose(V)));
  // mlfPrintf("AA:\n"); mlfPrintMatrix(AA);

  // copy the weights
  double *w = mxGetPr(W);
  for(i=0; i<channels; i++)
    for(j=0; j<channels; j++)
      weightout[i*channels+j] = w[j*channels+i];

  // copy the representative vectors
  double *v = mxGetPr(V);
  for(i=0; i<n; i++)
    for(j=0; j<channels; j++)
      dataout[i*channels+j] = v[j*n+i];

  mlfRestorePreviousContext(0, 0);
}
#endif // PLAY_WITH_MATLAB


// --------------------------------------------------------------------------
//  CRCache::PresampleCache()
// --------------------------------------------------------------------------
CRCache::EQueryResult 
CRCache::PresampleCache(const CHitPointInfo& inHitInfo,
                        const float        &footprintArea,
                        CHitPointInfo&       outHitInfo,
                        const CSurface*      surface,
                        const float        &a)
{
  assert(surface);
  const CSphBRDF<float,float> *shRep = surface->GetSHRep();
  if(!shRep) { return EE_Failed; }

  const CVector3D& P = inHitInfo.GetPoint();
  CVector3D  U, V, N;   // perturbed   = lighting
  inHitInfo.GetNormal(N);
  N.RightHandedBase(U,V);

  CVector3D inDir = ToLocalFrame(-inHitInfo.GetRay()->GetDir(), U, V, N);
	
  const CCoefVec<float>* shbrdf = shRep->GetCoeffs(inDir);
  if(!shbrdf) { return EE_Failed; }
  
  if(shbrdf->GetOrder() <= 0) { return EE_Interpolated; }

  // OK. Radiance caching would be used for this point and dir.

  ////////////////////////
  // look-up the cache
  if( CCachingMode::LookIntoCache(GetCachingMode()) ) {
    if( InterpolationPossible(P,N,a,shbrdf->GetOrder()))
      return EE_Interpolated;
  }

  /////////////////////////////////////
  // nothing found, sample hemisphere

  if( !CCachingMode::SampleHemisphere(GetCachingMode()) )
    return EE_Failed;

  EQueryResult r = AddRadiance(P,U,V,N,
      inHitInfo.GetGeometricNormal(),
      inHitInfo.GetObject(),
      inHitInfo.GetRay()->GetTermination(),
      footprintArea,
      outHitInfo,
      a,
      Max(shRep->GetStats().order99Quantile, shbrdf->GetOrder()),
      shbrdf,0,0);
  
  return r;
}


// --------------------------------------------------------------------------
//  CRCache::GetRadiance()
// --------------------------------------------------------------------------
CRCache::EQueryResult
CRCache::GetRadiance(const CHitPointInfo &inHitInfo,
                     const float        &footprintArea,
                     CHitPointInfo       &outHitInfo,
                     const CSurface      *surface,
                     const float        &a,
                     CValues             &result)
{
  assert(surface);
  const CSphBRDF<float,float> *shRep = surface->GetSHRep();
  if(!shRep) return EE_Failed;

  const CVector3D& P = inHitInfo.GetPoint();
  // (U,V,N) is a right-handed coord system;  U x V = normal
  CVector3D  U, V, N;   // perturbed   = lighting
  inHitInfo.GetNormal(N);

  /////////////////////////////////////////////////
  // Try to use the last computed color.
  //  Caching last used color is useful for flat surfaces when 
  //  antialiasing or computing very high resolution images.
  //////
  if(_cacheLastResult) {
    if(_last.surface == surface && 
      EpsilonEqualV3(_last.normal, N, 0.01) ) {  
        if(SqrMagnitude(_last.point - P) <= _last.minR)
        {
          result = _last.result;
          return EE_Interpolated;
        }
      }
  }
  /// end caching

  N.RightHandedBase(U,V);

  // transform incoming dir to (perturbed) local coords 
  CVector3D inDir = ToLocalFrame(-inHitInfo.GetRay()->GetDir(), U, V, N);

#if 0
  v3d = inDir;
#endif

  const CCoefVec<float>* shbrdf = shRep->GetCoeffs(inDir);
  // const CCoefVec<float>* shbrdf = shRep->GetCoeffs(CVector3D(0,0,1));
  if(!shbrdf) return EE_Failed;

  // reset the result
  result.setz();

#if 0
  // visualizes the pathological case, when the ray comes from 'behind'
  // with respect to the interpolated normal
  if(inDir.z<0)  {
    result.Set(1,100);
    return true; // ok
  }
#endif

  // return black if brdf is zero
  if(shbrdf->GetOrder() <= 0) { return EE_Interpolated; }

  ////////////////////////////////////////////////////////////////////////
  // at this point we have the non-zero coefficient vector for the brdf
  ////////////////////////////////////////////////////////////////////////

  // return value
  EQueryResult retval = EE_Failed;

  if( CCachingMode::LookIntoCache(GetCachingMode()) ) {
    // look-up the cache
    _tmpCoeffs._setOrder(shbrdf->GetOrder());
    retval = InterpolateRadianceDispatch(P,U,V,N,a,_last.minR,_tmpCoeffs);
  }

  // bail-out if found
  if (retval==EE_Interpolated) {
    _computeOutgoingRadiance(U,V,N,_tmpCoeffs,*shbrdf,result);
    if(_cacheLastResult) {
      // cache the last value
      _last.surface = surface;
      _last.normal  = N;
      _last.point   = P;
      _last.result  = result;
      _last.minR *= (a * 0.001);
    }
  }
  else {
  // ------------------------------------------------------------------------
  // no irradiance value found in the cache or no caching - sample hemisphere
  // ------------------------------------------------------------------------
    retval = AddRadiance(P,U,V,N,
      inHitInfo.GetGeometricNormal(),
      inHitInfo.GetObject(),
      inHitInfo.GetRay()->GetTermination(),
      footprintArea,
      outHitInfo,
      a,
      Max(shRep->GetStats().order99Quantile, shbrdf->GetOrder()),
      shbrdf,&result,0);
  }

  if(_visualizeRecPos) {
    float w = VisualizeNearestRecords(P,N,sqrtf(footprintArea)*_visualRecSize);
    result += _visualRecColor*w;
  }

  return retval;
}


// --------------------------------------------------------------------------
//  CRCache::AddRadiance()
// --------------------------------------------------------------------------
CRCache::EQueryResult
CRCache::AddRadiance(const CVector3D &P,
                     const CVector3D &U,
                     const CVector3D &V,
                     const CVector3D &N,
                     const CVector3D &uN,
                     const CObject3D *object,
                     const void      *termination,
                     const float    &footprintArea,
                     CHitPointInfo   &outHitInfo,
                     const float    &aparam,
                     int desiredOrder,
                     const CCoefVec<float>* shbrdf,
                     CValues            *resColor,
                     CRadianceRecord **newRecRet)
{
  if(newRecRet) *newRecRet = 0;

  if( !CCachingMode::SampleHemisphere(GetCachingMode()) )
    return EE_Failed;

  float a = _determineNewRecA(P,N,aparam);

  // new samples always computed using uperturbed (geometric) local frame

  // desactivate last value caching for next time
  static CDummySurface dummySurface;
  _last.surface = &dummySurface;

  float pixelSize = sqrt(footprintArea);
  float stdDev=0;
  float R = 1; // this is a temporary value, it will change
  float Rminus,Rplus;
  _setRminusRplus(Rminus,Rplus,pixelSize);

  // create new record
  CRadianceRecord *radRecord = new CRadianceRecord(P,U,V,N,_channels,_order,
    _useGradients, a, R, R, Rminus, Rplus, _rotationMultiplier(), stdDev, 0);

  // move the P half a pixel away from the polygon edge
  CVector3D modifiedP = P;
  bool  posModified = _displaceCenter(modifiedP,object,pixelSize);

  float harmMeanDist, minHitDist;
  bool r = _radSampler->ComputeRadiance(
    modifiedP, U, V, N, uN, object, termination, footprintArea, 
    outHitInfo, desiredOrder, radRecord->GetCoeffs(),
    _useGradients ? &radRecord->GetDX() : 0,
    _useGradients ? &radRecord->GetDY() : 0,
    harmMeanDist,minHitDist,stdDev,0);
  radRecord->SetStdDev(stdDev);

  /// hemisphere sampling failed
  if(!r) {
    delete radRecord;
    return EE_Failed;
  }

  /// compute resulting color (if required)
  if(resColor)
    _computeOutgoingRadiance(U,V,N,radRecord->GetCoeffs(),*shbrdf,*resColor);

  /// only for measuring
  // RotErrAnalysis(radRecord->GetCoeffs(),*shbrdf,_errAverages,_errMaxima);

  if( !CCachingMode::InsertIntoCache(GetCachingMode()) ) {
    delete radRecord;
    return EE_NewRecord;
  }

  // Radius assignment (minHitDist vs. HMD)
  float unclampedR = _setupUnclampedR(minHitDist,harmMeanDist);

  // PRINTVAR(_clampRadiusByGradients);

  // reduce unclampedR if gradient large
  if(_useGradients && _clampRadiusByGradients) {
    // maximum ratio between upper bound and the actual gradient
    const float gradRatio = _useMinHitDist ? aRatio : 1.0;
    // const float gradRatio =  1.0;
    // maximum gradient over all channels
    float sqrGradMag = 0;
    // int order = radRecord->GetCoeffs().GetOrder();
    // int gradClampVecLen = PAD_FOUR(order*order);
    int gradClampVecLen = radRecord->GetCoeffs().GetVecLen();
    for(int ch=0; ch<_channels; ch++) {
      float imagL = 1.0 / shvf::L24(gradClampVecLen,radRecord->GetCoeffs().GetChannel(ch));
      float magDX = shvf::L24(gradClampVecLen,radRecord->GetDX().GetChannel(ch));
      float magDY = shvf::L24(gradClampVecLen,radRecord->GetDY().GetChannel(ch));
      sqrGradMag = max<float>(sqrGradMag,0.5*(magDX+magDY)*imagL);
    }
    if ( sqrGradMag * sqr(unclampedR) > sqr(gradRatio) ) {
      // DEBUG << unclampedR << " ==> " << gradRatio/sqrt(sqrGradMag) << std::endl;
      unclampedR = gradRatio/sqrt(sqrGradMag);
    }
  }

  // clamp by minDist/maxDist
  R = Clamp(unclampedR, Rminus, Rplus);

  if(_useGradients) {
    // Rectify the gradients if radius enlarged.
    // Greg Ward does this gradient capping by the upper bound derived
    // from the split sphere model (i.e. by E/R), but this test works
    // better.
    if(unclampedR < R) {
      float m = unclampedR / R;
      radRecord->GetDX().mul(m);
      radRecord->GetDY().mul(m);
    }

    // If the center was displaced, modify irradiance by the translational 
    // gradient.
    if(posModified)
      ExtrapolateRadiance(P-modifiedP,radRecord,radRecord->GetCoeffs());
  }

  radRecord->SetUnclampedR(unclampedR);
  radRecord->SetR(R);

  // --------------------
  // radius clamping by the presence of overlapping records
  // (this is a heuristic to rectify some problems near corners)
  // --------------------
  if(_clampNeighbors || _clampByNeighbors)
    _neighborClamping(radRecord);

  // in case R was changed in _neigborClamping()
  radRecord->UnsetChanged();

  InsertRecord(radRecord);

  if(newRecRet)
    *newRecRet = radRecord;

  // visualize record positions
  if( _visualizeNewRecPos && resColor)
    *resColor = _visualNewRecColor; 

  // randomize radiance (for debugging)
  bool _randomizeRadiances=false;
  if(_randomizeRadiances) {
    int ic = GetNumRecords()%radRecord->GetCoeffs().GetNumChannels();
    int ig = GetNumRecords()%radRecord->GetDX().GetNumChannels();
    shvf::setz(radRecord->GetCoeffs().GetVecLen(), radRecord->GetCoeffs().GetChannel(ic));
    shvf::setz(radRecord->GetDX().GetVecLen(), radRecord->GetDX().GetChannel(ig));
    shvf::setz(radRecord->GetDY().GetVecLen(), radRecord->GetDY().GetChannel(ig));
  }

  return EE_NewRecord;
}

// --------------------------------------------------------------------------
//  CRCache::InterpolationPossible()
// --------------------------------------------------------------------------
bool
CRCache::InterpolationPossible(const CVector3D  &P,
                               const CVector3D  &N,
                               const float      &a,
                               int   desiredOrder) const
{
  _carray.clear();
  if(!GetContributions<false,false>(P,N,_carray,a))
    return false; // nothing found

  for(unsigned i=0; i<_carray.size(); i++) {
    const CRadianceRecord *rad = _carray[i].Rec();
    if(rad->GetCoeffs().GetOrder() >= desiredOrder)
      return true;
  }
  return false;
}

// --------------------------------------------------------------------------
//  rotZMad()
// --------------------------------------------------------------------------
/// rotate around Z, multiply and add 
/** 
   cosn and sinn are supposed to already been premultiplied by weight)
 */
void rotZMad(float       *dest,
             const float *src,
             int          order,
             const float *cosn, 
             const float *sinn, 
             float        weight)
{
  for(int l=0, index=0; l<order; l++, index+=2*l) {
    dest[index] += weight * src[index];
    for(int m=1; m<=l; ++m) {
      dest[index-m] += (cosn[m] * src[index-m] - sinn[m] * src[index+m]);
      dest[index+m] += (sinn[m] * src[index-m] + cosn[m] * src[index+m]);
    }
  }
}

#ifdef  D3DXROTATION
static inline void D3DXToGlobalFrame(D3DXMATRIX &mat,
				     const CVector3D &U,
				     const CVector3D &V, 
				     const CVector3D &N)
{
  mat(0,0) = U.x;	mat(0,1) = U.y;	mat(0,2) = U.z;	mat(0,3) = 0;
  mat(1,0) = V.x;	mat(1,1) = V.y;	mat(1,2) = V.z;	mat(1,3) = 0;
  mat(2,0) = N.x;	mat(2,1) = N.y;	mat(2,2) = N.z;	mat(2,3) = 0;
  mat(3,0) = 0;	mat(3,1) = 0;	mat(3,2) = 0;	mat(3,3) = 1;
}

#endif //  D3DXROTATION

// --------------------------------------------------------------------------
//  CRCache::_rotateAndAdd()
// --------------------------------------------------------------------------
template<bool MOV>
void
CRCache::RotateAndAdd(const CRadianceRecord *rad,			  
                       const CCoefVec<float> *vec,
                       const CVector3D  &U,
                       const CVector3D  &V,
                       const CVector3D  &N,
                       float weight,
                       CCoefVec<float>& ret) const
{
  assert(ret.GetOrder()<=MAX_FIXED_SH_ORDER); 
  int order    = ret.GetOrder();
  int sqrOrder = order * order;
  int vecLen   = ret.GetVecLen();
  int i;

#if 0
  /// ignore all rotation (PROFILE, DEBUG)
  for(int ch=0; ch<ret.GetNumChannels(); ch++) {
    if(MOV)
      shvf::movvv(vecLen,ret.GetChannel(ch),vec->GetChannel(ch));
    else // mad
      shvf::madvsip4(vecLen, ret.GetChannel(ch), vec->GetChannel(ch), weight);
  }
  return;
#endif

  ALIGN16 float tmpvec1  [MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
  ALIGN16 float tmpvec2  [MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
  ALIGN16 float cnz1[MAX_FIXED_SH_ORDER];
  ALIGN16 float cnz2[MAX_FIXED_SH_ORDER];
  ALIGN16 float cny[MAX_FIXED_SH_ORDER];
  ALIGN16 float snz1[MAX_FIXED_SH_ORDER];
  ALIGN16 float snz2[MAX_FIXED_SH_ORDER];
  ALIGN16 float sny[MAX_FIXED_SH_ORDER];

  // Euler angles decomposition we use here is ZYZ, angles z1, y, z2
  CVector3D Np = ToLocalFrame(rad->GetN(), U,V,N);

  // Np.z = cos(y)
  if(Np.z >= _cosIgnoreYRotUpTo) 
  {
    // ------------------ IGNORE ROTATION AROUND Y ----------------
    _numIgnoredYRotations++;
    
    // sine of the Z-rotation angle
    float s = DotProd(rad->GetV(), U);
    float c = DotProd(rad->GetU(), U);

    if(fabsf(s)<=_sinIgnoreYRotUpTo && c>=_cosIgnoreYRotUpTo) 
    // if(fabsf(s)<=0.0001 && c>=0.9999) 
    {
      /// ignore even the Z rotation (no tangent alignment)
      for(int ch=0; ch<ret.GetNumChannels(); ch++) {
        if(MOV)
          shvf::movvv(ret.GetVecLen(),ret.GetChannel(ch),vec->GetChannel(ch));
        else
          shvf::madvsip4(ret.GetVecLen(),ret.GetChannel(ch),
          vec->GetChannel(ch), weight);
      }
    }
    else 
    {
      /// tangents should be aligned
      
      // evaluate sin(n*phi) and cos(n*phi)
      recurrentSinCos<float>(order,c,s,cnz1,snz1);
      // multiply the cosines to save multiplication in the loop
      if(!MOV)
        for(i=0; i<order; i++) { cnz1[i] *= weight; snz1[i] *= weight; }
      // rotate around Z to align tangents, multiply by weight & add to result
      for(int ch=0; ch<ret.GetNumChannels(); ch++) {
        const float *in = vec->GetChannel(ch);
        float *out = ret.GetChannel(ch);
        if(MOV)
          shRotZ<float>(out,in,cnz1,snz1,order);
        else
          rotZMad(out,in,order,cnz1,snz1,weight);
      }
    }
  }
  else {
    // y-rotation cannot be ignored
    CVector3D Up = ToLocalFrame(rad->GetU(), U,V,N);
    CVector3D Vp = ToLocalFrame(rad->GetV(), U,V,N);

    float cz1,sz1,cy,sy,cz2,sz2;
    if( EulerAnglesZYZ<float>(Up,Vp,Np,cz1,sz1,cy,sy,cz2,sz2,1) ) {
      FATAL << "Error in decomposition" << std::endl;
      FATAL_EXIT;
    }

    // for HSH
    // cy = 2.0f * cy - 1.0f;
    // sy = sqrtf(1-cy*cy);

    // evaluate sin(n*z1), cos(n*z1) and sin(n*z2), cos(n*z2)
    recurrentSinCos<float>(order,cz1,sz1,cnz1,snz1);
    recurrentSinCos<float>(order,cz2,sz2,cnz2,snz2);
    // multiply the cosines to save multiplication in the loop
    if(!MOV)
      for(i=0; i<order; i++) { cnz2[i] *= weight; snz2[i] *= weight; }

    if( Np.z >= _cosDifferentiateYRotUpTo ) {

      // ------------ APPROXIMATE Y ROTATION BY DIFFERENTIATION --------------
      _numDifferentiatedYRotations++;

      // the angle of rotation can be approximated by its sine (small angle)
      float y = sy;
      // float y  = acos(cy);

      for(int ch=0; ch<ret.GetNumChannels(); ch++) {
        // Z1-rotation
        shRotZ<float>(tmpvec1,vec->GetChannel(ch),cnz1,snz1,order);
       // Y-rotation (by differential approximation)
#if 1  // first + partial second order
        rotGenerator.RotYdiff1(order,tmpvec2,tmpvec1,y);
#else  //full second order
        rotGenerator.RotYdiff2(order,tmpvec2,tmpvec1,y);
#endif
        // shRotZ<float>(tmpvec1,tmpvec2,cnz2,snz2,order);
        // exact rotation for band l=1
        // RotateBand1<float>(tmpvec1,vec->GetChannel(ch),Up,Vp,Np);
        // shvf::madvsip(sqr(order),ret.GetChannel(ch),tmpvec1,weight);

        // Z2-rotation, multiply and add
        if(MOV)
          shRotZ<float>(ret.GetChannel(ch),tmpvec2,cnz2,snz2,order);
        else
          rotZMad(ret.GetChannel(ch),tmpvec2,order,cnz2,snz2,weight);
      }
    }
    else {
      // --------------- FULL SH ROTATION (ajaj, this is slow) ---------------
      _numFullRotations++;

#if 0
      if(0) 
      {
        /*
        // D3DX
        D3DXMATRIX matrix;
        D3DXToGlobalFrame(matrix, Up, Vp, Np);
        // rotate & compute the dot product
        for(int ch = 0; ch < ret.GetNumChannels(); ch++) {
          if(MOV) 
            D3DXSHRotate(ret.GetChannel(ch),order,&matrix,vec->GetChannel(ch));
          else {
            D3DXSHRotate(tmpvec1, order, &matrix, vec->GetChannel(ch));
            shvf::madvsip4(vecLen,ret.GetChannel(ch),tmpvec1,weight);
          }
        }
        */
      }
      else {
        /// IVANIC
        rotGenerator.BuildMatrix(_shRotMatrix,order,
          Up.x, Up.y, Up.z,
          Vp.x, Vp.y, Vp.z,
          Np.x, Np.y, Np.z);

        for(i=sqrOrder; i<vecLen; i++) tmpvec1[i]=0;

        for(int ch=0; ch<ret.GetNumChannels(); ch++) {
          if(MOV) 
            _shRotMatrix.TransformVector(ret.GetChannel(ch),
            vec->GetChannel(ch),order);
          else {
            _shRotMatrix.TransformVector(tmpvec1,vec->GetChannel(ch),order);
            shvf::madvsip4(vecLen,ret.GetChannel(ch),tmpvec1,weight);
          }
        }
      }
#else
      /// ZXZXZ decomposition
      recurrentSinCos<float>(order,cy,sy,cny,sny);

      // rotate & madd
      for(int ch=0; ch<ret.GetNumChannels(); ch++) {
        // Z (alpha)
        shRotZ<float>(tmpvec1,vec->GetChannel(ch),cnz1,snz1,order);
        // X (-90)
        rotGenerator.TransformVectorXm90(tmpvec2,tmpvec1,order);
        // Z (beta)
        shRotZ<float>(tmpvec1,tmpvec2,cny,sny,order);
        // X (90)
        rotGenerator.TransformVectorXp90(tmpvec2,tmpvec1,order);
        // Z(gamma) & madd
        if(MOV) 
          shRotZ<float>(ret.GetChannel(ch),tmpvec2,cnz2,snz2,order);
        else
          rotZMad(ret.GetChannel(ch),tmpvec2,order,cnz2,snz2,weight);
      }
#endif
    }
  }
}

// --------------------------------------------------------------------------
//  CRCache::ExtrapolateRadiance()
// --------------------------------------------------------------------------
void
CRCache::ExtrapolateRadiance(const CVector3D &globDiff,
                             const CRadianceRecord *rad,
                             CCoefVec<float>  &ret) const
{
  // translation in the local frame of the considered sample
  // CVector3D diff = P - rad->GetPoint();
  float dx = DotProd(globDiff,rad->GetU());
  float dy = DotProd(globDiff,rad->GetV());

  // int maxVecLen  = rad->GetCoeffs().GetVecLen();
  int maxVecLen  = ret.GetVecLen();
  int gradVecLen = maxVecLen;

  if(_multichannelGrad) {
#if 1
    // separate gradients for each color channels
    for(int ch=0; ch<ret.GetNumChannels(); ch++) {
      shvf::movvv4  ( maxVecLen, ret.GetChannel(ch), 
        rad->GetCoeffs().GetChannel(ch) );
      shvf::madvsip4( gradVecLen, ret.GetChannel(ch), 
        rad->GetDX().GetChannel(ch), dx );
      shvf::madvsip4( gradVecLen, ret.GetChannel(ch), 
        rad->GetDY().GetChannel(ch), dy );
    }
#else
    for(int ch=0; ch<ret.GetNumChannels(); ch++) {
      shvf::movvv4  ( maxVecLen, ret.GetChannel(ch), 
        rad->GetCoeffs().GetChannel(ch) );
      shvf::madvsip4( gradVecLen, ret.GetChannel(ch), 
        rad->GetDX().GetChannel(0), dx * rad->xdweights[ch] );
      shvf::madvsip4( gradVecLen, ret.GetChannel(ch), 
        rad->GetDY().GetChannel(0), dy * rad->ydweights[ch] );
    }
#endif
  }
  else {
    // one common luminance gradient
    float ALIGN16 tmpvec1[SQR_MAX_FIXED_SH_ORDER];
    shvf::mulvs4  (gradVecLen,tmpvec1,rad->GetDX().GetChannel(0),dx);
    shvf::madvsip4(gradVecLen,tmpvec1,rad->GetDY().GetChannel(0),dy);
    for(int ch=0; ch<ret.GetNumChannels(); ch++) {
      shvf::movvv4 (maxVecLen, ret.GetChannel(ch), 
        rad->GetCoeffs().GetChannel(ch) );
      shvf::addvvip4( gradVecLen, ret.GetChannel(ch), tmpvec1 );
    }
  }
}

// --------------------------------------------------------------------------
//  CRCache::InterpolateRadiance()
// --------------------------------------------------------------------------
template<bool useStoredA,bool jitterWeight>
CRCache::EQueryResult
CRCache::InterpolateRadiance(const CVector3D  &P,
                             const CVector3D  &U,
                             const CVector3D  &V,
                             const CVector3D  &N,
                             const float     &a,
                             float           &minR,
                             CCoefVec<float>  &ret) const
{
  // set all entries to zero TODO: use setz only if more than one contribution, otherwise, it isn't required
  ret.setz();

  _carray.clear();
  if(!GetContributions<useStoredA,jitterWeight>(P,N,_carray,a))
    return EE_Failed; // nothing found

  int desiredOrder    = ret.GetOrder();
  int maxVecLen       = ret.GetVecLen();
  float weightSum  = 0.0f;
  const CCoefVec<float> *inVec;

  unsigned validContribs[100];
  unsigned numValidContribs = 0;

  for(unsigned i=0; i<_carray.size(); i++) {
    const CRadianceRecord *rad = _carray[i].Rec();
    if(rad->GetCoeffs().GetOrder() >= desiredOrder)
      validContribs[numValidContribs++]=i;
  }

  if(numValidContribs==0)
    return EE_Failed;
  else if(numValidContribs==1) {
    // one contribution - weighting can be avoided
    unsigned i = validContribs[0];
    const CRadianceRecord *rad = _carray[i].Rec();
    minR = rad->GetR();

    // apply translational gradients
    if(rad->UseGradients()) {
      ExtrapolateRadiance(P-rad->GetPoint(),rad,*_tmpVec);
      inVec = _tmpVec;
    }
    else {
      // use the sample's coefficient directly without gradients
      inVec = &rad->GetCoeffs(); 
    }
    if(_localFrame) {
      RotateAndAdd<true>(rad,inVec,U,V,N,0/*not used*/,ret);
    }
    else {
      // no rotation in global frame
      for(int ch=0; ch<ret.GetNumChannels(); ch++)
        shvf::movvv(maxVecLen,ret.GetChannel(ch),inVec->GetChannel(ch));
    }
    return EE_Interpolated;
  }

  minR = 1e8;
  // more than one contribution - weighting is necessary
  for(unsigned k=0; k<numValidContribs; k++) {
    unsigned i = validContribs[k];
    const CRadianceRecord *rad = _carray[i].Rec();
    float weight  = _carray[i].W();
    weightSum += weight;
    if(rad->GetR()<minR)minR=rad->GetR();

    // apply translational gradients
    if(rad->UseGradients()) {
      ExtrapolateRadiance(P-rad->GetPoint(),rad,*_tmpVec);
      inVec = _tmpVec;
    }
    else {
      // use the sample's coefficient directly without gradients
      inVec = &rad->GetCoeffs(); 
    }

    if(_localFrame) {
      RotateAndAdd<false>(rad,inVec,U,V,N,weight,ret);
    }
    else {
      // no rotation in global frame
      for(int ch=0; ch<ret.GetNumChannels(); ch++)
        shvf::madvsip4(maxVecLen,ret.GetChannel(ch),
        inVec->GetChannel(ch),weight);
    }
  }

  ret.mul(1.0f/weightSum);

  return EE_Interpolated;
}

__END_GOLEM_SOURCE

