// ===================================================================
// $Id: rcache.h,v 1.9 2005/11/29 16:04:04 xkrivanj Exp $
//
// rcache.h
//     Header file for the CRCache class.
//     Encapsulation of the radiance caching and (hemi)spherical harmonic
//      hemisphere sampling.
//
// Class: CRCache
//
// Initial coding by Jaroslav Krivanek, December 2003.

#ifndef __RCACHE_H__
#define __RCACHE_H__

// GOLEM headers
#include "configh.h"
#include "cachmode.h"
#include "lrcache.h"
#include "rcoctree.h"
#include "icache.h"

// LIBSPHAR header
#include "shrotmat.h"
#include "shrot.h"

// standard headers
#include <vector>

__BEGIN_GOLEM_HEADER

class CRadianceSampler;
class CHitPointInfo;
class CSurface;
template<class TV,class TA> class CSphBasis;

// for debugging - pass the viewing direction to _computeOutgoingRadiance
extern CVector3D v3d;

/// Data for error analysis of the approximate rotation
class CRotErrAnalysisData {
public:
  enum {
    EE_Ignore=0,
    EE_d10R,
    EE_d15R,
    EE_d20R
  };

  unsigned int n;
  double degrees;
  double lighterr[4];
  double brdferr[4];
  double doterrlight[4];
  double doterrbrdf[4];

  void Reset() {
    n=0;
    degrees = 0;
    for(int i=0; i<4; i++) {
      lighterr[i] = 0;
      brdferr[i]  = 0;
      doterrlight[i] = 0;
      doterrbrdf[i] = 0;
    }
  }

  const CRotErrAnalysisData& operator+=(const CRotErrAnalysisData& d) {
    degrees=d.degrees;
    n++;
    for(int i=0; i<4; i++) {
      lighterr[i] += d.lighterr[i];
      brdferr[i]  += d.brdferr[i];
      doterrlight[i] += d.doterrlight[i];
      doterrbrdf[i]  += d.doterrbrdf[i];
    }
    return *this;
  }

  const CRotErrAnalysisData& Maximize(const CRotErrAnalysisData& d) {
    degrees=d.degrees;
    n++;
    for(int i=0; i<4; i++) {
      if(lighterr[i]   < d.lighterr[i])   lighterr[i]    = d.lighterr[i];
      if(brdferr[i]    < d.brdferr[i])    brdferr[i]     = d.brdferr[i];
      if(doterrlight[i]< d.doterrlight[i])doterrlight[i] = d.doterrlight[i];
      if(doterrbrdf[i] < d.doterrbrdf[i]) doterrbrdf[i]  = d.doterrbrdf[i];
    }
    return *this;
  }

  void mul(double m) {
    for(int i=0; i<4; i++) {
      lighterr[i]   *= m;
      brdferr[i]    *= m;
      doterrlight[i]*= m;
      doterrbrdf[i] *= m;
    }
  }
};

/// Encapsulation of the radiance caching and hemisphere sampling.
class CRCache : public CCachingBase<CRadianceRecord>
{ 
public:    // methods
  
  /// Constructor. 
  /**
    First three parameters are passed to the CRadianceCache constructor. 
    'diffSampler' is basically an implementation of the 'primary method', i.e.
    sampling the diffuse hemisphere.
  
    @param sphBasis (in) Const pointer to an interface for evaluation of (hemi)spherical 
    functions.
  */
  CRCache (
    // Octree parameters
    const CAABox &sceneBox,
    float octreeScale,
    // common radiance/irradiance caching parameters
    float minDist,
    float maxDist,
    float centerDisplaceDist,
    float rotWeightMultiplier,
    float nearestRecordAMul,
    bool  useGradients,
    bool  cacheLastResult,
    bool  clampInScreenSpace,
    bool  useMinHitDist,
    bool  useGeometricNormal,
    bool  clampRadiusByGradients,
    bool  clampByNeighbors,
    bool  clampNeighbors,
    int   depth,
    // specific radiance caching parameters
    float ignoreYRotUpTo,
    float differentiateYRotUpTo,
    int   order,
    int   channels,
    bool  localFrame,
    bool  multichannelGrad,
    const CRadianceSampler         * radSampler,
    const CSphBasis<float,float>  * sphBasis)
    : 
    CCachingBase<CRadianceRecord>(
    sceneBox,
    octreeScale,
    minDist,
    maxDist,
    centerDisplaceDist,
    rotWeightMultiplier,
    nearestRecordAMul,
    useGradients,
    cacheLastResult,
    clampInScreenSpace,
    useMinHitDist,
    useGeometricNormal,
    clampRadiusByGradients,
    clampByNeighbors,
    clampNeighbors,
    depth),
    // init radiance caching specific
    _shRotMatrix(order),
    _zxzxzRotObject(order),
    _tmpCoeffs(channels, order),
    _radSampler(radSampler),
    _sphBasis(sphBasis),
    _channels(channels),
    _order(order),
    _sqrOrder(order*order),
    _localFrame(localFrame),
    _multichannelGrad(multichannelGrad)
  {
    assert( radSampler != 0);
    assert( sphBasis   != 0);
    assert( channels   >  0);
    assert( order      >  0);

    SetCachingMode(CCachingMode::EE_Normal);

    _tmpVec = new CCoefVec<float>(channels,_order);

    // _Rxp90.TransformVectorGenCode(cout,"out","in",order);
    rotGenerator.LiftOrder(_order);

    /// these angles are between 0 and PI, hence cos() is 1-to-1 and we're happy
    float minAngle = 0.1*M_PI/180.0;
    _ignoreYRotUpTo = Max(ignoreYRotUpTo, minAngle);
    _cosIgnoreYRotUpTo = cos(_ignoreYRotUpTo);
    _sinIgnoreYRotUpTo = sin(_ignoreYRotUpTo);
    _differentiateYRotUpTo = differentiateYRotUpTo;
    _cosDifferentiateYRotUpTo = cos(_differentiateYRotUpTo);

    _numIgnoredYRotations = 0;
    _numDifferentiatedYRotations = 0;
    _numFullRotations = 0;
  }

  /// Destructor.
  ~CRCache ();

  /// Apply translational gradients along globDiff and store the result in
  /// ret  (translation vector 'globDiff' given in global coordinate frame).
  void ExtrapolateRadiance(
    const CVector3D &globDiff,
    const CRadianceRecord *rad,
    CCoefVec<float>  &ret) const;

  /// Is interpolation possible at this point?
  /**
    @return 'true' is available records found and interpoilation is possible.
       'false' if nothing to interpolate.
    */
  bool InterpolationPossible(
    const CVector3D  &point,
    const CVector3D  &N,
    const float     &a,
    int   desiredOrder) const;

  /// Get radiance at a given point.
  /**
    @param point    (in) 
      Query location.
    @param U,V,N    (in) 
      Local coordinate frame at the interpolated point.
    @param sqrdist  (out) 
      Squared distance to the closest record used for interpolation.
    @param a        (int)  
      Approximation error.
    @param ret      (in,out) 
      The interpolated vector is set here. The order of 'ret' is used as the 
      desired order of the interpolation. Cache records shorter than desired 
      order will be ignored.
    @return 
      'false' if no usable radiance record is found in the cache.
			'true' if at least one record was found and radiance data has been set
      to 'ret' parameter.
  */
  template<bool useStoredA,bool jitterWeight>
  EQueryResult InterpolateRadiance(
    const CVector3D  &P,
    const CVector3D  &U,
    const CVector3D  &V,
    const CVector3D  &N,
    const float     &a,
    float           &minR,
    CCoefVec<float>  &ret) const;

  EQueryResult InterpolateRadianceDispatch(
    const CVector3D  &P,
    const CVector3D  &U,
    const CVector3D  &V,
    const CVector3D  &N,
    const float     &a,
    float           &minR,
    CCoefVec<float>  &ret) const
  {
    if(_useStoredA) {
      if(_jitterWeight) return InterpolateRadiance<true,true>(P,U,V,N,a,minR,ret);
      else return InterpolateRadiance<true,false>(P,U,V,N,a,minR,ret);
    }
    else {
      if(_jitterWeight) return InterpolateRadiance<false,true>(P,U,V,N,a,minR,ret);
      else return InterpolateRadiance<false,false>(P,U,V,N,a,minR,ret);
    }
  }
  
  /// Return radiance at a given point.
  /**
    @param a      (in)
      Approximation error.
    @param inHitInfo (in) 
      Represents the point at which we want to compute the radiance
    @param outHitInfo (in,out) 
      This the hitinfo that HAS to be passed to the radiance sampler if the 
      primary method is used.      
    @param result (out)  
      Store the radiance here. 
    @return 
      'true' if radiance has been evaluated and radiance value stored at 
      'result'. 'false' means that the (H)SH approximation of BRDF is not
      available or its error is too high. Returning 'false' indicates that 
      a different strategy should be used to determine the indirect specular 
      term (e.g. monte carlo importance sampling). 
  */
  EQueryResult GetRadiance(
    const CHitPointInfo &inHitInfo,
    const float        &footprintArea,
    CHitPointInfo       &outHitInfo,
    const CSurface      *surface,
    const float        &a,
    CValues             &result);

  /// Adds a new cache record at this point if needed.
  /**
    This is used to presample screen and fill the cache.
    Looks into the cache. If nothing found, hemisphere is sampled, new 
    record is created and added to the cache. For parameter meaning, 
    @see GetRadiance().

    @return 
   */
  EQueryResult PresampleCache(
    const CHitPointInfo& inHitInfo,
    const float        &footprintArea,
    CHitPointInfo&       outHitInfo,
    const CSurface*      surface,
    const float        &a);

  EQueryResult AddRadiance(
    const CVector3D &P,
    const CVector3D &U,
    const CVector3D &V,
    const CVector3D &N,
    const CVector3D &uN,
    const CObject3D *object,
    const void      *termination,
    const float    &footprintArea,
    CHitPointInfo   &outHitInfo,
    const float    &a,
    int desiredOrder,
    const CCoefVec<float>* shbrdf,
    CValues            *resColor,
    CRadianceRecord **newRecRet);

  /// align the record frame with the resulting frame and add
  /**
    @param MOV (template argumanet)
      if MOV=true, the rotated vector is moved to output (without weighting)
      otherwise, it is weighted and added to the result.
   */
  template<bool MOV>
    void RotateAndAdd(
    const CRadianceRecord *rad,
    const CCoefVec<float> *vec,
    const CVector3D  &U,
    const CVector3D  &V,
    const CVector3D  &N,
    float weight,
    CCoefVec<float>& ret) const;


  unsigned long GetNumIgnoredYRotations() const { return _numIgnoredYRotations; }
  unsigned long GetNumDifferentiatedYRotations() const 
  { return _numDifferentiatedYRotations; }
  unsigned long GetNumFullRotations() const { return _numFullRotations; }

  const CRadianceSampler* GetSampler() const {return _radSampler;}

  bool LocalFrame() const {return _localFrame;}

protected: // methods

  /// Computes the final dot product between brdf and incoming radiance.
  void _computeOutgoingRadiance(
    const CVector3D& U,
    const CVector3D& V,
    const CVector3D& N,
    const CCoefVec<float>& coeffs,
    const CCoefVec<float>& shbrdf,
    CValues &result);

protected: // data
  
  /// sampler to be used for the 'primary method' (i.e. when nothing is found in the cache)
  const CRadianceSampler* _radSampler;
  
  /// pointer to a (hemi)spherical basis interface
  const CSphBasis<float,float>        *_sphBasis;
 
  /// coefficient vector used for searching octree
  CCoefVec<float> _tmpCoeffs;
 
  /// Cached last query result.
  CLastIRCacheResult _last;

  /// for SH rotation through ZXZXZ decomposition
  CZXZXZRotationObject _zxzxzRotObject;

  // for error analysis of the apprximate rotation
  vector<CRotErrAnalysisData> _errMaxima;
  vector<CRotErrAnalysisData> _errAverages;

  /// number of color channels
  const int _channels;
  /// order of the (hemi)spherical harmonics
  const int _order;
  /// squared order of the (hemi)spherical harmonics
  const int _sqrOrder;
  /// Store radiance in local or global coordinates.
  const bool _localFrame;
  /// Compute gradient for all color channel or only for luminance?
  const bool _multichannelGrad;

  /// auxiliary (H)SH rotation matrix
  mutable CSHRotMatrix _shRotMatrix;

  /// auxiliary (H)SH vector
  mutable CCoefVec<float> *_tmpVec;

  /// just to avoid allocating every time GetContributions
  mutable CRCContribArray _carray;

  /// ignore rotation around Y for angles below this
  float _ignoreYRotUpTo;

  /// cos(_ignoreYRotUpTo)
  float _cosIgnoreYRotUpTo;

  /// sin(_ignoreYRotUpTo)
  float _sinIgnoreYRotUpTo;

  /// replace rotation around Y by its derivative for angles below this
  float _differentiateYRotUpTo;

  /// cos(_cosDifferentiateYRotUpTo)
  float _cosDifferentiateYRotUpTo;

  /// number of times tha rotation around Y was ignored
  mutable unsigned long  _numIgnoredYRotations;
  /// number of times the rotation around Y was approximated by its derivative
  mutable unsigned long  _numDifferentiatedYRotations;
  /// number of times the Y rotation could not be ignored or approximated
  mutable unsigned long  _numFullRotations;

};

__END_GOLEM_HEADER

#endif // __RCACHE_H__
