// ===================================================================
// $Id: rcalgo.cpp,v 1.6 2005/11/29 16:04:05 xkrivanj Exp $
//
// rcalgo.cpp
//     Implementation file for the CRCAlgo class.
//     A high level interface for radiance caching.
//
// Initial coding by Jaroslav Krivanek, Oct 2003.

// GOLEM headers
#include "config.h"
#include "radsampl.h"
#include "rcalgo.h"
#include "environ.h"

// LIBSPHAR headers
#include "shbasis.h"

// standard headers
#include <iomanip>

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  CRCAlgo::CRCAlgo()
// --------------------------------------------------------------------------
CRCAlgo::CRCAlgo(const CEnvironment& env,
                 int   depthLimit, 
                 CIncidentRadiance* incRad,
                 const CAABox &bbox,
                 const int channels):
_channels(channels),
CCachingAlgoBase<CRCache>("RadianceCache",env,depthLimit,incRad)
{
	assert( incRad != 0 );
  assert(channels>0);

  _presampleTotalTimer.Reset();
  _renderTotalTimer.Reset();

  _radSampler = 0;

	// don't init anything for depthLimit = 0, 
  // since we're not going to compute anything anyway
	if(_depthLimit <1)
		return;

	string str;
	// --------------------------------------------------------------------
	//  USE RADIANCE CACHE?
	// --------------------------------------------------------------------
  _radSampler  = new CRadianceSampler* [_depthLimit];
  _cache    = new CRCache* [_depthLimit];
	if(!_useCaching) {
		for( unsigned i=0; i < _depthLimit; i++) {
			_radSampler [i] = 0;
			_cache[i] = 0;
		}
		return;
	}

	// --------------------------------------------------------------------
	//  (H)SH ORDER
	// --------------------------------------------------------------------
	env.GetInt("RadianceCache.order", _order);

  if(_order<1) {
    FATAL << "'RadianceCache.order' must be higher than 1." << std::endl;
    FATAL_EXIT;
  }

	// --------------------------------------------------------------------
	//  TYPE OF (HEMI)SPHERICAL BASIS
	// --------------------------------------------------------------------
	int tabulatedBasisRes;
	env.GetString("RadianceCache.basisImplementation", str);
	env.GetInt("RadianceCache.tabulatedBasisRes", tabulatedBasisRes);
  if(tabulatedBasisRes<1)
    tabulatedBasisRes=1;
  CSphBasis<float,float>::EImplementation implementation;

  if(str=="analytic")
    implementation = CSphBasis<float,float>::EAnalytic;
  else if(str=="tabLegendre")
    implementation = CSphBasis<float,float>::ETabLegendre;
  else if(str=="tabulated")
    implementation = CSphBasis<float,float>::ETabulated;
  else {
    FATAL << "Invalid 'RadianceCache.basisImplementation' (" << str << ")\n";
    FATAL << "Possible values are:\n";
    FATAL << "  'analytic'\n";
    FATAL << "  'tabLegendre'\n";
    FATAL << "  'tabulated'" << std::endl;
    FATAL_EXIT;
  }

  env.GetString("RadianceCache.basisType", str);
  if( str == "HSH" ) {
    string cfg_sampleCoordFrame;
    env.GetString("RadianceCache.sampleCoordFrame", cfg_sampleCoordFrame);
    if(cfg_sampleCoordFrame=="global") {
      FATAL 
        << "HSH basis cannot be used for global frame.\n"
        << "Set 'RadianceCache.sampleCoordFrame' to 'local' or\n"
        << "set 'RadianceCache.basisType' to 'SH'.\n";
      FATAL_EXIT;
    }

		// hemispherical harmonics
    switch(implementation) {
    case CSphBasis<float,float>::EAnalytic:
      _sphBasis = new CHSHBasis<float,float>(_order);
      break;
    case CSphBasis<float,float>::ETabLegendre:
      _sphBasis = new CHSHBasisLTab<float,float>(_order);
      break;
    case CSphBasis<float,float>::ETabulated:
      _sphBasis = new CHSHBasisTab<float,float>(_order,tabulatedBasisRes);
      break;
    }
	}
	else {
	   // spherical harmonics
		// hemispherical harmonics
    switch(implementation) {
    case CSphBasis<float,float>::EAnalytic:
      _sphBasis = new CSHBasis<float,float>(_order);
      break;
    case CSphBasis<float,float>::ETabLegendre:
      _sphBasis = new CSHBasisLTab<float,float>(_order);
      break;
    case CSphBasis<float,float>::ETabulated:
      _sphBasis = new CSHBasisTab<float,float>(_order,tabulatedBasisRes);
      break;
    }
	}
	assert(_sphBasis);

	// --------------------------------------------------------------------
	//  RADIANCE CACHE
	// --------------------------------------------------------------------
	bool localFrame;
	env.GetString("RadianceCache.sampleCoordFrame",str);
	if(str=="local")
		localFrame = true;
	else if (str=="global")
		localFrame = false;
	else {
		FATAL << "Unrecognized value '" << str << "' of 'RadianceCache.sampleCoordFrame'." << std::endl;
		FATAL << "Possible values are 'local', 'global'. " << std::endl;
		FATAL_EXIT;
	}

	CRadianceSampler::EType rsType;
	env.GetString("RadianceCache.samplerType",str);
  if(str=="displace") {
		rsType = CRadianceSampler::ERadianceSamplerDisplace;
  }
  else if(str=="deriv") {
		rsType = CRadianceSampler::ERadianceSamplerDeriv;
  }
  else if(str=="ward" || str=="uni_ward") {
		rsType = CRadianceSampler::ERadianceSamplerWard;
  }
	else {
		FATAL 
      << "Unrecognized radiance sampler type '" << str << "'.\n" ;
		FATAL 
      << "'RadianceCache.samplerType' can be 'displace', 'deriv', 'derivdir' \n"
      << "or 'ward'." << std::endl;
		FATAL_EXIT;
	}

	float ignoreYRotUpTo, differentiateYRotUpTo;
	env.GetFloat("RadianceCache.ignoreYRotUpTo",ignoreYRotUpTo);
	env.GetFloat("RadianceCache.differentiateYRotUpTo",differentiateYRotUpTo);
	if(ignoreYRotUpTo<0)    ignoreYRotUpTo=0;
	if(ignoreYRotUpTo>180)  ignoreYRotUpTo=180;
	ignoreYRotUpTo *= (M_PI/180.0);
	if(differentiateYRotUpTo<0)    differentiateYRotUpTo=0;
	if(differentiateYRotUpTo>180)  differentiateYRotUpTo=180;
	differentiateYRotUpTo *= (M_PI/180.0);
	
	float infRayLength = Magnitude(bbox.Diagonal());

  // number of rays per cell
  int raysPerCell;
  env.GetInt("RadianceCache.raysPerCell",raysPerCell);
  if(raysPerCell<1) raysPerCell=1;

	bool multichannelGrad;
	env.GetBool("RadianceCache.multichannelGrad",multichannelGrad);

	for( unsigned i=0; i < _depthLimit; i++) {
		if(_numRays[i]>0) {
      _radSampler[i] = 0;

      // compute number of hemisphere cells
      int numCells = (int)((float)_numRays[i]/raysPerCell + 0.5);
      if(numCells<1) numCells = 1;

      switch(rsType) {
      case CRadianceSampler::ERadianceSamplerDisplace:
        _radSampler[i] = new CRadianceSamplerDisplace
          (i, numCells, raysPerCell, _order, infRayLength, _minAllowedHitDist,
          _incRad, _sphBasis, localFrame, multichannelGrad, 0);
        break;
      case CRadianceSampler::ERadianceSamplerDeriv:
				_radSampler[i] = new CRadianceSamplerDeriv
          (i, numCells, raysPerCell, _order, infRayLength, _minAllowedHitDist,
          _incRad, _sphBasis, localFrame, multichannelGrad, 0);
        break;
      case CRadianceSampler::ERadianceSamplerWard:
				_radSampler[i] = new CRadianceSamplerWard
          (i, numCells, raysPerCell, _order, infRayLength, _minAllowedHitDist,
          _incRad, _sphBasis, localFrame, multichannelGrad, 0);
        break;
      default:
        FATAL << "Unrecognized radiance sampler type. " << std::endl;
        FATAL_EXIT;
        break;
      }
		
      assert(_radSampler[i]);
      _cache[i] = 
        new CRCache(
        // octree params
        bbox,
        _octreeScale,
        // commong caching params
        _minDist,
        _maxDist, 
        _centerDisplaceDist,
        _rotWeightMultiplier,
        _nearestRecordAMul,
        (i==0) ? _useGradients : false,   // gradients only for primary rays
        (i==0) ? _cacheLastResult : false,
        _clampInScreenSpace,
        _useMinHitDist, 
        _useGeometricNormal,
        _clampRadiusByGradients, 
        _clampByNeighbors,
        _clampNeighbors,
        i,
        // radiance caching specific params
        ignoreYRotUpTo, 
        differentiateYRotUpTo, 
        _order, 
        _channels,
        localFrame, 
        multichannelGrad,
        _radSampler[i], 
        _sphBasis);


      assert(_cache[i]);
		}
		else {
			_radSampler [i] = 0;
			_cache[i] = 0;
		}
	}
	
  // rendering mode is the default
  SetRenderMode();
}

// --------------------------------------------------------------------------
//  CRCAlgo::~CRCAlgo()
// --------------------------------------------------------------------------
CRCAlgo::~CRCAlgo()
{
	for( unsigned i=0; i < _depthLimit; i++) {
		SAFE_DELETE(_radSampler[i]);
		SAFE_DELETE(_cache[i]);
	}
	delete _sphBasis;
	delete [] _radSampler;
	delete [] _cache;
}

// --------------------------------------------------------------------------
//  CRCAlgo::DoReport()
// --------------------------------------------------------------------------
bool
CRCAlgo::DoReport(ostream &app) 
{

  unsigned long numRecords       = 0;
  unsigned long numQueriesSuc    = 0;
  unsigned long numQueriesFail   = 0;
  unsigned long numWeightZero    = 0;
  unsigned long numWeightNonzero = 0;
  unsigned long numIgnoredYRotations = 0;
  unsigned long numDifferentiatedYRotations = 0;
  unsigned long numFullRotations = 0;
  unsigned long numRotations     = 0;
  unsigned      numCaches        = 0;
  float        radSamplerTime = 0;
  float        sampleTime = 0;
  float        projectTime = 0;
  float        gradTime = 0;

  unsigned i;

  assert(_cache);

  for(i=0; i<_depthLimit; i++) {
	  if(_cache[i]) {
		  numCaches ++;

		  numRecords       += _cache[i]->GetNumRecords();
		  numQueriesSuc    += _cache[i]->GetNumQueriesSuc();
		  numQueriesFail   += _cache[i]->GetNumQueriesFail();
		  numWeightZero    += _cache[i]->GetNumWeightZero();
		  numWeightNonzero += _cache[i]->GetNumWeightNonzero();
		  numIgnoredYRotations          += 
        _cache[i]->GetNumIgnoredYRotations();
		  numDifferentiatedYRotations   += 
        _cache[i]->GetNumDifferentiatedYRotations();
		  numFullRotations              += 
        _cache[i]->GetNumFullRotations();

      sampleTime  += _cache[i]->GetSampler()->GetSampleTime();
      projectTime += _cache[i]->GetSampler()->GetProjectTime();
      gradTime    += _cache[i]->GetSampler()->GetGradTime();
	  }
  }

  float renderTotal    = _renderTotalTimer.UserTime();
  float presampleTotal = _presampleTotalTimer.UserTime();

  radSamplerTime = sampleTime + projectTime + gradTime;
  numRotations = 
    numIgnoredYRotations + numDifferentiatedYRotations + numFullRotations;

  app
    << "\n"
    << "+----------------------------------------------------------------+\n"
    << "|        Radiance caching statistics                             |\n"
    << "+----------------------------------------------------------------+\n"
	<< "| # |  Records  |   Queries   | Q-succ | Weight Evals | W-succ   |\n";

  for(i=0; i<_depthLimit; i++) {
	  if(_cache[i]) {
		  app << "|" << setw(2) << i;
		  app << " |" << setw(10) << _cache[i]->GetNumRecords();
		  app << " |" << setw(12) << _cache[i]->GetNumQueriesSuc()+
			  _cache[i]->GetNumQueriesFail();
		  app << " |" << setw(7) << setprecision(4); 
		  if( _cache[i]->GetNumQueriesSuc()+_cache[i]->GetNumQueriesFail() 
        > 0 )
			  app << ((float)_cache[i]->GetNumQueriesSuc() / 
			  (float)(_cache[i]->GetNumQueriesSuc()+
			  _cache[i]->GetNumQueriesFail()));
		  else
			  app << "    -  ";
		  app << " |" << setw(13) << _cache[i]->GetNumWeightZero() +
			  _cache[i]->GetNumWeightNonzero();
		  app << " |" << setw(9);
		  if( (_cache[i]->GetNumWeightZero() + 
        _cache[i]->GetNumWeightNonzero()) > 0)
			  app  <<  (float)_cache[i]->GetNumWeightNonzero() / 
			  (float)(_cache[i]->GetNumWeightZero() +
			  _cache[i]->GetNumWeightNonzero());
		  else
			  app << "      -  ";
		  app << " |\n";
	  }
	  else
		  app << "| - |           |             |        |              |          |\n";
  }
 app  << "+----------------------------------------------------------------+\n";
		  app << "|TOT|" << setw(10) << numRecords;
		  app << " |" << setw(12) << numQueriesSuc+numQueriesFail;
		  app << " |" << setw(7) << setprecision(4);
		  if(numQueriesSuc+numQueriesFail > 0)
			  app << (float)numQueriesSuc / (float)(numQueriesSuc+numQueriesFail);
		  else
			  app << "    -  ";
		  app << " |" << setw(13) << numWeightZero + numWeightNonzero;
		  app << " |" << setw(9);
		  if ( numWeightZero + numWeightNonzero > 0)
			  app <<  (float)numWeightNonzero / 
        (float)(numWeightZero + numWeightNonzero);
		  else
			  app << "      -  ";
		  app << " |\n";
		  app  
        << "+----------------------------------------------------------------+\n";
		  
		  app << "| Rotations:            "  << setw(7) <<  numRotations 
        << std::endl;
		  if(numRotations>0) {
			  app 
          << "|   - ignored Y:        "  << setw(7) 
          <<  numIgnoredYRotations << " (= " 
				  << (float)numIgnoredYRotations/(float)numRotations*100 << "%)" 
          << std::endl;
			  app 
          << "|   - differentiated Y: "  << setw(7) 
          <<  numDifferentiatedYRotations << " (= " 
				  << (float)numDifferentiatedYRotations/(float)numRotations*100 
          << "%)" << std::endl;
			  app 
          << "|   - full Y:           "  << setw(7) 
          <<  numFullRotations << " (= " 
				  << (float)numFullRotations/(float)numRotations*100 << "%)" 
          << std::endl;
		  }
      app 
        << "| Radiance Sampler Time:   "  << radSamplerTime 
        << " sec. " << std::endl;
      if(radSamplerTime) {
        app 
          << "|   - hemisphere sampling:  "  << sampleTime
          << " sec = " << sampleTime/radSamplerTime*100 << "%." << std::endl;
        app 
          << "|   - projection:           "  << projectTime
          << " sec = " <<  projectTime/radSamplerTime*100 << "%." << std::endl;
        app 
          << "|   - gradient computation: " << gradTime
          << " sec = " <<  gradTime/radSamplerTime*100 << "%." << std::endl;
      }

      app << "| Total Caching Time:      "  << renderTotal + presampleTotal 
        << " sec. " << std::endl;

      app << "|   - in presampling:      "  << presampleTotal << " sec. ";
      if(renderTotal + presampleTotal > 0)
        app << " = " << presampleTotal/(presampleTotal+renderTotal)*100 << "%.";
      app << std::endl;

      app << "|   - in rendering:        "  << renderTotal << " sec. ";
      if(renderTotal + presampleTotal > 0)
        app << " = " << renderTotal/(presampleTotal+renderTotal)*100 << "%.";
      app << std::endl;

      app 
        << "| Interpolation time:      "  
        << renderTotal + presampleTotal - radSamplerTime << " sec. " << std::endl;

		  return false; // ok
}
__END_GOLEM_SOURCE
