// ===================================================================
// $Id: rcalgo.h,v 1.6 2005/11/29 16:04:05 xkrivanj Exp $ 
//
// rcalgo.h
//     Header file for the CRCAlgo class.
//     A high level interface for radiance caching.
//
// Class: CRCAlgo
//
// Initial coding by Jaroslav Krivanek, December 2003.

#ifndef __RCALGO_H__
#define __RCALGO_H__

// GOLEM headers
#include "configh.h"
#include "surface.h"
#include "rcache.h"
#include "icalgo.h"
#include "timer.h"

__BEGIN_GOLEM_HEADER

class CRadianceSampler;
class CRCache;
class CIncidentRadiance;
class CAABox;
class CSurface;
class CContributionArray;
class CEnvironment;
template<class TV,class TA> class CSphBasis;

/// Encapsulation of all that is related to radiance caching.
/**
  This class should be used by every app that wants to use radiance caching.
  The constructor creates everything that's needed according to the environment.
  Once an instance is created, it suffices to call IndirectLighting() to get 
  the (directional) indirect lighting at a given point.
*/
class CRCAlgo: public CCachingAlgoBase<CRCache> {
  
protected: // data
  /// Array of diffuse samplers to be used for hemisphere sampling.
  CRadianceSampler          **_radSampler;
  /// pointer to a (hemi)spherical basis interface for (possibly) approximative evaluation
  CSphBasis<float,float>          *_sphBasis;
  /// number of color channels
  const int _channels;
  /// SH order
  int _order;
  /// total caching time on presampling
  CTimer _presampleTotalTimer;
  /// total cacching time on rendering
  CTimer _renderTotalTimer;

public:  // methods
  /// Create all that is needed for radiance caching according to the environment.
  /**
    @param depthLimit (in) maximal depth of recursion
    @param incRad     (in) implementation of the CIncidentRadiance interface
    @param bbox       (in) bounding box of the whole scene
  */
  CRCAlgo(
    const CEnvironment& env,
    int depthLimit, 
    CIncidentRadiance* incRad,
    const CAABox &bbox,
    const int channels);
  
  /// Destructor.
  ~CRCAlgo();

  /// return the number of color channels
  int GetNumChannels() const { return _channels; }
  /// return SH order
  int GetOrder() const { return _order; }
  
  /// Compute indirect directional lighting at a given point.
  /**
    This function calls the radiance cache at a given recursion depth to see 
    if a cached sample can be used. If yes, use it. If not, sample the 
    hemisphere, store the sample into the cache and return it. The resulting 
    color (already weighted with BRDF) is stored in 'result.
  
    @param inHitInfo 
      (in) describes the point at which we want to compute the 
      indirect diffuse lighting

    @param surface         
      (in) surface interface of the object being hit by  'inHitInfo'.

    @param depth           (in) 
      current recursion depth

    @param outHitInfo 
      (in,out) this hitinfo has to be passed to the 
      radiance sampler in case that the hemisphere sampling will be used.

    @param result          
      (out) stores the radiance here.
    
    @return   
      'true' if radiance has been computed and 'result' set (even if it is 
       zero). 'false' means that a different strategy should be applied at 
       this point to compute the indirect glossy term.
  */
  inline bool IndirectLighting(
    const CHitPointInfo&      inHitInfo,
    const float          & footprintArea,
    const CSurface*           surface,
    const int                 depth,
    CHitPointInfo&            outHitInfo,
    CValues&                  result);

  /// Adds a new cache record at this point if needed.
  /**
    Looks into the cache. If nothing found, hemisphere is sampled, new 
    record is created and added to the cache.
    Nothing is returned. 
    This is used to presample screen and fill the cache.
   */
  inline bool PresampleCache(
    const CHitPointInfo&      inHitInfo,
    const float          & footprintArea,
    const CSurface*           surface,
    const int                 depth,
    CHitPointInfo&            outHitInfo
    );
  
  /// Returns the (hemi)spherical basis interface (possibly approximate evaluation).
  const CSphBasis<float,float>* GetSphBasis() const 
  { return _sphBasis; }
 
  /// Print a report.
  bool DoReport(ostream &app);

  /// Return pointer to irradiance cache interface
  CRCache *GetRCache(unsigned i) 
  {assert(i<_depthLimit); assert(_cache[i]); return _cache[i]; }
};

// --------------------------------------------------------------------------
//  CRCAlgo::IndirectLighting()
// --------------------------------------------------------------------------
inline bool 
CRCAlgo::IndirectLighting(const CHitPointInfo&   inHitInfo,
                          const float          & footprintArea,
                          const CSurface*        surface,
                          const int              depth,
                          CHitPointInfo&         outHitInfo,
                          CValues&                result)
{
  assert(surface->GetSupportedTypes() & EE_REFLECTED_GLOSSY);
  assert(depth < (int)_depthLimit);
  assert(_approxErrorCurrent);
  assert(_approxErrorCurrent==_approxErrorRender||_approxErrorCurrent==_approxErrorPresample);
  
  if(!(_radSampler[depth] && inHitInfo.GetUDotProd()<0) ) 
  { result.setz(); return false; }

  // start measuring
  _renderTotalTimer.Start();

  CRCache::EQueryResult queryResult = 
    _cache[depth]->GetRadiance(inHitInfo,footprintArea,outHitInfo,
    surface,_approxErrorCurrent[depth],result);

  // stop measuring
  _renderTotalTimer.Stop();

  if(queryResult == CRCache::EE_Failed)
  { result.setz(); return false; }

  // record visualization
  if(depth==0) {
    // visualize records
    if(_visualizeRecPosCurrent)
      VisualizeNearestRecords(inHitInfo.GetPoint(),inHitInfo.GetNormal(),
      depth, sqrt(footprintArea)*_visualRecSize, result);
    // visualize newly added record
    if(_visualizeNewRecPosCurrent&&queryResult==CRCache::EE_NewRecord)
      result = _visualNewRecColor;
  }

  // indirect glossy lighting computed and set in result
  return true;
}

// --------------------------------------------------------------------------
//  CRCAlgo::PresampleCache()
// --------------------------------------------------------------------------
inline bool 
CRCAlgo::PresampleCache(const CHitPointInfo&      inHitInfo,
                        const float          & footprintArea,
                        const CSurface*           surface,
                        const int                 depth,
                        CHitPointInfo&            outHitInfo)
{
  assert(surface->GetSupportedTypes() & EE_REFLECTED_GLOSSY);
  assert(depth < (int)_depthLimit);
  assert(_approxErrorPresample);
  
  if(_radSampler[depth] && inHitInfo.GetUDotProd()<0) 
  {
    // start measuring
    _presampleTotalTimer.Start();

    bool b = _cache[depth]->PresampleCache(inHitInfo,footprintArea,
      outHitInfo,surface,_approxErrorPresample[depth]);

    // stop measusing
    _presampleTotalTimer.Stop();

    if(b)
      return true;
    else
      return false;
  }
  return false;
}

__END_GOLEM_HEADER

#endif // __RCALGO_H__
