// ===================================================================
// $Id: rcfaced3d.cpp,v 1.4 2005/11/02 12:59:02 xkrivanj Exp $
//
// rcface3d.cpp
//     Implementation of the CRCache class - part that requires Direct3D
//     math library.
//
// Initial coding by Jaroslav Krivanek, December 2003.

// GOLEM headers
#include "config.h"
#include "rcache.h"

// SPHAR headers
#include "shbasis.h"

// #define D3DXROTATION

// D3DX headers (for SH-rotation)
#ifdef D3DXROTATION
#include "d3dx9math.h"
#endif

__BEGIN_GOLEM_SOURCE

CVector3D v3d;

#ifdef D3DXROTATION
inline void D3DXToLocalFrame(D3DXMATRIX &mat, 
						   const CVector3D &U,
						   const CVector3D &V, 
						   const CVector3D &N)
{
	mat(0,0) = U.x;	mat(1,0) = U.y;	mat(2,0) = U.z;	mat(3,0) = 0;
	mat(0,1) = V.x;	mat(1,1) = V.y;	mat(2,1) = V.z;	mat(3,1) = 0;
	mat(0,2) = N.x;	mat(1,2) = N.y;	mat(2,2) = N.z;	mat(3,2) = 0;
	mat(0,3) = 0;	mat(1,3) = 0;	mat(2,3) = 0;	mat(3,3) = 1;
}

static inline void D3DXToGlobalFrame(D3DXMATRIX &mat,
									 const CVector3D &U,
									 const CVector3D &V, 
									 const CVector3D &N)
{
	mat(0,0) = U.x;	mat(0,1) = U.y;	mat(0,2) = U.z;	mat(0,3) = 0;
	mat(1,0) = V.x;	mat(1,1) = V.y;	mat(1,2) = V.z;	mat(1,3) = 0;
	mat(2,0) = N.x;	mat(2,1) = N.y;	mat(2,2) = N.z;	mat(2,3) = 0;
	mat(3,0) = 0;	mat(3,1) = 0;	mat(3,2) = 0;	mat(3,3) = 1;
}
#endif


/// debugging function 
static void printL2(const float* v, int order)
{
	DEBUG << "";
	for(int l=0; l<order; l++) {
		float l2=0;
		for(int m=-l; m<=l; m++) {
			l2 += sqr(v[shIndex(l,m)]);
		}
		DEBUGW << l2 << ", ";
	}
	DEBUGW << "\n";
}


// --------------------------------------------------------------------------
//  CRCache::_computeOutgoingRadiance()
// --------------------------------------------------------------------------
void
CRCache::_computeOutgoingRadiance(const CVector3D& U,
                                  const CVector3D& V,
                                  const CVector3D& N,
                                  const CCoefVec<float>& Li,
                                  const CCoefVec<float>& brdf,
                                  CValues &result)
{
	//-----------------------------
	//  dot product lighting
	//-----------------------------
	int order    = Min( brdf.GetOrder(), _order);
  int vecLen   = PAD_FOUR(order*order);
  int channels = Li.GetNumChannels();
	
	// --------------------------------------
	//  Incoming radiance in the local frame
	// --------------------------------------
	if(_localFrame)  {		
		// dot product of brdf and incoming radiance
#if 1 // use the brdf
    for(int ch = 0; ch < channels; ch++) {
      float d = shvf::dot4( vecLen, brdf.GetChannel(ch), Li.GetChannel(ch));
			result.Set(ch, d);
    }
#else // use the basis - visualize directly the incoming radiance
    // CVector3D v3d(0,0,1);
    const float *H;
    _sphBasis->EvalBasisCartVInt(order, v3d, &H);  
    for(int ch = 0; ch < _channels; ch++)
			result.Set(ch, shvf::dot4( vecLen,H, Li.GetChannel(ch)) );
#endif

		result.ClipLower(0);
		return;
	}

	// --------------------------------------
	//  Incoming radiance in the global frame
	// --------------------------------------
	ALIGN16 float tmpvec1 [MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
	assert(order<=MAX_FIXED_SH_ORDER);

#ifdef D3DXROTATION
	if( order <= 6 )
	{
		// D3DX library's used to be significantly faster than ours,
    // but now it isn't anymore, so don't use it
		// (D3DX supports rotations only up to order 6)
		D3DXMATRIX matrix;
		D3DXToGlobalFrame(matrix, U, V, N);
		// rotate & compute the dot product
		for(int ch = 0; ch < _channels; ch++) {
			D3DXSHRotate(tmpvec1, order, &matrix, brdf.GetChannel(ch));
			result.Set(ch, shvf::dot4( vecLen, tmpvec1, Li.GetChannel(ch)) );
		}
	}
	else 
#endif
	{
#if 0
    //////////////////////////////////////////////
    // Ivanic's SH rotation matrix construction
    //////////////////////////////////////////////
		rotGenerator.ToGlobalFrame(_shRotMatrix,order,U,V,N);
		// rotate & compute the dot product
		for(int ch=0; ch<_channels; ch++) {		
			_shRotMatrix.
        TransformVector(tmpvec1,brdf.GetChannel(ch),order);
			result.Set(ch, shvf::dot4(vecLen, tmpvec1, Li.GetChannel(ch)) );
		}
#else
    //////////////////////////////////////////////
    // ZXZXZ decomposition
    //////////////////////////////////////////////
    _zxzxzRotObject.ToGlobalFrame<float>(U,V,N,order);
    // rotate & compute the dot product
    for(int ch=0; ch<_channels; ch++) {
      _zxzxzRotObject.
        TransformVector(tmpvec1,brdf.GetChannel(ch),order,rotGenerator);
      result.Set(ch, shvf::dot4(vecLen, tmpvec1, Li.GetChannel(ch)));
    }
#endif
	}

	result.ClipLower(0);
}

__END_GOLEM_SOURCE
