// ===================================================================
// $Id: rcoctree.cpp,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// rcoctree.cpp
//   Implementation of the radiance and irradiance cache octree.
//
// Licence: the use and distribution of this file is severely limited, please
// see the file 'doc/Licence.txt'. Any non-authorized use can be prosecuted under
// International Law. For further questions, please, e-mail to VHavran@seznam.cz
// or mail to Vlastimil Havran, Pohodli 27, 57001 Litomysl, the Czech Republic.
// REPLACEMENT_STRING
//
// Initial coding by Jaroslav Krivanek (oct 2000).

// GOLEM headers
#include "config.h"
#include "basmacr.h"
#include "growvec.h"
#include "icontrib.h"
#include "icrecord.h"
#include "rcoctree.h"
#include "rcrecord.h"

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  CRCOctree::CRCOctree()
// --------------------------------------------------------------------------
template<class TRecord>
CRCOctree<TRecord>::CRCOctree (const CAABox &sceneBox,
                      float octreeScale)
{
  assert(octreeScale>0);
  // at least slightly, otherwise unbounded recursion can occur if 
  // ( harmMeanDistance==0 )
  
  _octreeScale = octreeScale;
  _octreeScaleDiv2 = octreeScale * 0.5;
  _octreeScalePlusOne = octreeScale + 1.0;
  
  // set the `_bbox'
  CVector3D boxSize     = sceneBox.Size();
  
  // use a cubic box, slightly bigger than the scene
  _boxSideLen = 1.1 * boxSize[boxSize.DrivingAxis()];
  CVector3D boxCenter  = 0.5 * (sceneBox.Min() + sceneBox.Max());
  
  _box.min = boxCenter-CVector3D(_boxSideLen/2,_boxSideLen/2,_boxSideLen/2);
  _box.max = boxCenter+CVector3D(_boxSideLen/2,_boxSideLen/2,_boxSideLen/2);

  _root = new CRCOctreeNode<TRecord>(_boxSideLen, _box.min, 0);
  
  _numRecords = 0;
  _octreeDepth = 0;
  
  _searchOctreeStack = 0;
  _reallocateStack(20);
  
  ResetStats();
}

// --------------------------------------------------------------------------
//  CRCOctree::~CRCOctree()
// --------------------------------------------------------------------------
template<class TRecord>
CRCOctree<TRecord>::~CRCOctree ()
{
  delete [] _searchOctreeStack;
  delete _root;
}

// --------------------------------------------------------------------------
//  CRCOctree::ResetStats()
// --------------------------------------------------------------------------
template<class TRecord>
void
CRCOctree<TRecord>::ResetStats()
{
  _numQueriesSuc     = 0;
  _numQueriesFail    = 0;
  _numWeightZero     = 0;
  _numWeightNonzero  = 0;
  _numDecreaseA      = 0;
  _numDecreaseR      = 0;
  _numMoveInOctree   = 0;
}

#define _searchOctree_HELPER(j) \
    { boxOriginSon[j] = boxOrigin[j];  \
    if (1<<j & i) boxOriginSon[j] += sideLength;  \
    if (P[j] < boxOriginSon[j] - _octreeScale * sideLength) \
	   goto next_son;  \
    if (P[j] > boxOriginSon[j] + _octreeScalePlusOne * sideLength) \
     goto next_son; }

// --------------------------------------------------------------------------
//  CRCOctree::GetContributions()
// --------------------------------------------------------------------------
template<class TRecord>
template<bool useStoredA,bool jitterWeight>
bool
CRCOctree<TRecord>::GetContributions(const CVector3D  &P,
                                     const CVector3D  &N,
                                     CRecordContribArray<TRecord> &carray,
                                     const float     a) const
{
  const int channels = 3;
  unsigned itemsBefore = carray.size();
 
  _searchOctreeStack[0].Set(GetRootNode());
  int stackEnd=1;
  
  // contribution item
  CRecordContrib<TRecord> item;
  item.SetEInvalid();
  
  while(stackEnd>0) {
    stackEnd--;
    
    // pop from the stack
    CRCOctreeNode<TRecord>* node = 
      (CRCOctreeNode<TRecord>*)_searchOctreeStack[stackEnd].node;

    const CVector3D  boxOrigin   = node->GetBoxOrigin();
    float           sideLength  = node->GetSideLength();
    
    // run throug the list of the current node
    CRecordList<TRecord> &nodeList = node->GetList();
    typename CRecordList<TRecord>::iterator iter = nodeList.begin();
    
    for ( ; iter != nodeList.end(); iter++) {
      item.rec = *iter;
      if( item.rec->template ComputeWeightTemplate<jitterWeight>
        (P,N,useStoredA?item.rec->GetA():a,item.w) )
      {
        // accepted record
        _numWeightNonzero++;
        // record the clock state with the contribution
        item.changeClock = item.rec->GetChangeClock();
        // store the contribution
        carray.push_back(item);
      }
      else {
        // unaccepted cache sample
        _numWeightZero++;
      }
      
    } // for
    
    if (node->IsLeaf())
      continue;
    
    // for all descendants
    const CRCOctreeNode<TRecord> *son;
    CVector3D boxOriginSon;
    sideLength *= 0.5;
    for (int i = 0; i < 8; i++) {
      if (son = node->GetSon(i) ) {
        _searchOctree_HELPER(0);
        _searchOctree_HELPER(1);
        _searchOctree_HELPER(2);
        _searchOctreeStack[stackEnd++].Set(son);      
      }
next_son: ;
    } 
  }

  // end of tree traversal

  if(carray.size()>itemsBefore) {
    _numQueriesSuc ++;
    return true;
  }
  else {
    _numQueriesFail ++;
    return false;
  }
}

template bool CRCOctree<CIrradianceRecord>::GetContributions<true,true>(
  const CVector3D  &P, const CVector3D  &N,
  CICContribArray &carray, const float a) const;
template bool CRCOctree<CIrradianceRecord>::GetContributions<false,true>(
  const CVector3D  &P, const CVector3D  &N,
  CICContribArray &carray,const float a) const;
template bool CRCOctree<CIrradianceRecord>::GetContributions<true,false>(
  const CVector3D  &P, const CVector3D  &N,
  CICContribArray &carray, const float a) const;
template bool CRCOctree<CIrradianceRecord>::GetContributions<false,false>(
  const CVector3D  &P, const CVector3D  &N,
  CICContribArray &carray,const float a) const;

template bool CRCOctree<CRadianceRecord>::GetContributions<true,true>(
  const CVector3D  &P, const CVector3D  &N,
  CRCContribArray &carray, const float a) const;
template bool CRCOctree<CRadianceRecord>::GetContributions<false,true>(
  const CVector3D  &P, const CVector3D  &N,
  CRCContribArray &carray,const float a) const;
template bool CRCOctree<CRadianceRecord>::GetContributions<true,false>(
  const CVector3D  &P, const CVector3D  &N,
  CRCContribArray &carray, const float a) const;
template bool CRCOctree<CRadianceRecord>::GetContributions<false,false>(
  const CVector3D  &P, const CVector3D  &N,
  CRCContribArray &carray,const float a) const;

// --------------------------------------------------------------------------
//  CRCOctree::ValidateOctree()
// --------------------------------------------------------------------------
template<class TRecord>
void  
CRCOctree<TRecord>::ValidateOctree() const
{
  DEBUG << "Validating octree...";
  _validateOctree(GetRootNode(), _boxSideLen);
  DEBUGW << "done." << std::endl;
}

// --------------------------------------------------------------------------
//  CRCOctree::_validateOctree()
// --------------------------------------------------------------------------
template<class TRecord>
void
CRCOctree<TRecord>::_validateOctree(const CRCOctreeNode<TRecord> *node, 
                              float sideLength) const
{
  const CRecordList<TRecord> &nodeList = node->GetList();
  
  typename CRecordList<TRecord>::const_iterator it = nodeList.begin();
  
  while (it != nodeList.end()) {
    
    float validDomain = (*it)->GetValidDomain();
    
    // hd \in (
    
    if(!(
      ( sideLength * (_octreeScaleDiv2) <= validDomain ) &&
      ( sideLength * (_octreeScale)     >  validDomain ) 
      ) )
    {
      DEBUG << "relace neplati" << std::endl;
      PRINTVAR(sideLength);
      PRINTVAR(validDomain);
    }
    
    it++;
  } // while
  
  for(int i=0;i<8;i++) {
    if(node->GetSon(i))
      _validateOctree(node->GetSon(i), sideLength*0.5);
  }
}

// --------------------------------------------------------------------------
//  CRCOctree::_locateNodeForStoring()
// --------------------------------------------------------------------------
template<class TRecord>
inline CRCOctreeNode<TRecord>*
CRCOctree<TRecord>::_locateNodeForStoring(CRCOctreeNode<TRecord>   *node,
                                    TRecord  *rec,
                                    const float        validDomain)
{
  unsigned son; 
  float     sideLength  = node->GetSideLength();
  CVector3D boxOrigin    = node->GetBoxOrigin();
 
  // find the tree location to store the record
  while ( sideLength * (_octreeScaleDiv2) > validDomain) {
    sideLength *= 0.5f;
    son = 0;
    for (int i = 0; i < 3; i++)
      if (rec->GetPoint()[i] > boxOrigin[i] + sideLength) {
        boxOrigin[i] += sideLength;
        son |= 1 << i;
      }
      
      node = node->GetSonAlloc(son,sideLength,boxOrigin);
  }
  return node;
}

// --------------------------------------------------------------------------
//  CRCOctree::_storeInNode()
// --------------------------------------------------------------------------
template<class TRecord>
inline void
CRCOctree<TRecord>::_storeInNode(CRCOctreeNode<TRecord> *node, TRecord *rec)
{
  // add the record to the tree
  node->GetList().push_back(rec);
  // save the location of the record in the tree
  rec->SetTreeLoc(node, --node->GetList().end());
  // extend the recursion stack if needed
  int depth = node->GetDepth();
  if(depth > _octreeDepth) {
    _octreeDepth = depth;
    if(depth > _stackDepth)
      _reallocateStack(2*depth);
  }
}

// --------------------------------------------------------------------------
//  CRCOctree::InsertRecord()
// --------------------------------------------------------------------------
template<class TRecord>
void
CRCOctree<TRecord>::InsertRecord(TRecord *rec)
{
  assert(rec);
  // locate the node to save the record to
  CRCOctreeNode<TRecord> *node = _locateNodeForStoring(GetRootNode(),rec,rec->GetValidDomain());
  // store in the located node
  _storeInNode(node, rec);
  // add to the linear record list
  _linList.push_back(rec);
  // increase record counter
  _numRecords++;
}

// --------------------------------------------------------------------------
//  CRCOctree::DecreaseApproxError()
// --------------------------------------------------------------------------
template<class TRecord>
void
CRCOctree<TRecord>::DecreaseApproxError(TRecord *rec, float newA, bool updateOctree)
{
  if(newA<rec->GetA())  {
    rec->SetA(newA);
    _numDecreaseA ++;
    if(updateOctree) {

      /*
      CRCOctreeNode<TRecord>* node = 
        _locateNodeForStoring(GetRootNode(), rec, rec->GetValidDomain());
      rec->GetTreeNode()->GetList().erase(rec->GetTreeIterator());
        _storeInNode(node,rec);
        _numMoveInOctree++;
        */

      CRCOctreeNode<TRecord>* node = 
        _locateNodeForStoring(rec->GetTreeNode(), rec, rec->GetValidDomain());

      if(node != rec->GetTreeNode()) 
      {
        rec->GetTreeNode()->GetList().erase(rec->GetTreeIterator());
        _storeInNode(node,rec);
        _numMoveInOctree++;
      }

    }
  }
}

// --------------------------------------------------------------------------
//  CRCOctree::DecreaseRadius()
// --------------------------------------------------------------------------
template<class TRecord>
void
CRCOctree<TRecord>::DecreaseRadius(TRecord *rec, float newR, bool updateOctree)
{
  if( newR < rec->GetR() )  {
    rec->SetR(newR);
    _numDecreaseR ++;
    if(updateOctree) {
      CRCOctreeNode<TRecord>* node = 
        _locateNodeForStoring(rec->GetTreeNode(), rec, rec->GetValidDomain());
      if(node != rec->GetTreeNode()) 
      {
        rec->GetTreeNode()->GetList().erase(rec->GetTreeIterator());
        _storeInNode(node,rec);
        _numMoveInOctree++;
      }
    }
  }
}

// --------------------------------------------------------------------------
//  CRCOctree::_reallocateStack()
// --------------------------------------------------------------------------
template<class TRecord>
void
CRCOctree<TRecord>::_reallocateStack(unsigned depth)
{
  assert(depth>0);
  
  delete [] _searchOctreeStack;
  _searchOctreeStack = new CSearchOctreeRecord[8*depth+1];
  _stackDepth = depth;
}

// --------------------------------------------------------------------------
//  CRCOctree::FindNearest()
// --------------------------------------------------------------------------
template<class TRecord>
const TRecord*
CRCOctree<TRecord>::FindNearest(const CVector3D& P, 
                       const CVector3D& N, 
                       const float maxDist) const
{
  float sqrtMinDist = 
    maxDist>0 ? maxDist : 1.74*GetRootNode()->GetSideLength();
  float minDist = sqr(sqrtMinDist);
  const TRecord *closestRec=0;
  const CRCOctreeNode<TRecord> *son;

  _searchOctreeStack[0].Set(GetRootNode(),-1);
  int i, stackEnd=1;
 
  while(stackEnd>0) {  
    // pop from the stack
    CSearchOctreeRecord       &top  = _searchOctreeStack[stackEnd-1];
    const CRCOctreeNode<TRecord>    *node = top.node;
    // index of the child that contais P
    int                     myIndex = top.idx;

    // first visit of this node ?
    if(myIndex==-1) {
      const CVector3D      &boxOrigin = node->GetBoxOrigin();
      const float        &sideLength = node->GetSideLength();

      // reject if no possible candidates
      if (P.x < boxOrigin.x - sqrtMinDist ) {stackEnd--;continue;}
      if (P.x > boxOrigin.x + sideLength + sqrtMinDist) {stackEnd--;continue;}
      if (P.y < boxOrigin.y - sqrtMinDist ) {stackEnd--;continue;}
      if (P.y > boxOrigin.y + sideLength + sqrtMinDist) {stackEnd--;continue;}
      if (P.z < boxOrigin.z - sqrtMinDist ) {stackEnd--;continue;}
      if (P.z > boxOrigin.z + sideLength + sqrtMinDist) {stackEnd--;continue;}

      if( !node->IsLeaf()) {
        float sideLength2 = sideLength * 0.5;
        myIndex     = 0;
        // determine index of the child in which P is located
        if(P.x > boxOrigin.x + sideLength2) myIndex |= (1<<0);
        if(P.y > boxOrigin.y + sideLength2) myIndex |= (1<<1);
        if(P.z > boxOrigin.z + sideLength2) myIndex |= (1<<2);
        top.idx = myIndex;

        // recurse into the closest descendant
        if (son = node->GetSon(myIndex) ) {
          _searchOctreeStack[stackEnd++].Set(son,-1);
          continue;
        }
      }
    }

    // second visit or no children

    // stack pop
    stackEnd --;

    // run through the list of the current node
    const CRecordList<TRecord> &nodeList = node->GetList();
    typename CRecordList<TRecord>::const_iterator it;
    for (it = nodeList.begin(); it!=nodeList.end(); it++) {
      const TRecord *rec = *it;
      // accept only records facing in a similar direction
      float dist = SqrMagnitude(P - rec->GetPoint());
      if(dist<minDist) {
        if( DotProd(N, rec->GetNormal()) > TRecord::maxAllowedDot ) {
          minDist     = dist;
          sqrtMinDist = sqrtf(minDist);
          closestRec  = rec;
        }
      }      
    } // for
   
    if (node->IsLeaf())
      continue;
    
    // for all descendants
    for (i = 0; i < myIndex; i++) {
      if (son = node->GetSon(i) )
        _searchOctreeStack[stackEnd++].Set(son,-1);
    }
    for (i = myIndex+1; i < 8; i++) {
      if (son = node->GetSon(i) )
        _searchOctreeStack[stackEnd++].Set(son,-1);
    }
  }

  return closestRec;
}

// --------------------------------------------------------------------------
//  CRCOctree::FindNearestFromNodeRecursive()
// --------------------------------------------------------------------------
template<class TRecord>
void
CRCOctree<TRecord>::
FindNearestFromNodeRecursive(const CVector3D& P, 
                             const CVector3D& N,
                             const CRCOctreeNode<TRecord> *node,
                             const TRecord *&closestRec,
                             float &sqrtMinDist) const
{
    const CVector3D  &boxOrigin   = node->GetBoxOrigin();
    const float      sideLength  = node->GetSideLength();

    // reject if no possible candidates
    if (P.x < boxOrigin.x - sqrtMinDist ) return;
    if (P.x > boxOrigin.x + sideLength + sqrtMinDist) return;
    if (P.y < boxOrigin.y - sqrtMinDist ) return;
    if (P.y > boxOrigin.y + sideLength + sqrtMinDist) return;
    if (P.z < boxOrigin.z - sqrtMinDist ) return;
    if (P.z > boxOrigin.z + sideLength + sqrtMinDist) return;

    int myIndex = 0, i;
    const CRCOctreeNode<TRecord> *son;
    if (!node->IsLeaf()) {
      // determine index of the child containig P
      float sideLength2 = sideLength*0.5;
      if(P.x > boxOrigin.x + sideLength2) myIndex |= (1<<0);
      if(P.y > boxOrigin.y + sideLength2) myIndex |= (1<<1);
      if(P.z > boxOrigin.z + sideLength2) myIndex |= (1<<2);
      // recurse into the child containig P
      if (son = node->GetSon(myIndex) )
        FindNearestFromNodeRecursive(P,N,son,closestRec,sqrtMinDist);
    }

    // run throug the list of the current node
    float minDist = sqr(sqrtMinDist);
    const CRecordList<TRecord> &nodeList = node->GetList();
    typename CRecordList<TRecord>::const_iterator it;
    for (it = nodeList.begin(); it!=nodeList.end(); it++) {
      const TRecord *rec = *it;
      // accept only records facing in a similar direction
      float dist = SqrMagnitude(P - rec->GetPoint());
      if(dist<minDist) {
        if( DotProd(N, rec->GetNormal()) > TRecord::maxAllowedDot ) {
          minDist     = dist;
          sqrtMinDist = sqrtf(minDist);
          closestRec  = rec;
        }
      }      
    } // for

    // recurse into the remaining descendants
    if (!node->IsLeaf()) {
      for (i = 0; i < myIndex; i++) {
        if (son = node->GetSon(i) )
          FindNearestFromNodeRecursive(P,N,son,closestRec,sqrtMinDist);
      }
      for (i = myIndex+1; i < 8; i++) {
        if (son = node->GetSon(i) )
          FindNearestFromNodeRecursive(P,N,son,closestRec,sqrtMinDist);
      }
    }
}

// --------------------------------------------------------------------------
//  CRCOctree::FindInRange()
// --------------------------------------------------------------------------
template<class TRecord>
int
CRCOctree<TRecord>::FindInRange(const CVector3D& P, 
                              const CVector3D& N,
                              float R,
                              growvector<const TRecord*> &carray) const
{
  // testing all records
  float R2 = R*R;
  int   nfound=0;

  const CRCOctreeNode<TRecord> *son;
  _searchOctreeStack[0].Set(GetRootNode());
  int i, stackEnd=1;
 
  while(stackEnd>0) {
    stackEnd--;

    // pop from the stack
    CSearchOctreeRecord       &top  = _searchOctreeStack[stackEnd];
    const CRCOctreeNode<TRecord>    *node = top.node;
    const CVector3D      &boxOrigin = node->GetBoxOrigin();
    const float        &sideLength = node->GetSideLength();

    // reject if no possible candidates
    if (P.x < boxOrigin.x - R ) continue;
    if (P.x > boxOrigin.x + sideLength + R) continue;
    if (P.y < boxOrigin.y - R ) continue;
    if (P.y > boxOrigin.y + sideLength + R) continue;
    if (P.z < boxOrigin.z - R ) continue;
    if (P.z > boxOrigin.z + sideLength + R) continue;

    // run through the list of the current node
    const CRecordList<TRecord> &nodeList = node->GetList();
    typename CRecordList<TRecord>::const_iterator it;
    for(it=nodeList.begin(); it!=nodeList.end(); it++) {
      const TRecord *rec = *it;
      // accept only records facing in a similar direction
      if( DotProd(N, rec->GetNormal()) > TRecord::maxAllowedDot ) {
        if(SqrMagnitude(P - rec->GetPoint()) < R2) {
          carray.push_back(rec);
          nfound++;
        }
      }
    }
   
    if (node->IsLeaf())
      continue;
    
    // for all descendants
    for (i = 0; i < 8; i++) {
      if (son = node->GetSon(i) )
        _searchOctreeStack[stackEnd++].Set(son);
    }
  }

  return nfound;
}

// --------------------------------------------------------------------------
//  CRCOctree::MulApproxError()
// --------------------------------------------------------------------------
template<class TRecord>
void 
CRCOctree<TRecord>::MulApproxError(float m)
{
  typename CRecordList<TRecord>::const_iterator it;
  for (it = _linList.begin(); it != _linList.end(); it++) {
    TRecord *rec = *it;
    assert(rec);
    rec->SetA( rec->GetA()*m );
  }
}

// --------------------------------------------------------------------------
//  CRCOctree::FindInInfluenceArea()
// --------------------------------------------------------------------------
template<class TRecord>
template<bool useStoredA>
unsigned
CRCOctree<TRecord>::FindInInfluenceArea(const CVector3D  &P,
                                      const CVector3D  &N,
                                      const float     Radd,
                                      CRecordContribArray<TRecord> &carray,
                                      const float     a)const
{
  unsigned nFound=0;
 
  _searchOctreeStack[0].Set(GetRootNode());
  int stackEnd=1;
  
  // contribution item
  CRecordContrib<TRecord> item;
  item.SetEInvalid();
  
  while(stackEnd>0) {
    stackEnd--;
    // pop from the stack
    CRCOctreeNode<TRecord>* node = 
      (CRCOctreeNode<TRecord>*)_searchOctreeStack[stackEnd].node;

    const CVector3D  boxOrigin   = node->GetBoxOrigin();
    float           sideLength  = node->GetSideLength();
    
    // run throug the list of the current node
    CRecordList<TRecord> &nodeList = node->GetList();
    typename CRecordList<TRecord>::const_iterator iter = nodeList.begin();
    for ( ; iter != nodeList.end(); iter++) {
      item.rec = *iter;
      if( DotProd(N,item.rec->GetNormal()) < TRecord::maxAllowedDot ) continue;
      float maxDist = (useStoredA?item.rec->GetA():a)*item.rec->GetR() + Radd;
      if( SqrDistance(P,item.rec->GetPoint()) < sqr(maxDist) ) {
        // record the clock state with the contribution
        item.changeClock = item.rec->GetChangeClock();
        // store the contribution
        carray.push_back(item);
        nFound++;
      }     
    } // for
    
    if (node->IsLeaf())
      continue;
    
    // for all descendants
    const CRCOctreeNode<TRecord> *son;
    CVector3D boxOriginSon;
    sideLength *= 0.5;

#define _findInInfluenceArea_HELPER(j) \
    { boxOriginSon[j] = boxOrigin[j];  \
    if (1<<j & i) boxOriginSon[j] += sideLength;  \
    if (P[j] < boxOriginSon[j] - _octreeScale * sideLength - Radd) \
	   goto next_son;  \
    if (P[j] > boxOriginSon[j] + _octreeScalePlusOne * sideLength + Radd) \
     goto next_son; }

    for (int i = 0; i < 8; i++) {
      if (son = node->GetSon(i) ) {
        _findInInfluenceArea_HELPER(0);
        _findInInfluenceArea_HELPER(1);
        _findInInfluenceArea_HELPER(2);
        _searchOctreeStack[stackEnd++].Set(son);      
      }
next_son: ;
    } 
  }

  // end of tree traversal
  return nFound;
}

template unsigned CRCOctree<CIrradianceRecord>::FindInInfluenceArea<true>(
  const CVector3D &P, const CVector3D &N, const float Radd, 
  CICContribArray &carray, const float a)const;
template unsigned CRCOctree<CIrradianceRecord>::FindInInfluenceArea<false>(
  const CVector3D &P, const CVector3D &N, const float Radd, 
  CICContribArray &carray, const float a)const;

template unsigned CRCOctree<CRadianceRecord>::FindInInfluenceArea<true>(
  const CVector3D &P, const CVector3D &N, const float Radd, 
  CRCContribArray &carray, const float a)const;
template unsigned CRCOctree<CRadianceRecord>::FindInInfluenceArea<false>(
  const CVector3D &P, const CVector3D &N, const float Radd, 
  CRCContribArray &carray, const float a)const;

template class CRCOctree<CIrradianceRecord>;
template class CRCOctree<CRadianceRecord>;

__END_GOLEM_SOURCE
