// ===================================================================
// $Id: rcoctree.h,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// rcoctree.h
//     Header file for the radiance and irradiance cache octree.
//
// G. J. Ward, F. M. Rubinstein, R. D. Clear: "A ray tracing solution for
// diffuse interreflection" Computer graphics,  Vol 22, Num 4, August 1988.
// G. J. Ward, P. S. Heckbert "Irradiance gradients"
//
//
// Licence: the use and distribution of this file is severely limited, please
// see the file 'doc/Licence.txt'. Any non-authorized use can be prosecuted under
// International Law. For further questions, please, e-mail to VHavran@seznam.cz
// or mail to Vlastimil Havran, Pohodli 27, 57001 Litomysl, the Czech Republic.
// REPLACEMENT_STRING
//
// Initial coding by Jaroslav Krivanek (Oct 2000).
// Based on the implementation in Minimal Rendering Toolkit

#ifndef __RCOCTREE_H__
#define __RCOCTREE_H__

// GOLEM headers
#include "configh.h"
#include "aabox.h"
#include "color.h"
#include "growvec.h"
#include "icontrib.h"
#include "icrecord.h"

// standard headers
#include <list>

__BEGIN_GOLEM_HEADER

// forward declarations

/****************************************************************************
*
*              CRCOctreeNode class definition
*
****************************************************************************/

/// This class represents one node of the (ir)radiance cache octree.
template<class TRecord>
class CRCOctreeNode
{
public:
  // friend class CRCOctree;
  
  // Constructor resets the children.
  CRCOctreeNode (
    const float& sideLength, 
    const CVector3D& boxOrigin,
    int depth)
    : _sideLength(sideLength), _boxOrigin(boxOrigin), _depth(depth)
  {
    for (int i=0; i<8; i++) _children[i] = 0;
    _isLeaf = true;
  }
  
  /// Destructor also destructs all descendants.
  ~CRCOctreeNode();
  
  /// Returns the const i-th son.
  const CRCOctreeNode* GetSon(int i) const { return _children[i]; }
  
  /// Returns the i-th son. If it does not exist, it is created.
  CRCOctreeNode* GetSonAlloc(int i,float sideLen,const CVector3D& boxOrig)
  {
    if (_children[i] == 0) {
      _children[i] = new CRCOctreeNode(sideLen,boxOrig,_depth+1);
      _isLeaf = false;
    }
    return _children[i];
  }
  
  /// Returns the list of (ir)radiance entries in this node.
  const CRecordList<TRecord>& GetList() const { return _list;}
  /// Returns the list of (ir)radiance entries in this node.
  CRecordList<TRecord>& GetList() { return _list;} 
  /// Is this node leaf
  bool IsLeaf () const { return _isLeaf; }
  /// Return the side length of the octree node.
  const float& GetSideLength() const { return _sideLength; }
  /// Return the side length of the octree node.
  const CVector3D& GetBoxOrigin() const { return _boxOrigin; }
  /// return the depth of the octree node in the tree
  int GetDepth() const { return _depth; }
  
private:
  /// list of records in this node
  CRecordList<TRecord>    _list;
  /// pointer to the array containing pointers to eight sons
  CRCOctreeNode*    _children[8];
  /// set if the node is leaf
  bool              _isLeaf;
  /// side length of the octree node
  float            _sideLength;
  /// lower left corner of the octree node box
  CVector3D         _boxOrigin;
  /// depth of the octree node
  int               _depth;
};


/****************************************************************************
*
*                    CIrrOctreeNode  inline methods
*
****************************************************************************/

template<class TRecord>
inline 
CRCOctreeNode<TRecord>::~CRCOctreeNode()
{
  // delete sons
  if(!IsLeaf()) 
    for(int i = 8; i--;)
      delete _children[i];

  // delete (ir)radiance data
  typename CRecordList<TRecord>::const_iterator it;
  for(it=_list.begin(); it!=_list.end(); it++)
    delete *it;
}

/// Octree for (ir)radiance cache
template<class TRecord>
class CRCOctree
{
public:
  /// Constructor
  CRCOctree (
    const CAABox &sceneBox,
    float octreeScale);
  
  virtual ~CRCOctree ();
  
  /// Insert new record in the cache.
  /// Set octree location to newRecID if non-null.
  void InsertRecord(TRecord *rec);
  
  /// DEBUG: check if all cache entries lie in the correctly sized node 
  /// w.r.t. their size
  void  ValidateOctree() const;
  
  /// Get records' contributions at a given point.
  /**
     @param useStoredA (template parameter)
       If true, the value of approximation error is used for weight evaluation.
       If false, the value 'a' passed to the function will be used for weight 
       evaluation.
     @param jitterWeight (template parameter)
       If true, weight evaluation will be jittered, otherwise, weight will
       be deterministic.
     @param P (in) 
       Query point.
     @param N (in) 
       Surface (shading) normal at P.
     @param carray (out) 
       List of records that contribute to the query point.
     @param a (in) 
       Approximation error to use. 
       
     @return 'false' if no usable record is found in the cache
  */
  template<bool useStoredA,bool jitterWeight>
    bool GetContributions(
    const CVector3D  &P,
    const CVector3D  &N,
    CRecordContribArray<TRecord> &carray,
    const float     a) const;

  /// Find the record nearest to a given point.
  /**
     A record is considered only if the dot product of its normal with a given
     normal is greater than CIrradianceRecord::maxAllowedDot.
     Only records closer than maxDist from P are considered.

     @return The nearest record. May return NULL if nothing is found.
    */
  const TRecord* FindNearest(
    const CVector3D& P, 
    const CVector3D& N,
    const float maxDist=0) const;

  /// The same as FindNearest() but using recursion instead of stack.
  inline const TRecord* FindNearestRecursive(
    const CVector3D& P, 
    const CVector3D& N,
    const float maxDist=0) const 
  {
    const TRecord* rec=0;
    float sqrtMinDist = 
      maxDist>0 ? maxDist : 1.74*GetRootNode()->GetSideLength();
    FindNearestFromNodeRecursive(P,N,GetRootNode(),rec,sqrtMinDist);
    return rec;
  }

  /// The recursive part of FindNearestRecursive().
  void FindNearestFromNodeRecursive(
    const CVector3D& P, 
    const CVector3D& N,
    const CRCOctreeNode<TRecord> *node,
    const TRecord *&closestRec,
    float &sqrtMinDist) const;

  /// Find records within a given radius from the given point.
  /**
     A record is considered only if the dot product of its 
     normal with a given normal is greater than zero.
     @return The number of records found. The records are pushed back to the
      array 'carray' (the array is not cleared at the beginning).
    */
  int FindInRange(
    const CVector3D& P, 
    const CVector3D& N,
    float R,
    growvector<const TRecord*> &carray) const;

  /// Find all records having a given point in their influence area.
  /** 
    This query finds exactly the records ri = {Pi, Ni, Ri, ai} for which:
     a)  N . Ni <= CIrradianceRecord::maxAllowedDot
     b) ||P-Pi|| <= ai*Ri + Radd
   The records identification is stored in the contribution array carray.
   The fields CICContrib::E and CICContrib::w stay unfilled.
   */
  template<bool useStoredA>
    unsigned FindInInfluenceArea(
    const CVector3D  &P,
    const CVector3D  &N,
    const float     Radd,
    CRecordContribArray<TRecord> &carray,
    const float     a)const;

  /// Decreases the approximation error of a record and re-inserts the 
  /// record in the octree.
  void DecreaseApproxError(TRecord *rec, float newA, bool updateOctree);

  /// Decreases the radius of a record and re-inserts the 
  /// record in the octree.
  void DecreaseRadius(TRecord *rec, float newA, bool updateOctree);

  /// Sets all statistics to zero values.
  void ResetStats(void);
  
  /// Returns the number of irradiance records in the cache.
  unsigned long GetNumRecords() const { return _numRecords;}
  /// Returns the number of successful queries.
  unsigned long GetNumQueriesSuc() const { return _numQueriesSuc; }
  /// Returns the number of unsuccessful queries.
  unsigned long GetNumQueriesFail() const { return _numQueriesFail; }
  /// Returns the number of weight calculations with zero response.
  unsigned long GetNumWeightZero() const { return _numWeightZero; }
  /// Returns the number of weight calculations with nonzero response.
  unsigned long GetNumWeightNonzero() const { return _numWeightNonzero; }

  /// number of time a record's approx. error has been decreased
  unsigned long GetNumDecreaseA() const { return _numDecreaseA; }
  /// number of time a record's radius has been decreased
  unsigned long GetNumDecreaseR() const { return _numDecreaseR; }
  /// number of times a record hes been moved in octree
  unsigned long GetNumMoveInOctree() const { return _numMoveInOctree; }
  
  /// Return the list of all cached records
  const CRecordList<TRecord>& GetRecordList() const { return _linList; }
  CRecordList<TRecord>& GetRecordList() { return _linList; }
 
  /// return the root node of the octree
  const CRCOctreeNode<TRecord>* GetRootNode() const {return _root;}
  CRCOctreeNode<TRecord>* GetRootNode() {return _root;}

  /// multiply the value of 'a' of all records by the given number
  /**
     Notice!!! The octree is not altered by this change. 
     Use only for small changes of approximation error.
   */
  void MulApproxError(float m);

protected:
 
  /// Locate the octree node to store the record to, starting from 'startNode'.
  inline CRCOctreeNode<TRecord>* _locateNodeForStoring(
    CRCOctreeNode<TRecord> *node,
    TRecord *rec,
    const float     validDomain);

  /// Store irradiance record in the given octree node.
  inline void _storeInNode(CRCOctreeNode<TRecord> *node, TRecord *rec);
  
  /// DEBUG: check if all cache entries lie in the correctly sized node w.r.t. 
  /// their size
  void  _validateOctree(const CRCOctreeNode<TRecord> *node, float sideLength) const;
 
protected: // data
  
  /// bounding box of the root node
  CAABox          _box;
  /// bounding box sidelength
  float           _boxSideLen;
  /// scaling factor for octree (ratio between node size and size of stored 
  /// records)
  float           _octreeScale;
  float           _octreeScaleDiv2;
  float           _octreeScalePlusOne;
  /// root of the octree
  CRCOctreeNode<TRecord> *_root;
  /// linear list of records contais the same entries as the octree
  CRecordList<TRecord>   _linList;

  /// number of irradiance records in the cache
  unsigned long _numRecords;
  /// octree depth
  int _octreeDepth;
  /// number of successful queries
  mutable unsigned long _numQueriesSuc;
  /// number of unsuccessful queries
  mutable unsigned long _numQueriesFail;
  /// number of weight calculations with zero response
  mutable unsigned long _numWeightZero;
  /// number of weight calculations with nonzero response
  mutable unsigned long _numWeightNonzero;
  /// number of time a record's weight has been decreased
  mutable unsigned long _numDecreaseA;
  /// number of time a record's radius has been decreased
  mutable unsigned long _numDecreaseR;
  /// number of times a record hes been moved in octree
  mutable unsigned long _numMoveInOctree;

  /// record of the stack for simulation of recursion in _searchOctree
  struct CSearchOctreeRecord {   
    const CRCOctreeNode<TRecord>  *node;
    /// index of the nearest child for FindNearest
    int                      idx;

    void Set(const CRCOctreeNode<TRecord> *n) { node=n; }
    void Set(const CRCOctreeNode<TRecord> *n, int i) { node=n; idx=i; }
  };
  
  /// stack for simulation of recursion in _searchOctree
  CSearchOctreeRecord *_searchOctreeStack;
  
  /// depth, for which the stack capacity is sufficient
  int _stackDepth;
  
  /// reallocate '_searchOctreeStack' to a given depth
  void _reallocateStack(unsigned depth);

private:

};

__END_GOLEM_HEADER

#endif //  __RCOCTREE_H__
