// ===================================================================
// $Id: rt17app.cpp,v 1.6 2005/11/02 12:58:51 xkrivanj Exp $
//
// rt17app.cpp
//    View dependent radiosity with irradiance caching.
//     * Primary rays from camera.
//     * Irradiance caching for diffuse reflection.
//     * All other terms ignored.
//      
// Initial coding by Jaroslav Krivanek, November 2003

// GOLEM headers
#include "config.h"
#include "basmacr.h"
#include "dirlish.h"
#include "environ.h"
#include "icalgo.h"
#include "rt17app.h"
#include "updanim.h"
#include "world.h"

// standard headers
#include <cstdlib>

__BEGIN_GOLEM_SOURCE

// forward definitions
class CWorld;

// --------------------------------------------------------------------------
//  CRT_17_App::CRT_17_App()
// --------------------------------------------------------------------------
CRT_17_App::CRT_17_App():
  CRT_1_App()
{
  // set all data pointers to NULL (important!)
  _irrCacheAlgo           = 0;
  _presampling            = false;
}

// --------------------------------------------------------------------------
//  CRT_17_App::~CRT_17_App()
// --------------------------------------------------------------------------
CRT_17_App::~CRT_17_App()
{
  CleanUp();
}

// --------------------------------------------------------------------------
//  CRT_17_App::Init()
// --------------------------------------------------------------------------
bool
CRT_17_App::Init(CWorld *world, string &outputName)
{
  if( InitApplication(world,outputName) )
    return true;
  if( InitMovie(outputName) )
    return true;
  if( InitFrame(outputName, 0) )
    return true;
  
  return false;
}


// --------------------------------------------------------------------------
//  CRT_17_App::InitApplication()
// --------------------------------------------------------------------------
bool 
CRT_17_App::InitApplication(CWorld *world, string &outputName)
{
  // init previous classes
  if (CRT_1_App::InitApplication(world, outputName))
    return true; // error occured in super

  // No transparency for shadow rays
  if (_directLighting)
    _directLighting->SetShadowRaysTransparency(false);
  
  return false;
}


// --------------------------------------------------------------------------
//  CRT_17_App::InitMovie()
// --------------------------------------------------------------------------
bool 
CRT_17_App::InitMovie(string &outputName)
{
  if (CRT_1_App::InitMovie(outputName)) {
    return true; // error
  }

  // -------------------------------------------
  //  initialize irradiance caching
  // -------------------------------------------
  if(!_irrCacheAlgo && _depthLimit>0) {
    bool use_ic;
    _environment->GetBool("IrradianceCache.use", use_ic);
    if(use_ic) {
      _irrCacheAlgo = 
        new CICAlgo(*_environment, _depthLimit, this, _asds->BBox() );
      assert(_irrCacheAlgo);
    }
  }

  return false; // ok
}


// --------------------------------------------------------------------------
//  CRT_17_App::InitFrame()
// --------------------------------------------------------------------------
bool 
CRT_17_App::InitFrame(string          &outputName,
		      CUpdateAnimData *updateAnimData)
{
  if (CRT_1_App::InitFrame(outputName, updateAnimData)) {
    return true; // error
  }

  // Check if the camera was moved
  bool cameraAnimated = true;
  if (updateAnimData)
    cameraAnimated = updateAnimData->cameraMoved;

  // The image must exist
  assert(_image);

  // open the file for new image
  if (AllocateFileIO(outputName.c_str()))
    return true; // error occured

  // assign the file to the image
  _image->SetFileIO(_file);
  _image->SetFileName(outputName);

  // reinitialize camera
  if (cameraAnimated) {
    if( InitCamera() == true ) return true;
  }

  // rebuild the ray shooting structure
  if ( _asds == 0 ) {
    if ( _asdsInit() == true )
      return true;
  }

  return false;
}


// --------------------------------------------------------------------------
//  CRT_17_App::CleanUpFrame()
// --------------------------------------------------------------------------
bool 
CRT_17_App::CleanUpFrame()
{
  // this should be called if abjects are animated
  // SAFE_DELETE(_asds);

  if (CRT_1_App::CleanUpFrame())
    return true; // error

  return false;
}

// --------------------------------------------------------------------------
//  CRT_17_App::CleanUpMovie()
// --------------------------------------------------------------------------
bool 
CRT_17_App::CleanUpMovie()
{
  // delete (ir)radiance caches
  SAFE_DELETE(_irrCacheAlgo);

  if (CRT_1_App::CleanUpMovie()) {
    return true; // error
  }

  return false; // ok
}

// --------------------------------------------------------------------------
//  CRT_17_App::CleanUpApplication()
// --------------------------------------------------------------------------
bool 
CRT_17_App::CleanUpApplication()
{
  // the clean up predecessor
  return CRT_1_App::CleanUpApplication();
}

// --------------------------------------------------------------------------
//  CRT_17_App::CleanUp()
// --------------------------------------------------------------------------
bool
CRT_17_App::CleanUp()
{
  // make some cleaning of the local data
  if( CleanUpFrame() ) return true;
  if( CleanUpMovie() ) return true;
  if( CleanUpApplication() ) return true;

  return false;
}

// --------------------------------------------------------------------------
//  CRT_17_App::NeedImagePresampling()
// --------------------------------------------------------------------------
bool 
CRT_17_App::NeedImagePresampling() const
{
  // presampling is needed only if irradiance caching is on
  return _irrCacheAlgo!=0;
}

// --------------------------------------------------------------------------
//  CRT_17_App::StartingPresampleImage()
// --------------------------------------------------------------------------
void
CRT_17_App::StartingPresampleImage()
{
  SAFE_CALL( _irrCacheAlgo, SetPresampleMode() );
}

// --------------------------------------------------------------------------
//  CRT_17_App::PresampleImageFinished()
// --------------------------------------------------------------------------
void 
CRT_17_App::PresampleImageFinished()
{
  if (_irrCacheAlgo) {
    _irrCacheAlgo->SetRenderMode();
    // @@@ zrusit
    // _irrCacheAlgo->GetICache(0)->RandomizeIrradiances();
  }
}

// --------------------------------------------------------------------------
//  CRT_17_App::Run()
// --------------------------------------------------------------------------
bool
CRT_17_App::Run()
{  
  STATUS << "Rendering ... irradiance chaching (Ward '88)" << endl;

  // compute the image itself, the common part
  if ( RunCommon() == true )
    return true; // error

  // our stuff goes here

  // save one scanline of the image in textual form (readable by gnuplot)
  int scanlineToSave = -1;
  _environment->GetInt("RT17.saveScanline", scanlineToSave);
  if (scanlineToSave>=0 && _image->GetHeight()-scanlineToSave-1 >=0 ) {
    FILE *fout;
    char buff[100];
    std::string s,txtImgName,envName;
    _environment->GetString("RT17.saveScanlineName", envName);
    ChangeFilenameExtension(world->outputFileName, "txt", s);
    sprintf(buff,"-scanline%d", scanlineToSave);
    PrefixExtensionByName(s, envName+buff, txtImgName);
    if ( (fout = fopen(txtImgName.c_str(), "wt"))==NULL ) {
      WARNING << "Could not open '" << txtImgName << "' for writing." 
        << std::endl;
    }
    if( _image->SaveImageScanlineTxt(_image->GetHeight()-scanlineToSave-1, 
      CValues::EE_Luminance, fout, "\n") == true ) 
    {
      WARNING << "Could not save image scanline '" << txtImgName
        << "'." << std::endl;
    }
    fclose(fout);
  }

  return false;
}


// --------------------------------------------------------------------------
//  CRT_17_App::ComputeColor()
// --------------------------------------------------------------------------
void
CRT_17_App::ComputeColor(const CRay    &ray, 
                         CHitPointInfo &hitInfo,
                         CValues       &color, 
                         bool         /*saveForReuse*/)
{
  if (hitInfo.GetObject() != NULL) 
  {
    INC_HIT_COUNTER(info.GetObject());
    INC_PRIMARY_HIT;

    // save the footprint area for this pixel (will be used for all recursion levels)
    hitInfo.Extrap();
    hitInfo.GetNormal();
    _currPixelFootprintArea = GetFootprintArea(hitInfo);

#if 0
    // Visualize the distance from polygon edge.
    EObjectType type = hitInfo.GetObject()->ID();
    switch(type) {
      case EE_Triangle:
      case EE_TriangleInt:
      case EE_Polygon:
      case EE_PolygonInt:
        const CPolygon *polygon = (const CPolygon*)hitInfo.GetObject();
        float d2 = polygon->SqrDistFromEdge(hitInfo.GetPoint());
        color.Set(sqrt(d2));
        return;
    };
#endif

    // get the item to assign the primary ray and info
    _cnts[0].Reset();
    CContributionItem *citem = _cnts[0].AllocateNext();
    // info must be copied
    _cnts[0].hitInfo = hitInfo;
    
    // the initial color and weigth format for the primary rays
    citem->weigth.SetRepresentation(_defaultWeightType);
    citem->color.SetRepresentation(_defaultColorType);
    citem->weigth.Set(1.0);
    citem->diffuseness = 0.0;

    if(_presampling) {
      // fill the irradiance cache
      _shadeObjectPresample(*citem);
    }
    else {
      // perform the actual shading
      ShadeObject2(*citem, 0);
    }
    // copy out the computed pixel color
    color = citem->color;   
    return;
  }
  else 
  {
    INC_PRIMARY_MISS;
    if(!_presampling)
      _scene->background->GetColor(ray, color);
    return;
  }
}

// --------------------------------------------------------------------------
//  CRT_17_App::_shadeObjectPresample()
// --------------------------------------------------------------------------
void
CRT_17_App::_shadeObjectPresample(CContributionItem &contrib)
{
  contrib.color.setz();

  // depth is always 0 in this function
  if(_depthLimit<1) return;

  // -----------------------------------------------------------
  //  SET UP THIS RECURSION LEVEL
  // -----------------------------------------------------------
  UpdateCummulativeVars(contrib,0);
  // This is the array used for all contributions of higher order.
  CContributionArray &carray = _cnts[1];
  carray.Reset();
  CHitPointInfo      &nexthitInfo = carray.hitInfo;
  // Get hitInfo for the incoming ray and compute the point of intersection.
  CHitPointInfo &hitInfo = _cnts[0].hitInfo;
  hitInfo.Extrap();
  hitInfo.GetNormal(); 
  // Get object's surface material.
  CAppearance *apr=(CAppearance*)hitInfo.GetObject()->GetAttr(indexAppearance);
  CSurface    *surf = apr->GetSurface();
  int surfaceType = surf->GetSupportedTypes();
    
  // -------------------------------------------------------------
  //  IRRADIANCE CACHE
  // -------------------------------------------------------------
  if( (surfaceType & EE_REFLECTED_DIFFUSE) && _irrCacheAlgo ) {
    _irrCacheAlgo -> 
    IndirectDiffuseLighting(hitInfo, _currPixelFootprintArea, 
    surf, 0, nexthitInfo, carray.value2);
  }
}

// --------------------------------------------------------------------------
//  CRT_17_App::ShadeObject2()
// --------------------------------------------------------------------------

// This function assumes that a "cur" contains a valid ContributionItem with
// a ray hitting the object and that _cnts[depth].hitInfo corresponds to that
// ray. This _cnts[depth].hitInfo is not changed during the computation.
// Function computes the energy on the point of incidence according to lights
// and then cast the rays (reflected/refracted) to compute further contribution
// form secondary rays, but these are stored at _cnts[depth+1]. When such
// rays hit the objects, then the function goes into recursion, until a limit
// for maximum recursion depth is reached.
void
CRT_17_App::ShadeObject2(CContributionItem &contrib, int depth)
{
#ifdef _DEBUG
  static int maxDepth=-1;
  if(depth>maxDepth) {
    maxDepth=depth;
    DEBUG << "maximal recursion depth: " << maxDepth << endl;
  }
#endif

  // --------------------------------------------------------------------
  //  SET UP THIS RECURSION LEVEL
  // --------------------------------------------------------------------

  UpdateCummulativeVars(contrib,depth);
  
  // This is the array used for all contributions of higher order.
  CContributionArray &carray = _cnts[depth+1];
  carray.Reset();
  CHitPointInfo      &nexthitInfo = carray.hitInfo;

  // Get hitInfo for the incoming ray and compute the point of intersection.
  CHitPointInfo &hitInfo = _cnts[depth].hitInfo;
  hitInfo.Extrap();
  // compute the normal
  hitInfo.GetNormal();

  // Get object's surface material.
  CAppearance *apr=
    (CAppearance*)hitInfo.GetObject()->GetAttr(indexAppearance);
  CSurface    *surf = apr->GetSurface();

  // Reset the resulting color. All further steps will add to this value.
  contrib.color.setz();

  // get the surface type
  int surfaceType = surf->GetSupportedTypes();
   
  // -------------------------------------------------------------------
  // SELF EMITTED RADIANCE
  // -------------------------------------------------------------------
  AddObjectSelfRadiance(contrib, hitInfo);

  // -------------------------------------------------------------------
  // DIRECT LIGHTING
  // -------------------------------------------------------------------
  if ( _directLight && (depth >= _firstLevelForDirectLights) )
    DirectLights(hitInfo, surf, carray, depth);

  // -------------------------------------------------------------------
  // AMBIENT COMPONENT
  // -------------------------------------------------------------------
  if (surf->HasAmbientTerm() && _scene->ambientLight) {
    _scene->ambientLight->GetIllumination(carray, hitInfo, _samplingInfo);
    surf->GetAmbientTerm(carray.GetLastItem()->weigth, hitInfo);
  }

  // -------------------------------------------------------------------
  // SUM CONTRIBUTIONS
  // -------------------------------------------------------------------
  carray.WeightAndSum(contrib.color);

  // --------------------------------------------------------------------
  // DIFFUSE REFLECTION
  // --------------------------------------------------------------------
  if(_irrCacheAlgo &&  depth < _depthLimit &&
	  surf->GetSupportedTypes() & EE_REFLECTED_DIFFUSE ) {
#if 1
	  if( _irrCacheAlgo -> IndirectDiffuseLighting(hitInfo, _currPixelFootprintArea, 
      surf, depth, contrib.color, nexthitInfo, carray.value2) )
      contrib.color += carray.value2;
#else
	  if( _irrCacheAlgo -> Irradiance(hitInfo, _currPixelFootprintArea, 
      depth, contrib.color, nexthitInfo, carray.value2) )
      contrib.color += carray.value2;
#endif
  }


#if 0
    // Visualize the distance from polygon edge.
    EObjectType type = hitInfo.GetObject()->ID();
    switch(type) {
      case EE_Triangle:
      case EE_TriangleInt:
      case EE_Polygon:
      case EE_PolygonInt:
        const CPolygon *polygon = (const CPolygon*)hitInfo.GetObject();
        float d2 = polygon->SqrDistFromEdge(hitInfo.GetPoint());
        contrib.color *= 100*sqrt(d2);
    };
#endif


  return;
}

// --------------------------------------------------------------------------
//  CRT_17_App::IncidentRadiance()
// --------------------------------------------------------------------------
bool 
CRT_17_App::IncidentRadiance(CHitPointInfo      &hitInfo,
	                           CContributionItem  *contrib,
	                           const int           depth,
                             const float        footprintArea)
{
  // prepare for ray-cast
  hitInfo.SetForFindNearest(contrib->ray);
  // cast the ray
  if (_asds->FindNearest(contrib->ray, hitInfo)) {
    _currPixelFootprintArea  = footprintArea;
    // now compute the energy from the spawned ray (virtual call)
    ShadeObject2(*contrib, depth+1);
    // ok - object hit
    return true;
  }

  // no object hit, use background color for higher order rays
  BgColor(*contrib);
  // no object hit
  return false;
}


// --------------------------------------------------------------------------
//  CRT_17_App::DoReport()
// --------------------------------------------------------------------------
bool
CRT_17_App::DoReport(const CWorld &world, ostream &app,
		    CStatistics::EType reportType)
{
  // first provide the report of predecessor
  if (CRT_1_App::DoReport(world, app, reportType))
    return true; // some error

  app << "#---------------------------------------------------------\n";
  app << "#                  CRT_17_App - REPORT                  \n";
  app << "#    (diffuse interreflections with irradiance caching)   \n";
  app << "#---------------------------------------------------------\n";

  if(_irrCacheAlgo)
	  _irrCacheAlgo->DoReport(app);
  else
	  app << "| No irradiance caches                                   |\n";

  app << "#---------------------------------------------------------\n";

  return false; // ok
}

// --------------------------------------------------------------------------
//  CRT_17_App::SetPreferences()
// --------------------------------------------------------------------------
void
CRT_17_App::SetPreferences(CEnvironment &env)
{
  CRT_1_App::SetPreferences(env);

  bool detRefl;
  env.GetBool("Shading.phongDeterministicReflection", detRefl);
  
  if (detRefl)
    env.SetString("Surface.typePhong", "physicalDR");
  else
    env.SetString("Surface.typePhong", "physical");

  return;
}

// --------------------------------------------------------------------------
//  AllocRT_17_App()
// --------------------------------------------------------------------------
extern CSynthesisApp*  AllocRT_17_App()
{
  return new CRT_17_App();
}

__END_GOLEM_SOURCE

