// ===================================================================
// $Id: rt17app.h,v 1.6 2005/11/02 12:58:46 xkrivanj Exp $
//
// rt17app.h
//     Header file for the renderer using irradiance caching.
//
// Class: CRT_17_App
//
// Initial coding by Jaroslav Krivanek, October 2003

#ifndef __RT17APP_H__
#define __RT17APP_H__

// GOLEM headers
#include "configh.h"
#include "rt1app.h"
#include "incrad.h"

__BEGIN_GOLEM_HEADER

// Forward declarations
class CICAlgo;

/// application allocation
extern CSynthesisApp*  AllocRT_17_App();

/// An example application for irradiance caching.
/**
  This application is a demo for irradiance caching and showcases 
  the architecture used to implement it. All construction of
  irradiance caching related functionality is encapsulated
  in the constructor CICAlgo::CICAlgo().
  It sets up the irradiance caching according to the environment
  (CEnvironment - the Irradiance.* parameters). 
  The only noteworthy parameter passed to this
  constructor is a pointer to an implementation of the
  CIncidentRadiance interface. The CIncidentRadiance::IncidentRadiance()
  method of this interface is used by irradiance caching to compute the 
  incoming radiance if a new record is added to the cache. In our solution,
  the CIncidentRadiance is implemented by the application CRT_17_App
  itself: the call to CIncidentRadiance::IncidentRadiance() sends a
  sampling ray and if it hits something, the call is forwarded to
  the ShadeObject2() method (the recursive ray tracing method).
*/
class CRT_17_App:
  public CRT_1_App, public CIncidentRadiance
{
protected: // methods

  /// Compute the color for a primary ray.
  /**
    This is called by the camera.
    The primary ray/scene intersection query has already been
    executed before calling this routine. The info about the intersection
    is passed in hitInfo parameter.

    Our implementation calls either _shadeObjectPresample() or
    ShadeObject2() method, depending whether we're presampling the image
    (to fill the irradiance cache) or actually computing the image colors.
   */
  virtual void ComputeColor(
    const CRay    &ray, 
    CHitPointInfo &hitInfo,
    CValues       &color, 
    bool         /*saveForReuse*/);

  /// Shading method for the presampling pass.
  /**
     This method calls the irradiance cache to compute the
     indirect lighting with the sole purpose to fill the cache.
     The color returned in contrib is always zero, however.
   */
  void  _shadeObjectPresample(CContributionItem &contrib);

  /// Shading method for the rendering pass.
  virtual void ShadeObject2(CContributionItem &cur, int depth);

  virtual void StartingPresampleImage();
  virtual void PresampleImageFinished();
  /// Returns true if irradiance caching is on, indicating that image
  /// presampling is desirable.
  virtual bool NeedImagePresampling() const;

public:

  /// Default constructor sets everything to zero, no construction.
  /**
    All pointers to internal data are set to zero here.    
    The actual construction takes place in the InitApplication(), 
    InitMovie() or InitFrame() methods. Those methods test the pointers,
    and perform the construction only if they are NULL. It ts therefore
    essential to set them to zero in the constructor.
   */
  CRT_17_App();

  /// Virtual constructor calls the 'CleanUp()' method
  virtual ~CRT_17_App();

  /// Dispatching init method. 
  /**
    This method calls InitApplication() InitMovie() and InitFrame().
    Do not put any initialization code in here - this method will be 
    abolished. Use only InitApplication() InitMovie() and InitFrame()
    to initialize your data.
   */
  virtual bool Init(CWorld *world, string &outputName);

  /// @name Initialization functions (e.g. allocates irrad. caches)
  /* Use these functions to place you initialization code.

    For some reason, GOLEM might call those method twice before the rendering
    starts. Therefore, care must be taken not to allocate the internal 
    structures twice.
    Typically, one would set all internal pointers to NULL in the 
    constructor CRT_17_App(). In the Init() method, the construction 
    of a structure takes place if and only if the poiner to the 
    structure is still zero.
   */
  //@{
  /// Initialization called once for the application.
  virtual bool InitApplication(CWorld *world, string &outputName);
  /// Initialization called once before each video sequence.
  virtual bool InitMovie(string &outputName);
  /// Initialization called once before each frame.
  virtual bool InitFrame(string &outputName,CUpdateAnimData *updateAnimdata);
  //@}

  /// Dispatching deallocation function.
  /**
    This method calls CleanUpApplication() CleanUpMovie() and CleanUpFrame().
    Do not put any initialization code in here - this method will be 
    abolished. Uve only InitApplication() InitMovie() and InitFrame()
    to initialize your data.
   */
  virtual bool CleanUp();

   /// @name Data deallocation methods.
  /**
     Similarly to the Init*() methods, those methods can be called twice. 
     You HAVE to set all deallocated pointers to NULL and avoid their 
     deallocation the second time. To this end, use SAFE_DELETE() and 
     SAFE_DELETE_ARRAY() macros instead of delete and delete [] 
     operators, respectively.
   */
  //@{
  /// Deallocation called once before the application ends.
  virtual bool CleanUpApplication();
  /// Deallocation called once after finishing rendering of a video sequence.
  virtual bool CleanUpMovie();
  /// Deallocation called onec after finishing rendering of a single frame.
  virtual bool CleanUpFrame();
  //@}

  /// The main function for this app
  virtual bool Run();
  
  /// returns the application type (EE_DiffuseIrrCache)
  virtual EApplicationType GetApplicationType() const { return EE_DiffuseIrrCache; }
  
  /// reports the statistics result of the specific application
  virtual bool DoReport(const CWorld &world, ostream &app,
			CStatistics::EType reportType);

  void SetPreferences(CEnvironment &env);
  
  /**
     Utility function to compute radiance visible in ray's direction.

     Inherited from CIncidentRadiance class. Used for irradiance caching.

     @param hitInfo    Hit info to be used for ray cast.
     @param contrib    Save the color to contrib.color.
     @param currDepth  Current recursion level.

     @return 'true' if object was hit, 'false' otherwise. Even if the return value is 
             'false' there might be (and usually is) a contribution from the background.
  */
  inline virtual bool IncidentRadiance(
    CHitPointInfo      &hitInfo,
    CContributionItem  *contrib,
    const int           depth,
    const float        footprintArea);

protected: // data

  /// Encapsulated irradiance caching algorithm.
  CICAlgo *_irrCacheAlgo;

  /// footprint area for the pixel being currently processed
  float _currPixelFootprintArea;
};

__END_GOLEM_HEADER

#endif // __RT17APP_H__
