// ===================================================================

// $Id: rt19app.cpp,v 1.10 2005/11/30 15:46:45 havran Exp $

//

// rt19app.cpp

//    Radiance caching renderer.

//     * Primary rays from camera.

//     * Deterministic rays for perfect specular reflections/refractions.

//     * Irradiance caching for diffuse reflections.

//     * Radiance caching for low-frequency glossy reflections.

//     * MC importance sampling for highe-frequncy glossy reflections.

//      

// Initial coding by Jaroslav Krivanek, December 2003



// GOLEM headers

#include "config.h"

#include "brdfman.h"

#include "brdfwrap.h"

#include "dirlish.h"

#include "environ.h"

#include "icalgo.h"

#include "rcalgo.h"

#include "rt19app.h"

#include "updanim.h"

#include "world.h"



// This ordering is required for some 3rd party library

#include "fileiotp.h"

#include "colors.h"

#include "tonemap.h"



// SPHAR headers

#include "sheval.h"

#include "shbasis.h"

#include "shbrdf.h"



// standard headers

#include <cstdlib>

#include <set>

#include <cmath>

#include <cfloat>



// include purely diffuse surfaces in radiance caching?

// #define RCACHE_DIFFUSE



__BEGIN_GOLEM_SOURCE



// VH !!!! - isfinite does now work for ICC7.1 compiler 2005-11-25

//static inline bool _weird(double f) { return !isfinite(f) || isnan(f); }

static inline bool _weird(double f) { return !_finite(f) || _isnan(f); }

// static inline bool _weird(double f) { return !isnan(f); }



static inline bool _weird(const CVector3D &v) {

  return _weird(v.x) || _weird(v.y) || _weird(v.z);

}



/// Fill the images with per-texel BRDF representation error and number 

/// of coefficients.

/**

  Uniquely for presentation purposes (for the paper) and debugging.

  Returns true on error (one of the images is not big enough); false if ok.

  If the image pointers is null, it is not manipulated.

*/

bool 

FillStatsImages(const CSphBRDF<float,float>& sphBrdf, CImage *errImage, CImage *numImage)

{

  CValues errVal, numVal;

  float error;

  int   order, sqrOrder;

  int xRes = sphBrdf.GetXRes();

  int yRes = sphBrdf.GetYRes();

  

  if(errImage) {

    if(errImage->GetWidth() < xRes+1) return true;

    if(errImage->GetHeight() < yRes+1)  return true;

    errVal.SetRepresentation(errImage->GetStandardValue()->GetType());

  }

  if(numImage) {

    if(numImage->GetWidth() < xRes+1) return true;

    if(numImage->GetHeight() < yRes+1)  return true;

    numVal.SetRepresentation(numImage->GetStandardValue()->GetType());

  }

  

  for(int x=0; x <= xRes; x++) {	

    for(int y=0; y <= yRes; y++) {

      const CBRDFRec<float> *rec = sphBrdf.GetCoeffsIdx(x,y);

    

      error    = rec->GetError();

      order    = rec->GetOrder();

      sqrOrder = sqr(order);

      if(sphBrdf.GetAdaptiveDeletion() && error > sphBrdf.GetMaxAllowedError() ) {

        // no error, since radiance caching is not used for this dir

        error    = 0;  

        sqrOrder = 0;

      }

      

      if(errImage) { errVal.Set(error);  errImage->SetValue(x,y,errVal); }

      if(numImage) { numVal.Set(sqrOrder);  numImage->SetValue(x,y,numVal); }

    }

  }

  

  return false;

}



// --------------------------------------------------------------------------

//  CRT_19_App::CRT_19_App()

// --------------------------------------------------------------------------

CRT_19_App::CRT_19_App():

CRT_1_App()

{

  _icAlgo           = 0;

  _rcAlgo           = 0;

  _reflGlossyBounces      = 0;

  _reflGlossyBouncesRC    = 0;

  _presampleWithOverview  = 0;

  _rcDepthLimit           = 0;

  _glossyRays             = 0;

  _brdfRepMan             = 0;

}



CRT_19_App::~CRT_19_App()

{

  CleanUp();

}



// --------------------------------------------------------------------------

//  CRT_19_App::Run()

// --------------------------------------------------------------------------

// This ray traces the image using distributed ray tracing algorithm

bool

CRT_19_App::Run()

{  

  STATUS << "Rendering ... radiance chaching" << endl;

  

  for(int i=0;i<=_depthLimit;i++)

    _cnts[i].value2.SetRepresentation(_defaultColorType);

  

#if 0

  // generate image for all cameras

  unsigned ccc = _scene->cameras.size();

  for(unsigned cidx=0; cidx<ccc; cidx++) {

    _environment->SetInt("SetCamera.cameraIndex", 0);

    DEBUG << "Camera #" << cidx << std::endl;

    

    if( CRT_Core_App::InitCamera() ) { 

      FATAL << "InitCamera() failed." << std::endl; 

      return true; 

    }

    

    if ( RunCommon() ) {

      FATAL << "RunCommon() failed." << std::endl;

      return true;

    }

    

    DoReportRT19(DEBUGW);

  }

  return false; // ok

#else

  return RunCommon();

#endif

}



// --------------------------------------------------------------------------

//  CRT_19_App::_createBRDFFileName()

// --------------------------------------------------------------------------

string 

CRT_19_App::_createBRDFFileName(CSphBasis<float,float>::EType basisType)

{

  string ext, fname;



  switch(basisType) {

		case CSphBasis<float,float>::ESH:

      ext = "sh-brdf";

      break;

    case CSphBasis<float,float>::EHSH:

      ext = "hsh-brdf";

      break;

    default:

      FATAL << "Unrecognized spherical basis type." << std::endl;

      FATAL_EXIT;

  }

  

  ChangeFilenameExtension(world->outputFileName, ext, fname);

  

  return fname;

}



// --------------------------------------------------------------------------

//  CRT_19_App::_loadBRDFSHRep()

// --------------------------------------------------------------------------

bool 

CRT_19_App::_loadBRDFSHRep(CSphBasis<float,float>::EType basisType)

{

  string errReason;

  string s;

  char s1[2000];

  long sceneFileLength = 0;

  long sceneFileParity = 0;

  long sceneFileCRC = 0;

  long numSurfaces = 0;

  CValues c(world->_scene->stdValues.GetDefaultColorType());

  int fchannels = -1;

  int forder    = -1, eorder;

  int fxres = -1, exres;

  int	fyres   = -1, eyres;

  int fnumsamples = -1, enumsamples;

  float fmaxerror = -1, emaxerror;

  

  s = _createBRDFFileName(basisType);

  FILE *fin = fopen(s.c_str(), "rt");

  if(!fin) {

    errReason = "can't open input file";

    goto err_return;

  }

  

  fscanf(fin, " %s %ld %ld %ld %ld ", 

    s1,

    &sceneFileLength,

    &sceneFileParity,

    &sceneFileCRC,

    &numSurfaces);

  

  if(string(s1) != world->outputFileName) {

    errReason = "different output file name";

    goto close_err_return;

  }

  

  if(sceneFileLength != world->_scene->sceneFileLength) {

    errReason = "different scene file length";

    goto close_err_return;

  }

  

  if(sceneFileParity != world->_scene->sceneFileParity) {

    errReason = "different scene file parity";

    goto close_err_return;

  }

  

  if(sceneFileCRC != world->_scene->sceneFileCRC ) {

    errReason = "different scene file CRC";

    goto close_err_return;

  }

  

  if(numSurfaces != world->_scene->surfaces.size()) {

    errReason = "different number of scene BRDFs";

    goto close_err_return;

  }

  

  fscanf(fin, " %d %d %d %d %d %f ", 

    &fchannels,

    &forder,

    &fxres,

    &fyres,

    &fnumsamples,

    &fmaxerror);

  

  world->_environment.GetInt("RadianceCache.BRDF.xRes", exres);

  world->_environment.GetInt("RadianceCache.BRDF.yRes", eyres);

  world->_environment.GetInt("RadianceCache.BRDF.numSamples", enumsamples);

  world->_environment.GetInt("RadianceCache.order", eorder );

  world->_environment.GetFloat("RadianceCache.BRDF.maxError", emaxerror );

  

  if(c.GetCount() != fchannels) {

    errReason = "different number of channels";

    goto close_err_return;

  }

  if(forder != eorder) {

    errReason = "different order";

    goto close_err_return;

  }

  if(fxres != exres) {

    errReason = "different x resolution";

    goto close_err_return;

  }

  if(fyres != eyres) {

    errReason = "different y resolution";

    goto close_err_return;

  }

  if(fnumsamples != enumsamples) {

    errReason = "different number of brdf samples";

    goto close_err_return;

  }

  if(fmaxerror != emaxerror) {

    errReason = "different maximum error";

    goto close_err_return;

  }

  

  

  bool adaptiveDeletion;

  world->_environment.

    GetBool("RadianceCache.BRDF.adaptiveDeletion", adaptiveDeletion);

  

  // header was ok, let's load the data

  {

    int i=0;

    for(CSurfaceList::iterator siter = world->_scene->surfaces.begin() ; 

    siter != world->_scene->surfaces.end(); siter++, i++) {

      CSurface *s = *siter;

      if(!s) continue;

#ifndef RCACHE_DIFFUSE

      if(!(s->GetSupportedTypes() & EE_REFLECTED_GLOSSY))

        continue;

#endif

      

      // check the type

      int type= -1;

      fscanf(fin, " %d ", &type);

      if(type != (int)s->GetType()) {

        errReason = "different BRDF type for CSurface # i";

        goto close_err_return;

      }



      // if some representation exist, replace it

      if( s->GetSHRep() ) {

        delete s->GetSHRep();

        (*siter)->SetSHRep(0);

      }

      

      CSphBRDF<float,float> *shRep = new CSphBRDF<float,float>(fin, adaptiveDeletion);

      if(shRep && shRep->OK())

        s->SetSHRep(shRep);

      else {

        errReason = "could not load BRDF coefficients";

        goto close_err_return;

      }

    }

  }

  

  fclose(fin);

  return false; // ok

  

close_err_return:

  fclose(fin);

err_return:

  STATUS 

    << " - could not load BRDF rep. from file (" 

    << errReason << ")." << std::endl;

  return true;

}



// --------------------------------------------------------------------------

//  CRT_19_App::_saveBRDFSHRep()

// --------------------------------------------------------------------------

bool 

CRT_19_App::_saveBRDFSHRep(CSphBasis<float,float>::EType basisType)

{

  // write the file header

  string errReason;

  string s;

  long sceneFileLength = 0;

  long sceneFileParity = 0;

  long sceneFileCRC = 0;

  long numSurfaces = 0;

  CValues c(world->_scene->stdValues.GetDefaultColorType());

  

  s = _createBRDFFileName(basisType);

  FILE *fout = fopen(s.c_str(), "wt");

  if(!fout) {

    errReason = "can't open output file";

    goto err_return;

  }

  

  if ( fprintf(fout, "%s %ld %ld %ld %ld \n", 

    world->outputFileName.c_str(), 

    world->_scene->sceneFileLength, 

    world->_scene->sceneFileParity,

    world->_scene->sceneFileCRC,

    world->_scene->surfaces.size() ) == EOF ) {

    errReason = "can't write header";

    goto close_err_return;

  }

  

  int exres, eyres, eorder, enumsamples;

  float emaxerror;

  world->_environment.GetInt("RadianceCache.BRDF.xRes",   exres  );

  world->_environment.GetInt("RadianceCache.BRDF.yRes",     eyres    );

  world->_environment.GetInt("RadianceCache.BRDF.numSamples", enumsamples);

  world->_environment.GetInt("RadianceCache.order",           eorder     );

  world->_environment.GetFloat("RadianceCache.BRDF.maxError",  emaxerror  );

  

  if ( fprintf(fout, "%d %d %d %d %d %f \n", 

    c.GetCount(),

    eorder,

    exres,

    eyres,

    enumsamples,

    emaxerror) == EOF ) {

    errReason = "can't write second part of header";

    goto close_err_return;

  }

  

  // header written, let's write the data

  {

    CSurfaceList::const_iterator siter = world->_scene->surfaces.begin();

    

    for( ; siter != world->_scene->surfaces.end(); siter++ ) {

      const CSurface* s = *siter;

      if(!s) continue;

      // const CSHBRDF *shRep = s->GetSHRep();

      const CSphBRDF<float,float> *shRep = s->GetSHRep();

      if(!shRep) continue;

      

      if( fprintf(fout, " %d \n", (int)s->GetType() ) == EOF) {

        errReason = "can't write BRDF data";

        goto close_err_return;

      }

      shRep->SaveToFile(fout);

    }

  }

  

  fclose(fout);

  return false; // ok

  

close_err_return:

  fclose(fout);

err_return:

  STATUS 

    << " - could not save BRDF rep. to file (" 

    << errReason << ")." << std::endl;

  return true; // error

}









// --------------------------------------------------------------------------

//  CRT_19_App::_writeStatImages()

// --------------------------------------------------------------------------

void 

CRT_19_App::_writeStatImages(CSphBRDF<float,float> *shRep,

                             int  i,

                             bool adaptiveOrder)

{

  assert(shRep);

  float maxError = shRep->GetMaxAllowedError();

  int   order    = shRep->GetStats().maxOrder;

  int   xRes     = shRep->GetXRes();

  int   yRes     = shRep->GetYRes();



  string basisTypeStr;

  world->_environment.GetString("RadianceCache.basisType",basisTypeStr);



  CValues stdValue(CValues::EE_Weigth1);

  char numstr[20]; sprintf(numstr,"%d",i);

  string errImgName, numImgName, adapStr;

  adapStr = adaptiveOrder ? "-adap-" : "-nonadap-";

  string errImgNBase = string("-errimage-") + basisTypeStr + adapStr + numstr;

  string numImgNBase = string("-numimage-") + basisTypeStr + adapStr + numstr;

  string s;

  ChangeFilenameExtension(world->outputFileName, "tiff", s);

  PrefixExtensionByName(s, errImgNBase, errImgName);

  PrefixExtensionByName(s, numImgNBase, numImgName);

  CTIFFIO errImgIO (errImgName.c_str());

  CTIFFIO numImgIO (numImgName.c_str());

  CImageReal errImage(xRes+1, yRes+1, 1, &errImgIO, &stdValue);

  CImageReal numImage(xRes+1, yRes+1, 1, &numImgIO, &stdValue);

  FillStatsImages(*shRep,&errImage, &numImage);

  

  vector<CValues> map;

  map.push_back(Black);  map.push_back(Blue);   map.push_back(Cyan);

  map.push_back(Green);  map.push_back(Yellow); map.push_back(Red);

  

  CValues stdValueRemap(CValues::EE_Color3);

  CImageReal remapErrImg(xRes+1, yRes+1, 3,&errImgIO,&stdValueRemap);

  CImageReal remapNumImg(xRes+1, yRes+1, 3,&numImgIO,&stdValueRemap);

  CValues dummy(-1,-1,-1);

  

  CToneMapping::ColorCodeImage(remapErrImg, errImage, 1e-12, maxError, map,

    White, 

    map[map.size()-1], dummy, dummy, 

    CValues::EE_Channel0);

  CToneMapping::ColorCodeImage(remapNumImg, numImage, 1e-12, sqr(order), map, 

    White, 

    map[map.size()-1], dummy, dummy, 

    CValues::EE_Channel0);

  

  if (!remapErrImg.SaveImage()) {

    WARNING 

      << "Could not write error image for surface # " << i << "." 

      << std::endl;

  }

  if (!remapNumImg.SaveImage()) {

    WARNING 

      << "Could not write order image for surface # " << i << "." 

      << std::endl;

  }

  

  // write in txt form

  ChangeFilenameExtension(world->outputFileName, "txt", s);

  PrefixExtensionByName(s, errImgNBase, errImgName);

  PrefixExtensionByName(s, numImgNBase, numImgName);

  int scanline = (yRes+1)/2;

  FILE *fout;

  // error image

  if ( (fout = fopen(errImgName.c_str(), "wt"))==NULL ) {

    WARNING << "Could not open '" << errImgName << "' for writing." 

      << std::endl;

  }

  else {

    if( errImage.SaveImageScanlineTxt(scanline, CValues::EE_Channel0, fout, 

      "\n") == true ) 

    {

      WARNING << "Could not write error image '" << errImgName 

        << "' for surface # " << i << "." << std::endl;

    }

    fclose(fout);

  }

  

  // (H)SH order image

  if ( (fout = fopen(numImgName.c_str(), "wt"))==NULL ) {

    WARNING << "Could not open '" << numImgName << "' for writing." 

      << std::endl;

  }

  else {

    if( numImage.SaveImageScanlineTxt(scanline, CValues::EE_Channel0, fout,

      "\n") == true ) 

    {

      WARNING << "Could not write order image '" << numImgName 

        << "' for surface # " << i << "." << std::endl;

    }

    fclose(fout);

  }     

}





// --------------------------------------------------------------------------

//  CRT_19_App::_computeBRDFSHRep()

// --------------------------------------------------------------------------

bool 

CRT_19_App::_computeBRDFSHRep(const CSphBasis<float,float> *shBasis)

{

  assert(shBasis);



  CSurfaceList::iterator siter = world->_scene->surfaces.begin();

  int i=0;

  

  int   xRes  = -1;

  int   yRes    = -1;

  int   order     = -1;

  int   numSamples= -1;

  int   channels  = -1;

  bool  adaptiveOrder    = false;

  bool  adaptiveDeletion = false;

  bool  brdfImageStats   = false;

  float maxError  = -1;

  

  world->_environment.GetInt("RadianceCache.BRDF.xRes", xRes);

  world->_environment.GetInt("RadianceCache.BRDF.yRes", yRes);

  world->_environment.GetInt("RadianceCache.BRDF.numSamples", numSamples);

  world->_environment.GetInt("RadianceCache.order", order );

  CValues c(world->_scene->stdValues.GetDefaultColorType());

  channels = c.GetCount();

  

  world->_environment.

    GetFloat("RadianceCache.BRDF.maxError", maxError);

  world->_environment.

    GetBool("RadianceCache.BRDF.adaptiveOrder", adaptiveOrder);

  world->_environment.

    GetBool("RadianceCache.BRDF.adaptiveDeletion", adaptiveDeletion);

  world->_environment.

    GetBool("RadianceCache.BRDF.brdfImageStats", brdfImageStats);

  

  if(xRes<1)       xRes       = 1;

  if(yRes<1)       yRes       = 1;

  if(order<1)      order      = 1;

  if(numSamples<1) numSamples = 1;

  if(maxError>1)   maxError   = 1;

  if(maxError<0)   maxError   = 0;

  

  STATUS << "   - precomputing directions...";

  

  // we want to have an analytic (not approximate tabulated) evaluation

  // for brdf approximation

  CSphBasisEvalCache<float,float>

    cache(numSamples, order, shBasis->GetAnalytic() );

  STATUS << "done." << std::endl;

  

  CTimer timer;

  

  for( ; siter != world->_scene->surfaces.end(); siter++, i++ ) {

    CSurface* s = *siter;

    if(!s) continue;

#ifndef RCACHE_DIFFUSE

    if(!(s->GetSupportedTypes() & EE_REFLECTED_GLOSSY))

      continue;

#endif



    // if some representation exist, replace it

    if( s->GetSHRep() ) {

      delete s->GetSHRep();

      (*siter)->SetSHRep(0);

    }

    

    STATUS << "   - converting surface # " << i << "\n";



    s->Describe(STATUSW,4);

    

    timer.Reset();

    timer.Start();

    

    CBRDFSurfaceWrapper wrapper(s);

    

    CSphBRDF<float,float> *shRep = new CSphBRDF<float,float>(

      wrapper,           // brdf

      channels,          // num of color channels

      order,             // max allowd sh-expansion order

      xRes,              // texture resolution

      yRes,              // texture resolution

      maxError,          // max. allowd error

      adaptiveOrder,     // adaptive length of sh-expansion

      adaptiveDeletion,  // delete sh-evectors with too high error

      cache);            // precomputed directions

    

    assert(shRep);

    

    s->SetSHRep(shRep);

    

    if(brdfImageStats)

      _writeStatImages(shRep,i,adaptiveOrder);

       

    timer.Stop();

    STATUS << "done in " << timer.UserTime() << " secs." << std::endl;

  }

  

  return false;

}





// --------------------------------------------------------------------------

//  CRT_19_App::_cleanUpBRDFSHRep()

// --------------------------------------------------------------------------

bool 

CRT_19_App::_cleanUpBRDFSHRep()

{

  if( world!=0 && world->_scene!=0 ) {

    CSurfaceList::iterator siter = world->_scene->surfaces.begin();

    for( ; siter != world->_scene->surfaces.end(); siter++) {

      CSurface* s = *siter;

      assert(s);

      CSphBRDF<float,float>* shRep = (*siter)->GetSHRep();

      delete shRep;

      (*siter)->SetSHRep(0);

    }

  }

  return false;

}





// --------------------------------------------------------------------------

//  CRT_19_App::_setupBRDFSHRep()

// --------------------------------------------------------------------------

bool 

CRT_19_App::_setupBRDFSHRep(const CSphBasis<float,float> *shBasis) 

{

  assert(shBasis);

  CTimer timer;



  timer.Reset();

  

  bool loadBrdfRep=true;

  world->_environment.

    GetBool("RadianceCache.BRDF.loadBrdfRep", loadBrdfRep);



  if(loadBrdfRep) {

    STATUS << " - loading BRDF rep from file..." << std::endl;

    timer.Start();

    if ( _loadBRDFSHRep(shBasis->GetType()) == false) {

      timer.Stop();

      STATUS << " - done in " << timer.UserTime() << " secs." << std::endl;

      return false;

    }

  }

  else {

    STATUS << " - skipping BRDF loading (RadianceCache.BRDF.loadBrdfRep=OFF)" << std::endl;

  }

  

  STATUS << " - precomputing BRDF representation..." << std::endl;

  timer.Start();

  if ( _computeBRDFSHRep(shBasis) ) {

    STATUS << " - failed!!!" << std::endl;

    return true;

  }

  timer.Stop();

  STATUS 

    << " - BRDF precomputation done in " << timer.UserTime() << " secs." 

    << std::endl;

  

  bool saveBrdfRep=true;

  world->_environment.

    GetBool("RadianceCache.BRDF.saveBrdfRep", saveBrdfRep);



  if(saveBrdfRep) {

    STATUS << " - saving BRDF rep to file..." << std::endl;

    timer.Reset();

    timer.Start();

    if( _saveBRDFSHRep(shBasis->GetType()) ) {

      STATUS << " - failed!!!" << std::endl;		

      return true;

    }

    else {

      timer.Stop();

      STATUS << " - done in " << timer.UserTime() << " secs." << std::endl;

      return false;

    }

  }

  else {

    STATUS << " - skipping BRDF saving  (RadianceCache.BRDF.saveBrdfRep=OFF)" << std::endl;

    return false;

  }

}



/// String with modified coparison - prefix match means equal strings.

class CStringPrefix : public std::string {

public:

  CStringPrefix(const char *s) : std::string(s) {}

  CStringPrefix(const string &s) : std::string(s) {}

  int compare(const CStringPrefix& s) const {

    return strncmp(c_str(),s.c_str(),length());

  }

  bool operator<(const CStringPrefix& s) const { return compare(s)<0; }

  bool operator<=(const CStringPrefix& s) const { return compare(s)<=0; }

  bool operator>(const CStringPrefix& s) const { return compare(s)>0; }

  bool operator>=(const CStringPrefix& s) const { return compare(s)>=0; }

};



typedef std::set<CStringPrefix> TSurfTypeSet;

inline TSurfTypeSet operator += (TSurfTypeSet &list, const string& s) 

{

  list.insert(s);

  return list;

}



// --------------------------------------------------------------------------

//  CRT_19_App::_setupBRDFSHRep()

// --------------------------------------------------------------------------

bool 

CRT_19_App::_setupBrdfRepMan()

{

  _brdfRepMan = 0;



  if(!_rcAlgo)

    return true; // error



  int   xRes;

  int   yRes;

  int   numSamples;

  float maxError;

  bool  adaptiveOrder;

  bool  adaptiveDeletion;

  bool  brdfImageStats;

  bool  loadBrdfRep;

  bool  saveBrdfRep;

  string rcTypeAllowStr;

  string rcTypeDenyStr;



  // proxy for environment

  const CEnvironment &env = world->_environment;



  env.GetInt  ("RadianceCache.BRDF.xRes",            xRes);

  env.GetInt  ("RadianceCache.BRDF.yRes",            yRes);

  env.GetInt  ("RadianceCache.BRDF.numSamples",      numSamples); 

  env.GetFloat("RadianceCache.BRDF.maxError",        maxError);

  env.GetBool ("RadianceCache.BRDF.adaptiveOrder",   adaptiveOrder);

  env.GetBool ("RadianceCache.BRDF.adaptiveDeletion",adaptiveDeletion);

  env.GetBool ("RadianceCache.BRDF.brdfImageStats",  brdfImageStats);

  env.GetBool ("RadianceCache.BRDF.loadBrdfRep",     loadBrdfRep);

  env.GetBool ("RadianceCache.BRDF.saveBrdfRep",     saveBrdfRep);

  env.GetString("RadianceCache.BRDF.rcTypeAllow",    rcTypeAllowStr);

  env.GetString("RadianceCache.BRDF.rcTypeDeny",     rcTypeDenyStr);



  // The following are sets of surface types for radiance caching.

  // The semantics should be like in UNIX hostsAllow and hostsDeny

  TSurfTypeSet rcTypeAllow, rcTypeDeny;



  ParseStringRow(rcTypeAllowStr,rcTypeAllow);

  ParseStringRow(rcTypeDenyStr,rcTypeDeny);

 

  // @@@ <JK> i should finish  allow and deny, so that it can be used 

  // from the environment

  // rcTypeAllow.insert(CSurface::EE_BSDFLafortune);

  // rcTypeDeny.insert("CSurfacePhong");



  DEBUG << "rcTypeAllow\n";

  for(TSurfTypeSet::const_iterator ii=rcTypeAllow.begin(); ii!=rcTypeAllow.end(); ii++) {

    PRINTVAR(*ii);

  }

  DEBUG << "rcTypeDeny\n";

  for(TSurfTypeSet::const_iterator ii=rcTypeDeny.begin(); ii!=rcTypeDeny.end(); ii++) {

    PRINTVAR(*ii);

  }



  // create BRDF representation manager

  _brdfRepMan = new CBrdfRepManager(

    _rcAlgo->GetSphBasis(),

    _rcAlgo->GetNumChannels(), 

    _rcAlgo->GetOrder(),

    xRes, 

    yRes, 

    maxError, 

    adaptiveOrder, 

    adaptiveDeletion, 

    numSamples,

    loadBrdfRep, 

    saveBrdfRep);



  if(!_brdfRepMan)

    return true; // error



  // conver scene surfaces to SH via the BRDF rep manager

  int i=0;

  for(CSurfaceList::iterator siter = world->_scene->surfaces.begin() ; 

    siter != world->_scene->surfaces.end(); siter++, i++) 

  {

    CSurface *s = *siter;

    if(!s) continue;

#ifndef RCACHE_DIFFUSE

    if(!(s->GetSupportedTypes() & EE_REFLECTED_GLOSSY))

      continue;

#endif

    // handle rcTypeAllow and rcTypeDeny

    // CSurface::EType stype = s->GetType();

    string stype = s->GetSignature();

    if(rcTypeDeny.find(stype) != rcTypeDeny.end() )

      continue; // found in deny, no rc caching

    if(rcTypeAllow.size()>0 && rcTypeAllow.find(stype) == rcTypeAllow.end())

      continue; // allow is non empty, but it does not contain this type



    _brdfRepMan->SetUpBrdfRep(s);

  }



  return false; // ok

}



// --------------------------------------------------------------------------

//  CRT_19_App::_setupBRDFSHRep()

// --------------------------------------------------------------------------

bool 

CRT_19_App::_cleanUpBrdfRepMan()

{

  // delete the representation manager (which also deletes all cached BRDF

  // representations)

  SAFE_DELETE(_brdfRepMan);



  // set the BRDF representation pointers to zero

  if(world && world->_scene) {

    for(CSurfaceList::iterator siter = world->_scene->surfaces.begin() ; 

      siter != world->_scene->surfaces.end(); siter++) 

    {

      CSurface *s = *siter;

      if(s) s->SetSHRep(0);

    }

  }



  return false; // ok 

}



// --------------------------------------------------------------------------

//  CRT_19_App::Init()

// --------------------------------------------------------------------------

bool

CRT_19_App::Init(CWorld *world, string &outputName)

{

  if( InitApplication(world,outputName) )

    return true;

  if( InitMovie(outputName) )

    return true;

  if( InitFrame(outputName, 0) )

    return true;

  

  return false;

}





// --------------------------------------------------------------------------

//  CRT_19_App::InitApplication()

// --------------------------------------------------------------------------

bool 

CRT_19_App::InitApplication(CWorld *world, string &outputName)

{

  // init previous classes

  if (CRT_1_App::InitApplication(world, outputName))

    return true; // error occured in super



  // No transparency for shadow rays

  if (_directLighting)

    _directLighting->SetShadowRaysTransparency(false);

  

  return false;

}





// --------------------------------------------------------------------------

//  CRT_19_App::InitMovie()

// --------------------------------------------------------------------------

bool 

CRT_19_App::InitMovie(string &outputName)

{

  if (CRT_1_App::InitMovie(outputName)) {

    return true; // error

  }



  // -------------------------------------------

  //  initialize irradiance caching

  // -------------------------------------------

  if(!_icAlgo && _depthLimit>0) {

    bool use_ic;

    world->_environment.GetBool("IrradianceCache.use", use_ic);

    if(use_ic) {

      _icAlgo = 

        new CICAlgo(*_environment,_depthLimit, this, _asds->BBox() );

      assert(_icAlgo);

    }

  }

  

  // -------------------------------------------

  //  initialize radiance caching

  // -------------------------------------------



  // maximum recursion depth for using radiance cache

  _environment->GetInt("RadianceCache.depthLimit",_rcDepthLimit);

  if(_rcDepthLimit<0) _rcDepthLimit = 0;

  if(_rcDepthLimit>_depthLimit) _rcDepthLimit = _depthLimit;



  // @@@ adding new recursion levels would be more reasonable 

  // than rebuilding the whole cache

  if(_rcAlgo && (int)_rcAlgo->GetDepthLimit()<_rcDepthLimit) {

    SAFE_DELETE(_rcAlgo);

  }



  if(!_rcAlgo && _rcDepthLimit>0) {

    bool use_rc;

    _environment->GetBool("RadianceCache.use",use_rc);

    if(use_rc) {

      CValues c(_defaultColorType);

      _rcAlgo = new CRCAlgo(*_environment, _rcDepthLimit, this, 

        _asds->BBox(), c.GetCount() );

      assert(_rcAlgo);



      // -----------------------------------------------

      //  transform BRDF's to the (H)SH representation

      // -----------------------------------------------

      _brdfRepMan = 0;



      bool useBrdfRepMan;

      world->_environment.

        GetBool("RadianceCache.BRDF.useBrdfRepMan",useBrdfRepMan);

      bool brdfSetUpError;



      if(useBrdfRepMan)

        brdfSetUpError = _setupBrdfRepMan();

      else

        brdfSetUpError = _setupBRDFSHRep(_rcAlgo->GetSphBasis());



      if ( brdfSetUpError ) {

        FATAL 

          << "Could not setup the SH representation of scene BRDF's. " 

          << std::endl;

        return true;

      }



      // remove BRDF representation for those BRDFs, for which the maximum number

      // of errorneous directions is exceeded

      float maxErrorneousDirs;

      world->_environment.

        GetFloat("RadianceCache.BRDF.maxErrorneousDirs",maxErrorneousDirs);

      if(maxErrorneousDirs<1) {

        for(CSurfaceList::iterator siter = world->_scene->surfaces.begin() ; 

          siter != world->_scene->surfaces.end(); siter++) 

        {

          CSurface *s = *siter;

          if(!s) continue;

          CSphBRDF<float,float>* shRep = s->GetSHRep();

          if(!shRep) continue;

          int nErrDir = shRep->GetStats().numErrorDirections;

          int nDir    = shRep->GetStats().numDirections;

          if(maxErrorneousDirs*nDir < nErrDir) {

            // too many errorneous dirs

            // DEBUG << "removing BRDF representation" << std::endl;

            s->SetSHRep(0);

            // @@@ i should setup usage counters for BRDF reps, because this is

            // really dirty

            // delete if not managed by BrdfRepMan

            if(!useBrdfRepMan)

              delete shRep;

          }

        }

      }

    }

  }



  // this must be done - ShadeObject relies on this

  if (!_rcAlgo)

    _rcDepthLimit = 0;



  // number of glossy rays

  string s;

  _environment->GetString("RT10RC.glossyRays",s);

  if(!_glossyRays) _glossyRays = new int[Max(1,_depthLimit)];

  assert(_glossyRays);

  ParseIntegerRow(s,_glossyRays,_depthLimit);



  return false;

}





// --------------------------------------------------------------------------

//  CRT_19_App::InitFrame()

// --------------------------------------------------------------------------

bool 

CRT_19_App::InitFrame(string          &outputName,

		      CUpdateAnimData *updateAnimData)

{

  if (CRT_1_App::InitFrame(outputName, updateAnimData)) {

    return true; // error

  }





  // Check if the camera was moved

  bool cameraAnimated = true;

  if (updateAnimData)

    cameraAnimated = updateAnimData->cameraMoved;



  // The image must exist

  assert(_image);



  // open the file for new image

  if (AllocateFileIO(outputName.c_str()))

    return true; // error occured



  // assign the file to the image

  _image->SetFileIO(_file);

  _image->SetFileName(outputName);



  // reinitialize camera

  if (cameraAnimated) {

    if( InitCamera() == true ) return true;

  }



  // rebuild the ray shooting structure

  if ( _asds == 0 ) {

    if ( _asdsInit() == true )

      return true;

  }



  return false;

}





// --------------------------------------------------------------------------

//  CRT_19_App::CleanUpFrame()

// --------------------------------------------------------------------------

bool 

CRT_19_App::CleanUpFrame()

{

  // read number of records in radiance cache

  if ( _rcAlgo )

    _rcRecords.push_back( _rcAlgo->GetNumRecords() );



  // this should be called if abjects are animated

  // SAFE_DELETE(_asds);



  if (CRT_1_App::CleanUpFrame())

    return true; // error



  return false;

}



// --------------------------------------------------------------------------

//  CRT_19_App::CleanUpMovie()

// --------------------------------------------------------------------------

bool 

CRT_19_App::CleanUpMovie()

{

#if 1

  /// write the number of radiance cache samples to a file

  if(_rcAlgo) {

    FILE *f = fopen("rcrecords.txt","wt");

    if(f) {

      for(int i=0; i<(int)_rcRecords.size(); i++)

        fprintf(f,"%d\n",_rcRecords[i]);

      fclose(f);

    }

  }

#endif



  // delete (H)SH representation for scene BRDFs

  bool useBrdfRepMan;

  world->_environment.

    GetBool ("RadianceCache.BRDF.useBrdfRepMan",useBrdfRepMan);



  bool brdfCleanUpError;



  if(useBrdfRepMan)

    brdfCleanUpError = _cleanUpBrdfRepMan();

  else

    brdfCleanUpError = _cleanUpBRDFSHRep();



   if( brdfCleanUpError ) 

     return true; // error



  // delete (ir)radiance caches

  SAFE_DELETE(_icAlgo);

  SAFE_DELETE(_rcAlgo);



  SAFE_DELETE_ARRAY(_glossyRays);



  // ------------------------------



  if (CRT_1_App::CleanUpMovie()) {

    return true; // error

  }



  return false; // ok

}



// --------------------------------------------------------------------------

//  CRT_19_App::CleanUpApplication()

// --------------------------------------------------------------------------

bool 

CRT_19_App::CleanUpApplication()

{

  // the clean up predecessor

  return CRT_1_App::CleanUpApplication();

}



// --------------------------------------------------------------------------

//  CRT_19_App::CleanUp()

// --------------------------------------------------------------------------

bool

CRT_19_App::CleanUp()

{

  // make some cleaning of the local data

  if( CleanUpFrame() ) return true;

  if( CleanUpMovie() ) return true;

  if( CleanUpApplication() ) return true;



  return false;

}



// --------------------------------------------------------------------------

//  CRT_19_App::NeedImagePresampling()

// --------------------------------------------------------------------------

bool CRT_19_App::NeedImagePresampling() const

{

  // presampling is needed only if irradiance or radiance caching is on

  return _icAlgo!=0 || _rcAlgo!=0;

}



// --------------------------------------------------------------------------

//  CRT_19_App::StartingPresampleImage()

// --------------------------------------------------------------------------

void CRT_19_App::StartingPresampleImage()

{

  SAFE_CALL( _icAlgo, SetPresampleMode() );

  SAFE_CALL( _rcAlgo, SetPresampleMode() );

}



// --------------------------------------------------------------------------

//  CRT_19_App::PresampleImageFinished()

// --------------------------------------------------------------------------

void CRT_19_App::PresampleImageFinished()

{

  SAFE_CALL(_rcAlgo, SetRenderMode());

  SAFE_CALL(_icAlgo, SetRenderMode());

}



// --------------------------------------------------------------------------

//  CRT_19_App::ComputeColor()

// --------------------------------------------------------------------------

void

CRT_19_App::ComputeColor(const CRay &ray, CHitPointInfo &info,

			CValues &color, bool /*saveForReuse*/)

{

  if (info.GetObject() != NULL) 

  {

    INC_HIT_COUNTER(info.GetObject());

    INC_PRIMARY_HIT;



    // save the footprint area for this pixel (will be used for all recursion levels)

    info.Extrap();

    info.GetNormal();

    _currPixelFootprintArea = GetFootprintArea(info);



    // get the item to assign the primary ray and info

    _cnts[0].Reset();

    CContributionItem *citem = _cnts[0].AllocateNext();

    

    // info must be copied

    _cnts[0].hitInfo = info;

    

    // the initial color and weigth format for the primary rays

    citem->weigth.SetRepresentation(_defaultWeightType);

    citem->color.SetRepresentation(_defaultColorType);

    citem->weigth.Set(1.0);

    citem->diffuseness = 0.0;



    if(_presampling)

      _shadeObjectPresample(*citem);

    else

      ShadeObject2(*citem, 0);



    // copy out the computed pixel color

    color = citem->color;

    

    return;

  }

  else 

  {

    INC_PRIMARY_MISS;

    if(!_presampling)

      _scene->background->GetColor(ray, color);

    return;

  }

}



// --------------------------------------------------------------------------

//  CRT_19_App::_shadeObjectPresample()

// --------------------------------------------------------------------------

void

CRT_19_App::_shadeObjectPresample(CContributionItem &contrib)

{

  contrib.color.setz();



  // depth is always 0 in this function

  if(_depthLimit<1) return;



  // -----------------------------------------------------------

  //  SET UP THIS RECURSION LEVEL

  // -----------------------------------------------------------

  UpdateCummulativeVars(contrib,0);

  // This is the array used for all contributions of higher order.

  CContributionArray &carray = _cnts[1];

  carray.Reset();

  CHitPointInfo      &nexthitInfo = carray.hitInfo;

  // Get hitInfo for the incoming ray and compute the point of intersection.

  CHitPointInfo &hitInfo = _cnts[0].hitInfo;

  hitInfo.Extrap();

  hitInfo.GetNormal(); 

  // Get object's surface material.

  CAppearance *apr=(CAppearance*)hitInfo.GetObject()->GetAttr(indexAppearance);

  CSurface    *surf = apr->GetSurface();

  int surfaceType = surf->GetSupportedTypes();



  // -------------------------------------------------------------

  //  RADIANCE CACHE

  // -------------------------------------------------------------

  if( (surfaceType & EE_REFLECTED_GLOSSY) && (_rcDepthLimit>0) ) 

  {

    // radiance caching

    assert(_rcAlgo);

    _rcAlgo->PresampleCache(hitInfo,_currPixelFootprintArea,surf,0,

      nexthitInfo);

  }

  else {

    // -------------------------------------------------------------

    //  IRRADIANCE CACHE

    // -------------------------------------------------------------

    if( (surfaceType & EE_REFLECTED_DIFFUSE) && _icAlgo ) {

      // if(hitInfo.GetObject()->ID()==EE_Sphere)

        _icAlgo->IndirectDiffuseLighting(hitInfo, _currPixelFootprintArea,

        surf, 0, nexthitInfo, carray.value2);

    }

  }

}



// --------------------------------------------------------------------------

//  CRT_19_App::IncidentRadiance()

// --------------------------------------------------------------------------

bool 

CRT_19_App::IncidentRadiance(CHitPointInfo      &hitInfo,

                             CContributionItem  *contrib,

                             const int           depth,

                             const float        footprintArea)

{

  // prepare for ray-cast

  hitInfo.SetForFindNearest(contrib->ray);

  // cast the ray

  if (_asds->FindNearest(contrib->ray, hitInfo)) {

    _currPixelFootprintArea = footprintArea;

    // compute radiance along the spawned ray (virtual call)

    ShadeObject2(*contrib, depth+1);

    // ok - object hit

    return true;

  }

  // no object hit, use background color for higher order rays

  BgColor(*contrib);

  // no object hit

  return false;

}



static int a;



// --------------------------------------------------------------------------

//  CRT_19_App::ShadeObject2()

// --------------------------------------------------------------------------

// This function assumes that a "cur" contains a valid ContributionItem with

// a ray hitting the object and that _cnts[depth].hitInfo corresponds to that

// ray. This _cnts[depth].hitInfo is not changed during the computation.

// Function computes the energy on the point of incidence according to lights

// and then cast the rays (reflected/refracted) to compute further contribution

// form secondary rays, but these are stored at _cnts[depth+1]. When such

// rays hit the objects, then the function goes into recursion, until a limit

// for maximum recursion depth is reached.

void

CRT_19_App::ShadeObject2(CContributionItem &contrib, int depth)

{

#ifdef _DEBUG

  static int maxDepth=-1;

  if(depth>maxDepth) {

    maxDepth=depth;

    DEBUG << "maximal recursion depth: " << maxDepth << endl;

  }

#endif



  // -----------------------------------------------------------

  //  SET UP THIS RECURSION LEVEL

  // -----------------------------------------------------------

  

  UpdateCummulativeVars(contrib,depth);

  

  // This is the array used for all contributions of higher order.

  CContributionArray &carray = _cnts[depth+1];

  carray.Reset();

  CHitPointInfo      &nexthitInfo = carray.hitInfo;

  

  // Get hitInfo for the incoming ray and compute the point of intersection.

  CHitPointInfo &hitInfo = _cnts[depth].hitInfo;

  hitInfo.Extrap();

  // compute the normal

  hitInfo.GetNormal();

  

  // Get object's surface material.

  CAppearance *apr=(CAppearance*)hitInfo.GetObject()->GetAttr(indexAppearance);

  CSurface    *surf = apr->GetSurface();

  

  // Reset the resulting color. All further steps will add to this value.

  contrib.color.setz();

  

  // get the surface type

  EDistributionType surfaceType = surf->GetSupportedTypes();



  // -----------------------------------------------------------------------

  // SELF EMITTED RADIANCE

  // -----------------------------------------------------------------------

  if(depth==0)

    AddObjectSelfRadiance(contrib, hitInfo);

  

  // -----------------------------------------------------------------------

  // DIRECT LIGHTING

  // -----------------------------------------------------------------------

  //  if( !((surfaceType & EE_REFLECTED_GLOSSY) && (depth==0)) ) 

  {    

    if ( _directLight && (depth >= _firstLevelForDirectLights) )

      DirectLights(hitInfo, surf, carray, depth);

  }

  

  // -----------------------------------------------------------------------

  // AMBIENT COMPONENT

  // -----------------------------------------------------------------------

  if (surf->HasAmbientTerm() && _scene->ambientLight) {

    _scene->ambientLight->GetIllumination(carray, hitInfo, _samplingInfo);

    surf->GetAmbientTerm(carray.GetLastItem()->weigth, hitInfo);

  }

  

  int firstToTrace = carray.GetUsedItems();

   

  bool radianceCachingPerformed = false;



  // -----------------------------------------------------------

  //  HIGHER ORDER LIGHTING

  // -----------------------------------------------------------

  if( depth < _depthLimit )

  {

#if 1

    // -----------------------------------------------------------

    // DETERMINISTIC REFLECTION / TRANSMISSION

    // -----------------------------------------------------------

    EDistributionType determDist = 

      (EDistributionType)(surfaceType & _distributionType & EE_PERFECT);

    if(determDist)

      int numRays = surf->SpawnDeterministicRays(carray,hitInfo,determDist);



    // -----------------------------------------------------------

    // GLOSSY REFLECTION

    // -----------------------------------------------------------

    if( surfaceType & EE_REFLECTED_GLOSSY) {

      _reflGlossyBounces ++;

      _samplingInfo.SetMaxSamples(_glossyRays[depth]);



      if( depth < _rcDepthLimit ) {

        // radiance caching

        assert(_rcAlgo);

        CValues glossyCol = carray.value2;



        if ( _rcAlgo->IndirectLighting(hitInfo, _currPixelFootprintArea,

          surf,depth,nexthitInfo,glossyCol) ) {

            contrib.color += glossyCol;

            _reflGlossyBouncesRC ++;

            radianceCachingPerformed = true;

            _samplingInfo.SetMaxSamples(0);

            if(_weird(glossyCol.Sum())) {PRINTVAR(glossyCol);FATAL_EXIT;}

        }

      }

      if( _samplingInfo.GetMaxSamples()>0 ) {

        surf->SpawnStochasticRays(carray, hitInfo, _samplingInfo,

          EE_REFLECTED_GLOSSY);

      }

    } // GLOSSY

#else



    /*

    ////////

    // path tracing - experimental

    ////////

    _samplingInfo.SetMaxSamples(_glossyRays[depth]);

#if 0 // importance sampling

    surf->SpawnStochasticRays(carray, hitInfo, _samplingInfo,

      EE_ALL);

#else // uniform sampling

    surf->CSurface::SpawnStochasticRays(carray, hitInfo, _samplingInfo,

      EE_ALL);

#endif

    */



#endif

   

    // ------------------------------------------------------------------

    // TRACE HIGHER ORDER RAYS

    // ------------------------------------------------------------------

    int endToTrace = carray.GetUsedItems();

    

    for (int i = firstToTrace; i < endToTrace; i++)

      IncidentRadiance(nexthitInfo, carray.GetItem(i), depth, 

      _currPixelFootprintArea);

    

  } // depth



  if ( surfaceType & EE_REFRACTED_GLOSSY ) {

    FATAL << "Glossy refraction not yet implemented." << endl;

    FATAL_ABORT;

  }

  

  // -----------------------------------------------------------------------

  // SUM CONTRIBUTIONS

  // -----------------------------------------------------------------------

  carray.WeightAndSum(contrib.color);





  // -----------------------------------------------------------

  // DIFFUSE REFLECTION

  // -----------------------------------------------------------

  if(depth<_depthLimit && !radianceCachingPerformed) {

    if(_icAlgo && (surfaceType & EE_REFLECTED_DIFFUSE) ) {

      // if(hitInfo.GetObject()->ID()==EE_Sphere) 

      {

        CValues &diffuseCol=carray.value2;

        if( _icAlgo -> IndirectDiffuseLighting(hitInfo,_currPixelFootprintArea,

          surf, depth, contrib.color, nexthitInfo, diffuseCol) )

          contrib.color += diffuseCol;

        if(_weird(diffuseCol.Sum())) {PRINTVAR(diffuseCol);FATAL_EXIT;}

      }

    }

  }





  return;

}



// --------------------------------------------------------------------------

//  CRT_19_App::DoReportRT19()

// --------------------------------------------------------------------------

void

CRT_19_App::DoReportRT19(ostream &app)

{

  app

    << "+--------------------------------------------------------------------------+\n"

    << "|                           CRT_19_App - REPORT                            |\n"

    << "|                  (interreflections with radiance caching)                |\n"

    << "+--------------------------------------------------------------------------+\n";

  

  if(_icAlgo)

    _icAlgo->DoReport(app);

  else

    app << "| No irradiance caches                                                     |\n";

  

  if(_rcAlgo)

    _rcAlgo->DoReport(app);

  else

    app << "| No radiance caches                                                       |\n";

  

  app << "| Reflective glossy bounces :   " << _reflGlossyBounces << '\n';

  app << "| Reflective glossy bounces RC: " << _reflGlossyBouncesRC;

  if(_reflGlossyBounces>0) 

    app 

    << " (= " <<  (float)_reflGlossyBouncesRC/(float)_reflGlossyBounces*100 

    <<  "%)";

  app << "\n";

  

  app << "+--------------------------------------------------------------------------+\n";

}



// --------------------------------------------------------------------------

//  CRT_19_App::DoReport()

// --------------------------------------------------------------------------

bool

CRT_19_App::DoReport(const CWorld &world, ostream &app,

                       CStatistics::EType reportType)

{

  // first provide the report of predecessor

  if (CRT_1_App::DoReport(world, app, reportType))

    return true; // some error

  

  DoReportRT19(app);

  

  return false; // ok

}



// --------------------------------------------------------------------------

//  CRT_19_App::SetPreferences()

// --------------------------------------------------------------------------

void

CRT_19_App::SetPreferences(CEnvironment &env)

{

  CRT_1_App::SetPreferences(env);



  bool detRefl;

  env.GetBool("Shading.phongDeterministicReflection", detRefl);

  

  if (detRefl)

    env.SetString("Surface.typePhong", "physicalDR");

  else

    env.SetString("Surface.typePhong", "physical");



  return;

}



// --------------------------------------------------------------------------

//  AllocRT_19_App()

// --------------------------------------------------------------------------

extern CSynthesisApp*  AllocRT_19_App()

{

  return new CRT_19_App();

}



__END_GOLEM_SOURCE

