// ===================================================================
// $Id: rt19app.h,v 1.10 2005/11/02 12:58:47 xkrivanj Exp $
//
// rt19app.h
//     Header file for the renderer using radiance caching.
//
// Class: CRT_19_App
//
// Initial coding by Jaroslav Krivanek, December 2003

#ifndef __RT19APP_H__
#define __RT19APP_H__

// GOLEM headers
#include "configh.h"
#include "rt1app.h"
#include "incrad.h"

// SPHAR headers
#include "shbasis.h"

__BEGIN_GOLEM_HEADER

// Forward declarations
class CICAlgo;
class CRCAlgo;
class CBrdfRepManager;

/// application allocation
extern CSynthesisApp*  AllocRT_19_App();

/// A ray tracing application that full global illumination (except caustics)
/// using radiance caching.
/**
  CRCAlgo class is used for radiance caching, and this app
  stands  mainly as a demo for how to use it.  In order to use
  CRCAlgo, this class must implement the CIncidentRadiance
  interface by defining  the function
  CIncidentRadiance::IncidentRadiance(). Pointer to this interface
  must be passe to the radiance caching algo in order to evaluate the
  primary method (hemisphere sampling) whenever it is needed.

*/
class CRT_19_App:
  public CRT_1_App, public CIncidentRadiance
{
protected: // methods

  virtual void ComputeColor(
    const CRay &ray, 
    CHitPointInfo &info,
    CValues &color, 
    bool saveForReuse = false);

  ///  Recursive part of distributed raytracing.
  virtual void ShadeObject2(CContributionItem &cur, int depth);

  /// This handles eye rays in presampling mode (just fill the cache).
  virtual void _shadeObjectPresample(CContributionItem &cur);

public:
  /// Default constructor sets everything to zero, no construction.
  CRT_19_App();

  /// Virtual constructor calls the 'CleanUp()' method
  virtual ~CRT_19_App();

  /// Precompute anything what is needed (e.g. allocates irrad. caches)
  virtual bool Init(CWorld *world, string &outputName);
  virtual bool InitApplication(CWorld *world, string &outputName);
  virtual bool InitMovie(string &outputName);
  virtual bool InitFrame(string &outputName,CUpdateAnimData *updateAnimdata);

  /// Deletes all the allocated data structures during Init() and Run()
  virtual bool CleanUp();
  virtual bool CleanUpApplication();
  virtual bool CleanUpMovie();
  virtual bool CleanUpFrame();

  /// The main function for this app
  virtual bool Run();
  
  /// returns the application type (EE_DiffuseIrrCache)
  virtual EApplicationType GetApplicationType() const {
    return EE_RadianceCacheRenderer;
  }

  /// reports the statistics result of the specific application
  virtual bool DoReport(const CWorld &world, ostream &app,
			CStatistics::EType reportType);
  
  /// Report only 19 app related statistics.
  void DoReportRT19(ostream &app);

  void SetPreferences(CEnvironment &env);

/**
   Utility function to compute radiance visible in ray's direction.

   Inherited from CIncidentRadiance class. Used for irradiance caching.

   @param hitInfo    Hit info to be used for ray cast.
   @param contrib    Save the color to contrib.color.
   @param currDepth  Current recursion level.

   @return 'true' if object was hit, 'false' otherwise. Even if the return value is 
    'false' there might be (and usually is) a contribution from the background.
*/ 

  inline virtual bool IncidentRadiance(
    CHitPointInfo      &hitInfo,
    CContributionItem  *contrib,
    const int           depth,
    const float        footprintArea);

protected: // methods

  /// Computes a RME of surface's SH approximation (works only for SH!, ni HSH)
  float _approxRME(const CSurface *s);

  /// Loads SH representation of scene BRDFs from a file. 
  /**
     The SH representation is attached to scene's BRDFs.
     @return 'true' if impossible, 'false' if ok.
  */
  bool _loadBRDFSHRep(CSphBasis<float,float>::EType basisType);
  
  /// Computes the SH representation of scene BRDFs.
  /**
     @return 'true' on error, 'false' otherwise.
  */
  bool _computeBRDFSHRep(const CSphBasis<float,float> *shBasis);
  
  /// Saves SH representation of scene BRDFs to a file. 
  /**
     @return 'true' on error, 'false' otherwise.
  */
  bool _saveBRDFSHRep(CSphBasis<float,float>::EType basisType);

  /// Sets up the SH representation of scene BRDFs.
  /**
     First the function tries to load the data from file.
     If it fails, it computes the BRDF rep and saves it to a file.

     @return 'true' if the BRDF computation or saving fails, 'false' otherwise.
  */
  bool _setupBRDFSHRep(const CSphBasis<float,float> *shBasis);

  /// Sets up the SH representation of scene BRDFs via BRDF rep manager.
  /**
     Allocates _brdfRepMan and then uses it to create SH representation
     for scene BRDFs.

     @return 'true' on error, 'false' otherwise.
   */
  bool _setupBrdfRepMan();

  /// Deletes the _brdfRepMan and (H)SH representation of all scene BRDFs.
  /**
    @return 'true' on error, 'false' on success.
  */
  bool _cleanUpBrdfRepMan();

  /// Deletes the (H)SH representation of all scene BRDFs.
  /**
    @return 'true' on error, 'false' on success.
  */
  bool _cleanUpBRDFSHRep();

  void _writeStatImages(
    CSphBRDF<float,float> *shRep,
    int       i,
    bool      adaptiveOrder);

  /// Returns a filename of the sh brdf representation.
  std::string _createBRDFFileName(CSphBasis<float,float>::EType basisType);
  
  virtual void StartingPresampleImage();
  virtual void PresampleImageFinished();
  /// Returns true if irradiance or radiance caching is on, indicating that 
  /// image presampling is desirable.
  virtual bool NeedImagePresampling() const;
  
protected: // data

  /// Encapsulated irradiance caching algorithm.
  CICAlgo *_icAlgo;

  /// Encapsulated radiance caching algorithm.
  CRCAlgo *_rcAlgo;

  /// footprint area for the pixel being currently processed
  float _currPixelFootprintArea;

  /// total number of relective glossy bounces
  unsigned long _reflGlossyBounces;

  /// numbar of reflective glossy bounces handled by the radiace cache
  unsigned long _reflGlossyBouncesRC;

  /// number of radiance cache records (accumulating)
  vector<unsigned long>  _rcRecords;

  /// maximum depth for using radiance cache
  int _rcDepthLimit;

  int *_glossyRays;

  /// BRDF representation manager
  CBrdfRepManager *_brdfRepMan;
};

__END_GOLEM_HEADER

#endif //  __RT19APP_H__

