// ===================================================================
// $Id: rt26app.cpp,v 1.1 2005/11/02 12:58:53 xkrivanj Exp $
//
// rt26app.cpp
//    Irradiance caching guided by a perceptual metric
//      
// Initial coding by Jaroslav Krivanek, March 2005

// GOLEM headers
#include "config.h"
#include "dirliba.h"
#include "icadapt.h"
#include "rt26app.h"
#include "fileio.h"
#include "world.h"
#include "icalgo.h"
#include "hinflite.h"
#include "matrix.h"

#include "icache.h"

// SPHAR headers
#include "shbrdf.h"

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  CRT_26_App::CRT_26_App()
// --------------------------------------------------------------------------
CRT_26_App::CRT_26_App()
:
_hdrFileType("pic")
{
  _hitInfoImage     = 0;
  _icAdapt          = 0;

  _saveIndirectIllumImage  = false;
  _saveDirectIllumImage    = false;
  _saveRecPosImage         = false;
  _saveApproxImage         = false;
  _saveStdDevImage         = false;
  _numTiles = 1;
}

// --------------------------------------------------------------------------
//  CRT_26_App::~CRT_26_App()
// --------------------------------------------------------------------------
CRT_26_App::~CRT_26_App()
{
  CleanUp();
}

// --------------------------------------------------------------------------
//  CRT_26_App::_getPixIdx()
// --------------------------------------------------------------------------
inline int
CRT_26_App::_getPixIdx() const
{
  int w = _image->GetWidth();
  int h = _image->GetHeight();
  int x = _colorMapper->GetCurrX();
  int y = _colorMapper->GetCurrY();
  return y*w + x;
}

// --------------------------------------------------------------------------
//  AddImages()
// --------------------------------------------------------------------------
void AddImages(float *destImage,const CImage &image)
{
  CValues c;
  int w = image.GetWidth();
  int h = image.GetHeight();
  for(int y=0; y<h; y++) {
    for(int x=0; x<w; x++) {
      int idx = w*y+x;
      image.GetValue(idx,c);
      const float *cc = c.data();
      for(int ch=0; ch<3; ch++)
        destImage[3*idx+ch] += cc[ch];
    }
  }
}

// --------------------------------------------------------------------------
//  CRT_26_App::ExtVarFSSFinishedPresampling()
// --------------------------------------------------------------------------
bool
CRT_26_App::ExtVarFSSFinishedPresampling()
{
  // adaptive indirect illumination computation
  if(!_icAdapt || _icAdaptDone)
    return false;

  _icAdaptDone = true;

  // get io for image
  CImageFileIO* imgfileio = _image->GetFileIO();
  if(!imgfileio) {
    FATAL << "_image->GetFileIO() = 0 !!!" << std::endl;
    FATAL_EXIT;
  }

  int w = _image->GetWidth();
  int h = _image->GetHeight();
  int channels = _image->GetChannelNum();

  float *resultImage = new float[w*h*channels];
  assert(resultImage);
  memset(resultImage, 0, 3*w*h*sizeof(float));

  CContributionArray &carray = _cnts[1];
  carray.Reset();
  CHitPointInfo &outHitInfo = carray.hitInfo;

  // dig out the modelview and projection matrices for current camera
  CMatrix4 projMatrix, modelviewMatrix;
  _camera->GetCameraMatrices(modelviewMatrix,projMatrix);
  SAFE_CALL(_icAlgo, SetRenderMode());
  SAFE_CALL(_rcAlgo, SetRenderMode());
  // set base name for saving images
  _icAdapt->SetImgBaseName(imgfileio->GetFilename());
  _icAdapt->SetHdrFileType(_hdrFileType);

  // process in tiles (horizontal bands)
  int tileHeight = h/_numTiles;
  int lastTileHeight = h%_numTiles;
  int tileOffset = w*tileHeight;

  int i=0;
  if(tileHeight>0) {
    for( ; i<_numTiles; i++) {
      _icAdapt->IndirectIlluminationOverlap(resultImage+3*i*tileOffset,
        _hitInfoImage+i*tileOffset,w,tileHeight,outHitInfo,projMatrix,modelviewMatrix);
    }
  }
  if(lastTileHeight>0) {
    _icAdapt->IndirectIlluminationOverlap(resultImage+3*i*tileOffset,
      _hitInfoImage+i*tileOffset,w,lastTileHeight,outHitInfo,projMatrix,
      modelviewMatrix);
  }

  // --------------------------------------
  //  save images with separate components
  // --------------------------------------

  /// save image components separately in HDR format
  string baseName, hdrName;
  ChangeFilenameExtension(imgfileio->GetFilename(),"tiff",baseName);

  // save the indirect illumination HDR image
  CValues stdValue(channels);
  CImageReal image(w,h,3,0,&stdValue);

  if(_saveIndirectIllumImage) {
    PrefixExtensionByName(baseName,string("-indirect"),hdrName);
    image.SetData(resultImage);
    SaveHDRImage(&image,hdrName);
  }

  // save the direct illumination HDR image
  if(_saveDirectIllumImage) {
    PrefixExtensionByName(baseName, string("-direct"), hdrName);
    // image.SetData(_directIllumImage);
    SaveHDRImage(_image,hdrName);
  }

  // save the record positions to HDR image
  if(_saveRecPosImage || _saveApproxImage || _saveStdDevImage) {
    // create the record position image
    float *_auxImage = new float[w*h*3];
    assert(_auxImage);
    // splat record positions
    CICSplatterRegular splatter(projMatrix,modelviewMatrix,w,h,_hitInfoImage,
      CHitInfoLite::EE_IrrCaching);
    splatter.SetIrradImage(_auxImage);

    if(_saveRecPosImage) {
      // splatter.Clear(CICSplatterRegular::EE_Irradiance);
      CTriCol black(0,0,0);
      CTriCol white(1,1,1);
      splatter.ClearIrradiance(white.data());
      if(_icAlgo) {
        splatter.SetRecPosColor(black.data()/*_icAdapt->GetRecPosCol<CIrradianceRecord>()*/);
        splatter.SplatRecordPos(_icAlgo->GetICache(0)->GetRecordList());
      }
      if(_rcAlgo) {
        splatter.SetRecPosColor(black.data()/*_icAdapt->GetRecPosCol<CRadianceRecord>()*/);
        splatter.SplatRecordPos(_rcAlgo->GetRCache(0)->GetRecordList());
      }
      // save the image
      PrefixExtensionByName(baseName, string("-recpos"), hdrName);
      image.SetData(_auxImage);
      SaveHDRImage(&image,hdrName);
    }
    if(_saveApproxImage || _saveStdDevImage) {
      float *weightImage = new float[w*h];
      assert(weightImage);
      splatter.SetWeightImage(weightImage);
      if(_saveApproxImage) {
        splatter.Clear(CICSplatterRegular::EE_Irradiance|CICSplatterRegular::EE_Weight);
        if(_icAlgo) {
          splatter.SetRecPosColor(_icAdapt->GetRecPosCol<CIrradianceRecord>());
          splatter.SplatA(_icAlgo->GetICache(0)->GetRecordList());
        }
        if(_rcAlgo) {
          splatter.SetRecPosColor(_icAdapt->GetRecPosCol<CRadianceRecord>());
          splatter.SplatA(_rcAlgo->GetRCache(0)->GetRecordList());
        }
        splatter.NormalizeBuffer();
        // save the image
        PrefixExtensionByName(baseName, string("-approx"), hdrName);
        image.SetData(_auxImage);
        SaveHDRImage(&image,hdrName);
      }
      if(_saveStdDevImage) {
        splatter.Clear(CICSplatterRegular::EE_Irradiance|CICSplatterRegular::EE_Weight);
        if(_icAlgo)
          splatter.SplatStdDev(_icAlgo->GetICache(0)->GetRecordList());
        if(_rcAlgo)
          splatter.SplatStdDev(_rcAlgo->GetRCache(0)->GetRecordList());
        splatter.NormalizeBuffer();
        PrefixExtensionByName(baseName, string("-stddev"), hdrName);
        image.SetData(_auxImage);
        SaveHDRImage(&image,hdrName);
      }
      delete [] weightImage;
    }
    // delete the record position image
    delete [] _auxImage;
  }

  // add direct illumination
  AddImages(resultImage,*_image);
  // set to output image
  _image->SetData(resultImage);

  delete [] resultImage;

  // save & enlarge the approximation error
  if(_icAlgo && _icAlgo->GetICache(0)) {
    CIrradianceList &list = _icAlgo->GetICache(0)->GetRecordList();
    _icA.resize(list.size());
    float multiplier = 1.0 / _icAlgo->GetPresampleMultiplier();
    CIrradianceList::iterator it = list.begin();
    for( int i=0; it != list.end(); it++, i++) {
      // save old value of 'a'
      _icA[i] = (*it)->GetA();
      (*it)->SetA(_icA[i]*multiplier);
    }
  }

  if(_rcAlgo && _rcAlgo->GetRCache(0)) {
    CRadianceList &list = _rcAlgo->GetRCache(0)->GetRecordList();
    _rcA.resize(list.size());
    float multiplier = 1.0 / _rcAlgo->GetPresampleMultiplier();
    CRadianceList::iterator it = list.begin();
    for( int i=0; it != list.end(); it++, i++) {
      // save old value of 'a'
      _rcA[i] = (*it)->GetA();
      (*it)->SetA(_rcA[i]*multiplier);
    }
  }

  return true;
}

// --------------------------------------------------------------------------
//  CRT_26_App::Run()
// --------------------------------------------------------------------------
// This ray traces the image using distributed ray tracing algorithm
bool
CRT_26_App::Run()
{  
  STATUS << "Rendering ... (ir)radiance caching with a peceptual metric" << endl;
  
  for(int i=0;i<=_depthLimit;i++)
    _cnts[i].value2.SetRepresentation(_defaultColorType);

  // not yet done for this pixel
  _icAdaptDone = false;
  
  // prepare the data structures etc.
  if (RunCommonPrepare())
    return true; // failure

  // compute the image itself
  if (RunCommonCompute())
    return true; // problem

  ExtVarFSSFinishedPresampling();

  // Save the image to the file and deallocate the dynamic variables
  if (RunCommonFinish())
    return true; // problem

  // restore the saved values of 'a'
  // save & enlarge the approximation error
  if(_icAlgo && _icAlgo->GetICache(0)) {
    CIrradianceList &list = _icAlgo->GetICache(0)->GetRecordList();
    CIrradianceList::iterator it = list.begin();
    for( unsigned i=0; it!=list.end() && i<_icA.size(); it++, i++)
      (*it)->SetA(_icA[i]);
  }

  if(_rcAlgo && _rcAlgo->GetRCache(0)) {
    CRadianceList &list = _rcAlgo->GetRCache(0)->GetRecordList();
    CRadianceList::iterator it = list.begin();
    for( unsigned i=0; it!=list.end() && i<_rcA.size(); it++, i++)
      (*it)->SetA(_rcA[i]);
  }

  return false; // ok
}

// --------------------------------------------------------------------------
//  CRT_26_App::InitApplication()
// --------------------------------------------------------------------------
bool 
CRT_26_App::InitApplication(CWorld *world, string &outputName)
{
  // init previous classes
  if (CRT_19_App::InitApplication(world, outputName))
    return true; // error occured in super

  return false;
}

// --------------------------------------------------------------------------
//  CRT_26_App::InitMovie()
// --------------------------------------------------------------------------
bool 
CRT_26_App::InitMovie(string &outputName)
{
  if( CRT_19_App::InitMovie(outputName) ) 
    return true; // error

  // alloc image-sized buffers
  int w = _image->GetWidth();
  int h = _image->GetHeight();
  int channels = _image->GetChannelNum();
  assert(w>0);
  assert(h>0);

  const CEnvironment &env = *_environment;
  env.GetBool("RT26.saveIndirectIllumImage",_saveIndirectIllumImage);
  env.GetBool("RT26.saveDirectIllumImage",_saveDirectIllumImage);
  env.GetBool("RT26.saveRecPosImage",_saveRecPosImage);
  env.GetBool("RT26.saveApproxImage",_saveApproxImage);
  env.GetBool("RT26.saveStdDevImage",_saveStdDevImage);
  env.GetString("Image.HDR.fileType",_hdrFileType);

  string envPrefix="AdaptiveCaching";

  if(_icAlgo || _rcDepthLimit>0) {

    if(_icAlgo && _icAlgo->GetICache(0)) {
      _icAlgo->GetICache(0)->SetUseStoredA(true);
      _icAlgo->GetICache(0)->SetUseNearestRecordA(true);
    }
    if(_rcAlgo && _rcAlgo->GetRCache(0)) {
      _rcAlgo->GetRCache(0)->SetUseStoredA(true);
      _rcAlgo->GetRCache(0)->SetUseNearestRecordA(true);
    }

    if(!_icAdapt) {
      // read params from environment
      float maxErrorIC,maxErrorRC;
      string s;
      int debugImageSaveFrequency;
      bool useThresholdElevation;
      bool useMasking;
      bool saveDebugImages;
      bool saveVDMDebugImages;
      bool saveVDMInOut;
      bool adaptive;
      float mappingCoeff, mappingGamma;

      env.GetString("ToneMapping.algorithm",s);
      if(s=="None")
        mappingCoeff = mappingGamma = 1;
      else if(s=="Linear") {
        env.GetFloat("ToneMapping.linear.coeff",mappingCoeff);
        env.GetFloat("ToneMapping.linear.gamma",mappingGamma);
      }
      else {
        FATAL
          << "AdaptiveCaching only works with linear tone mapping."
          << std::endl;
        FATAL_EXIT;
      }

      env.GetFloat(envPrefix+".maxErrorIC", maxErrorIC);
      env.GetFloat(envPrefix+".maxErrorRC", maxErrorRC);
      env.GetBool (envPrefix+".adaptive",adaptive);
      env.GetBool (envPrefix+".useThresholdElevation",useThresholdElevation);
      env.GetBool (envPrefix+".useMasking",useMasking);
      env.GetBool (envPrefix+".saveDebugImages", saveDebugImages);
      env.GetBool (envPrefix+".saveVDMDebugImages", saveVDMDebugImages);
      env.GetBool (envPrefix+".saveVDMInOut", saveVDMInOut);
      env.GetInt  (envPrefix+".debugImageSaveFrequency",
        debugImageSaveFrequency);
      if(debugImageSaveFrequency<1)debugImageSaveFrequency=1;
      env.GetInt(envPrefix+".numTiles",_numTiles);
      if(_numTiles<1) _numTiles = 1;
      if(_numTiles>h) _numTiles = h;

      env.GetString(envPrefix+".convergenceAlgo",s);
      CAdaptiveCaching::EConvergenceAlgo convergenceAlgo;
      if(s == "non_adaptive")
        convergenceAlgo = CAdaptiveCaching::EE_ConvergenceAlgoNonAdaptive;
      else if(s == "per_pixel")
        convergenceAlgo = CAdaptiveCaching::EE_ConvergenceAlgoPerPixel;
      else if(s == "one_pass")
        convergenceAlgo = CAdaptiveCaching::EE_ConvergenceAlgoOnePass;
      else if(s == "list")
        convergenceAlgo = CAdaptiveCaching::EE_ConvergenceAlgoList;
      else if(s == "tree")
        convergenceAlgo = CAdaptiveCaching::EE_ConvergenceAlgoTree;
      else if(s == "bbox")
        convergenceAlgo = CAdaptiveCaching::EE_ConvergenceAlgoBBox;
      else if(s == "bbox_irreg")
        convergenceAlgo = CAdaptiveCaching::EE_ConvergenceAlgoBBoxIrreg;
      else {
        FATAL << "Unknown value for '" << envPrefix << ".convergenceAlgo': '" 
          <<  s << "'.\n";
        FATAL << "Possible values are:\n"
          << "\t'non_adaptive'\n"
          << "\t'per_pixel'\n"
          << "\t'one_pass'\n"
          << "\t'list'\n"
          << "\t'tree'\n"
          << "\t'bbox'\n"
          << "\t'bbox_irreg'.\n"
          << std::endl;
        FATAL_EXIT;
      }

      // create object
      _icAdapt = new CAdaptiveCaching(convergenceAlgo,
        _icAlgo,maxErrorIC,
        _rcAlgo,maxErrorRC,adaptive,mappingCoeff, mappingGamma);

      assert(_icAdapt);
      // set further params
      _icAdapt->SetDebugImageSaveFrequency(debugImageSaveFrequency);
      _icAdapt->SetSaveDebugImages(saveDebugImages);
      _icAdapt->SetSaveVDMDebugImages(saveVDMDebugImages);
      _icAdapt->SetSaveVDMInOut(saveVDMInOut);
      _icAdapt->SetUseThresholdElevation(useThresholdElevation);
      _icAdapt->SetUseMasking(useMasking);
    }

    if(!_hitInfoImage)     _hitInfoImage     = new CHitInfoLite[w*h];
    assert(_hitInfoImage);
    // buffers are cleared in InitFrame()
  }

  return false;
}


// --------------------------------------------------------------------------
//  CRT_26_App::InitFrame()
// --------------------------------------------------------------------------
bool 
CRT_26_App::InitFrame(string &outputName,
		      CUpdateAnimData *updateAnimdata)
{
  if( CRT_19_App::InitFrame(outputName, updateAnimdata) )
    return true; // error

  // clear the image buffers
  int w = _image->GetWidth();
  int h = _image->GetHeight();
  int channels = _image->GetChannelNum();
  int len = w*h*channels;

  if(_hitInfoImage) memset(_hitInfoImage, 0, w*h*sizeof(CHitInfoLite));

  return false;
}

// --------------------------------------------------------------------------
//  CRT_26_App::CleanUpFrame()
// --------------------------------------------------------------------------
bool 
CRT_26_App::CleanUpFrame()
{
  /*
  if(_image && _directIllumImage && _kdImage) {
    int w = _image->GetWidth();
    int h = _image->GetHeight();
    int channels = _image->GetChannelNum();
    WriteImageTiff(w,h,channels,_directIllumImage,"directIllumImage.tif");
    WriteImageTiff(w,h,channels,_kdImage,"kdImage.tif");
  }
  */
  return CRT_19_App::CleanUpFrame();
}

// --------------------------------------------------------------------------
//  CRT_26_App::CleanUpMovie()
// --------------------------------------------------------------------------
bool 
CRT_26_App::CleanUpMovie()
{
  SAFE_DELETE_ARRAY(_hitInfoImage);

  SAFE_DELETE(_icAdapt);

  return CRT_19_App::CleanUpMovie();
}

// --------------------------------------------------------------------------
//  CRT_26_App::CleanUpApplication()
// --------------------------------------------------------------------------
bool 
CRT_26_App::CleanUpApplication()
{

  return CRT_19_App::CleanUpApplication();
}


// --------------------------------------------------------------------------
//  CRT_26_App::ComputeColor()
// --------------------------------------------------------------------------
void
CRT_26_App::ComputeColor(const CRay &ray, CHitPointInfo &hitInfo,
			CValues &color, bool /*saveForReuse*/)
{
  unsigned handleFlag = 0;
  if(_icAdapt && !_icAdaptDone && !_presampling && hitInfo.GetObject()) {
      hitInfo.Extrap();
      handleFlag = _savePixelInfo(_hitInfoImage[_getPixIdx()],hitInfo);
  }

  if (hitInfo.GetObject() != NULL) 
  {
    INC_HIT_COUNTER(hitInfo.GetObject());
    INC_PRIMARY_HIT;

    if(_presampling)
      return;

    // save the footprint area for this pixel (will be used for all recursion levels)
    hitInfo.Extrap();
    hitInfo.GetNormal();
    _currPixelFootprintArea = GetFootprintArea(hitInfo);

    // get the item to assign the primary ray and hitInfo
    _cnts[0].Reset();
    CContributionItem *citem = _cnts[0].AllocateNext();
    
    // hitInfo must be copied
    _cnts[0].hitInfo = hitInfo;
    
    // the initial color and weigth format for the primary rays
    citem->weigth.SetRepresentation(_defaultWeightType);
    citem->color.SetRepresentation(_defaultColorType);
    citem->weigth.Set(1.0);
    citem->diffuseness = 0.0;

    // handle everything except irradiance/radiance cache
    if(_icAdaptDone)
      ShadeObject2(*citem, 0);
    else
      _shadeObjectDirect(*citem, 0, handleFlag);
   
    color = citem->color;
  }
  else 
  {
    INC_PRIMARY_MISS;
    if(!_presampling)
      _scene->background->GetColor(ray, color);
  }

  if(handleFlag)
    color.CopyData(_hitInfoImage[_getPixIdx()].directIllum.data());
}

// --------------------------------------------------------------------------
//  CRT_26_App::NeedImagePresampling()
// --------------------------------------------------------------------------
bool 
CRT_26_App::NeedImagePresampling() const
{
  // presampling is NOT needed, bacause everything is handled in
  // CAdaptiveCaching
  return false;
}

// --------------------------------------------------------------------------
//  CRT_26_App::_savePixelInfo()
// --------------------------------------------------------------------------
unsigned
CRT_26_App::_savePixelInfo(CHitInfoLite &hitInfoLite,
                           const CHitPointInfo &hitInfo)
{
  int channels = _image->GetChannelNum();
  unsigned handleFlag = 0;
  // leave the rest blank if no object was hit
  if(!hitInfo.GetObject()) return handleFlag;
  // make sure the dot product is computed
  CVector3D N = hitInfo.GetNormal();
  CVector3D uN = hitInfo.GetGeometricNormal();
  // is the object hit from front?
  if(hitInfo.GetUDotProd()>=0) return handleFlag;
  // Get object's surface material.
  CAppearance *apr=(CAppearance*)hitInfo.GetObject()->GetAttr(indexAppearance);
  if(!apr) return handleFlag;
  CSurface    *surf = apr->GetSurface();
  if(!surf) return handleFlag;

  // try if radiance caching works
  const CSphBRDF<float,float> *shRep = surf->GetSHRep();

  if(shRep && _rcDepthLimit>0 && surf->IsTypeSupported(EE_REFLECTED_GLOSSY)) {
    // construct the local coordinate system
    CVector3D U, V;
    N.RightHandedBase(U,V);
    // transform incoming dir to (perturbed) local coords 
    CVector3D inDir = ToLocalFrame(-hitInfo.GetRay()->GetDir(), U, V, N); 
    float u,v;
    shRep->GetUV(inDir,u,v);
    if (u>=0 && u<=1 && v>=0 && v<=1) {
      if(!shRep->GetAdaptiveDeletion() || 
        (shRep->GetErrorUV(u,v)<=shRep->GetMaxAllowedError()))
      {
        // YES to radiance caching!
        handleFlag = CHitInfoLite::EE_RadCaching;
        hitInfoLite.U = U;
        hitInfoLite.V = V;
        hitInfoLite.shRep = shRep;
        hitInfoLite.brdf_u = u;
        hitInfoLite.brdf_v = v;
      }
    }
  }

  // try if irradiance caching works
  if(handleFlag==0 && _icAlgo && surf->IsTypeSupported(EE_REFLECTED_DIFFUSE)) {
    CValues kd;
    if(surf->GetAlbedo(kd,hitInfo,EE_REFLECTED_DIFFUSE) && kd.Luminance()>1e-6) {
      handleFlag = CHitInfoLite::EE_IrrCaching;
      kd *= M_1_PI;
      kd.CopyData(hitInfoLite.kd);
    }
  }

  if(handleFlag) {
    // common for irradiance and radiance caching
    hitInfoLite.flags = handleFlag;
    hitInfoLite.SetObject(hitInfo.GetObject());
    hitInfoLite.P      = hitInfo.GetPoint();
    hitInfoLite.N      = N;
    hitInfoLite.uN     = uN;
    hitInfoLite.SetDot( -hitInfo.GetDotProd() );
    hitInfoLite.footprintArea = GetFootprintArea(hitInfo);
    hitInfoLite.SetTermination(hitInfo.GetRay()->GetTermination());
  }
  return handleFlag;
}

// --------------------------------------------------------------------------
//  CRT_26_App::_shadeObjectDirect()
// --------------------------------------------------------------------------
void
CRT_26_App::_shadeObjectDirect(CContributionItem &contrib,
                               int depth,
                               int handleFlag)
{  
  // -----------------------------------------------------------
  //  SET UP THIS RECURSION LEVEL
  // ----------------------------------------------------------- 
  UpdateCummulativeVars(contrib,depth);
  
  // This is the array used for all contributions of higher order.
  CContributionArray &carray = _cnts[depth+1];
  carray.Reset();
  CHitPointInfo      &nexthitInfo = carray.hitInfo;
  
  // Get hitInfo for the incoming ray and compute the point of intersection.
  CHitPointInfo &hitInfo = _cnts[depth].hitInfo;
  hitInfo.Extrap();
  // compute the normal
  hitInfo.GetNormal();
  
  // Get object's surface material.
  CAppearance *apr=(CAppearance*)hitInfo.GetObject()->GetAttr(indexAppearance);
  CSurface    *surf = apr->GetSurface();
  
  // Reset the resulting color. All further steps will add to this value.
  contrib.color.Set(0.0);
  
  // get the surface type
  EDistributionType surfaceType = surf->GetSupportedTypes();
  
  // -----------------------------------------------------------------------
  // SELF EMITTED RADIANCE
  // -----------------------------------------------------------------------
  if(depth==0) AddObjectSelfRadiance(contrib, hitInfo);
  
  // -----------------------------------------------------------------------
  // DIRECT LIGHTING
  // -----------------------------------------------------------------------
  if ( _directLight && (depth >= _firstLevelForDirectLights) )
    DirectLights(hitInfo, surf, carray, depth);
  
  // -----------------------------------------------------------------------
  // AMBIENT COMPONENT
  // -----------------------------------------------------------------------
  if (surf->HasAmbientTerm() && _scene->ambientLight) {
    if(_scene->ambientLight->GetIllumination(carray, hitInfo, _samplingInfo))
      surf->GetAmbientTerm(carray.GetLastItem()->weigth, hitInfo);
  }

  // -----------------------------------------------------------------------
  // GLOSSY REFLECTION VIA MONTE CARLO
  // -----------------------------------------------------------------------
  if( ( surfaceType & EE_REFLECTED_GLOSSY ) && 
     !( handleFlag&CHitInfoLite::EE_RadCaching ) &&
      ( depth<_depthLimit ) &&
      ( _glossyRays[depth] > 0 ))
  {
    int firstToTrace = carray.GetUsedItems();
    _samplingInfo.SetMaxSamples(_glossyRays[depth]);
    surf->SpawnStochasticRays(carray,hitInfo,_samplingInfo,EE_REFLECTED_GLOSSY);
    int endToTrace = carray.GetUsedItems();
    for (int i = firstToTrace; i < endToTrace; i++)
      IncidentRadiance(nexthitInfo, carray.GetItem(i), depth, 
      _currPixelFootprintArea);
  }
 
  // -----------------------------------------------------------------------
  // SUM CONTRIBUTIONS
  // -----------------------------------------------------------------------
  carray.WeightAndSum(contrib.color);

  return;
}

// --------------------------------------------------------------------------
//  CRT_26_App::DoReportRT26()
// --------------------------------------------------------------------------
void
CRT_26_App::DoReportRT26(ostream &app)
{
  app
    << "+--------------------------------------------------------------------------+\n"
    << "|                           CRT_26_App - REPORT                            |\n"
    << "|               perceptually guided adaptive irradiance caching            |\n"
    << "+--------------------------------------------------------------------------+\n";
  if(_icAdapt)
    _icAdapt->DoReport(app);
  else
    app 
    << "| Nothing to report.                                                       |\n";
  app 
    << "+--------------------------------------------------------------------------+\n";
}

// --------------------------------------------------------------------------
//  CRT_26_App::DoReport()
// --------------------------------------------------------------------------
bool
CRT_26_App::DoReport(const CWorld &world, ostream &app,
                       CStatistics::EType reportType)
{
  // first provide the report of predecessor
  if (CRT_19_App::DoReport(world, app, reportType))
    return true; // some error
  
  DoReportRT26(app);
  
  return false; // ok
}

// --------------------------------------------------------------------------
//  AllocRT_26_App()
// --------------------------------------------------------------------------
extern CSynthesisApp*  AllocRT_26_App()
{
  return new CRT_26_App();
}

__END_GOLEM_SOURCE
