// ===================================================================
// $Id: vdm.cpp,v 1.1 2005/11/02 12:59:02 xkrivanj Exp $
//
// vdm.cpp
//   Visible Difference Metric
//
// Initial coding by Jaroslav Krivanek (March, 2005).

// GOLEM headers
#include "config.h"
#include "vdm.h"
#include "imgpyram.h"
#include "jkimage.h"

#include "basmacr.h"
#include "fileio.h"
#include "timer.h"

// standard headers
#include <cmath>
#include <iomanip>

__BEGIN_GOLEM_SOURCE

// --------------------------------------------------------------------------
//  SavePyramid()
// --------------------------------------------------------------------------
void SavePyramid(const CImagePyramid &pyramid, const char *basename)
{
  assert(basename);
  char imgName[1000];
  for(int i=0; i<pyramid.size(); i++) {
    const jkimage<float> &img = pyramid[i];
    sprintf(imgName, "vdm-%s-%03d.tif", basename, i);
    WriteImageTiff(img.w(),img.h(),1,img.data(),imgName);
  }
}

// --------------------------------------------------------------------------
//  CVDMWeber::CVDMWeber()
// --------------------------------------------------------------------------
void 
CVDMWeber::Diff(float* diffMap, const float* img1, const float* img2) const
{
  assert(diffMap);
  assert(img1);
  assert(img2);

  float invMaxContrast = 1.0f/_maxContrast;

  for(int i=0; i<_w*_h; i++) {
    // luminance contrast
    float C = LumContrast(LumFromRGB(&img1[3*i]),LumFromRGB(&img2[3*i]));     
    // michelson constrast
    // float C = MichelsonContrast(LumFromRGB(&img1[3*i]),LumFromRGB(&img2[3*i]));
    // how many times the contrast is bigger than maximum allowed contrast
    diffMap[i] = C * invMaxContrast;
  }
}

// --------------------------------------------------------------------------
//  CVDMRamasubramanian::CVDMRamasubramanian()
// --------------------------------------------------------------------------
CVDMRamasubramanian::CVDMRamasubramanian(const jkimage<float> &src,
                                         bool useMasking,
                                         bool verbatim,
                                         bool writeDebugImages)
                                         :CVDM(src.w(), src.h())
{
  const char *imgext="hdr";

  CTimer timer, timerTotal;

  if(verbatim) {
    STATUS << "- precomputing ramasubramanian spatial component..." << std::endl;
    STATUS << setprecision(3);
    timerTotal.Start();
  }

  CImagePyramidBurtAdelson gaussPyramid;
  CImagePyramidBurtAdelson contrastPyramid;
  CImagePyramidBurtAdelson fractionPyramid;
  int width   = src.width();
  int height  = src.height();
  int datalen = src.datalen();
  _Fspatial = new jkimage<float>(width,height);
  assert(_Fspatial);

  // frequency in cycles per degree
  float CPD[8] = {32,16,8,4,2,1,0.5,0.25};
  // normalized threshold elevation due to CSF. Fcsf[i]=514/CSF(CPD[i],100)
  float Fcsf[8] = {31.32,4.2,1.57,1.02,1,1,1,1};

  // -------------------------------------------
  //  a) generate 8 levels of Gaussian pyramid
  // -------------------------------------------

  if(verbatim) {
    STATUS << " - creating multires...                     ";
    timer.Restart();
  }

  gaussPyramid.CreateMultires(src, 8);
  // gaussPyramid.CreatePyramid(src, 8);

  if(verbatim) {
    timer.Stop();
    STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
  }

  if(writeDebugImages) {
    SavePyramid(gaussPyramid,"gauss");
  }

  //FATAL_EXIT;

  // --------------------------------------------------
  //  c) compute bandlimited Weber contrast
  // --------------------------------------------------
  if(verbatim) {
    STATUS << " - computing bandlimited Weber contrast... ";
    timer.Restart();
  }

  int numContrastPyramidLevels = Min(8,gaussPyramid.size()-2);

  /* 
  // debugging code
  int offset = 500*width;
  float *Gi  = gaussPyramid[0].data()+offset;   assert(Gi);
  float *Gi1 = gaussPyramid[1].data()+offset; assert(Gi1);
  float *Gi2 = gaussPyramid[2].data()+offset; assert(Gi2);
  char buff[500];
  DEBUGW << std::endl;
  for(int k=0; k<3; k++) {
    sprintf(buff,"Gi %.5f, Gi1 %.5f, Gi2 %.5f | Gi-Gi1 %.5f | (Gi-Gi1)/Gi2 %.5f\n",
      Gi[k], Gi1[k], Gi2[k], fabsf(Gi[k]-Gi1[k]), fabsf(Gi[k]-Gi1[k])/Gi2[k] );
    DEBUGW << buff;
  }
  //FATAL_EXIT;
  */
     
  for(int i=0; i<numContrastPyramidLevels; i++) {
    jkimage<float> *img = new jkimage<float>(width,height);
    assert(img);
    contrastPyramid.push_back(*img);
    float *Ci  = img->data();
    const float *Gi  = gaussPyramid[i].data();   assert(Gi);
    const float *Gi1 = gaussPyramid[i+1].data(); assert(Gi1);
    const float *Gi2 = gaussPyramid[i+2].data(); assert(Gi2);

    // compute the contrast
    for(int k=0; k<datalen; k++)
      Ci[k] = fabsf( Gi[k] - Gi1[k] ) / (Gi2[k] + 1e-5);
  }

  if(verbatim) {
    timer.Stop();
    STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
  }

  if(writeDebugImages) {
    SavePyramid(contrastPyramid,"contrast");
  }

  // --------------------------------------------------
  //  d) sum contrasts
  // --------------------------------------------------
  if(verbatim) {
    STATUS << " - summing contrasts...                    ";
    timer.Restart();
  }

  jkimage<float> sumContrast(contrastPyramid[0]);
  float *sumC = sumContrast.data();

  for(int i=1; i<contrastPyramid.size(); i++)
    shvf::addvvip4(datalen, sumC, contrastPyramid[i].data());

  if(verbatim) {
    timer.Stop();
    STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
  }

  if(writeDebugImages) {
    WriteImageTiff(sumContrast.w(),sumContrast.h(),1,sumC,"vdm-contrastsum.tif");
  }

  // -------------------------------------------------------
  //  e) compute fraction_i(x,y) = C_i(x,y)/ SUM_i C_i(x,y)
  // -------------------------------------------------------
  if(verbatim) {
    STATUS << " - computing fraction_i...                 ";
    timer.Restart();
  }

  // sumC = 1/sumC
  for(int k=0; k<datalen; k++)
    sumC[k] = 1.0f/(sumC[k] + 1e-5f);

  for(int i=0; i<contrastPyramid.size(); i++) {
    // alloc new image
    jkimage<float> *img = new jkimage<float>(width,height);
    assert(img);
    fractionPyramid.push_back(*img);
    const float *Ci = contrastPyramid[i].data();
    // fraction_i = C_i / sumC
    shvf::mulvv4(datalen, img->data(), Ci, sumC);
  }

  if(verbatim) {
    timer.Stop();
    STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
  }

  if(writeDebugImages) {
    SavePyramid(fractionPyramid,"fraction");
  }

  // -------------------------------------------------------
  //  f) normalize constrast with the CSF & compute masking
  // -------------------------------------------------------
  if(useMasking) {
    if(verbatim) {
      STATUS << " - contrast normalization & masking...     ";
      timer.Restart();
    }

    for(int i=0; i<contrastPyramid.size(); i++) {
      const float *L  = gaussPyramid[i+1].data();
      float       *Ci = contrastPyramid[i].data();

      for(int k=0; k<datalen; k++) {
        // C_i / CSF  (for luminance one level down)
        float Scsf = CSFBarten( CPD[i], L[k] );
        Ci[k] = MaskingDaly( Ci[k]*Scsf );
      }
    }

    if(verbatim) {
      timer.Stop();
      STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
    }

    if(writeDebugImages) {
      SavePyramid(contrastPyramid,"normContrast");
    }
  }

  // -------------------------------------------------------
  //  f) compute Fspatial
  // -------------------------------------------------------
  if(verbatim) {
    STATUS << " - computing Fspatial...                   ";
    timer.Restart();
  }

  _Fspatial->setz();
  float *Fspatial = _Fspatial->data();

  for(int i=0; i<contrastPyramid.size(); i++) {
    const float *Fmasking_i = contrastPyramid[i].data();
    const float *fraction_i = fractionPyramid[i].data();

    if(useMasking) {
      for(int k=0; k<datalen; k++) {
        Fspatial[k] += Fcsf[i] * fraction_i[k] * Fmasking_i[k];
      }   
    }
    else
      shvf::madvsip4(datalen, Fspatial, fraction_i, Fcsf[i]);
  }

  if(verbatim) {
    timer.Stop();
    STATUS << "done in " << timer.UserTime() << " sec." << std::endl;
  }

  if(writeDebugImages) {
    WriteImageTiff(_Fspatial->w(),_Fspatial->h(),1,_Fspatial->data(),
      "vdm-Fspatial.tif");
    WriteImageRGBE(_Fspatial->w(),_Fspatial->h(),1,_Fspatial->data(),
      "vdm-Fspatial.hdr");
  }

  if(verbatim) {
    timerTotal.Stop();
    STATUS 
      << "- ramasubramanian spatial component precomputed in " 
      << timerTotal.UserTime() << " sec." << std::endl;
  }
}

// --------------------------------------------------------------------------
//  CVDMRamasubramanian::~CVDMRamasubramanian()
// --------------------------------------------------------------------------
CVDMRamasubramanian::~CVDMRamasubramanian()
{
  delete _Fspatial;
}


__END_GOLEM_SOURCE
