// ===================================================================
// $Id: vdm.h,v 1.2 2005/11/29 16:04:05 xkrivanj Exp $
//
// vdm.h
//   Visible Difference Metric
//   
// REPLACEMENT_STRING
//
// Initial coding by Jaroslav Krivanek (March, 2005).

#ifndef __VDM_H__
#define __VDM_H__

// GOLEM headers
#include "configh.h"

// standard headers
#include <cassert>
#include <cmath>

__BEGIN_GOLEM_HEADER

// forward declarations
// class CImageRedExp;
template<class T>
class jkimage;

// --------------------------------------------------------------------------
//  class CVDM
// --------------------------------------------------------------------------

/// Computes wisible difference between a pair of images.
class CVDM {

public: // methods

  inline CVDM(int w, int h);

  virtual ~CVDM() {}

  /// Evaluate the visual difference between the two images
  virtual void Diff(float* diffMap,const float* img1,const float* img2)const=0;

  /// Compute luminance value for a RGB triplet
  /**
  Copied from Erik Renhard's tone mapper. Maybe it should be computed in some
  better way. First I have to read up on color from Erik's book.
  */
  template<class T>
  static inline T LumFromRGB(const T* col)
  {
    assert(col);
    // from erik
    /*
    return 
      0.3278688f * col[0] + 
      0.6557377f * col[1] + 
      0.0163934f * col[2];
    */
    // from golem
    return
      0.299f * col[0] + 
      0.587f * col[1] + 
      0.114f * col[2];
    /*
    // average
    return
      0.33333f*(col[0]+col[1]+col[2]);
    */
  }

  /// Evaluate Barten CSF for a given frequency in CPD and luminance.
  /// Return the contrast sentisivity (=1/threshold contrast).
  static inline float CSFBarten(const float f, const float L)
  {
    float a = 440.0f * pow( (1.0 + 0.7/L), -0.2);
    float b =   0.3f * pow( (1.0 + 100.0/L), 0.15);
    return a*f * exp(-b*f) * sqrt(1.0 + 0.06*exp(b*f));
  }

  /// Return threshold elevation due to masking for a given normalized 
  /// contrast.
  static inline float MaskingDaly(const float Cn)
  {
    float a = 0.0153*pow(392.498*Cn, 0.7);
    return pow(1.0 + a*a*a*a, 0.25);
  }

  /// compute luminance contrast
  template<class T>
  static inline T LumContrast(const T& L1, const T& L2)
  {
    return fabs(L1-L2)/(L1+T(1e-3));
  }

  /// compute Michelson contrast
  template<class T>
  static inline T MichelsonContrast(const T& L1, const T& L2)
  {
    return fabs(L1-L2)/(L1+L2+T(1e-3));
  }

protected: // data

  /// image width
  int _w;
  /// image height
  int _h;
};

// --------------------------------------------------------------------------
//  CVDM::CVDM()
// --------------------------------------------------------------------------
inline 
CVDM::CVDM(int w, int h)
{
  assert(w>0);
  assert(h>0);

  _w = w;
  _h = h;
}

// --------------------------------------------------------------------------
//  class CVDMWeber
// --------------------------------------------------------------------------
/// Visible diference based on the pixelwise Weber contrast (relative 
/// difference between images)
class CVDMWeber: public CVDM {

protected:

  float _maxContrast;

public:

  CVDMWeber(int w, int h, float maxContrast);

  virtual ~CVDMWeber() {}

  virtual void Diff(float* diffMap,const float* img1,const float* img2)const;
};


// --------------------------------------------------------------------------
//  CVDMWeber::CVDMWeber()
// --------------------------------------------------------------------------
inline 
CVDMWeber::CVDMWeber(int w, int h, float maxContrast)
:CVDM(w,h)
{
  assert(maxContrast>0);
  _maxContrast = maxContrast;
}


// --------------------------------------------------------------------------
//  class CVDMRamasubramanian
// --------------------------------------------------------------------------
class CVDMRamasubramanian : public CVDM
{
public: // methods

  /// Constructor performs the spatial processing and precomputes
  /// the threshold elevation map.
  CVDMRamasubramanian(
    const jkimage<float> &spatial,
    bool useMasking,
    bool verbatim,
    bool writeDebugImages);

  virtual ~CVDMRamasubramanian();

  virtual void Diff(float* diffMap,const float* img1,const float* img2)const
  { }

  /// return the precomputed spatial component of the threshold elevation
  inline const jkimage<float>& GetElevationImage() const {return *_Fspatial;}

protected: // methods

public: // data

protected: // data

  jkimage<float> *_Fspatial;
};


__END_GOLEM_HEADER

#endif //  __VDM_H__
